#include "cashboxserviceapi.h"
#include "logmanip.h"
#include "cashboxbuildconfig.h"


#include <QElapsedTimer>
#include <QAndroidParcel>
#include <QAndroidJniExceptionCleaner>
#include <QCoreApplication>


CashboxServiceConnection::CashboxServiceConnection(CashboxServiceApi *parent) noexcept
    : QAndroidServiceConnection()
    , api_(parent)
{

}

CashboxServiceConnection::CashboxServiceConnection(CashboxServiceApi *parent, const QAndroidJniObject &serviceConnection) noexcept
    : QAndroidServiceConnection(serviceConnection)
    , api_(parent)
{

}

CashboxServiceConnection::~CashboxServiceConnection()
{

}

void CashboxServiceConnection::onServiceConnected(const QString &name, const QAndroidBinder &serviceBinder)
{
    Q_UNUSED(name)
//    lmWarning() << "CashboxServiceConnection::onServiceConnected" << name << api_;
    if(api_)api_->binder_ = QSharedPointer<QAndroidBinder>(new QAndroidBinder(serviceBinder));

}

void CashboxServiceConnection::onServiceDisconnected(const QString &name)
{
    Q_UNUSED(name)
//    lmWarning() << "CashboxServiceConnection::onServiceDisconnected" << name << api_;
    if(api_)api_->binder_ = QSharedPointer<QAndroidBinder>();
}
//-------------------------------------------------------------------------------------------
CashboxServiceApi::CashboxServiceApi(QObject *parent) noexcept
    : QObject(parent)
    , binder_()
    , connection_{this}
{
    try
    {
        QString serviceClass = SERVICE_CLASS;
        QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
        Q_UNUSED(cleaner)
        QAndroidIntent intent;
        QAndroidJniObject pn = QAndroidJniObject::fromString(CASHBOX_SERVICE_PACKAGE_NAME);
        QAndroidJniObject cn = QAndroidJniObject::fromString(serviceClass);
        QAndroidJniObject component("android/content/ComponentName", "(Ljava/lang/String;Ljava/lang/String;)V",
                                    pn.object<jstring>(), cn.object<jstring>());
//        lmWarning() << component.isValid();
        QAndroidJniObject resobj =
                intent.handle().callObjectMethod("setComponent",
                                                 "(Landroid/content/ComponentName;)Landroid/content/Intent;",
                                                 component.object());
//        lmWarning() << intent.handle().toString() << logtab << resobj.toString();
        QtAndroid::BindFlags flags;
        flags.setFlag(QtAndroid::BindFlag::AutoCreate, true);
        bool res = QtAndroid::bindService(intent, connection_, flags);
        waitBinder(2000);
        lmWarning() << "CashboxServiceApi " << res;
    }
    catch(...)
    {
        lmWarning() << "INVALID CONSTRUCTOR";
    }
}

CashboxServiceApi::~CashboxServiceApi()
{
    binder_ = QSharedPointer<QAndroidBinder>();
}

bool CashboxServiceApi::isReady() const
{
    return !binder_.isNull();
}

bool CashboxServiceApi::ping()
{
    QAndroidParcel out;
    return transact(CoreApiConst::RemoteCommand::PING_TRANSACTION, out).isOk();
}

CoreApiResult CashboxServiceApi::loadCashierNames()
{
    CoreTransaction t;
    t.setOperation(CoreTransaction::Operation::GET_CASHIER_NAMES);
    t.genUid();
    return exec(t);
}

CoreApiResult CashboxServiceApi::exec(CoreTransaction trans)
{
    QAndroidParcel reply, request;
    request.writeVariant(trans.toMap());
    CoreApiResult res = transact(trans.operation(), reply, &request);


    if(request.handle().isValid())
    {
        request.handle().callMethod<void>("recycle");
    }
    if(reply.handle().isValid())
    {
        reply.handle().callMethod<void>("recycle");
    }
    return res;
}

CoreApiResult CashboxServiceApi::writeSerial(CoreTransaction trans)
{
    CoreApiResult res = exec(trans);
    if(res.isOk())
    {
        CashboxStatus st(res.data());
        initGlobalStatus(st);
        lmWarning() << logvariant(res.toMap());
        return res;
    }
    return res;
}

CoreApiResult CashboxServiceApi::loadCashboxStatus(bool toExternal)
{
    CoreTransaction trans;
    trans.setOperation(CoreTransaction::Operation::CASHBOX_STATUS);
    trans.setFormatToExternal(toExternal);
    trans.genUid();
    CoreApiResult res = exec(trans);

    CashboxStatus st(res.data());
    initGlobalStatus(st);
//    lmWarning() << logvariant(res.toMap());
    return res;
}

ModelData CashboxServiceApi::loadModelData(bool toExternal)
{
    CoreTransaction trans;
    trans.setOperation(CoreTransaction::Operation::GET_CASHBOX_MODELDATA);
    trans.setFormatToExternal(toExternal);
    trans.genUid();
    CoreApiResult res = exec(trans);

    CashboxStatus st(res.data());
//    lmWarning() << logvariant(res.toMap());
    return ModelData(res.data());
}

CoreApiResult CashboxServiceApi::loadReceiptData()
{
    CoreTransaction trans;
    trans.setOperation(CoreTransaction::Operation::GET_DATA_FOR_RECEIPT);
    trans.genUid();
    CoreApiResult res = exec(trans);
    initCatalogsData(res.data());
//    lmWarning() << logvariant(res.toMap());
    return res;
}



void CashboxServiceApi::waitBinder(qint32 timeout)
{
    QElapsedTimer t;
    t.start();
    while(!binder_ && t.elapsed() < timeout)
    {
        QCoreApplication::processEvents();
    }
}

CoreApiResult CashboxServiceApi::transact(CoreApiConst::RemoteCommand code, QAndroidParcel &reply, QAndroidParcel *data)
{
    lmWarning() << QString::number(static_cast<qint32>(code), 16);
    QAndroidParcel in;
    if(data) in = *data;
    if(!binder_)
    {
        lmWarning() << QString::number(static_cast<qint32>(code), 16) << " BINDER IS NULL";
        return CoreApiResult(CoreApiConst::ErrorCode::ServiceAccessError);
    }
    if(!binder_->transact(static_cast<qint32>(code), in, &reply))
    {
        lmWarning() << "BINDER ERROR " << QString::number(static_cast<qint32>(code), 16)
                    << logvariant(in.readVariant())
                    << logtab<< logvariant(reply.readVariant())
                    << logtab << binder_->handle().toString()
                    << binder_->handle().callMethod<jboolean>("isBinderAlive")
                    << binder_->handle().callMethod<jboolean>("pingBinder");

        CoreApiResult res(reply.readVariant().toMap());
        if(res.code() != CoreApiConst::ErrorCode::UnknownError) return res;
        return CoreApiResult(CoreApiConst::ErrorCode::ServiceAccessError);
    }

   return CoreApiResult (reply.readVariant().toMap());
}
