#include "regdoc.h"
#include "formatutils.h"

#include <QJsonDocument>

BaseRegDoc::~BaseRegDoc()
{

}

const CoreRegData &BaseRegDoc::data() const
{
    return data_;
}

void BaseRegDoc::setData(const CoreRegData &newData)
{
    data_ = newData;
}

RegData BaseRegDoc::regData() const
{
    RegData res(data_);
    res.setDt(dt_);
    res.setDocNumber(fd_);
    res.setFsNumber(fsNumber_);
    res.setFiscalCode(fiscalCode_);
    return res;
}

void BaseRegDoc::clean()
{
    BaseDoc::clean();
    data_.clean();
}

QVariantMap BaseRegDoc::toMap() const
{
    QVariantMap res = BaseDoc::toMap();
    res = UNIT_MAPS(res, data_.toMap());
    return res;
}

QVariantMap BaseRegDoc::toExternalMap() const
{
    QVariantMap res = BaseDoc::toExternalMap();
    res = UNIT_MAPS(res, data_.toExternalMap());
    return res;
}

void BaseRegDoc::parseMap(const QVariantMap &map)
{
    BaseDoc::parseMap(map);
    data_.parseMap(map);
}

QVariantList BaseRegDoc::fieldsForSaving() const
{
    QVariantList  res;
    res << QVariantMap{{"flags", QString::number(static_cast<qint32>(data_.flags()), 16)}}
        << QVariantMap{{"extFlags",  QString::number(data_.extFlags().flags(), 16)}}
        << QVariantMap{{"taxes", QString::number(static_cast<qint32>(data_.taxes()), 16)}};
    if(!data_.ofdInn().isEmpty()) res << QVariantMap{{"ofdInn", data_.ofdInn()}};
    if(!data_.cashier().isEmpty()) res << QVariantMap{{"cashier", data_.cashier()}};
    if(!data_.boxId().isEmpty())res <<QVariantMap{{"boxId", data_.boxId()}};
    if(!data_.ofdName().isEmpty())res << QVariantMap{{"ofdName", data_.ofdName()}};
    if(!data_.ftsSite().isEmpty())res << QVariantMap{{"ftsSite", data_.ftsSite()}};
    if(!data_.email().isEmpty())res  << QVariantMap{{"email", data_.email()}};
    if(!data_.cashierInn().isEmpty())res<< QVariantMap{{"cashierInn", data_.cashierInn()}};
    if(!data_.additionalParam().isEmpty())res << QVariantMap{{"additionalParam", data_.additionalParam()}};
    if(!data_.additionalParamData().isEmpty())
    {
        res << QVariantMap{{"additionalParamData",
                            QString::fromLatin1(data_.additionalParamData().toHex())}};
    }
    if(type_ == fdf::DocType::ReRegistration)
    {
        res << QVariantMap{{"reason",  QString::number(data_.reasons().flags(), 16)}};
    }
    res << QVariantMap{{"registeredWith", data_.serial()}}
        << QVariantMap{{"fsFfd", QString::number(static_cast<qint32>(data_.fsFfd()))}}
        << QVariantMap{{"ffd", QString::number(static_cast<qint32>(data_.ffd()))}}
        << QVariantMap{{"cashboxFfd", QString::number(static_cast<qint32>(data_.cashboxFfd()))}}
        << QVariantMap{{"registeredWithVersion", data_.cashboxVersion()}}
        << QVariantMap{{"keysResource", QString::number(data_.keysResource())}};
    return res;
}

BaseRegDoc::BaseRegDoc(fdf::DocType docType, QObject *parent)
    : BaseDoc(docType, parent)
    , data_()
{

}

void BaseRegDoc::parseFromTlv(const Tlv &tlv)
{
    switch (tlv.tag()) {
    case fdf::Tag::AutoMode: data_.setAutoMode(tlv.toBool());break;
    case fdf::Tag::OfflineMode: data_.setOfflineMode(tlv.toBool());break;
    case fdf::Tag::Address: data_.setAddress(tlv.toString());break;
    case fdf::Tag::CashboxSerial: data_.setSerial(tlv.toString());break;
    case fdf::Tag::OfdInn: data_.setOfdInn(tlv.toString().trimmed());break;
    case fdf::Tag::OwnerInn: data_.setOwnerInn(tlv.toString().trimmed());break;
    case fdf::Tag::Cashier: data_.setCashier(tlv.toString().trimmed());break;
    case fdf::Tag::BoxId: data_.setBoxId(tlv.toString().trimmed());break;
    case fdf::Tag::CashboxRegNum: data_.setRegNumber(tlv.toString().trimmed());break;
    case fdf::Tag::OfdName: data_.setOfdName(tlv.toString().trimmed());break;
    case fdf::Tag::OwnerName: data_.setOwnerName(tlv.toString().trimmed());break;
    case fdf::Tag::EncryptionMode: data_.setEncriptionMode(tlv.toBool());break;
    case fdf::Tag::FnsSite: data_.setFtsSite(tlv.toString().trimmed());break;
    case fdf::Tag::TaxSystems: data_.setTaxes(fdf::Taxes(tlv.toByte()));break;
    case fdf::Tag::OwnerEmail: data_.setEmail(tlv.toString().trimmed());break;
    case fdf::Tag::Place: data_.setPlace(tlv.toString().trimmed());break;
    case fdf::Tag::CashboxVersion: data_.setCashboxVersion(tlv.toString().trimmed());break;
    case fdf::Tag::CashboxFFD: data_.setCashboxFfd(static_cast<fs::FFD>(tlv.toByte()));break;
    case fdf::Tag::FsFFD: data_.setFsFfd(static_cast<fs::FFD>(tlv.toByte()));break;
    case fdf::Tag::CashierInn: data_.setCashierInn(tlv.toString().trimmed());break;
    case fdf::Tag::FFD: data_.setFfd(static_cast<fs::FFD>(tlv.toByte()));break;
    case fdf::Tag::KeysResource: data_.setKeysResource(tlv.toUInt16());break;
    case fdf::Tag::RegAdditionalParam: data_.setAdditionalParam(tlv.toString().trimmed());break;
    case fdf::Tag::RegAdditionalData: data_.setAdditionalParamData(tlv.value());break;
    case fdf::Tag::CashboxUsageFlags: data_.setExtFlags(fdf::RegFlags1290(tlv.toUInt32()));break;
    case fdf::Tag::CashboxDataChangeReason: data_.setReasons(fdf::Reasons(tlv.toUInt32()));break;
    default: BaseDoc::parseFromTlv(tlv);
        break;
    }
}

void BaseRegDoc::setSavedField(const QString &key, const QString &value)
{
    if(key == "flags") data_.setFlags(fdf::RegFlags(value.toUInt(nullptr, 16)));
    else if(key == "extFlags") data_.setExtFlags(fdf::RegFlags1290(value.toUInt(nullptr, 16)));
    else if(key == "taxes") data_.setTaxes(fdf::Taxes(value.toUInt(nullptr, 16)));
    else if(key == "ofdInn") data_.setOfdInn(value);
    else if(key == "ofdName") data_.setOfdName(value);
    else if(key == "boxId") data_.setBoxId(value);
    else if(key == "cashier") data_.setCashier(value);
    else if(key == "cashierInn") data_.setCashierInn(value);
    else if(key == "ftsSite") data_.setFtsSite(value);
    else if(key == "email") data_.setEmail(value);
    else if(key == "additionalParam") data_.setAdditionalParam(value);
    else if(key == "additionalParamData") data_.setAdditionalParamData(QByteArray::fromHex(value.toLatin1()));
    else if(key == "reason") data_.setReasons(fdf::Reasons(value.toUInt(nullptr, 16)));
    else if(key == "registeredWith") data_.setSerial(value);
    else if(key == "fsFfd") data_.setFsFfd(static_cast<fs::FFD>(value.toUInt(nullptr, 16)));
    else if(key == "ffd") data_.setFfd(static_cast<fs::FFD>(value.toUInt(nullptr, 16)));
    else if(key == "cashboxFfd") data_.setCashboxFfd(static_cast<fs::FFD>(value.toUInt(nullptr, 16)));
    else if(key == "registeredWithVersion") data_.setCashboxVersion(value);
    else if(key == "keysResource") data_.setKeysResource(value.toUInt());
}

//--------------------------------------------------------------------------------------------------

RegDoc::RegDoc(QObject *parent)
    :BaseRegDoc(fdf::DocType::Registration, parent)
{

}

RegDoc::~RegDoc()
{

}
//--------------------------------------------------------------------------------------------------

ReRegDoc::ReRegDoc(QObject *parent)
    : BaseRegDoc(fdf::DocType::ReRegistration, parent)
    , fsCounters_()
{

}

ReRegDoc::~ReRegDoc()
{

}

const TaggedFsCounters &ReRegDoc::fsCounters() const
{
    return fsCounters_;
}

void ReRegDoc::setFsCounters(const TaggedFsCounters &newFsCounters)
{
    fsCounters_ = newFsCounters;
}

void ReRegDoc::clean()
{
    BaseRegDoc::clean();
    fsCounters_.clean();
}

QVariantMap ReRegDoc::toMap() const
{
    QVariantMap res = BaseRegDoc::toMap();
    res.insert("fsCounters", fsCounters_.toMap());
    return res;
}

QVariantMap ReRegDoc::toExternalMap() const
{
    QVariantMap res = BaseRegDoc::toExternalMap();
    res.insert("fsCounters", fsCounters_.toExternalMap());
    return res;
}

void ReRegDoc::parseMap(const QVariantMap &map)
{
    BaseRegDoc::parseMap(map);
    fsCounters_.parseMap(map["fsCounters"].toMap());
}

QVariantList ReRegDoc::fieldsForSaving() const
{
    QVariantList res = BaseRegDoc::fieldsForSaving();
    QJsonDocument jdoc = QJsonDocument::fromVariant(fsCounters_.toMap());
    res << QVariantMap{{"fsCounters", QString::fromUtf8(jdoc.toJson(QJsonDocument::Compact))}};
    return res;
}

void ReRegDoc::parseFromTlv(const Tlv &tlv)
{
    switch (tlv.tag())
    {
    case fdf::Tag::FsTotalsCounters: fsCounters_.parseTlv(tlv);break;

    default: BaseRegDoc::parseFromTlv(tlv);
    }
}

void ReRegDoc::setSavedField(const QString &key, const QString &value)
{
    QVariant buf;
    if(key == "fsCounters" && jsonToVariant(value, buf))
    {
        fsCounters_.parseMap(buf.toMap());
    }
    else BaseRegDoc::setSavedField(key, value);
}
