#include "bbcodeparser.h"

#include "bbcpputils.h"
#include "logmanip.h"

#include <QDebug>
#include <QRegularExpression>
#include <QImage>
#include <QDir>

const QMap<QString, DocumentBlock::EBarcodeType> BbCodeParser::BARCODE_TYPES = {
    {"AZTEC"             , DocumentBlock::AZTEC,           },
    {"CODABAR"           , DocumentBlock::CODABAR,         },
    {"CODE39"           , DocumentBlock::CODE_39,         },
    {"CODE93"           , DocumentBlock::CODE_93,         },
    {"CODE128"          , DocumentBlock::CODE_128,        },
    {"DATAMATRIX"       , DocumentBlock::DATA_MATRIX,     },
    {"EAN8"             , DocumentBlock::EAN_8,           },
    {"EAN13"            , DocumentBlock::EAN_13,          },
    {"ITF"               , DocumentBlock::ITF,             },
    {"MAXICODE"          , DocumentBlock::MAXICODE,        },
    {"PDF417"           , DocumentBlock::PDF_417,         },
    {"QRCODE"           , DocumentBlock::QR_CODE,         },
    {"RSS14"            , DocumentBlock::RSS_14,          },
    {"RSSEXPANDED"      , DocumentBlock::RSS_EXPANDED,    },
    {"UPCA"             , DocumentBlock::UPC_A,           },
    {"UPCE"             , DocumentBlock::UPC_E,           },
    {"UPCEANEXTENSION"  , DocumentBlock::UPC_EAN_EXTENSION}
};

BbCodeParser::BbCodeParser(QObject *parent) :
    QObject(parent),
    curField_(nullptr),
    curColumn_(nullptr),
    curHr_(nullptr),
    curImage_(nullptr),
    text_(),
    symbolsPerLine_(32),
    curFont_(0),
    curFormat_(0),
    curAlignment_(ALIGN_LEFT),
    blockType_(DocumentBlock::Text),
    barcodeType_(DocumentBlock::BARCODE_UNKNOWN),
    blocks_(),
    imagesPath_()

{

}

BbCodeParser::~BbCodeParser()
{
    if(curField_)
    {
        delete curField_;
        curField_ = nullptr;
    }
    if(curColumn_)
    {
        delete curColumn_;
        curColumn_ = nullptr;
    }
    if(curHr_)
    {
        delete curHr_;
        curHr_ = nullptr;
    }
    if(curImage_)
    {
        delete curImage_;
        curImage_ = nullptr;
    }
}

void BbCodeParser::parseText(QString text)
{
    test.clear();
//    QStringList sl = text.split("\n");
//    for(QString &s :sl)
//    {
//        if(s.isEmpty()) s = " ";
//        if(!s.trimmed().startsWith("["))
//        {
//            s = "[left]" + s + "[/left]";
//        }
//    }
//    text = sl.join("\n");
//    lmWarning() << loglist(text.split("\n"));
    QRegularExpression::PatternOptions opts;
    opts.setFlag(QRegularExpression::CaseInsensitiveOption);
    opts.setFlag(QRegularExpression::DotMatchesEverythingOption);
//	lmDebug() << loglist(text.split("\n")) << logtab << logbinary(text.toUtf8());
    blocks_.clear();
    auto doc = BBDocument::create();
    doc->load(text.toUtf8().toStdString());
    parseChildren(*doc);
    if(!text_.isEmpty())
    {
        blocks_.append(createCurrentBlock());
        blocks_.last().setType(blockType_);
        blocks_.last().setBarCodeType(barcodeType_);
        text_.clear();
        textFont_.clear();
        textFormat_.clear();
    }
//    else  blocks_ << createEmptyBlock(curAlignment_);
    curAlignment_ = ALIGN_LEFT;
//    lmWarning() << loglist(test);
}

void BbCodeParser::parseChildren(const BBNode &parent)
{
    for (const auto &node : parent.getChildren())
    {
        switch (node->getNodeType())
        {
            default:
                break;

            case BBNode::NodeType::ELEMENT:
            {
                const auto element = node->downCast<BBElementPtr>();
                switch(element->getElementType())
                {
                case BBElement::SIMPLE: openElement(element);break;
                case BBElement::VALUE:
                {
                    qWarning() << QByteArray::fromStdString(element->getNodeName());
                }break;
                case BBElement::PARAMETER: processParams(element);break;
                case BBElement::CLOSING: closeElement(element);
                }

            }
                break;

            case BBNode::NodeType::TEXT:
            {
                const auto textnode = node->downCast<BBTextPtr>();
                QString txt = QString::fromUtf8(QByteArray::fromStdString(textnode->getText()));
                txt = txt.replace("\r\n", "\n");
                txt = txt.replace("\r", "\n");

                //DEBUG

//                QStringList temp = txt.split("\n");
//                for(const QString &s:temp) test << ("    " + s);

                //DEBUG

                if(curField_)
                {
                    if(curField_->zerofill_ > 0) txt = QString("%1").arg(txt.toLongLong(), (qint32)curField_->zerofill_, 10, QLatin1Char('0') );
                    if(txt.size() < curField_->width_)
                    {
                        switch (curField_->alignment_) {
                        case ALIGN_CENTER:
                            txt = QString((curField_->width_ - txt.size()) / 2, QLatin1Char(' ')) + txt;
                            txt = txt + QString((curField_->width_ - txt.size()), QLatin1Char(' '));
                            break;
                        case ALIGN_RIGHT:
                            txt = QString((curField_->width_ - txt.size()), QLatin1Char(' ')) + txt;
                            break;
                        default:
                            txt = txt + QString((curField_->width_ - txt.size()), QLatin1Char(' '));
                            break;
                        }
                    }
                    if(!txt.endsWith("\n"))txt += "\t";
//                    lmWarning() << txt;
                }
                if(curColumn_)
                {
                    QStringList sl = txt.split("\n");
                    for(QString &s :sl)
                    {
                        if(s.size() < curColumn_->width_)
                        {
                            switch (curColumn_->alignment_) {
                            case ALIGN_CENTER:
                                s = QString((curColumn_->width_ - s.size()) / 2, QLatin1Char(' ')) + s;
                                s = s + QString((curColumn_->width_ - s.size()), QLatin1Char(' '));
                                break;
                            case ALIGN_RIGHT:
                                s = QString((curColumn_->width_ - s.size()), QLatin1Char(' ')) + s;
                                break;
                            default:
                                s = s + QString((curColumn_->width_ - s.size()), QLatin1Char(' '));
                                break;
                            }
                        }
                        if(!s.endsWith("\n"))s += "\t";
                    }
                    txt = sl.join("\n");
                }
                //NOTE удаляю возможные лишние переводы строк после [center],[left],[right]
                if(text_.isEmpty())
                {
//                    lmWarning() << txt << text_;
                    if(txt.startsWith("\n")) txt = txt.mid(1);
                    else if(txt.startsWith(" \n") || txt.startsWith("\t\n")) txt = txt.mid(2);
                    else if(txt.startsWith("  \n")) txt = txt.mid(3);
//                    lmWarning() << txt << text_;

                }
                if(!text_.isEmpty() && !text_.at(text_.size() - 1).isSpace() && !txt.isEmpty() && !txt[0].isSpace()) txt = " " + txt;
                if(!txt.isEmpty())
                {
                    text_ += txt;
                    textFont_.append(QByteArray(txt.size(), static_cast<char>(curFont_)));
                    textFormat_.append(QByteArray(txt.size(), static_cast<char>(curFormat_)));
                }
            }break;
        }

        parseChildren(*node);
    }

}

QList<DocumentBlock> BbCodeParser::blocks() const
{
    return blocks_;
}

TextPrinterDocument BbCodeParser::document() const
{
    TextPrinterDocument doc;

    //NOTE: костыль для выравнивания. возможно временно. объединяю блоки с одинаковым выравниванием
    DocumentBlocks blocks;
    if(!blocks_.isEmpty())
    {
        blocks << blocks_[0];
        for(int i = 1; i < blocks_.size();++i)
        {
            if(blocks.last().type() != DocumentBlock::Text ||
                    blocks_[i].type() != DocumentBlock::Text ||
                    blocks_[i].format().format() != blocks.last().format().format())
            {
                blocks << blocks_[i];
            }
            else
            {
                if(!blocks.last().append(blocks_[i]))
                {
                    blocks << blocks_[i];
                }

            }
        }
    }

    doc.setBlocks(blocks);
//    lmWarning() << logvariant(doc.toMap());
    return doc;


}

quint8 BbCodeParser::symbolsPerLine() const
{
    return symbolsPerLine_;
}

void BbCodeParser::setSymbolsPerLine(const quint8 &symbolsPerLine)
{
    symbolsPerLine_ = symbolsPerLine;
}

QString BbCodeParser::imagesPath() const
{
    return imagesPath_;
}

void BbCodeParser::setImagesPath(const QString &imagesPath)
{
    imagesPath_ = imagesPath;
}

void BbCodeParser::openElement(BBElementPtr elem)
{
    QString node = QString::fromUtf8(QByteArray::fromStdString(elem->getNodeName())).trimmed().toLower();
    test << "<" + node + ">";
    if(node == "center")
    {
        processCurentBlock();
        curAlignment_ = ALIGN_CENTER;
    }
    else if(node == "left")
    {
        processCurentBlock();
        curAlignment_ = ALIGN_LEFT;
    }
    else if(node == "right")
    {
        processCurentBlock();
        curAlignment_ = ALIGN_RIGHT;
    }
    else if(node == "b")
    {
        curFormat_ |= CF_BOLD;
    }
    else if(node == "i")
    {
        curFormat_ |= CF_ITALIC;
    }
    else if(node == "u")
    {
        curFormat_ |= CF_UNDERLINE;
    }
    else if(node == "hr")
    {
        if(curHr_) delete curHr_;
        curHr_ = new Hr;
    }
    else if(node == "img")
    {
        processCurentBlock();
        if(curImage_) delete curImage_;
        curImage_ = new MonochromeImage();
    }
    else if(node == "column")
    {
        processCurentBlock();
        if(curColumn_) delete curColumn_;
        curColumn_ = new Column;
    }
    else if(node == "barcode")
    {
        processCurentBlock();
        blockType_ = DocumentBlock::BarCode;
        barcodeType_ = DocumentBlock::CODE_128;
    }
    else if(node == "qrcode")
    {
        processCurentBlock();
        blockType_ = DocumentBlock::QrCode;
        barcodeType_ = DocumentBlock::BARCODE_UNKNOWN;
    }

    else if(node == "field")
    {
        if(curField_) delete curField_;
        curField_ = new Field;
    }
    else
    {
        qWarning() << node;
    }
}

void BbCodeParser::processParams(BBElementPtr elem)
{
    QString node = QString::fromUtf8(QByteArray::fromStdString(elem->getNodeName())).trimmed().toLower();
    for(auto const &x :elem->getParameters())
    {
        QString name = QString::fromUtf8(QByteArray::fromStdString(x.first)).trimmed().toLower();
        QString value= QString::fromUtf8(QByteArray::fromStdString(x.second)).trimmed().toLower();

        if(test.isEmpty() || !test.last().startsWith("<") || test.last().startsWith("</") ||
            !test.last().contains(node) || !test.last().endsWith(">"))
        {
            test << QStringLiteral("<!--ОШИБКА! %1=%2 (%3)-->")
                        .arg(name).arg(value).arg(node);
        }
        else
        {
            test.last().insert(test.last().indexOf(">"),
                               QStringLiteral(" %1=\"%2\"").arg(name).arg(value));
        }


        if(name == "size")
        {
            curFormat_ = curFormat_ & ~(CF_DOUBLE_WIDTH | CF_DOUBLE_HEIGHT);
            if(value == "small") curFont_ = 0;
            else if(value == "normal") curFont_ = 1;
            else if(value == "middle")
            {
                curFont_ = 2;
            }
            else if(value == "large")
            {
                curFont_ = 3;
                curFormat_ = curFormat_ | CF_DOUBLE_WIDTH | CF_DOUBLE_HEIGHT;
            }
        }
        else if(name == "hr")
        {
            if(!curHr_) curHr_ = new Hr;

            if(value == "dashed") curHr_->symbol_ = QLatin1Char('-');
            else if(value == "dotted") curHr_->symbol_ = QLatin1Char('.');
            else curHr_->symbol_ = QLatin1Char('_');
        }
        else if(name == "width")
        {
            if(curField_) curField_->width_ = value.toUInt();
            else if(curColumn_) curColumn_->width_= value.toUInt();
        }
        else if(name == "align")
        {
            if(curField_)
            {
                if(value == "center") curField_->alignment_ = ALIGN_CENTER;
                else if(value == "right")curField_->alignment_ = ALIGN_RIGHT;
                else curField_->alignment_ = ALIGN_LEFT;
            }
            else if(curColumn_)
            {
                if(value == "center") curColumn_->alignment_ = ALIGN_CENTER;
                else if(value == "right")curColumn_->alignment_ = ALIGN_RIGHT;
                else curColumn_->alignment_ = ALIGN_LEFT;
            }
        }
        else if(name == "zerofill")
        {
            if(curField_) curField_->zerofill_ = value.toUInt();
        }
        else if(name == "name")
        {
            if(curImage_)loadImage(value.trimmed());
        }
        else if(name == "bctype")
        {
            if(BARCODE_TYPES.contains(value.trimmed().toUpper()))
            {
                barcodeType_ = BARCODE_TYPES[value.trimmed().toUpper()];
            }
            else
            {
                barcodeType_ = DocumentBlock::EBarcodeType::BARCODE_UNKNOWN;
            }
        }
        else
        {
            qWarning() << name << value;
        }
    }
}

void BbCodeParser::closeElement(BBElementPtr elem)
{
    QString node = QString::fromUtf8(QByteArray::fromStdString(elem->getNodeName())).trimmed().toLower();
    test << "</" + node + ">";
    if(node == "center")
    {
        processCurentBlock();
        curAlignment_ = ALIGN_LEFT;
    }
    else if(node == "left")
    {
        processCurentBlock();
        curAlignment_ = ALIGN_LEFT;
    }
    else if(node == "right")
    {
        processCurentBlock();
        curAlignment_ = ALIGN_LEFT;
    }
    else if(node == "b")
    {
        curFormat_ &= ~CF_BOLD;
    }
    else if(node == "i")
    {
        curFormat_ &= ~CF_ITALIC;
    }
    else if(node == "u")
    {
        curFormat_ &= ~CF_UNDERLINE;
    }
    else if(node == "size")
    {
        curFont_ = 0;
        curFormat_ = curFormat_ & ~(CF_DOUBLE_WIDTH | CF_DOUBLE_HEIGHT);
    }
    else if(node == "img")
    {
        if(curImage_)
        {
            if(curImage_->width() <= 0 || curImage_->height() <=0)
            {
                blocks_.append(createCurrentBlock());
                blocks_.last().setType(blockType_);
                blocks_.last().setBarCodeType(barcodeType_);
                text_.clear();
                textFont_.clear();
                textFormat_.clear();
            }
            else
            {
                LineFormat f;
                f.setFlags(LF_CENTER);
                DocumentBlock block;
                block.setImg(*curImage_);
                block.setType(DocumentBlock::Image);
                block.setFormat(f);
                blocks_ << block;
//				lmWarning() << logvariant(blocks_.last().toMap());
                text_.clear();
                textFormat_.clear();
                textFont_.clear();
            }
            delete  curImage_;
            curImage_ = nullptr;
        }
    }
    else if(node == "column")
    {
        processCurentBlock();
        delete curColumn_;
        curColumn_ = nullptr;
    }
    else if(node == "barcode")
    {
        processCurentBlock(true);
        blockType_ = DocumentBlock::Text;
        barcodeType_ = DocumentBlock::BARCODE_UNKNOWN;
    }
    else if(node == "qrcode")
    {
        processCurentBlock(true);
        blockType_ = DocumentBlock::Text;
        barcodeType_ = DocumentBlock::BARCODE_UNKNOWN;
    }
    else if(node == "field")
    {
        if(curField_)
        {
            delete curField_;
            curField_ = nullptr;
        }
    }
    else if(node == "hr")
    {
        if(!curHr_) curHr_ = new Hr;
        if(curHr_)
        {
            QStringList sl;
            sl << QString(symbolsPerLine_, curHr_->symbol_);
            LineFormat f;
            f.setPrintGray(5000);
            f.setFlags(LF_CENTER);
            blocks_ << DocumentBlock::createSimpleBlock(sl, f);
            delete curHr_;
            curHr_ = nullptr;
        }
    }
    else
    {
        qWarning() << node;
    }
}

void BbCodeParser::loadImage(const QString &name)
{
//	lmWarning() << name;
    if(!curImage_) return;
    QImage img;

    QDir dir;
    if(!imagesPath_.isEmpty())
    {
        dir = QDir(imagesPath_);
    }
    else
    {
        dir = QDir::current();
    }


    QFileInfoList files = dir.entryInfoList(QStringList() << (name+"*"), QDir::Files, QDir::Name);
//	lmWarning() << imagesPath_ << logtab <<
//				   (files.isEmpty() ? QStringLiteral("EMPTY") : files.front().absoluteFilePath());

    if(curImage_  && !files.isEmpty() &&img.load(files.front().absoluteFilePath()))
    {
        img = img.convertToFormat(QImage::Format_Mono);
        if(img.height() > 384) img = img.scaledToWidth(384, Qt::SmoothTransformation);
        *curImage_ = MonochromeImage::createEmpty(img.width(), img.height(), 32);
        for(int x = 0; x < img.width(); ++x)
        {
            for(int y = 0; y < img.height(); ++y)
            {
                curImage_->setPixel(x, y, (img.pixel(x, y) &0xFFFFFF) == 0, 32);
            }
        }
        curImage_->setFileName(files.front().absoluteFilePath());
    }
}

DocumentBlocks BbCodeParser::createCurrentBlock() const
{
    static const QSet<DocumentBlock::EBlockType> SUPPORTED {
        DocumentBlock::Text,
        DocumentBlock::BarCode,
        DocumentBlock::QrCode
    };
    if(text_.isEmpty() || !SUPPORTED.contains(blockType_))
    {
//        lmWarning() << text_ << blockType_;
        return DocumentBlocks() << DocumentBlock::createEmpty();
    }
    QByteArrayList fonts, formats;
    QByteArray tfont = textFont_, tformat = textFormat_;
    QString str = text_;
    QStringList text;
//    lmWarning() << logbinary(textFont_) << logtab << logbinary(textFormat_);

    while(!str.isEmpty())
    {
        qint32 idx = str.indexOf("\n");
        if(idx < 0)
        {
            text << str;
            fonts << (tfont.isEmpty() ? QByteArray(str.size(), '\x00') : tfont);
            formats << (tfont.isEmpty() ? QByteArray(str.size(), '\x00') : tformat);
            str.clear();
        }
        else
        {
            QString s = str.mid(0, idx);
            str = str.mid(idx + 1);
            text << s;
            QByteArray fo, fa;
            if(tfont.size() >= idx)
            {
                fo = tfont.mid(0, idx);
                tfont = tfont.mid(idx + 1);
            }
            else
            {
                fo = tfont;
                tfont.clear();
            }
            if(tformat.size() >= idx)
            {
                fa = tformat.mid(0, idx);
                tformat = tformat.mid(idx + 1);
            }
            else
            {
                fa = tformat;
                tformat.clear();
            }
            fonts << (fo.isEmpty() ? QByteArray(s.size(), '\x00') : fo);
            formats << (fa.isEmpty() ? QByteArray(s.size(), '\x00') : fa);
        }
    }

    auto formatByLine = [text, fonts, formats, this](qint32 line) ->LineFormat {
        if(line < 0 || line >= text.size()) return LineFormat();
        LineFormat res;
        res.setPrintGray(5000);
        QByteArray fts = (line >=0 && line < fonts.size()) ? fonts[line] : QByteArray(1, '\x00');
        if(fts.isEmpty()) fts = QByteArray(1, '\x00');
        res.setFont(fonts.size() > line ? fts[0] : 0);
        qint32 format = 0;
        switch (curAlignment_) {
        case ALIGN_CENTER : format = format | LF_CENTER;break;
        case ALIGN_RIGHT  : format = format | LF_RIGHT;break;
        default: format = format | LF_LEFT;break;
        }
        QByteArray fms = (line >=0 && line < formats.size()) ? formats[line] : QByteArray(1, '\x00');
        if(fms.isEmpty()) fms = QByteArray(1, '\x00');

        if(fms[0] & CF_BOLD) format = format | LF_BOLD;
        if(fms[0] & CF_ITALIC) format = format | LF_ITALIC;
        if(fms[0] & CF_INVERSE) format = format | LF_INVERT;
        res.setFlags(format);
        return res;
    };

    DocumentBlocks blocks;
    LineFormat prev, cur;
    QStringList txt;
    QString line = text.front();
    if(line.endsWith("\t")) line = line.mid(0, line.size() - 1);
    else if (line.endsWith("\t\n"))line = line.mid(0, line.size() - 2) + "\n";
    txt << line;

    prev = formatByLine(0);
//    if(text_.contains("ID:"))
//    {
//        lmWarning() << loglist(text) << loglist(txt);
//    }
    for(int i = 1; i < text.size(); ++i)
    {
        cur = formatByLine(i);
        if(cur.format() != prev.format())
        {
            blocks << DocumentBlock::createSimpleBlock(txt, prev);
            txt.clear();
            prev = cur;
            line = text[i];
            if(line.endsWith("\t")) line = line.mid(0, line.size() - 1);
            else if (line.endsWith("\t\n"))line = line.mid(0, line.size() - 2) + "\n";
            txt << line;
        }
    }
    if(!txt.isEmpty())
    {
        blocks << DocumentBlock::createSimpleBlock(txt, prev);
    }
//    if(text_.contains("ID:"))
//    {
//        lmWarning() << loglist(text) << logtab << blocks.size() << loglist(txt);
//    }

    if(blocks.isEmpty()) blocks << DocumentBlock::createEmpty();
    return blocks;
    //    DocumentBlock::createSimpleBlock(text_, curAlignment_, textFont_, textFormat_, 0, symbolsPerLine_);
}

void BbCodeParser::processCurentBlock(bool addEmpty)
{
    if(!text_.isEmpty())
    {

        blocks_.append(createCurrentBlock());
        if(barcodeType_ == DocumentBlock::BARCODE_UNKNOWN && blockType_ == DocumentBlock::BarCode)
        {
            barcodeType_ = DocumentBlock::CODE_128;
        }
        blocks_.last().setBarCodeType(barcodeType_);
        blocks_.last().setType(blockType_);
        if(blockType_ == DocumentBlock::BarCode)blocks_.last().setBarCode(text_.toUtf8());
        else if(blockType_ == DocumentBlock::QrCode)blocks_.last().setQrCode(text_.toUtf8());
        text_.clear();
        textFont_.clear();
        textFormat_.clear();
        if(addEmpty)blocks_ << DocumentBlock::createEmpty();
    }
}
