#include "monochromeimage.h"


MonochromeImage::MonochromeImage() noexcept
    : width_(0)
    , height_(0)
    , images_()
    , fileName_()
{

}

MonochromeImage::MonochromeImage(const QVariantMap &data) noexcept
    : width_(0)
    , height_(0)
    , images_()
    , fileName_()
{
    setMap(data);
}
MonochromeImage::MonochromeImage(const MonochromeImage &other)noexcept
    : width_(other.width_)
    , height_(other.height_)
    , images_(other.images_)
    , fileName_(other.fileName_)
{

}

MonochromeImage::MonochromeImage(MonochromeImage &&other) noexcept
    : width_(other.width_)
    , height_(other.height_)
    , images_()
    , fileName_()
{
    images_.swap(other.images_);
    fileName_.swap(other.fileName_);
}
MonochromeImage::~MonochromeImage()
{

}

bool MonochromeImage::isValid() const
{
    return width_ > 0 && height_ > 0 && images_.size() >= (width_ * height_) / 8;
}

int MonochromeImage::width() const
{
    return width_;
}
void MonochromeImage::setWidth(int w)
{
    width_ = w;
}
int MonochromeImage::height() const
{
    return height_;
}
void MonochromeImage::setHeight(int h)
{
    height_ = h;
}
QByteArray MonochromeImage::image() const
{
    return images_;
}
void MonochromeImage::setImage(const QByteArray &im)
{
    images_ = im;
}

QString MonochromeImage::fileName() const
{
    return fileName_;
}

void MonochromeImage::setFileName(const QString &fileName)
{
    fileName_ = fileName;
}
bool MonochromeImage::pixel(qint32 x, qint32 y, qint32 rawword) const
{
    if(x < 0 || x >= width_ || y < 0 || y >= height_) return false;
    qint8 rawbit = x % 8;
    qint32 idx = x / 8;
    qint32 ww = width_ / rawword + ((width_% rawword) ? 1 : 0);
    ww *= (rawword / 8);
    idx = y * ww + idx;
    if(idx < 0 || idx > images_.size()) return false;
    return static_cast<quint8>(images_[idx]) & (0x80 >> rawbit);
}

bool MonochromeImage::setPixel(qint32 x, qint32 y, bool black, qint32 rawword)
{
    if(x < 0 || x >= width_ || y < 0 || y >= height_) return false;
    qint8 rawbit = x % 8;
    qint32 idx = x / 8;
    qint32 ww = width_ / rawword + ((width_% rawword) ? 1 : 0);
    ww *= (rawword / 8);
    idx = y * ww + idx;
    if(idx < 0 || idx > images_.size()) return false;
    if(black)
    {
        images_[idx] = images_[idx] | (0x80 >> rawbit);
    }
    else
    {
        images_[idx] = images_[idx] & (~(0x80 >> rawbit));
    }
    return static_cast<quint8>(images_[idx]) & (0x80 >> rawbit);
}

QByteArray MonochromeImage::raw(int y, qint32 rawword) const
{
    if(y < 0 || y >= height_) return QByteArray();
    qint32 wbytes = width_ / rawword + ((width_% rawword) ? 1 : 0);
    wbytes *= (rawword / 8);
    qint32 idx = y * wbytes;
    return images_.mid(idx, wbytes);
}

QVariantMap MonochromeImage::toMap() const
{
    QVariantMap res;
    res.insert("h", height_);
    res.insert("w", width_);
    res.insert("data", QString::fromLatin1(images_.toBase64()));
    if(!fileName_.trimmed().isEmpty()) res.insert("fn", fileName_);
    return res;
}

void MonochromeImage::setMap(const QVariantMap &data)
{
    width_ = data["w"].toInt();
    height_ = data["h"].toInt();
    images_ = QByteArray::fromBase64(data["data"].toString().toLatin1());
    fileName_.clear();
    if(data.contains("fn")) fileName_ = data["fn"].toString();
}

MonochromeImage &MonochromeImage::operator = (const MonochromeImage &other) noexcept
{
    width_ = other.width_;
    height_ = other.height_;
    images_ = other.images_;
    fileName_ = other.fileName_;
    return *this;
}

MonochromeImage &MonochromeImage::operator =(MonochromeImage &&other)noexcept
{
    width_ = other.width_;
    height_ = other.height_;
    images_.swap(other.images_);
    fileName_.swap(other.fileName_);
    return *this;
}

bool MonochromeImage::operator == (const MonochromeImage &other) const noexcept
{
    return width_ == other.width_ &&
            height_ == other.height_ &&
            images_ == other.images_ &&
            fileName_ == other.fileName_;;
}
bool MonochromeImage::operator != (const MonochromeImage &other) const noexcept
{
    return !(*this == other);
}

MonochromeImage MonochromeImage::createEmpty(qint32 w, qint32 h, qint32 rawword)
{
    if(rawword <= 0) rawword = 8;
    MonochromeImage res;
    res.setHeight(h);
    res.setWidth(w);
    qint32 wbytes = w / rawword + ((w% rawword) ? 1 : 0);
    wbytes *= (rawword / 8);
    qint32 sz = wbytes * h;
    res.setImage(QByteArray(sz, '\x00'));
    return res;
}


