#include "textprinterdocument.h"

#include "logmanip.h"

#include <QDir>
#include <QDebug>
#include <QDateTime>
#include <QUrl>
#include <QFile>
#include <QProcess>
#include <QDir>
#include <QFileInfo>
#include <QJsonParseError>
#include <QJsonDocument>
#include <QCryptographicHash>


DocumentBlock::DocumentBlock() noexcept
    : type_(Text)
    , barCodeType_(BARCODE_UNKNOWN)
    , text_()
    , img_()
    , format_()
{
}

DocumentBlock::DocumentBlock(const QVariantMap &data) noexcept
    : type_(Text)
    , barCodeType_(BARCODE_UNKNOWN)
    , text_()
    , img_()
    , format_()
{
    setMap(data);
}

DocumentBlock::DocumentBlock(const DocumentBlock &other)noexcept
    : type_(other.type_)
    , barCodeType_(other.barCodeType_)
    , text_(other.text_)
    , img_(other.img_)
    , format_(other.format_)
{
}

DocumentBlock::DocumentBlock(DocumentBlock &&other)noexcept
    : type_(other.type_)
    , barCodeType_(other.barCodeType_)
    , text_()
    , img_(other.img_)
    , format_(other.format_)
{
    text_.swap(other.text_);
}

DocumentBlock::~DocumentBlock()
{
}

bool DocumentBlock::isImage() const
{
    return type_ == Image && img_.isValid();
}

bool DocumentBlock::isText() const
{
    return type_ == Text && !text_.isEmpty();
}

bool DocumentBlock::isBarcode() const
{
    return type_ == BarCode && !text_.isEmpty() &&
            barCodeType_ != BARCODE_UNKNOWN;
}

bool DocumentBlock::isQrCode() const
{
    return type_ == QrCode && !text_.isEmpty();
}

bool DocumentBlock::isValid() const
{
    return isImage() || isText();
}

QByteArray DocumentBlock::barcode() const
{
    return text_.isEmpty() ?
                QByteArray() :
                QByteArray::fromBase64(text_.front().toLatin1());
}

void DocumentBlock::setBarCode(const QByteArray &data)
{
    text_.clear();
    text_ << QString::fromLatin1(data.toBase64());
    type_ = BarCode;
}

void DocumentBlock::setBarCode(const QByteArray &data, EBarcodeType type)
{
    text_.clear();
    text_ << QString::fromLatin1(data.toBase64());
    type_ = BarCode;
    barCodeType_ = type;
}

QByteArray DocumentBlock::qrCode() const
{
    return text_.isEmpty() ?
                QByteArray() :
                QByteArray::fromBase64(text_.front().toLatin1());
}

void DocumentBlock::setQrCode(const QByteArray &data)
{
    text_.clear();
    text_ << QString::fromLatin1(data.toBase64());
    type_ = QrCode;
}

DocumentBlock::EBlockType DocumentBlock::type() const
{
    return type_;
}

void DocumentBlock::setType(const EBlockType &type)
{
    type_ = type;
}

DocumentBlock::EBarcodeType DocumentBlock::barCodeType() const
{
    return barCodeType_;
}

QStringList DocumentBlock::text() const
{
    return text_;
}

void DocumentBlock::setText(const QStringList &text)
{
    text_ = text;
}

MonochromeImage DocumentBlock::img() const
{
    return img_;
}

void DocumentBlock::setImg(const MonochromeImage &img)
{
    img_ = img;
}

LineFormat DocumentBlock::format() const
{
    return format_;
}

void DocumentBlock::setFormat(const LineFormat &format)
{
    format_ = format;
}

QVariantMap DocumentBlock::toMap() const
{
    QVariantMap res;
    res.insert("type", type_);
    res.insert("format", QString::number(format_.format()));
    if(isImage())
    {
        res.insert("img", img_.toMap());
    }
    else if (isText())
    {
        res.insert("txt", text_.join("\n"));
    }
    else if(isBarcode())
    {
        res.insert("bar", text_.front());
        res.insert("bt", barCodeType_);
    }
    else if(isQrCode())
    {
        res.insert("qr", text_.front());
    }
    return res;
}

void DocumentBlock::setMap(const QVariantMap &data)
{
    type_ = static_cast<EBlockType>(data["type"].toInt());
    barCodeType_ = BARCODE_UNKNOWN;
    text_.clear();
    format_.setFormat(data["format"].toInt());
    img_ = MonochromeImage();
    switch (type_) {
    case Text:
    {
        text_ = data["txt"].toString().split("\n");
    }break;
    case Image:
        img_.setMap(data["img"].toMap());
        break;
    case BarCode:
    {
        barCodeType_ = static_cast<EBarcodeType>(data["bt"].toInt());
        text_ << data["bar"].toString();
    }break;
    case QrCode:
        text_ << data["qr"].toString();
        break;
    }
}

DocumentBlock &DocumentBlock::operator =(const DocumentBlock &other) noexcept
{
    type_ = other.type_;
    barCodeType_ = other.barCodeType_;
    text_ = other.text_;
    format_ = other.format_;
    img_ = other.img_;
    return *this;
}

DocumentBlock &DocumentBlock::operator=(DocumentBlock &&other)noexcept
{
    type_ = other.type_;
    barCodeType_ = other.barCodeType_;
    text_.swap(other.text_);
    format_ = other.format_;
    img_ = other.img_;
    return *this;
}

bool DocumentBlock::operator ==(const DocumentBlock &other) const noexcept
{
    return 	type_ == other.type_ &&
            barCodeType_ == other.barCodeType_ &&
            text_ == other.text_ &&
            format_ == other.format_ &&
            img_ == other.img_;
}

bool DocumentBlock::operator !=(const DocumentBlock &other) const noexcept
{
    return !(*this == other);
}

bool DocumentBlock::append(const DocumentBlock &other)
{
    bool res = isText() && other.isText() && format_ == other.format_;
    if(res)
    {
        text_.append(other.text_);
    }
    return res;
}


void DocumentBlock::toUpper()
{
    for(QString &s :text_) s = s.toUpper();
}

void DocumentBlock::toLower()
{
    for(QString &s :text_) s = s.toLower();
}

DocumentBlock DocumentBlock::emptyBlock(const LineFormat &format)
{
    DocumentBlock res;
    res.format_ = format;
    res.type_ = Text;
    return res;
}

DocumentBlock DocumentBlock::createSimpleBlock(const QStringList &txt, const LineFormat &format)
{
    DocumentBlock block;
    block.setText(txt);
    block.format_ = format;
    block.setType(DocumentBlock::Text);
    return block;
}

DocumentBlock DocumentBlock::createSimpleBlock(const QString &txt, const LineFormat &format)
{
    return createSimpleBlock(QStringList() << txt, format);
}

DocumentBlock DocumentBlock::createEmpty()
{
    return createSimpleBlock(QString(), LineFormat());
}

qint32 DocumentBlock::symbolsCount() const
{
    qint32 res = 0;
    if(type_ != Text || text_.isEmpty()) return res;
    res = std::accumulate(text_.constBegin(), text_.constEnd(), 0,
                          [](qint32 sum, const QString &s) -> qint32 {
        return sum + s.size();});
    return res;
}




void DocumentBlock::setBarCodeType(const EBarcodeType &barCodeType)
{
    barCodeType_ = barCodeType;
    if(barCodeType != BARCODE_UNKNOWN)
    {
        type_ = BarCode;
    }
}


//--------------------------------------------------------------------------------------------------

TextPrinterDocument::TextPrinterDocument() noexcept
    : blocks_()

{
}
TextPrinterDocument::TextPrinterDocument(const DocumentBlocks &blocks) noexcept
    : blocks_(blocks)
{
    simplify();
}
TextPrinterDocument::TextPrinterDocument(const TextPrinterDocument &other) noexcept
    : blocks_(other.blocks_)
{
}

TextPrinterDocument::TextPrinterDocument(TextPrinterDocument &&other) noexcept
    : blocks_(QList<DocumentBlock>())
{
    blocks_.swap(other.blocks_);
}

TextPrinterDocument::~TextPrinterDocument()
{
}


QList<DocumentBlock> TextPrinterDocument::blocks() const
{
    return blocks_;
}

void TextPrinterDocument::setBlocks(const QList<DocumentBlock> &blocks)
{
    blocks_ = blocks;
}

void TextPrinterDocument::appendBlock(const  DocumentBlock &block)
{
    if(blocks_.isEmpty() && !blocks_.back().append(block))
    {
        blocks_.push_back(block);
    }
}

void TextPrinterDocument::prependBlock(const DocumentBlock &block)
{
    DocumentBlock bl = block;
    if(blocks_.isEmpty() || !bl.append(blocks_.first())) blocks_.push_front(block);
    else blocks_[0] = bl;
}

void TextPrinterDocument::toUpper()
{
    std::for_each(blocks_.begin(), blocks_.end(), [](DocumentBlock &block) -> void{block.toUpper();});
}

void TextPrinterDocument::toLower()
{
    std::for_each(blocks_.begin(), blocks_.end(), [](DocumentBlock &block) -> void{block.toLower();});
}

QStringList TextPrinterDocument::text() const
{
    QStringList list;
    for(const DocumentBlock &b: blocks_) if(b.isText()) list.append(b.text());
    return list;
}

QVariantMap TextPrinterDocument::toMap() const
{
    QVariantMap res;
    QVariantList blocks;
    std::for_each(blocks_.begin(), blocks_.end(),
                  [&blocks](const DocumentBlock &block) -> void{blocks << block.toMap();});
    res.insert("blocks", blocks);
    return res;
}

void TextPrinterDocument::setMap(const QVariantMap &data)
{
    blocks_.clear();
    QVariantList blocks = data["blocks"].toList();
    std::for_each(blocks.begin(), blocks.end(),
                  [this](const QVariant &b) -> void{blocks_ << DocumentBlock(b.toMap());});
}

qint32 TextPrinterDocument::symbolsCount() const
{
    qint32 res = 0;
    if(blocks_.isEmpty()) return res;
    res = std::accumulate(blocks_.begin(), blocks_.end(), 0,
                          [](qint32 sum, const DocumentBlock &b) -> qint32 {
        return sum + b.symbolsCount();
    });
    return res * 1.2;
}

TextPrinterDocument &TextPrinterDocument::operator =(const TextPrinterDocument &other) noexcept
{
    blocks_ = other.blocks_;
    return *this;
}

TextPrinterDocument &TextPrinterDocument::operator=(TextPrinterDocument &&other)noexcept
{
    blocks_.swap(other.blocks_);
    return *this;
}

bool TextPrinterDocument::operator ==(const TextPrinterDocument &other) const noexcept
{
    return 	blocks_ == other.blocks_ ;
}

bool TextPrinterDocument::operator !=(const TextPrinterDocument &other) const noexcept
{
    return !(*this == other);
}

bool TextPrinterDocument::saveToFile(const QString &fileName)
{
    if(fileName.isEmpty()) return false;

    auto saveSingleDoc = [this](QString fileName) -> bool {
        QFileInfo fi(fileName);
        QFile f(fileName);
        if(!f.open(QIODevice::WriteOnly))
        {
            lmCritical() << fileName << f.errorString();
            return false;
        }
        QJsonDocument jdoc = QJsonDocument::fromVariant(toMap());
        QByteArray data = jdoc.toJson(QJsonDocument::Compact);
        f.write(data);
        f.flush();
        f.close();
        QFile csfile(fileName + QStringLiteral(".sha1"));
        if(!csfile.open(QIODevice::WriteOnly))
        {
            lmCritical() << fileName << f.errorString();
            return false;
        }
        csfile.write(QCryptographicHash::hash(data, QCryptographicHash::Sha1).toHex());
        csfile.close();
        system("sync");
        return true;
    };

    bool res = saveSingleDoc(fileName);
    res = saveSingleDoc(fileName + QStringLiteral(".bak")) && res;

    return res;
}

bool TextPrinterDocument::loadFromFile(const QString &fileName)
{
    if(fileName.isEmpty()) return false;
    auto checkFile = [](const QString &fileName, QByteArray &data) -> bool {
        QFileInfo fi(fileName);
        if(!fi.exists())
        {
            lmWarning() << fi.absoluteFilePath() << " does not exists";
            return false;
        }

        QFile fsha (fileName + ".sha1");
        if(!fsha.open(QIODevice::ReadOnly)) return false;
        QByteArray sha2 = QByteArray::fromHex(fsha.read(40));
        fsha.close();
        QFile f(fileName);
        if(!f.open(QIODevice::ReadOnly)) return false;
        data = f.readAll();
        QByteArray sha1 = QCryptographicHash::hash(data, QCryptographicHash::Sha1);
        f.close();
//        lmDebug() << logbinary(sha2) << logbinary(sha1);
        return sha2 == sha1;
    };

    QByteArray data;
    QString fn = fileName;
    if(!checkFile(fn, data)) fn = fileName + ".bak";
    if(!checkFile(fn, data)) return false;

    QJsonParseError jerr;
    QJsonDocument jdoc = QJsonDocument::fromJson(data, &jerr);
    setMap(jdoc.toVariant().toMap());
    if(jerr.error != QJsonParseError::NoError)
    {
        lmWarning() << jerr.error << jerr.errorString();
    }
    return jerr.error == QJsonParseError::NoError;
}

void TextPrinterDocument::removeFile(const QString &fileName)
{
    QFile::remove(fileName);
    QFile::remove(fileName + ".sha1");
    QFile::remove(fileName + ".bak");
    QFile::remove(fileName + ".bak.sha1");
    system("sync");
}

void TextPrinterDocument::simplify()
{
    QList<DocumentBlock> blocks = blocks_;
    blocks_.clear();
    for(DocumentBlock &b: blocks)
    {
        if(blocks_.isEmpty() || !blocks_.last().append(b)) blocks_ << b;
    }
}
