#ifndef TEXTPRINTERDOCUMENT_H
#define TEXTPRINTERDOCUMENT_H

#include "monochromeimage.h"
#include "lineformat.h"
#include <QtGlobal>
#include <QString>
#include <QByteArray>
#include <QList>
#include <QMap>
#include <QMultiMap>
#include <QPair>
#include <QStringList>
#include <QTemporaryFile>


class DocumentBlock
{
public:
    enum EBlockType
    {
        Text,
        Image,
        BarCode,
        QrCode
    };
    enum EBarcodeType
    {
        BARCODE_UNKNOWN = -1,
        AZTEC,
        CODABAR,
        CODE_39,
        CODE_93,
        CODE_128,
        DATA_MATRIX,
        EAN_8,
        EAN_13,
        ITF,
        MAXICODE,
        PDF_417,
        QR_CODE,
        RSS_14,
        RSS_EXPANDED,
        UPC_A,
        UPC_E,
        UPC_EAN_EXTENSION
    };

    DocumentBlock()noexcept;
    explicit DocumentBlock(const QVariantMap &data)noexcept;
    DocumentBlock(const DocumentBlock &other)noexcept;
    DocumentBlock(DocumentBlock &&other)noexcept;
    ~DocumentBlock();

    bool isImage() const;
    bool isText() const;
    bool isBarcode()const;
    bool isQrCode() const;
    bool isValid() const;

    QByteArray barcode() const;
    void setBarCode(const QByteArray &data);
    void setBarCode(const QByteArray &data, EBarcodeType type);

    QByteArray qrCode() const;
    void setQrCode(const QByteArray &data);

    EBlockType type() const;
    void setType(const EBlockType &type);

    EBarcodeType barCodeType() const;
    void setBarCodeType(const EBarcodeType &barCodeType);

    QStringList text() const;
    void setText(const QStringList &text);

    MonochromeImage img() const;
    void setImg(const MonochromeImage &img);

    LineFormat format() const;
    void setFormat(const LineFormat &format);

    QVariantMap toMap() const;
    void setMap(const QVariantMap &data);

    DocumentBlock &operator= (const DocumentBlock &other)noexcept;
    DocumentBlock &operator= (DocumentBlock &&other) noexcept;
    bool operator== (const DocumentBlock &other) const noexcept;
    bool operator!= (const DocumentBlock &other) const noexcept;

    bool append(const DocumentBlock &other);


    void toUpper();
    void toLower();

    static DocumentBlock emptyBlock(const LineFormat &format);

    static DocumentBlock createSimpleBlock(const QStringList &txt, const LineFormat &format);
    static DocumentBlock createSimpleBlock(const QString& txt, const LineFormat &format);
    static DocumentBlock createEmpty();

    qint32 symbolsCount() const;


private:
    EBlockType type_;
    EBarcodeType barCodeType_;
    QStringList text_;
    MonochromeImage img_;
    LineFormat format_;


};
typedef QList<DocumentBlock> DocumentBlocks;
//--------------------------------------------------------------------------------------------------
class TextPrinterDocument
{
public:
    TextPrinterDocument()noexcept;
    explicit TextPrinterDocument(const DocumentBlocks &blocks)noexcept;
    TextPrinterDocument(const TextPrinterDocument & other)noexcept;
    TextPrinterDocument(TextPrinterDocument && other)noexcept;
    ~TextPrinterDocument();

    QList<DocumentBlock> blocks() const;
    void setBlocks(const QList<DocumentBlock> &blocks);
    void appendBlock(const DocumentBlock &block);
    void prependBlock(const DocumentBlock &block);

    void toUpper();
    void toLower();
    QStringList text() const;

    QVariantMap toMap() const;
    void setMap(const QVariantMap &data);

    qint32 symbolsCount() const;

    TextPrinterDocument &operator= (const TextPrinterDocument &other) noexcept;
    TextPrinterDocument &operator= (TextPrinterDocument &&other)noexcept;
    bool operator== (const TextPrinterDocument &other) const noexcept;
    bool operator!= (const TextPrinterDocument &other) const noexcept;


    bool saveToFile(const QString &fileName);
    bool loadFromFile(const QString &fileName);
    void removeFile(const QString &fileName);

    void simplify();

private:
    QList<DocumentBlock> blocks_;
};



Q_DECLARE_METATYPE(DocumentBlock)
Q_DECLARE_METATYPE(TextPrinterDocument)

#define SIMPLEBL_(X,Y) DocumentBlock::createSimpleBlock(X, Y)


#endif // TEXTPRINTERDOCUMENT_H
