#include "basesqlitedbconnector.h"

#include "logmanip.h"

#include "dircreator.h"
#include "logmanip.h"

#include <QThread>

#include <QSqlQuery>
#include <QSqlError>
#include <QSqlDriver>
#include <QDebug>
#include <QJsonDocument>
#include <QFileInfo>
#include <QCryptographicHash>
#include <QUrl>
#include <QUrlQuery>
#include <QTextCodec>
#include <functional>
#include <QDateTime>
#include <QSqlRecord>
#include <algorithm>

BaseSqliteDbConnector::~BaseSqliteDbConnector()
{
}

void BaseSqliteDbConnector::clearAllConnections()
{
    release();
    if(!prefix_.isEmpty())
    {
        QStringList names = QSqlDatabase::connectionNames();
        std::for_each(names.cbegin(), names.cend(), [this](const QString &name)->void {
            if(name.startsWith(prefix_))
            {
                QSqlDatabase *db = new QSqlDatabase(QSqlDatabase::database(name));
                if(db->isOpen())db->close();
                delete db;
                QSqlDatabase::removeDatabase(name);
            }
        });
    }
}

bool BaseSqliteDbConnector::vacuum() const
{
    if(!db_.isOpen())
    {
        lmCritical() << "DATABASE IS NOT OPEN";
        return false;
    }
    QSqlQuery q(db_);
    if(!q.exec("reindex;"))
    {
        lmCritical() << q.lastQuery() << q.lastError();
        return false;
    }
    if(!q.exec("VACUUM;"))
    {
        lmCritical() << q.lastQuery() << q.lastError();
        return false;
    }
    return true;
}

bool BaseSqliteDbConnector::resetDb()
{
    QStringList tables  = db_.tables();

    if(!db_.transaction())
    {
        qCritical() << db_.lastError();
        return false;
    }
    if(!resetDbTables(tables) || !db_.commit())
    {
        lmCritical() << db_.lastError();
        db_.rollback();
        return  false;
    }
    return true;
}

BaseSqliteDbConnector::EDbVersionState BaseSqliteDbConnector::checkVersion(qint32 &dbVersion, QString &msg, bool updateDb)
{

    QString  dbName = databaseFileName();
    if(dbName.isEmpty())
    {
        msg = tr("Database name is empty");
        lmCritical()  << msg;
        return DATABASE_CHECKING_ERROR;
    }

    QFileInfo fi(dbName);
    if(!fi.exists() || fi.size() == 0) return DATABASE_IS_NOT_EXISTS;

    EMetadataState state;
    if(!getMetaData(dbVersion, state))
    {
        if(processMetadataState(state, msg)) return checkVersion(dbVersion, msg, updateDb);
        if(msg.isEmpty()) msg = tr("Database checking error. Try restart the application");
        lmCritical() << msg;
        return DATABASE_CHECKING_ERROR;
    }
    if(dbVersion == actualMetadataVersion()) return DATABASE_IS_VALID;
    if(dbVersion < actualMetadataVersion())
    {
        if(updateDb && updateDatabase(dbVersion))
        {
            return checkVersion(dbVersion, msg, false);
        }
        else
        {
            msg = tr("Database checking error. Try restart the application");
            return DATABASE_CHECKING_ERROR;
        }
    }
    if(dbVersion > actualMetadataVersion())
    {
        msg = tr("The database requires a application update.");
        return DATABASE_IS_NEW;
    }
    msg = tr("Database checking error. Try restart the application");
    return DATABASE_CHECKING_ERROR;
}

BaseSqliteDbConnector::EDbVersionState BaseSqliteDbConnector::checkVersion(qint32 &dbVersion, QString &msg)
{
    QString  dbName = databaseFileName();
    if(dbName.isEmpty())
    {
        msg = tr("Database name is empty");
        lmCritical()  << msg;
        return DATABASE_CHECKING_ERROR;
    }

    QFileInfo fi(dbName);
    if(!fi.exists() || fi.size() == 0) return DATABASE_IS_NOT_EXISTS;

    EMetadataState state;
    if(!getMetaData(dbVersion, state))
    {
        if(msg.isEmpty()) msg = tr("Database checking error. Try restart the application");
        return DATABASE_CHECKING_ERROR;
    }
    if(dbVersion == actualMetadataVersion()) return DATABASE_IS_VALID;
    if(dbVersion < actualMetadataVersion())
    {
        msg = tr("The database is too old.");
        return DATABASE_IS_OLD;
    }
    if(dbVersion > actualMetadataVersion())
    {
        msg = tr("The database requires a application update.");
        return DATABASE_IS_NEW;
    }
    msg = tr("Database checking error. Try restart the application");
    return DATABASE_CHECKING_ERROR;
}

BaseSqliteDbConnector::BaseSqliteDbConnector(const QString &connectionPrefix,
                                             ConnectionsCounter *opened, QObject *parent) noexcept :
    QObject(parent),
    readOnly_(false),
    prefix_(connectionPrefix),
    opened_(opened)
{
    QString suffix = readOnly_ ? QStringLiteral("-RO") : QStringLiteral("-RW");
    QString exid = QString::number(reinterpret_cast<quint64>(QThread::currentThreadId()), 16);
    exid += QThread::currentThread() ? QThread::currentThread()->objectName() : QString("========");
    exid += QString::number(reinterpret_cast<qint64>(this));
    setObjectName(prefix_+ exid + suffix);
}

BaseSqliteDbConnector::BaseSqliteDbConnector(const QString &connectionPrefix,
                                             ConnectionMode mode, ConnectionsCounter *opened,
                                             QObject *parent) noexcept :
    QObject(parent),
    readOnly_(mode == CM_READ_ONLY),
    prefix_(connectionPrefix),
    opened_(opened)
{
    QString suffix = readOnly_ ? QStringLiteral("-RO") : QStringLiteral("-RW");
    QString exid = QString::number(reinterpret_cast<quint64>(QThread::currentThreadId()), 16);
    exid += QThread::currentThread() ? QThread::currentThread()->objectName() : QString("========");
    setObjectName(prefix_+ exid + suffix);
}

void BaseSqliteDbConnector::init()
{
    QString dbName = databaseFileName();
//    lmWarning() << dbName;
    if(dbName.isEmpty()) return;
    QFileInfo fi(dbName);
    if(!DirCreator::checkAndCreate(fi.absolutePath()))
    {
        lmCritical() << tr("Can\'t create the database directory %1").arg(fi.absoluteFilePath());
        return;
    }
    if(!QSqlDatabase::contains(connectionName()))
    {
        db_ = QSqlDatabase::addDatabase(driver(), connectionName());
        QString opts = QStringLiteral("QSQLITE_BUSY_TIMEOUT=10000");
        if(readOnly_) opts += QStringLiteral(";QSQLITE_OPEN_READONLY;QSQLITE_OPEN_URI");
        db_.setConnectOptions(opts);

    }
    else
    {
        db_ = QSqlDatabase::database(connectionName());
//        lmWarning() << db_.databaseName() << dbName;
        if(db_.databaseName() != dbName)
        {
//            lmWarning() <<"RECONNECT: "<< dbName;
            db_.close();
            db_ = QSqlDatabase();
            QSqlDatabase::removeDatabase(connectionName());
            db_ = QSqlDatabase::addDatabase(driver(), connectionName());
            QString opts = QStringLiteral("QSQLITE_BUSY_TIMEOUT=10000");
            if(readOnly_) opts += QStringLiteral(";QSQLITE_OPEN_READONLY;QSQLITE_OPEN_URI");
            db_.setConnectOptions(opts);
        }
    }
//    lmWarning() << db_.isOpen();
    if(!db_.isOpen())
    {
        if(readOnly_)
        {
            QUrl url;
            url.setPath(dbName);
            QUrlQuery uq;
            uq.addQueryItem("mode", "ro");
            uq.addQueryItem("cache", "private");
            url.setQuery(uq);
            url.setScheme(QStringLiteral("file"));
            dbName = url.toString();
        }
        db_.setDatabaseName(dbName);
        db_.setHostName(host());
        db_.setPort(port());
        db_.setUserName(user());
        db_.setPassword(password());
        if(!db_.open())
        {
            QStringList msg;
            msg << tr("Database {") << (db_.hostName() + ":"+ QString::number(db_.port())
                                        + "//" + db_.databaseName()) << db_.userName() << db_.password()
                << db_.driverName() << tr("} opening error: ") << db_.lastError().databaseText();
            lmCritical() << loglist(msg);
        }
        else
        {
            QSqlQuery q = db_.exec("PRAGMA foreign_keys = true;");
            QSqlError qerr = q.lastError();
            if(qerr.isValid())lmWarning() << qerr;
        }
    }
    attacheDatabases();
    if(opened_->hasLocalData()) opened_->setLocalData(opened_->localData() + 1);
    else opened_->setLocalData(1);
//    lmWarning() << db_.isOpen();
}

void BaseSqliteDbConnector::release()
{
    bool remove = !opened_->hasLocalData();
    if(opened_->hasLocalData())
    {
        qint32 cnt = opened_->localData();
        if(cnt > 0)opened_->setLocalData(--cnt);
        if(cnt <= 0)
        {
            opened_->setLocalData(0);
            remove = true;
        }
    }
    if(remove)
    {
        if(db_.isOpen())db_.close();
        db_ = QSqlDatabase();
        QSqlDatabase::removeDatabase(connectionName());
    }
}

void BaseSqliteDbConnector::attacheDatabases()
{

}

QString BaseSqliteDbConnector::driver() const
{
    return QStringLiteral("QSQLITE");
}

QString BaseSqliteDbConnector::host() const
{
    return QStringLiteral("localhost");
}

qint16 BaseSqliteDbConnector::port() const
{
    return 0;
}

QString BaseSqliteDbConnector::user() const
{
    return QString();
}

QString BaseSqliteDbConnector::password() const
{
    return QString();
}

bool BaseSqliteDbConnector::resetDbTables(const QStringList &tables)
{
    QString sql1 = "delete from ";
    QString sql2 = QStringLiteral("UPDATE SQLITE_SEQUENCE SET SEQ=0 WHERE NAME=\'%1\'");
    QSqlQuery q(db_);

    return std::all_of(tables.constBegin(), tables.constEnd(), [&q, sql1, sql2](const QString &t) -> bool {
        if(!q.exec(sql1 + t))
        {
            lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << q.executedQuery()
                         << logvariant(q.boundValues());
            return false;
        }
        if(!q.exec(sql2.arg(t)))
        {
            lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << q.executedQuery()
                         << logvariant(q.boundValues());
        }
        return true;
    });
}


bool BaseSqliteDbConnector::processMetadataState(EMetadataState state, QString &msg)
{
    Q_UNUSED(state);
    Q_UNUSED(msg);
    return false;
}

