#include "loggermanager.h"
#include "dircreator.h"
#include "logmanip.h"

#include <stdio.h>


#include <QDir>
#include <QDate>
#include <QCoreApplication>
#include <QTimer>
#include <QTextCodec>
#include <QThread>
#include <QLoggingCategory>
#include <QJsonDocument>
#include <QtConcurrent/QtConcurrent>

//NOTE: Добавил возможность не закрывать файлы с логами.
//Нужно проверить, не поломалось ли все при этом и будут ли создаваться новые файлы в новый день

QString LoggerManager::_defaultSuffix = QStringLiteral("log");

LoggerManager::EQtDebugOuts LoggerManager::_debugOuts = LoggerManager::QDO_NO;
QMap<QString, LoggerMessageOutput> LoggerManager::_outputs;
QString LoggerManager::_logPath = QString();
quint32 LoggerManager::_keepingDays = 7;
QtMsgType LoggerManager::_minimumLogLevel = QtDebugMsg;
QMutex LoggerManager::_logMutex;
QTimer *LoggerManager::_eraceTimer = nullptr;
bool LoggerManager::_rotateLogs = false;
bool LoggerManager::_keepLogHandlers = true;
QStringList LoggerManager::_additionalEraseFilters;
QMap<QString, QSharedPointer<LoggerManager> > LoggerManager::_loggers;
QAtomicInt LoggerManager::_loggersCounter = 0;
bool LoggerManager::_cleanAnyOldLogs = false;

class logfilename
{
public:
    explicit logfilename(const QString &fn);
    explicit logfilename(const char * fn);
    logfilename(const char * fn, const char *enc);
    ~logfilename();

    friend QTextStream &operator << (QTextStream &stream, const logfilename& manip);

private:
    QString text_;
};

//--------------------------------------------------------------------------------------------------

class logprefix
{
public:
    logprefix(bool rotateLogs, const char *fn, const char *func,
              int line, int msgType, const char *enc = SRC_ENCODING);
    ~logprefix();

    friend QTextStream &operator << (QTextStream &stream, const logprefix& manip);

private:
    bool rotateLogs_;
    QString fileName_;
    QString functionName_;
    int line_;
    int msgType_;
};

//--------------------------------------------------------------------------------------------------

QTextStream &logcurdt(QTextStream &stream);

//--------------------------------------------------------------------------------------------------


bool LoggerManager::initLog(const QString &path, quint32 keepingDays, QtMsgType minimumLevel)
{
    return initLog(path, keepingDays, true, minimumLevel);
}

bool LoggerManager::initLog(const QString &path, quint32 keepingDays, bool rotateLogs, QtMsgType minimumLevel)
{
    if(path.isEmpty() || keepingDays < 1)
    {
        qInstallMessageHandler(LoggerManager::ignoreLogs);
        return true;
    }
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker);
#ifdef DEBUG_TO_STDOUT
    QLoggingCategory::setFilterRules(QStringLiteral("default.debug=true"));
#else
    QLoggingCategory::setFilterRules(QStringLiteral("default.debug=false"));
#endif
    _rotateLogs = rotateLogs;
    DirCreator::checkAndCreate(path);
    _logPath = path;
    _keepingDays = keepingDays;
    _minimumLogLevel = minimumLevel;

    QDir testLogs (path);
    if(testLogs.exists("no_more_logs"))
    {
        _debugOuts = QDO_NO;
    }

    LoggerManager logger(_defaultSuffix);
    logger.stream() << Qt::endl  << Qt::endl << QString(80, QChar('-')) << Qt::endl
                    << logprefix(rotateLogs, __FILE__, __FUNCTION__, __LINE__, QtDebugMsg)
                    << QObject::tr("Start logging") << Qt::endl;
    qInstallMessageHandler(LoggerManager::logMessageOutput);

    if(_rotateLogs)
    {
        auto eraceOldLogs = []() {
            QDir dir(LoggerManager::logsPath());
            if(!dir.exists()) return;
            QStringList filtersList = LoggerManager::additionalEraseFilters();
            filtersList  << QString("*%1.log*").arg(qApp->applicationName());
            QStringList el = dir.entryList(filtersList, QDir::Files, QDir::Name);
            for(QString const &fn: el)
            {
                QDate dt = QDate::fromString(fn.mid(0, fn.indexOf("_")), "yyyy-MM-dd");
                if(dt.isValid() && dt.addDays(LoggerManager::keepingDays()) <= QDate::currentDate())
                    dir.remove(fn);
                else if(dt.isValid() && dt.addDays(1) < QDate::currentDate() &&
                        dt.addDays(5) >= QDate::currentDate() &&
                        !fn.toLower().endsWith(".zip") && !fn.toLower().endsWith(".gz") )
                {
                    //NOTE: ТЕСТИРУЕМ
#if defined Q_OS_LINUX && !defined(Q_OS_ANDROID)
                    QString lfn =  QDir::toNativeSeparators(fn);
                    QProcess::startDetached("/bin/gzip", QStringList() << "-9" << "-f" << lfn, _logPath);
#else
                    dir.remove(fn);
#endif
                }
            }
            if(LoggerManager::cleanAnyOldLogs())
            {
                std::function<void(QDir)> cleanPath;
                cleanPath = [&cleanPath](QDir dir) -> void {
                    QFileInfoList fil = dir.entryInfoList(QStringList() << "*" << "*.*", QDir::Files, QDir::Name);
                    for(QFileInfo const &fi: fil)
                    {
                        QDate dt = fi.lastModified().date();
                        if(dt.isValid() && dt.addDays(LoggerManager::keepingDays()) <= QDate::currentDate())
                            dir.remove(fi.fileName());
                    }
                    fil = dir.entryInfoList(QStringList() << "*", QDir::Dirs | QDir::NoDotAndDotDot, QDir::Name);
                    for(QFileInfo const &fi: fil)
                    {
                        QDir dr(fi.absoluteFilePath());
                        cleanPath(dr);
                    }
                };
                cleanPath(dir);
            }
        };
        auto logEracer = [eraceOldLogs]() {
            QtConcurrent::run(eraceOldLogs);
        };

        _eraceTimer = new QTimer;
        _eraceTimer->setInterval(24 * 3600 * 1000);
        _eraceTimer->connect(_eraceTimer, &QTimer::timeout, logEracer);
        _eraceTimer->start();
        QtConcurrent::run(eraceOldLogs);
    }
    return true;
}

void LoggerManager::releaseLog()
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    if(_eraceTimer)
    {
        delete _eraceTimer;
        _eraceTimer = nullptr;
    }
    //UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU
    if(!_loggers.isEmpty())_loggers.clear();
    //UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU

    LoggerManager logger(_defaultSuffix);
    logger.stream() << logprefix(_rotateLogs, __FILE__, __FUNCTION__, __LINE__, QtDebugMsg)
                    << QObject::tr("Stop logging")  << Qt::endl;
    qInstallMessageHandler(0);
//FIXME:	_outputs.clear();
}

void LoggerManager::logMessageOutput(QtMsgType type, const QMessageLogContext &context, const QString &msg)
{
    _logMutex.lock();
    if(context.category && _outputs.contains(QString::fromUtf8(context.category)))
    {
        auto out = _outputs[QString::fromUtf8(context.category)];
        _logMutex.unlock();
        out(type, context, msg);
        return ;
    }
    _logMutex.unlock();
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    if(!_debugOuts || type < _minimumLogLevel) return;

//    if(_debugOuts & QDO_STDERR)
//    {
//        QTextStream ts(stderr);
//        ts << logprefix(_rotateLogs, context.file, context.function, context.line, type)
//           << msg << Qt::endl;
//    }

    if(_debugOuts & QDO_FILE)
    {
        QString suffix = _defaultSuffix.isEmpty() ? QStringLiteral("log") : _defaultSuffix;
        QString category = QLatin1String(context.category);
        if(category != QStringLiteral("default")) suffix += QStringLiteral(".") + category;
        if(_rotateLogs && _keepLogHandlers)
        {
            QSharedPointer<LoggerManager> logger;
            if(_loggers.contains(suffix))
            {
                logger = _loggers[suffix];
                logger->open();
            }
            else
            {
                logger = QSharedPointer<LoggerManager>(new LoggerManager(suffix));
            }
            logger->stream().setCodec(LOG_ENCODING);
            logger->stream() << logprefix(_rotateLogs, context.file, context.function, context.line, type)
                             << msg << Qt::endl;
            logger->stream().flush();
            _loggers.insert(suffix, logger);
        }
        else
        {
            LoggerManager logger(suffix);
            logger.stream().setCodec(LOG_ENCODING);
            logger.stream() << logprefix(_rotateLogs, context.file, context.function, context.line, type)
                             << msg << Qt::endl;
        }
    }
}

void LoggerManager::ignoreLogs(QtMsgType type, const QMessageLogContext &context, const QString &msg)
{
    Q_UNUSED(type)
    Q_UNUSED(context)
    Q_UNUSED(msg)
}

LoggerManager::EQtDebugOuts LoggerManager::debugOuts()
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    return _debugOuts;
}

void LoggerManager::setDebugOuts(const EQtDebugOuts &debugOuts)
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    _debugOuts = debugOuts;
    if(!_logPath.trimmed().isEmpty())
    {
        QDir testLogs (_logPath);
        if(testLogs.exists("no_more_logs"))
        {
            _debugOuts = QDO_NO;
        }
    }
}
QString LoggerManager::logsPath()
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    return _logPath;
}

QMap<QString, LoggerMessageOutput> LoggerManager::outputs()
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    return _outputs;
}

void LoggerManager::setOutputs(const QMap<QString, LoggerMessageOutput> &outputs)
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    _outputs = outputs;
}

void LoggerManager::setOutputs(const QString &category, LoggerMessageOutput output)
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    if(!_outputs.contains(category))
        _outputs.insert(category, output);
}

void LoggerManager::removeOutput(const QString &category)
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    if(_outputs.contains(category))_outputs.remove(category);
}

bool LoggerManager::rotateLogs()
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    return _rotateLogs;
}
bool LoggerManager::keepLogHandlers()
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    return _keepLogHandlers;
}

void LoggerManager::setKeepLogHandlers(bool keepLogHandlers)
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    _keepLogHandlers = keepLogHandlers;
}
QString LoggerManager::defaultSuffix()
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    return _defaultSuffix;
}

void LoggerManager::setDefaultSuffix(const QString &defaultSuffix)
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    _defaultSuffix = defaultSuffix;
}

QStringList LoggerManager::additionalEraseFilters()
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    return _additionalEraseFilters;
}

void LoggerManager::setAdditionalEraseFilters(const QStringList &additionalEraseFilters)
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    _additionalEraseFilters = additionalEraseFilters;
}

bool LoggerManager::cleanAnyOldLogs()
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    return _cleanAnyOldLogs;
}

void LoggerManager::setCleanAnyOldLogs(bool cleanAnyOldLogs)
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    _cleanAnyOldLogs = cleanAnyOldLogs;
}

quint32 LoggerManager::keepingDays()
{
    QMutexLocker locker(&_logMutex);
    Q_UNUSED(locker)
    return _keepingDays;
}





LoggerManager::LoggerManager(const QString &suffix) :
    suffix_(suffix),
    logFile_(nullptr),
    stream_(nullptr)
{
    _loggersCounter.ref();
    _loggersCounter ++;
    open();
}

LoggerManager::~LoggerManager()
{
    close();
    _loggersCounter --;
    _loggersCounter.deref();
}

bool LoggerManager::initEmptyLog()
{
    return initLog(QString(), 0, QtDebugMsg);
}

void LoggerManager::open()
{
//    QFileInfo fi = QFileInfo(_logPath);
//    if(fi.exists())
//    {
//        QDateTime cur = QDateTime::currentDateTime();
//        QString fileName;
//        if(_rotateLogs)
//        {
//            fileName = QStringLiteral("%1/%2_%3.%4").arg(fi.absoluteFilePath())
//                    .arg(cur.toString("yyyy-MM-dd"))
//                    .arg(qApp->applicationName())
//                    .arg(suffix_);
//        }
//        else
//        {
//            fileName = QStringLiteral("%1/%2.%3").arg(fi.absoluteFilePath())
//                    .arg(qApp->applicationName()).arg(suffix_);
//        }
//        if(logFile_ && logFile_->fileName() != fileName)
//        {
//            close();
//        }
//        if(!logFile_)
//        {
//            logFile_ = new QFile(fileName);
//        }
//        if(!logFile_->isOpen() && !logFile_->open(QIODevice::ReadWrite | QIODevice::Text | QIODevice::Append))
//        {
//            delete logFile_;
//            logFile_ = 0;
//        }
//        else
//        {
//            logFile_->setPermissions(QFileDevice::ReadOwner | QFileDevice::WriteOwner |
//                                     QFileDevice::ReadGroup | QFileDevice::WriteGroup |
//                                     QFileDevice::ReadOther);
//        }
//    }
//    if(!logFile_ || !logFile_->isOpen()) stream_ = new QTextStream(stderr);
//    else stream_ = new QTextStream(logFile_);
    QFileInfo fi = QFileInfo(_logPath);
    if(fi.exists())
    {
        QDateTime cur = QDateTime::currentDateTime();
        QString fileName;
        if(_rotateLogs)
        {
            fileName = QStringLiteral("%1/%2_%3.%4").arg(fi.absoluteFilePath())
                           .arg(cur.toString("yyyy-MM-dd"))
                           .arg(qApp->applicationName())
                           .arg(suffix_);
        }
        else
        {
            fileName = QStringLiteral("%1/%2.%3").arg(fi.absoluteFilePath())
                           .arg(qApp->applicationName()).arg(suffix_);
        }
        QFileInfo logfi(fileName);
        if(logfi.exists() && logfi.size() > MAX_FILE_SIZE)
        {
            if(logFile_ && logFile_->isOpen())
            {
                close();
            }
            QFile::rename(fileName, fileName + ".rm");
            QFile::remove(fileName + ".rm");
        }

        if(logFile_ && logFile_->fileName() != fileName)
        {
            close();
        }
        if(!logFile_)
        {
            logFile_ = new QFile(fileName);
        }
        if(!logFile_->isOpen() && !logFile_->open(QIODevice::ReadWrite | QIODevice::Text | QIODevice::Append))
        {
            delete logFile_;
            logFile_ = 0;
        }
        else
        {
            logFile_->setPermissions(QFileDevice::ReadOwner | QFileDevice::WriteOwner |
                                     QFileDevice::ReadGroup | QFileDevice::WriteGroup |
                                     QFileDevice::ReadOther);
        }
    }
    if(!logFile_ || !logFile_->isOpen()) stream_ = new QTextStream(stderr);
    else stream_ = new QTextStream(logFile_);
}

void LoggerManager::close()
{
    delete stream_;
    stream_ = nullptr;
    if(logFile_ && logFile_->isOpen())
    {
        logFile_->close();
    }
    delete logFile_;
    logFile_ = nullptr;
}

QTextStream &LoggerManager::stream()
{
    return *stream_;
}

//--------------------------------------------------------------------------------------------------

logfilename::logfilename(const QString &fn) :
    text_(fn)
{
}

logfilename::logfilename(const char *fn) :
    text_(QString::fromUtf8(fn))
{
}

logfilename::logfilename(const char *fn, const char *enc)
{
    if(fn && enc)
    {
        QByteArray ba(fn);
        QTextDecoder dec(QTextCodec::codecForName(enc));
        text_ = dec.toUnicode(ba);
    }
}

logfilename::~logfilename()
{
}

QTextStream &operator <<(QTextStream &stream, const logfilename &manip)
{
    QFileInfo fi(manip.text_);
    stream << fi.fileName();
    return stream;
}

//--------------------------------------------------------------------------------------------------

QTextStream &logcurdt(QTextStream &stream)
{
    stream << QTime::currentTime().toString("[hh:mm:ss.zzz] ");
    return stream;
}


//--------------------------------------------------------------------------------------------------


logprefix::logprefix(bool rotateLogs, const char *fn, const char *func, int line, int msgType, const char *enc):
    rotateLogs_(rotateLogs),
    functionName_(QString::fromUtf8(func)), line_(line),
    msgType_(msgType)
{
    if(fn && enc)
    {
        QByteArray ba(fn);
        QTextDecoder dec(QTextCodec::codecForName(enc));
        fileName_ = dec.toUnicode(ba);
    }

}

logprefix::~logprefix()
{
}

QTextStream &operator <<(QTextStream &stream, const logprefix &manip)
{
    if(!manip.rotateLogs_)
    {
        stream << QStringLiteral("[%1]").arg(QDate::currentDate().toString("yy-MM-dd")) << Qt::endl;
    }
    stream << logcurdt;

    if(manip.line_ == 0 && manip.fileName_.isEmpty())
    {
        QString msgType;
        switch (manip.msgType_)
        {
        case QtDebugMsg: msgType = QLatin1String("INFO:");break;
        case QtWarningMsg: msgType = QLatin1String("WARNING:");break;
        case QtCriticalMsg: msgType = QLatin1String("CRITICAL:");break;
        case QtFatalMsg: msgType = QLatin1String("FATAL:");break;
        default: msgType = QString();
        }
        if(!msgType.isEmpty())
        {
            stream << Qt::left << qSetFieldWidth(10) << msgType;
        }
        stream << Qt::dec << Qt::left << qSetFieldWidth(0);
    }
    else
    {
        QString msgType;
        switch (manip.msgType_)
        {
        case QtDebugMsg: msgType = QLatin1String("INFO IN:");break;
        case QtWarningMsg: msgType = QLatin1String("WARNING IN:");break;
        case QtCriticalMsg: msgType = QLatin1String("CRITICAL IN:");break;
        case QtFatalMsg: msgType = QLatin1String("FATAL IN:");break;
        default: msgType = QString();
        }
        if(!msgType.isEmpty())
        {
            stream << Qt::left << qSetFieldWidth(16) << msgType;
        }
        stream << Qt::dec << Qt::left << qSetFieldWidth(32)
               << logfilename(manip.fileName_) << qSetFieldWidth(5) << manip.line_
               << qSetFieldWidth(14) << QStringLiteral("THREAD_ID:")
               << qSetFieldWidth(16) << QString::number((qint64) QThread::currentThreadId(), 16)
               << qSetFieldWidth(0) << logtab << manip.functionName_ << logtab;
    }
    return stream;
}
