#include "logmanip.h"
#include <QDate>
#include <QTextCodec>
#include <QThread>
#include <QLoggingCategory>
#include <QJsonDocument>

//--------------------------------------------------------------------------------------------------

QTextStream &logtab(QTextStream &stream)
{
    stream  << Qt::endl << QString(15, QChar(' '));
    return stream;
}

//--------------------------------------------------------------------------------------------------

logbinary::logbinary(const QByteArray &data, const char *enc) :
    data_(data), encoding_(enc ? enc : LOG_ENCODING)
{
}

logbinary::logbinary(logbinary &&other) noexcept:
    data_(QByteArray()),
    encoding_(QByteArray())
{
    data_.swap(other.data_);
    encoding_.swap(other.encoding_);
}

logbinary::~logbinary()
{
}

logbinary &logbinary::operator=(logbinary &&other)
{
    data_.swap(other.data_);
    encoding_.swap(other.encoding_);
    return *this;
}

QTextStream  &operator <<(QTextStream  &stream, const logbinary &manip)
{
    QTextCodec *codec = QTextCodec::codecForName(manip.encoding_);
    QTextDecoder decoder(codec);
    QByteArray data;
    stream << QObject::tr("Bytearrray size = ") << manip.data_.size() << logtab;
    if (manip.data_.isEmpty())
    {
        stream << QObject::tr("Byte array is empty") << Qt::endl;
        return stream;
    }
    for(int i = 0; i < manip.data_.size(); i += 16)
    {
        data = manip.data_.mid(i, 16);
        for(int j = 0; j < data.size(); ++j)
        {
            stream << data.mid(j, 1).toHex() << " ";
            if((j % 8) == 7) stream << "   ";
            if(static_cast<quint8>(data[j]) < 0x20u) data[j] = '.';
        }
        if(data.size() < 8)
        {
            stream << QString(30 + 3 * (8 - data.size()), QChar(' '));
        }
        else if(data.size() < 16)
        {
            stream << QString(3 + 3 * (16 - data.size()), QChar(' '));
        }
        stream << decoder.toUnicode(data);
        if(i + 16 < manip.data_.size()) stream << Qt::endl << QString(15, QChar(' '));
    }
    return stream;
}


QDebug operator <<(QDebug debug, const logbinary &manip)
{
#if (QT_VERSION >= QT_VERSION_CHECK(5, 4, 0))
    QTextCodec *codec = QTextCodec::codecForName(manip.encoding_);
    QTextDecoder decoder(codec);
    QByteArray data;
    debug.noquote() << QObject::tr("Bytearrray size = ") << manip.data_.size() << logtab;
    if (manip.data_.isEmpty())
    {
        debug.noquote() << QObject::tr("Byte array is empty") << Qt::endl;
        return debug;
    }
    for(int i = 0; i < manip.data_.size(); i += 16)
    {
        data = manip.data_.mid(i, 16);
        for(int j = 0; j < data.size(); ++j)
        {
            debug.noquote() << data.mid(j, 1).toHex();
            if((j % 8) == 7) debug.noquote() << "  ";
            if(static_cast<quint8>(data[j]) < 0x20u) data[j] = '.';
        }
        if(data.size() < 8)
        {
            debug.noquote() << QString(30 + 3 * (8 - data.size()), QChar(' '));
        }
        else if(data.
                size() < 16)
        {
            debug.noquote() << QString(3 + 3 * (16 - data.size()), QChar(' '));
        }
        debug.noquote() << decoder.toUnicode(data);
        if(i + 16 < manip.data_.size()) debug.noquote() << Qt::endl << QString(14, QChar(' '));
    }
    return debug.noquote();
#else
    QString s;
    QTextStream ts(&s, QIODevice::WriteOnly);
    ts << manip;
    debug << s;
    return debug;
#endif

}
//--------------------------------------------------------------------------------------------------

logencoded::logencoded(const QByteArray &data, const char *enc) :
    data_(data), encoding_(enc ? enc : LOG_ENCODING)
{

}

logencoded::logencoded(const char *data, const char *enc) :
    data_(data ? QByteArray(data, strlen(data)) : QByteArray()), encoding_(enc ? enc : LOG_ENCODING)
{

}

logencoded::logencoded(logencoded &&other) noexcept:
    data_(QByteArray()),
    encoding_(QByteArray())
{
    data_.swap(other.data_);
    encoding_.swap(other.encoding_);
}

logencoded::~logencoded()
{

}

QTextStream &operator <<(QTextStream &stream, const logencoded &manip)
{
    QTextDecoder decoder(QTextCodec::codecForName(manip.encoding_.constData()));
    QString str = decoder.toUnicode(manip.data_);
    while(str.contains("\r\n")) str = str.replace("\r\n", "\n");
    if(str.contains("\n")) stream << loglist(str.split("\n"));
    else stream << str;
    return stream;
}

QDebug operator <<(QDebug debug, const logencoded &manip)
{
    QTextDecoder decoder(QTextCodec::codecForName(manip.encoding_.constData()));
    QString str = decoder.toUnicode(manip.data_);
    while(str.contains("\r\n")) str = str.replace("\r\n", "\n");

    QString s;
    QTextStream ts(&s, QIODevice::WriteOnly);
    ts << manip;
    debug << s;
    return debug;
}

logencoded &logencoded::operator=(logencoded &&other)
{
    data_.swap(other.data_);
    encoding_.swap(other.encoding_);
    return *this;
}

//--------------------------------------------------------------------------------------------------
loglist::loglist(const QStringList &data):
    data_(data)
{
}

loglist::loglist(loglist &&other) noexcept:
    data_(QStringList())
{
    data_.swap(other.data_);
}

loglist::~loglist()
{
}

loglist &loglist::operator =(loglist &&other)
{
    data_.swap(other.data_);
    return *this;
}

QTextStream & operator <<(QTextStream &stream, const loglist &manip)
{
    QStringList sl = manip.data_;
    for(QString const &s: sl) stream << logtab << s;
    return stream;
}


QDebug operator <<(QDebug debug, const loglist &manip)
{
    QStringList sl = manip.data_;
#if (QT_VERSION >= QT_VERSION_CHECK(5, 4, 0))
    for(QString const &s: sl) debug.noquote() << logtab << s;
    return debug.noquote();
#else
    QString s;
    QTextStream ts(&s);
    ts << loglist(sl);
    debug << s;
    return debug;
#endif
}

//--------------------------------------------------------------------------------------------------
logvariant::logvariant(const QVariant &data):
    data_(data)
{

}

logvariant::logvariant(logvariant &&other) noexcept:
    data_(QVariant())
{
    data_.swap(other.data_);

}

logvariant::~logvariant()
{

}

QTextStream &operator <<(QTextStream &stream, const logvariant &manip)
{
    QStringList sl = QString::fromUtf8(QJsonDocument::fromVariant(manip.data_).toJson()).split("\n");
    stream << loglist(sl);
    return stream;
}

logvariant &logvariant::operator =(logvariant &&other)
{
    data_.swap(other.data_);
    return *this;
}


QDebug operator <<(QDebug debug, const logvariant &manip)
{
    QStringList sl = QString::fromUtf8(QJsonDocument::fromVariant(manip.data_).toJson()).split("\n");
    debug << loglist(sl);
    return debug;
}
