#include "cashboxhttpserviceroot.h"
#include "localizationmgr.h"
#include "appdirectories.h"
#include "loggermanager.h"
#include "cashboxbuildconfig.h"
#include "logmanip.h"
#include "cashboxhttpservicebinder.h"
#include "androidapihelper.h"
#include "abstractworkerthread.h"
#include "httpworkersfactory.h"
#include "apiconfigstore.h"
#include "apiconfigstore.h"
#include "httpcashboxserviceapi.h"

#include <QLocale>
#include <QTranslator>
#include <QAndroidService>
#include <QTextCodec>
#include <QAndroidIntent>
#include <QTimer>
#include <QtAndroid>

#include <QtHttpServer>

CashboxHttpServiceRoot* CashboxHttpServiceRoot::_instance = nullptr;

int CashboxHttpServiceRoot::start(int &argc, char *argv[])
{
    ADD_LOCALISATION_FILE(APPNAME);
    CashboxHttpServiceRoot root(argc, argv);
    int res = root.init() ? root.exec() : -1;
    root.stop();
    return res;
}

void CashboxHttpServiceRoot::setUpLocale(const QLocale &locale)
{
     doSetupLocale(locale);
}

void CashboxHttpServiceRoot::setUpLocale()
{
    setUpLocale(QLocale::system());
}

void CashboxHttpServiceRoot::startHttp()
{
    stopHttp();
    ApiConfigStore st;
    ApiConfig cfg = st.apiConfig();
    if(!server_ && cfg.useHttpApi())
    {
        server_ = new QHttpServer(this);
        HttpWorkersFactory::initHttpServer(server_);
        auto port = server_->listen(QHostAddress::Any, cfg.httpPort());
        Q_UNUSED(port)
//        lmWarning() << port;
    }
}

void CashboxHttpServiceRoot::stopHttp()
{
    if(server_)
    {
        server_->deleteLater();
        server_ = nullptr;
    }
}

void CashboxHttpServiceRoot::startUdp()
{
    stopUdp();
    ApiConfigStore st;
    ApiConfig cfg = st.apiConfig();
    if(cfg.useUdp())
    {
        broadCastSocket_ = new QUdpSocket(this);
        connect(broadCastSocket_, SIGNAL(readyRead()), this, SLOT(processPendingDatagramm()));
        bool bindResult = broadCastSocket_->bind(cfg.udpPort(), QUdpSocket::ShareAddress);
        lmWarning() << bindResult << broadCastSocket_->errorString();
    }

}

void CashboxHttpServiceRoot::stopUdp()
{
    if(broadCastSocket_)
    {
        delete broadCastSocket_;
        broadCastSocket_ = nullptr;
    }
}

CashboxHttpServiceRoot::CashboxHttpServiceRoot(int &argc, char *argv[])
    : QObject(nullptr)
    , argc_(&argc)
    , argv_(argv)
    , server_(nullptr)
    , broadCastSocket_(nullptr)
    , lastModelData_()
{
    if (!_instance) _instance = this;
    Q_ASSERT(argc > 0 && argv);
}

CashboxHttpServiceRoot::~CashboxHttpServiceRoot()
{
    if(_instance == this) _instance = nullptr;
}

void CashboxHttpServiceRoot::doSetupLocale(const QLocale &locale)
{
    QStringList locales = LocalizationManager::localisations();
    locales.push_front("qtbase");
    locales.push_front("qt");
    QLocale::setDefault(locale);
    for(const QString &l :locales)
    {
        QTranslator* translator = new QTranslator(this);
        if(translator->load(locale, l, "_", APATH.translations(), ".qm"))
        {
            qApp->installTranslator(translator);
        }
    }

    QTextCodec::setCodecForLocale(QTextCodec::codecForName("UTF-8"));
}

bool CashboxHttpServiceRoot::createApplication()
{
    QAndroidService *app = new QAndroidService(*argc_, argv_,
                                           [](const QAndroidIntent &intent) -> QAndroidBinder *{
                                               lmWarning() << intent.handle().toString();
                                                QAndroidBinder *res = new CashboxHttpServiceBinder();
                                                return res;
                                            });
    app->setApplicationName(APPNAME);
    app->setOrganizationName(ORGANISATION_NAME);
    app->setOrganizationDomain(ORGANISATION_DOMAIN);
    app->setApplicationVersion(AndroidApiHelper().versionName());

    AndroidApiHelper().initForeGround();
    return true;
}

bool CashboxHttpServiceRoot::init()
{
#ifdef DEBUG_TO_STDOUT
    LoggerManager::setDebugOuts(LoggerManager::QDO_BOTH);
#else
    LoggerManager::setDebugOuts(LoggerManager::QDO_FILE);
#endif
    LoggerManager::setKeepLogHandlers(true);
    if(!createApplication()) return false;

    LoggerManager::initLog(APATH.logsPath(), 7, true);
    QTextCodec::setCodecForLocale(QTextCodec::codecForName("UTF-8"));
    setUpLocale();
    return true;
}

int CashboxHttpServiceRoot::exec()
{
    QTimer::singleShot(100, this, SLOT(startAfterExec()));
    return qApp->exec();
}

void CashboxHttpServiceRoot::stop()
{
    stopHttp();
    stopUdp();
//    if(coreWorker_)
//    {
//        coreWorker_->quit();
//    }
//    QElapsedTimer tout;
//    tout.start();
//    while((coreWorker_ && !coreWorker_->isFinished()) )
//    {
//        QCoreApplication::processEvents();
//        if(tout.elapsed() > 30000 && coreWorker_) coreWorker_->terminate();
//    }

    LoggerManager::releaseLog();
}

void CashboxHttpServiceRoot::startAfterExec()
{
    startHttp();
    startUdp();
    //    lmWarning() << qApp->applicationName() << qApp->organizationName() << qApp->organizationDomain() << logtab
//                << qApp->applicationVersion() << QLocale() << AndroidApiHelper().getDsn();
//    if(!coreWorker_)
//    {
//        FrWorkingThread *w = new FrWorkingThread();
//        AbstractWorkerThread * wt = new AbstractWorkerThread(w, this);
//        connect(wt, &QThread::finished, wt, &QThread::deleteLater);
//        coreWorker_ = wt;
//        wt->start();
//    }

}

void CashboxHttpServiceRoot::processPendingDatagramm()
{
    while (broadCastSocket_ && broadCastSocket_->hasPendingDatagrams())
    {
        QByteArray datagram;
        datagram.resize(broadCastSocket_->pendingDatagramSize());
        QHostAddress peer;
        quint16 peerPort = 0;
        broadCastSocket_->readDatagram(datagram.data(), datagram.size(), &peer, &peerPort);
        QString text = QString::fromUtf8(datagram);
        lmWarning() << text << QHostAddress(peer.toIPv4Address()).toString() << peerPort;
        if(text.trimmed().toLower().startsWith("searchcashbox012"))
        {
            sendBroadCast();
        }
    }
}

void CashboxHttpServiceRoot::sendBroadCast()
{
    if(broadCastSocket_)
    {
        if(!lastModelData_.isValid())
        {
            HttpCashboxServiceApi api;
            lastModelData_ = api.loadModelData();
        }
        QVariantMap content;
        QVariantMap obj;
        ApiConfigStore st;
        ApiConfig cfg = st.apiConfig();
        obj = lastModelData_.toExternalMap();
        obj.insert("http", cfg.httpPort());
        content.insert("cashbox012autoconf", obj);
        broadCastSocket_->writeDatagram(QJsonDocument::fromVariant(content).toJson(QJsonDocument::Compact),
                                        QHostAddress::Broadcast, cfg.broadcastPort());
        lmWarning() << cfg.broadcastPort() << QHostAddress(QHostAddress::Broadcast).toString()
                    << QString::fromUtf8(QJsonDocument::fromVariant(content).toJson(QJsonDocument::Compact));



    }

}

