#include "cashboxserviceroot.h"
#include "localizationmgr.h"
#include "appdirectories.h"
#include "loggermanager.h"
#include "cashboxbuildconfig.h"
#include "logmanip.h"
#include "cashboxservicebinder.h"
#include "androidapihelper.h"
#include "abstractworkerthread.h"
#include "frworkingthread.h"
#include "coreconfigdb.h"
#include "watcherdb.h"
#include "ofdworker.h"
#include "oismworker.h"

#include <QLocale>
#include <QTranslator>
#include <QAndroidService>
#include <QTextCodec>
#include <QAndroidIntent>
#include <QTimer>
#include <QtAndroid>
#include <QStorageInfo>

CashboxServiceRoot* CashboxServiceRoot::_instance = nullptr;

int CashboxServiceRoot::start(int &argc, char *argv[])
{
    ADD_LOCALISATION_FILE(APPNAME);
    CashboxServiceRoot root(argc, argv);
    int res = root.init() ? root.exec() : -1;
    root.stop();
    return res;
}

void CashboxServiceRoot::setUpLocale(const QLocale &locale)
{
     doSetupLocale(locale);
}

void CashboxServiceRoot::setUpLocale()
{
    setUpLocale(QLocale::system());
}

CashboxServiceRoot::CashboxServiceRoot(int &argc, char *argv[])
    : QObject(nullptr)
    , argc_(&argc)
    , argv_(argv)
    , coreWorker_(nullptr)
    , ofdWorker_(nullptr)
    , oismWorker_(nullptr)
{
    if (!_instance) _instance = this;
    Q_ASSERT(argc > 0 && argv);
}

CashboxServiceRoot::~CashboxServiceRoot()
{
    if(_instance == this) _instance = nullptr;
}

void CashboxServiceRoot::doSetupLocale(const QLocale &locale)
{
    QStringList locales = LocalizationManager::localisations();
    locales.push_front("qtbase");
    locales.push_front("qt");
    QLocale::setDefault(locale);
    for(const QString &l :locales)
    {
        QTranslator* translator = new QTranslator(this);
        if(translator->load(locale, l, "_", APATH.translations(), ".qm"))
        {
            qApp->installTranslator(translator);
        }
    }

    QTextCodec::setCodecForLocale(QTextCodec::codecForName("UTF-8"));
}

bool CashboxServiceRoot::createApplication()
{
    QAndroidService *app = new QAndroidService(*argc_, argv_,
                                           [](const QAndroidIntent &intent) -> QAndroidBinder *{
                                               lmWarning() << intent.handle().toString();
                                                QAndroidBinder *res = new CashboxServiceBinder();
                                                return res;
                                            });
    app->setApplicationName(APPNAME);
    app->setOrganizationName(ORGANISATION_NAME);
    app->setOrganizationDomain(ORGANISATION_DOMAIN);
    app->setApplicationVersion(AndroidApiHelper().versionName());
    AndroidApiHelper().initForeGround();

    return true;
}

bool CashboxServiceRoot::init()
{
#ifdef DEBUG_TO_STDOUT
    LoggerManager::setDebugOuts(LoggerManager::QDO_BOTH);
#else
    LoggerManager::setDebugOuts(LoggerManager::QDO_FILE);
#endif
    LoggerManager::setKeepLogHandlers(true);
    if(!createApplication()) return false;


    QString fontFile = QStringLiteral("%1/UbuntuMono-R.ttf")
            .arg(APATH.workDir());
    if(!QFile::exists(fontFile))
    {
        QFile f(":/fonts/UbuntuMono-R.ttf");
        if(f.open(QIODevice::ReadOnly))
        {
            QByteArray data = f.readAll();
            f.close();
            f.setFileName(fontFile);
            if(f.open(QIODevice::WriteOnly))
            {
                f.write(data);
                f.flush();
                f.close();
            }
        }
    }
    LoggerManager::initLog(APATH.logsPath(), 7, true);
    QTextCodec::setCodecForLocale(QTextCodec::codecForName("UTF-8"));
    setUpLocale();
    CoreConfigDBDbInitializer::initializeDatabase();
    WhatcherDbDbInitializer::initializeDatabase();
//    DocsDBDbInitializer::initializeDatabase();
    return true;
}

int CashboxServiceRoot::exec()
{
    QTimer::singleShot(100, this, SLOT(startAfterExec()));
    return qApp->exec();
}

void CashboxServiceRoot::stop()
{
//    qrConfigMgr_->stopWatcher();
    stopThread(coreWorker_);
    stopThread(ofdWorker_);
    stopThread(oismWorker_);


    LoggerManager::releaseLog();
    AndroidApiHelper().releaseApp();
}

void CashboxServiceRoot::stopThread(QPointer<QThread> &t)
{
    if(t)
    {
        t->quit();
    }
    QElapsedTimer tout;
    tout.start();
    while((t && !t->isFinished()) )
    {
        QCoreApplication::processEvents();
        if(tout.elapsed() > 30000 && t) t->terminate();
    }
}

void CashboxServiceRoot::startAfterExec()
{
    AndroidApiHelper().initApp();

    lmWarning() << qApp->applicationName() << qApp->organizationName() << qApp->organizationDomain() << logtab
                << qApp->applicationVersion() << QLocale() << AndroidApiHelper().getDsn();
    if(!coreWorker_)
    {
        FrWorkingThread *w = new FrWorkingThread();
        connect(w, &FrWorkingThread::sendToOfd, this, &CashboxServiceRoot::sendToOfd, Qt::QueuedConnection);
        AbstractWorkerThread * wt = new AbstractWorkerThread(w, this);
        connect(wt, &QThread::finished, wt, &QThread::deleteLater);
        coreWorker_ = wt;
        wt->start();
    }
    if(!ofdWorker_)
    {
        OfdWorker *w = new OfdWorker();
        connect(this, &CashboxServiceRoot::sendToOfd, w, &OfdWorker::send, Qt::QueuedConnection);
        AbstractWorkerThread * wt = new AbstractWorkerThread(w, this);
        connect(wt, &QThread::finished, wt, &QThread::deleteLater);
        ofdWorker_ = wt;
        wt->start();

    }
    if(!oismWorker_)
    {
        OismWorker *w = new OismWorker();
        connect(this, &CashboxServiceRoot::sendToOfd, w, &OismWorker::send, Qt::QueuedConnection);
        AbstractWorkerThread * wt = new AbstractWorkerThread(w, this);
        connect(wt, &QThread::finished, wt, &QThread::deleteLater);
        oismWorker_ = wt;
        wt->start();

    }
//https://medium.com/@vishtech36/installing-apps-programmatically-in-android-10-7e39cfe22b86
}
