#include "coreconfigdb.h"

#include "cashboxbuildconfig.h"
#include "appdirectories.h"
#include "dircreator.h"
#include "logmanip.h"
#include "fixnumber.h"

#include <QThread>
#include <QSqlDatabase>
#include <QSqlQuery>
#include <QSqlError>
#include <QDebug>
#include <QFile>
#include <QDir>
#include <QJsonDocument>
#include <QJsonParseError>


CoreConfigDB::ConnectionsCounter CoreConfigDB::_rwopened;
CoreConfigDB::ConnectionsCounter CoreConfigDB::_roopened;

CoreConfigDB::CoreConfigDB(QObject *parent)
    : BaseSqliteDbConnector("CoreConfigDB-", &_rwopened, parent)
{
    init();
}

CoreConfigDB::CoreConfigDB(ConnectionMode mode, QObject *parent)
    : BaseSqliteDbConnector("CoreConfigDB-", mode,
                            mode == CM_READ_ONLY ? &_roopened : &_rwopened,
                            parent)
{
    init();
}

CoreConfigDB::~CoreConfigDB()
{
    release();
}

bool CoreConfigDB::loadCashierNames(QVariantList &cashiers)
{
    cashiers.clear();
    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    QString sql = "select login, name from cashiers where perms <> 0";
    QSqlQuery q(db_);
    if(!q.exec(sql))
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }

    while(q.next())
    {
        QVariantMap m;
        m["login"] = q.value("login");
        m["name"] = q.value("name");
        cashiers << m;
    }
    return (db_.commit() || db_.rollback()) && !cashiers.isEmpty();
}

bool CoreConfigDB::loadCashiers(bool active, bool local, bool server, Cashiers &cashiers)
{
    QStringList filters;
    if(active) filters << "(perms <> 0)";
    if(local && !server) filters << "((serverId <= 0) or (serverId is null))";
    else if(!local && server) filters << "(serverId > 0)";
    return loadCashiers(filters.join(" and "), cashiers);
}

bool CoreConfigDB::loadCashiers(bool active, bool local, bool server, QVariantList &cashiers)
{
    QStringList filters;
    if(active) filters << "(perms <> 0)";
    if(local && !server) filters << "((serverId <= 0) or (serverId is null))";
    else if(!local && server) filters << "(serverId > 0)";
    return loadCashiers(filters.join(" and "), cashiers);
}

bool CoreConfigDB::loadCashiers(qint32 idFrom, bool active, bool local,
                                bool server, QVariantList &cashiers)
{
    QStringList filters;
    if(active) filters << "(perms <> 0)";
    filters << QStringLiteral("(id >= %1)").arg(idFrom);
    if(local && !server) filters << "((serverId <= 0) or (serverId is null))";
    else if(!local && server) filters << "(serverId > 0)";
    return loadCashiers(filters.join(" and "), cashiers);
}

bool CoreConfigDB::saveCashier(const Cashier &c, QVariantMap &out)
{
    out.clear();
    if(c.serverId() > 0) return false;
    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    qint32 cashierId = 0;
    QSqlQuery q(db_);
    if(c.isSuperUser())
    {
        q.prepare("update cashiers set name = :name, pass = :pass, inn = :inn where id = :id");
        QString pass = c.password().trimmed();
        if(pass.isEmpty()) pass = 99;
        q.bindValue(":name", c.name());
        q.bindValue(":pass", pass);
        q.bindValue(":inn", c.inn());
        q.bindValue(":id", Cashier::SUPER_USER_ID);
        if(!q.exec())
        {
            lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
            db_.rollback();
            return false;
        }
        cashierId = Cashier::SUPER_USER_ID;
    }
    else
    {
        if(c.id() > 0)
        {
            q.prepare("update cashiers set login = :login, name = :name, pass = :pass, "
                    "perms = :perms, inn = :inn where id = :id");
            q.bindValue(":id", c.id());
        }
        else
        {
            q.prepare("insert into cashiers (login, name, pass, perms, inn) "
                        "values (:login, :name, :pass, :perms, :inn)");
        }
        q.bindValue(":login", c.login());
        q.bindValue(":name", c.name());
        q.bindValue(":pass", c.password());
        q.bindValue(":perms", static_cast<qint32>(c.permissions()));
        q.bindValue(":inn", c.inn());
        if(!q.exec())
        {
            lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
            db_.rollback();
            return false;
        }
        cashierId = c.id() > 0 ? c.id() : q.lastInsertId().toInt();
    }
    q.prepare("select id, serverId, login, name, pass, perms, inn from cashiers where id = :id");
    q.bindValue(":id", cashierId);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    Cashier cashier;
    if(q.first())
    {
        cashier.setId(q.value("id").toInt());
        cashier.setServerId(q.value("serverId").toLongLong());
        cashier.setLogin(q.value("login").toString());
        cashier.setName(q.value("name").toString());
        cashier.setPassword(q.value("pass").toString());
        cashier.setPermissions(Cashier::Permission(q.value("perms").toUInt()));
        cashier.setInn(q.value("inn").toString().trimmed());
        out = cashier.toMap();
    }
    if(!db_.commit())
    {
        lmWarning() << db_.lastError();
        db_.rollback();
        return false;
    }
    return true;
}

bool CoreConfigDB::removeCashier(int id)
{
    if(id == Cashier::SUPER_USER_ID) return false;
    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    QSqlQuery q(db_);
    q.prepare("select serverId from cashiers where id = :id");
    q.bindValue(":id", id);
    if(!q.exec() || !q.first() || q.value("serverId").toLongLong() > 0)
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    q.prepare("delete from cashiers where id = :id");
    q.bindValue(":id", id);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        lmWarning() << db_.lastError();
        db_.rollback();
        return false;
    }
    return true;
}

bool CoreConfigDB::loadSupportedOfd(QVariantList &ofd)
{
    ofd.clear();
    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    QString sql = "select inn, address, port, dns, queryInterval, url, name, shortName,"
                  "labelAddr, labelPort, keysAddr, keysPort, testTout, keysTout from SUPPORTED_OFD";

    QSqlQuery q(db_);
    if(!q.exec(sql))
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    while(q.next())
    {
        OfdSettings o;
        o.setAddress(q.value("address").toString());
        o.setDns(q.value("dns").toString());
        o.setInn(q.value("inn").toString());
        o.setName(q.value("name").toString());
        o.setPort(q.value("port").toUInt());
        o.setShortName(q.value("shortName").toString());
        o.setUrl(q.value("url").toString());
        o.setQueryInterval(q.value("queryInterval").toUInt());
        o.setLabelsAddress(q.value("labelAddr").toString());
        o.setLabelsPort(q.value("labelPort").toUInt());
        o.setKeysAddress(q.value("keysAddr").toString());
        o.setKeysPort(q.value("keysPort").toUInt());
        o.setTestTimeout(q.value("testTout").toInt());
        o.setKeysTimeout(q.value("keysTout").toInt());
        ofd << o.toMap();
    }
    return (db_.commit() || db_.rollback()) && !ofd.isEmpty();
}

bool CoreConfigDB::loadSupportedOfd(QList<OfdSettings> &ofd)
{
    ofd.clear();
    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    QString sql = "select inn, address, port, dns, queryInterval, url, name, shortName,"
                    "labelAddr, labelPort, keysAddr, keysPort, testTout, keysTout from SUPPORTED_OFD";
    QSqlQuery q(db_);
    if(!q.exec(sql))
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    while(q.next())
    {
        OfdSettings o;
        o.setAddress(q.value("address").toString());
        o.setDns(q.value("dns").toString());
        o.setInn(q.value("inn").toString());
        o.setName(q.value("name").toString());
        o.setPort(q.value("port").toUInt());
        o.setShortName(q.value("shortName").toString());
        o.setUrl(q.value("url").toString());
        o.setQueryInterval(q.value("queryInterval").toUInt());
        o.setLabelsAddress(q.value("labelAddr").toString());
        o.setLabelsPort(q.value("labelPort").toUInt());
        o.setKeysAddress(q.value("keysAddr").toString());
        o.setKeysPort(q.value("keysPort").toUInt());
        o.setTestTimeout(q.value("testTout").toInt());
        o.setKeysTimeout(q.value("keysTout").toInt());
        ofd << o;
    }
    return (db_.commit() || db_.rollback()) && !ofd.isEmpty();
}

bool CoreConfigDB::loadCurrentOfd(const QString &inn, OfdSettings &ofd)
{
    lmWarning() << inn;
    ofd = OfdSettings();
    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    QString sql = "select inn, address, port, dns, queryInterval, url, "
                    "labelAddr, labelPort, keysAddr, keysPort, testTout, keysTout from CURRENT_OFD";
    QSqlQuery q(db_);
    if(!q.exec(sql))
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    if(q.first())
    {
        ofd.setAddress(q.value("address").toString());
        ofd.setDns(q.value("dns").toString());
        ofd.setInn(q.value("inn").toString());
        ofd.setPort(q.value("port").toUInt());
        ofd.setUrl(q.value("url").toString());
        ofd.setQueryInterval(q.value("queryInterval").toUInt());
        ofd.setLabelsAddress(q.value("labelAddr").toString());
        ofd.setLabelsPort(q.value("labelPort").toUInt());
        ofd.setKeysAddress(q.value("keysAddr").toString());
        ofd.setKeysPort(q.value("keysPort").toUInt());
        ofd.setTestTimeout(q.value("testTout").toInt());
        ofd.setKeysTimeout(q.value("keysTout").toInt());
    }
    else
    {
        q.prepare("select inn, address, port, dns, queryInterval, url, name, shortName,"
                    "labelAddr, labelPort, keysAddr, keysPort, testTout, keysTout from SUPPORTED_OFD "
                        "where inn = :inn");
        q.bindValue(":inn", inn.trimmed());
        if(!q.exec() ||! q.first())
        {
            lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
            db_.rollback();
            return false;
        }
        ofd.setAddress(q.value("address").toString());
        ofd.setDns(q.value("dns").toString());
        ofd.setInn(q.value("inn").toString());
        ofd.setName(q.value("name").toString());
        ofd.setPort(q.value("port").toUInt());
        ofd.setShortName(q.value("shortName").toString());
        ofd.setUrl(q.value("url").toString());
        ofd.setQueryInterval(q.value("queryInterval").toUInt());
        ofd.setLabelsAddress(q.value("labelAddr").toString());
        ofd.setLabelsPort(q.value("labelPort").toUInt());
        ofd.setKeysAddress(q.value("keysAddr").toString());
        ofd.setKeysPort(q.value("keysPort").toUInt());
        ofd.setTestTimeout(q.value("testTout").toInt());
        ofd.setKeysTimeout(q.value("keysTout").toInt());
    }
    return (db_.commit() || db_.rollback()) && ofd.isValid();
}

bool CoreConfigDB::saveCurrentOfd(const OfdSettings &ofd, bool forceReset)
{
    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    QSqlQuery q(db_);
    if(!forceReset)
    {
        if(!q.exec("select inn from CURRENT_OFD"))
        {
            lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
            db_.rollback();
            return false;
        }
        if(q.first())
        {
            QString inn = q.value("inn").toString();
            if(inn == ofd.inn())
            {
                db_.rollback();
                return true;
            }
        }
    }

    q.prepare("insert or replace into CURRENT_OFD (id, inn, address, port, dns, queryInterval, url,"
              "labelAddr, labelPort, keysAddr, keysPort, testTout, keysTout) values"
              "(:id, :inn, :address, :port, :dns, :queryInterval, :url, "
                ":labelAddr, :labelPort, :keysAddr, :keysPort, :testTout, :keysTout)");
    q.bindValue(":id", 1);
    q.bindValue(":inn", ofd.inn());
    q.bindValue(":address", ofd.address());
    q.bindValue(":port", ofd.port());
    q.bindValue(":dns", ofd.dns());
    q.bindValue(":queryInterval", ofd.queryInterval());
    q.bindValue(":url", ofd.url());
    q.bindValue(":labelAddr", ofd.labelsAddress());
    q.bindValue(":labelPort", ofd.labelsPort());
    q.bindValue(":keysAddr", ofd.keysAddress());
    q.bindValue(":keysPort", ofd.keysPort());
    q.bindValue(":testTout", ofd.testTimeout());
    q.bindValue(":keysTout", ofd.keysTimeout());
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        lmCritical() << db_.lastError();
        db_.rollback();
        return  false;
    }
    return true;
}

bool CoreConfigDB::resetDb()
{
    QStringList tables  = db_.tables();

    for(int i = tables.size() - 1; i >= 0; --i)
    {
        if(tables[i].toLower() == "metadata")
        {
            tables.removeAt(i);
            break;
        }
    }

    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    bool res =  resetDbTables(tables);


    if(res && !db_.commit())
    {
        lmCritical() << db_.lastError();
        db_.rollback();
        return  false;
    }
    return  CoreConfigDBDbInitializer::initCashiers(&db_)  ;
}



QString CoreConfigDB::databaseFileName() const
{
    return APATH.configDir() +"/config.db";
}

QString CoreConfigDB::connectionName() const
{
    return objectName();
}

bool CoreConfigDB::getMetaData(qint32 &metadataVersion, BaseSqliteDbConnector::EMetadataState &state)
{
    Q_UNUSED(state)
    metadataVersion = 0;
    if(!db_.isOpen()) return false;
    QSqlQuery q(db_);
    if(!q.exec("select metadataVersion from metadata") )
    {
        qCritical() << q.lastError();
        return false;
    }
    if(!q.first())
    {
        qCritical() << tr("Касса не зарегистрирована");
        return false;
    }
    metadataVersion =  q.value(0).toInt();
    return true;
}

qint32 CoreConfigDB::actualMetadataVersion() const
{
    return DB_METADATA_VERSION;
}

bool CoreConfigDB::updateDatabase(qint32 from)
{
    CoreConfigDBDbInitializer dbi;
    bool res = true;
    for(int i = from + 1; i <= DB_METADATA_VERSION;++i)
    {
        res = res && dbi.updateDatabase(&db_, i);
    }
    return res;
}

bool CoreConfigDB::loadCashiers(QString filter, QMap<qint32, Cashier> &cashiers)
{
    cashiers.clear();
    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    filter = filter.trimmed();
    QString sql = "select id, serverId, login, name, pass, perms, inn from cashiers ";
    if(!filter.isEmpty()) sql += QStringLiteral("where ") + filter;
    QSqlQuery q(db_);
    if(!q.exec(sql))
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    Cashier cashier;
    while(q.next())
    {
        cashier.setId(q.value("id").toInt());
        cashier.setServerId(q.value("serverId").toLongLong());
        cashier.setLogin(q.value("login").toString());
        cashier.setName(q.value("name").toString());
        cashier.setPassword(q.value("pass").toString());
        cashier.setPermissions(Cashier::Permission(q.value("perms").toUInt()));
        cashier.setInn(q.value("inn").toString().trimmed());
        cashiers.insert(cashier.id(), cashier);
    }
    return (db_.commit() || db_.rollback()) && !cashiers.isEmpty();
}

bool CoreConfigDB::loadCashiers(QString filter, Cashiers &cashiers)
{
    cashiers.clear();
    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    filter = filter.trimmed();
    QString sql = "select id, serverId, login, name, pass, perms, inn from cashiers ";
    if(!filter.isEmpty()) sql += QStringLiteral("where ") + filter;
    QSqlQuery q(db_);
    if(!q.exec(sql))
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    Cashier cashier;
    while(q.next())
    {
        cashier.setId(q.value("id").toInt());
        cashier.setServerId(q.value("serverId").toLongLong());
        cashier.setLogin(q.value("login").toString());
        cashier.setName(q.value("name").toString());
        cashier.setPassword(q.value("pass").toString());
        cashier.setPermissions(Cashier::Permission(q.value("perms").toUInt()));
        cashier.setInn(q.value("inn").toString().trimmed());
        cashiers << cashier;
    }
    return (db_.commit() || db_.rollback()) && !cashiers.isEmpty();
}

bool CoreConfigDB::loadCashiers(QString filter, QVariantList &cashiers)
{
    cashiers.clear();
    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    filter = filter.trimmed();
    QString sql = "select id, serverId, login, name, pass, perms, inn from cashiers ";
    if(!filter.isEmpty()) sql += QStringLiteral("where ") + filter;
    QSqlQuery q(db_);
    if(!q.exec(sql))
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    Cashier cashier;
    while(q.next())
    {
        cashier.setId(q.value("id").toInt());
        cashier.setServerId(q.value("serverId").toLongLong());
        cashier.setLogin(q.value("login").toString());
        cashier.setName(q.value("name").toString());
        cashier.setPassword(q.value("pass").toString());
        cashier.setPermissions(Cashier::Permission(q.value("perms").toUInt()));
        cashier.setInn(q.value("inn").toString().trimmed());
        cashiers << cashier.toMap();
    }
    return (db_.commit() || db_.rollback()) && !cashiers.isEmpty();
}


//-----------------------------------------------------------------------------------------

CoreConfigDBDbInitializer::CoreConfigDBDbInitializer(QObject *parent)
    : QObject(parent)
{
    QString thrId = QString::number(reinterpret_cast<qint64>(QThread::currentThreadId()), 16);
    setObjectName(QString("CoreConfigDBDbInitializer-")+ thrId);
}

CoreConfigDBDbInitializer::~CoreConfigDBDbInitializer()
{

}

bool CoreConfigDBDbInitializer::initializeDatabase()
{
    AppDirectories cfg;
    QString fileName = cfg.configDir() +"/config.db";
    lmWarning() << fileName;
    bool valid = false;
    if(QFile::exists(fileName))
    {
        CoreConfigDB db;
        QString msg;
        qint32 ver = 0;
        valid = db.db().isOpen() && db.checkVersion(ver, msg, true) == CoreConfigDB::DATABASE_IS_VALID;
        if(valid)
        {
            QSqlQuery q(db.db());
            valid = q.exec("pragma integrity_check;");
            q.first();
        }
    }
    if(!valid)
    {
        QDir dir(cfg.configDir());
        QFileInfo fi(fileName);
        QFileInfoList fil = dir.entryInfoList(QStringList() << fi.fileName() + "*", QDir::Files);

        for(const QFileInfo &fi :fil)
        {
            QFile::remove(fi.absoluteFilePath());
        }
    }
    lmWarning() << valid;
    if(!QFile::exists(fileName))
    {
        CoreConfigDB dbc;
        QSqlDatabase db = dbc.db();
        CoreConfigDBDbInitializer dbi;
        bool res = dbi.createDatabase(&db, ":/configdb/configdb.sql");
        initCashiers(&db);
        initOfd(&db);
        return res;
    }
    CoreConfigDB dbc;
    QSqlDatabase db = dbc.db();
    initCashiers(&db);
    initOfd(&db);
    return true;
}

bool CoreConfigDBDbInitializer::createDatabase(QSqlDatabase * const db, const QString &fileName)
{
    QString sql = getScript(fileName);
    if(sql.isEmpty() || !(db->isOpen() || db->open()))
    {
        QSqlError err = db->lastError();
        qCritical() << (err.isValid() ? err.text() : tr("Не задан скрипт для создания базы!"));
        return false;
    }
    db->exec("PRAGMA journal_mode = \"WAL\"");
//    lmCritical() << db->exec(QStringLiteral("").arg(CORE_CONFIG_DB_METADATA_VERSION)).lastError();
    return applyScript(db, sql);
}

bool CoreConfigDBDbInitializer::updateDatabase(QSqlDatabase * const db, qint32 toMetaData)
{
    QString fileName = QString(":/configdb/md_%1.sql").arg(toMetaData);
    QString sql = getScript(fileName);
    if(sql.isEmpty()) return true;
    if(!(db->isOpen() || db->open()))
    {
        QSqlError err = db->lastError();
        qCritical() << (err.isValid() ? err.text() : tr("Не задан скрипт для создания базы!"));
        return false;
    }
    return applyScript(db, sql);
}

QString CoreConfigDBDbInitializer::getScript(const QString &fileName)
{
    QFile f(fileName);
    if(!f.exists() || !f.open(QIODevice::ReadOnly))
    {
        qCritical() << tr("Ошибка открытия файла %1 (%2)").arg(fileName).arg(f.errorString());
        return QString();
    }
    QTextStream ts(&f);
    ts.setCodec("UTF-8");
    QString res = ts.readAll();
    f.close();
    return res;
}

bool CoreConfigDBDbInitializer::applyScript(QSqlDatabase * const db, const QString &sql)
{
    auto checkFk = [db](bool off) -> bool{
        QSqlQuery q(*db);
        bool res = q.exec(QStringLiteral("PRAGMA foreign_keys = %1").arg(off ? "off" : "on"));
        if(!res)
        {
            qCritical() << q.lastError() << logtab << q.lastQuery();
        }
        return res;
    };
    if(!checkFk(true))
    {
        return false;
    }
    bool res = db->transaction();
    QStringList script = sql.split(";;");
    QSqlQuery q(*db);

    for(QString s: script)
    {
        if(s.size() < 7) continue;
        res = res && q.exec(s);
        if(!res)
        {
            s.replace("\r\n","\n");
            qWarning() << loglist(s.split("\n"));
            qCritical() << q.lastError();
            break;
        }
    }
    res = res && db->commit();
    if(res)
    {
        return checkFk(false);
    }
    qCritical() << db->lastError();
    db->rollback();
    checkFk(false);
    return false;
}

bool CoreConfigDBDbInitializer::initCashiers(QSqlDatabase *db)
{
    QSqlQuery qc(*db);
    if(!qc.exec("select count(id) from cashiers") || !qc.first())
    {
        lmCritical() << qc.executedQuery() << qc.lastError();
        return false;
    }
    if(qc.value(0).toInt() > 0) return true;
    if(!db->transaction())
    {
        lmCritical() << db->lastError();
        return false;
    }
    QSqlQuery q(*db);
    q.prepare("insert into cashiers (id, serverId, login, name, pass, perms, inn)"
              "values (:id, :serverId, :login, :name, :pass, :perms, :inn)");

    q.bindValue(":id", Cashier::SUPER_USER_ID);
    q.bindValue(":serverId", QVariant());
    q.bindValue(":login", "su");
    q.bindValue(":name", "СИС. АДМИН");
    q.bindValue(":pass", "99");
    q.bindValue(":perms", static_cast<quint32>(Cashier::SU_PERMS));
    q.bindValue(":inn", QVariant());

    if(!q.exec())
    {
        lmCritical() << q.executedQuery() << logtab << q.lastQuery()  << q.lastError() << logtab
                    << logvariant(q.boundValues());
        db->rollback();
        return false;
    }
    if(!db->commit())
    {
        lmCritical() << db->lastError();
        db->rollback();
        return  false;
    }

    return true;
}

bool CoreConfigDBDbInitializer::initOfd(QSqlDatabase *db)
{
    QSqlQuery qc(*db);
    if(!qc.exec("select ofdSettingsVersion from metadata") || !qc.first())
    {
        lmCritical() << qc.executedQuery() << qc.lastError();
        return false;
    }
    if(qc.value(0).toUInt() >= OfdSettings::OFD_SETTINGS_VERSION) return true;
    if(!db->transaction())
    {
        lmCritical() << db->lastError();
        return false;
    }
    QSqlQuery q(*db);
    q.prepare("insert or replace into SUPPORTED_OFD (inn, address, port, dns, queryInterval, url,"
              " name, shortName, labelAddr, labelPort, keysAddr, keysPort, testTout, keysTout) values "
              "(:inn, :address, :port, :dns, :queryInterval, :url, :name, :shortName, "
              ":labelAddr, :labelPort, :keysAddr, :keysPort, :testTout, :keysTout)");

    for(auto it = OfdSettings::SUPPORTED_OFD.constBegin(); it != OfdSettings::SUPPORTED_OFD.constEnd(); ++it)
    {
        const OfdSettings &ofd = it.value();
        q.bindValue(":inn", ofd.inn());
        q.bindValue(":address", ofd.address());
        q.bindValue(":port", ofd.port());
        q.bindValue(":dns", ofd.dns());
        q.bindValue(":queryInterval", ofd.queryInterval());
        q.bindValue(":url", ofd.url());
        q.bindValue(":name", ofd.name());
        q.bindValue(":shortName", ofd.shortName());
        q.bindValue(":labelAddr", ofd.labelsAddress());
        q.bindValue(":labelPort", ofd.labelsPort());
        q.bindValue(":keysAddr", ofd.keysAddress());
        q.bindValue(":keysPort", ofd.keysPort());
        if(!q.exec())
        {
            lmCritical() << q.executedQuery() << logtab << q.lastQuery()  << q.lastError() << logtab
                        << logvariant(q.boundValues());
            db->rollback();
            return false;
        }
    }
    q.prepare("update metadata set ofdSettingsVersion = :ofdSettingsVersion");
    q.bindValue(":ofdSettingsVersion", OfdSettings::OFD_SETTINGS_VERSION);
    if(!q.exec())
    {
        lmCritical() << q.executedQuery() << logtab << q.lastQuery()  << q.lastError() << logtab
                    << logvariant(q.boundValues());
        db->rollback();
        return false;
    }
    if(!db->commit())
    {
        lmCritical() << db->lastError();
        db->rollback();
        return  false;
    }

    return true;
}
