#include "docsdb.h"

#include "cashboxbuildconfig.h"
#include "appdirectories.h"
#include "dircreator.h"
#include "logmanip.h"
#include "formatutils.h"
#include "formatutils.h"

#include <QThread>
#include <QSqlDatabase>
#include <QSqlQuery>
#include <QSqlError>
#include <QDebug>
#include <QFile>
#include <QDir>
#include <QJsonDocument>
#include <QJsonParseError>


DocsDb::ConnectionsCounter DocsDb::_rwopened;
DocsDb::ConnectionsCounter DocsDb::_roopened;

DocsDb::DocsDb(const QString &regNumber, QObject *parent)
    : BaseSqliteDbConnector("DocsDB-", &_rwopened, parent)
    , cashboxRegNumber_(regNumber.trimmed())
{
    init();
}

DocsDb::DocsDb(const QString &regNumber, BaseSqliteDbConnector::ConnectionMode mode, QObject *parent)
    : BaseSqliteDbConnector("DocsDB-", mode,
                            mode == CM_READ_ONLY ? &_roopened : &_rwopened,
                            parent)
    , cashboxRegNumber_(regNumber.trimmed())
{
    init();
}

DocsDb::~DocsDb()
{
    release();
}

bool DocsDb::resetDb()
{
    QStringList tables  = db_.tables();

    for(int i = tables.size() - 1; i >= 0; --i)
    {
        if(tables[i].toLower() == "metadata")
        {
            tables.removeAt(i);
            break;
        }
    }

    if(!db_.transaction())
    {
        qCritical() << db_.lastError();
        return false;
    }
    bool res =  resetDbTables(tables);


    if(res && !db_.commit())
    {
        lmCritical() << db_.lastError();
        db_.rollback();
        return  false;
    }
    return true;
}

bool DocsDb::saveDocument(const BaseDocDbSaver &doc)
{
    if(!db_.transaction())
    {
        lmCritical() << "TRANSACTION_ERROR"<< db_.lastError();
        saveError(doc, db_.lastError(), tr("TRANSACTION ERROR"));
        return false;
    }
    QVariant footerId, headerId, regDataId, placeId;
    //Результат не критичен. пропускаем далее
    saveDocText(BaseDocDbSaver::Footer, doc.footer(), footerId);
    saveDocText(BaseDocDbSaver::Header, doc.header(), headerId);
    if(!saveDocRegData(doc, regDataId) || !saveDocPlace(doc, placeId))
    {
        db_.rollback();
        return false;
    }
    QSqlQuery q(db_);
    q.prepare("insert into documents (docType, fd, cycle, dt, externalId, clientId, regdataId, "
              "headerId, footerId, placeDataId) values (:docType, :fd, :cycle, :dt, :externalId, "
              ":clientId, :regdataId, :headerId, :footerId, :placeDataId)");
    q.bindValue(":docType",     static_cast<qint32>(doc.type()));
    q.bindValue(":fd",          doc.fd());
    q.bindValue(":cycle",       doc.cycle());
    q.bindValue(":dt",          DT2STR_(doc.dt()));
    q.bindValue(":externalId",  doc.externalId());
    q.bindValue(":clientId",    doc.clientId());
    q.bindValue(":regdataId",   regDataId);
    q.bindValue(":headerId",    headerId);
    q.bindValue(":footerId",    footerId);
    q.bindValue(":placeDataId", placeId);
    if(!q.exec())
    {
        saveError(doc, q.lastError(), q.boundValues(), q.lastQuery());
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    QVariant docId = q.lastInsertId();
    q.prepare("insert into docsfields (docId, name, val) values (:docId, :name, :val)");
    q.bindValue(":docId", docId);
    QVariantList fields = doc.fields();
    if(!std::all_of(fields.constBegin(), fields.constEnd(), [&q](const QVariant &v){
            QVariantMap m = v.toMap();
            for(auto v = m.constBegin(); v != m.constEnd(); ++v)
            {
                q.bindValue(":name", v.key());
                q.bindValue(":val", v.value().toString());
                if(!q.exec()) return false;
            }
            return true;
        }))
    {
        saveError(doc, q.lastError(), q.boundValues(), q.lastQuery());
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        saveError(doc, db_.lastError(), tr("COMMIT ERROR"));
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    return true;
}

BaseDoc *DocsDb::loadDocument(quint32 fd)
{
    BaseDocDbSaver bds(nullptr);

    if(!db_.transaction())
    {
        lmCritical() << "TRANSACTION ERROR";
        return  nullptr;
    }
    QSqlQuery q(db_);
    if(!q.exec("select max(id) from docsregdata") || !q.first())
    {
        lmCritical() << q.lastError() << q.lastQuery();
        db_.rollback();
        return nullptr;
    }
    qint64 rdId = q.value(0).toLongLong();
    db_.commit() || db_.rollback();
    if(!loadBaseDocDbSaver(bds, QStringLiteral("(fd = %1) and (regdataId = %2)").arg(fd).arg(rdId))) return nullptr;
    BaseDoc *res = bds.create(nullptr);
    return res;
}

BaseDoc *DocsDb::loadDocument(const QString &extId)
{
    BaseDocDbSaver bds(nullptr);
    if(!loadBaseDocDbSaver(bds, QStringLiteral("(externalId = '%1')").arg(extId))) return nullptr;
    BaseDoc *res = bds.create(nullptr);
    return res;

}

QString DocsDb::databaseFileName() const
{
    if(cashboxRegNumber_.isEmpty()) return QString();
    return APATH.dbDir() +QStringLiteral("/%1.db").arg(cashboxRegNumber_);
}

QString DocsDb::connectionName() const
{
    return objectName();
}

bool DocsDb::getMetaData(qint32 &metadataVersion, BaseSqliteDbConnector::EMetadataState &state)
{
    Q_UNUSED(state)
    metadataVersion = 0;
    if(!db_.isOpen()) return false;
    QSqlQuery q(db_);
    if(!q.exec("select metadataVersion from metadata") )
    {
        qCritical() << q.lastError();
        return false;
    }
    if(!q.first())
    {
        qCritical() << tr("Касса не зарегистрирована");
        return false;
    }
    metadataVersion =  q.value(0).toInt();
    return true;
}

qint32 DocsDb::actualMetadataVersion() const
{
    return DB_METADATA_VERSION;
}

bool DocsDb::updateDatabase(qint32 from)
{
    DocsDBDbInitializer dbi;
    bool res = true;
    for(int i = from + 1; i <= DB_METADATA_VERSION;++i)
    {
        res = res && dbi.updateDatabase(&db_, i);
    }
    return res;
}

bool DocsDb::loadBaseDocDbSaver(BaseDocDbSaver &bds, const QString &filter)
{
    if(!db_.transaction())
    {
        lmCritical() << "TRANSACTION ERROR";
        return false;
    }
    QSqlQuery q(db_);
    if(!q.exec(QLatin1String("select id, docType, fd, cycle, dt, externalId, clientId, regdataId, headerId,"
                              " footerId, placeDataId from documents where ") + filter ) || !q.first())
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery();
        db_.rollback();
        return false;
    }
    QVariant id = q.value("id");
    bds.setType(static_cast<fdf::DocType>(q.value("docType").toUInt()));
    bds.setFd(q.value("fd").toUInt());
    bds.setCycle(q.value("cycle").toUInt());
    bds.setDt(STR2DT_(q.value("dt").toString()));
    bds.setExternalId(q.value("externalId").toString());
    bds.setClientId(q.value("clientId").toString());
    QVariant regDataId = q.value("regdataId");
    QVariant headerId = q.value("headerId");
    QVariant footerId = q.value("footerId");
    QVariant placeDataId = q.value("placeDataId");
    if(!loadDocRegData(bds, regDataId) || !loadDocPlace(bds, placeDataId) ||
        !loadDocText(bds, headerId) || !loadDocText(bds, footerId) || !loadFields(bds, id))
    {
        db_.rollback();
        return false;
    }
    return true;
}

bool DocsDb::loadDocText(BaseDocDbSaver &bds, const QVariant &id)
{
    if(!id.isValid() || id.isNull()) return true;
    QSqlQuery q(db_);
    q.prepare("select textType, txt from docstext where id = :id");
    q.bindValue(":id", id);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << logvariant(q.boundValues());
        return false;
    }
    if(q.first())
    {
        int type = q.value("textType").toInt();
        QString text = q.value("txt").toString();
        if(type  == BaseDocDbSaver::Footer)bds.setFooter(text);
        else if(type  == BaseDocDbSaver::Header)bds.setHeader(text);
        else
        {
            lmWarning() << "UNKNOWN TEXT" << logtab << text;
        }
    }
    return true;
}

bool DocsDb::loadDocRegData(BaseDocDbSaver &doc, const QVariant &id)
{
    if(!id.isValid() || id.isNull()) return false;
    QSqlQuery q(db_);
    q.prepare("select fsNumber, regNumber, ownerInn, ownerName, ffd from docsregdata where id = :id");
    q.bindValue(":id", id);
    if(!q.exec() || !q.first())
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << logvariant(q.boundValues());
        return false;
    }
    doc.setFsNumber(q.value("fsNumber").toString());
    doc.setRegNumber(q.value("regNumber").toString());
    doc.setOwnerInn(q.value("ownerInn").toString());
    doc.setOwnerName(q.value("ownerName").toString());
    doc.setFfd(static_cast<fs::FFD>(q.value("ffd").toUInt()));
    return true;
}

bool DocsDb::loadDocPlace(BaseDocDbSaver &doc, const QVariant &id)
{
    if(!id.isValid() || id.isNull()) return true;
    QSqlQuery q(db_);
    q.prepare("select address, place from docsplace where id = :id");
    q.bindValue(":id", id);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << logvariant(q.boundValues());
        return false;
    }
    if(q.first())
    {
        doc.setAddress(q.value("address").toString());
        doc.setPlace(q.value("place").toString());
    }
    return true;
}

bool DocsDb::loadFields(BaseDocDbSaver &doc, const QVariant &id)
{
    if(!id.isValid() || id.isNull()) return false;
    QSqlQuery q(db_);
    q.prepare("select name, val from docsfields where docId = :id");
    q.bindValue(":id", id);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << logvariant(q.boundValues());
        return false;
    }
    QVariantList fields;
    while(q.next())
    {
        fields <<  QVariantMap{
            {q.value("name").toString(), q.value("val").toString()}
        };
    }
    lmWarning() << logvariant(fields);
    doc.setFields(fields);
    return true;
}

bool DocsDb::saveDocText(qint32 docTextType, const QString &text, QVariant &id)
{
    id.clear();
    if(text.isEmpty()) return  true;
    QSqlQuery q(db_);
    q.prepare("select id from docstext where (textType = :textType) and (txt = :txt)");
    q.bindValue(":textType", docTextType);
    q.bindValue(":txt", text);
    if(q.exec() && q.first())
    {
        id = q.value("id");
        return true;
    }
    q.prepare("insert into docstext (textType, txt) values (:textType, :txt)");
    q.bindValue(":textType", docTextType);
    q.bindValue(":txt", text);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << logvariant(q.boundValues());
        return false;
    }
    id = q.lastInsertId();
    return true;
}

bool DocsDb::saveDocRegData(const BaseDocDbSaver &doc, QVariant &id)
{
    id.clear();
    QSqlQuery q(db_);
    q.prepare("select id from docsregdata where (fsNumber = :fsNumber) and (regNumber = :regNumber)"
              "and (ownerInn = :ownerInn) and (ownerName = :ownerName) and (ffd = :ffd)");
    q.bindValue(":fsNumber",  doc.fsNumber());
    q.bindValue(":regNumber", doc.regNumber());
    q.bindValue(":ownerInn",  doc.ownerInn());
    q.bindValue(":ownerName", doc.ownerName());
    q.bindValue(":ffd",       static_cast<qint32>(doc.ffd()));
    if(q.exec() && q.first())
    {
        id = q.value("id");
        return true;
    }
    q.prepare("insert into docsregdata (fsNumber, regNumber, ownerInn, ownerName, ffd)"
              "values (:fsNumber, :regNumber, :ownerInn, :ownerName, :ffd)");
    q.bindValue(":fsNumber",  doc.fsNumber());
    q.bindValue(":regNumber", doc.regNumber());
    q.bindValue(":ownerInn",  doc.ownerInn());
    q.bindValue(":ownerName", doc.ownerName());
    q.bindValue(":ffd",       static_cast<qint32>(doc.ffd()));
    if(!q.exec())
    {
        saveError(doc, q.lastError(), q.boundValues(), q.lastQuery());
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << logvariant(q.boundValues());
        return false;
    }
    id = q.lastInsertId();
    return true;
}

bool DocsDb::saveDocPlace(const BaseDocDbSaver &doc, QVariant &id)
{
    id.clear();
    QSqlQuery q(db_);
    q.prepare("select id from docsplace where (address = :address) and (place = :place)");
    q.bindValue(":address",  doc.address());
    q.bindValue(":place", doc.place());
    if(q.exec() && q.first())
    {
        id = q.value("id");
        return true;
    }
    q.prepare("insert into docsplace (address, place)"
              "values (:address, :place)");
    q.bindValue(":address",  doc.address());
    q.bindValue(":place", doc.place());

    if(!q.exec())
    {
        saveError(doc, q.lastError(), q.boundValues(), q.lastQuery());
        lmCritical() << q.lastError() << logtab << q.lastQuery() << logtab << logvariant(q.boundValues());
        return false;
    }
    id = q.lastInsertId();
    return true;
}

void DocsDb::saveError(const BaseDocDbSaver &doc, const QSqlError &error, const QVariantMap &bounds,
                       const QString &sql, const QString &descr)
{
    QString fn = APATH.dbErrorsDir() +
                 QStringLiteral("/%1_%2_%3.json")
                     .arg(doc.dt().toString("yyyy-MM-dd_hh_mm_ss"))
                     .arg(doc.fsNumber())
                     .arg(doc.fd());
    QVariantMap data;
    data["doc"] = doc.toMap();
    data["error"] = error.text();
    if(!bounds.isEmpty()) data["boundValues"] = bounds;
    if(!sql.isEmpty()) data["sql"] = sql;
    if(!descr.isEmpty()) data["descr"] = descr;
    QFile f(fn);
    if(f.open(QIODevice::WriteOnly))
    {
        f.write(QJsonDocument::fromVariant(data).toJson());
        f.close();
    }
    else
    {
        lmCritical() << fn << logtab << f.error() << f.errorString();
    }
}

void DocsDb::saveError(const BaseDocDbSaver &doc, const QSqlError &error, const QString &descr)
{
    saveError(doc, error, QVariantMap(), QString(), descr);
}


//--------------------------------------------------------------------------------------------------

DocsDBDbInitializer::DocsDBDbInitializer(QObject *parent)
    : QObject(parent)
{
    QString thrId = QString::number(reinterpret_cast<qint64>(QThread::currentThreadId()), 16);
    setObjectName(QString("DocsDBDbInitializer-")+ thrId);
}

DocsDBDbInitializer::~DocsDBDbInitializer()
{

}

bool DocsDBDbInitializer::initializeDatabase(const QString &regNumber)
{
    if(regNumber.trimmed().isEmpty()) return false;
    AppDirectories cfg;
    QString fileName = cfg.dbDir() +QStringLiteral("/%1.db").arg(regNumber.trimmed());
    lmWarning() << fileName;
    bool valid = false;
    if(QFile::exists(fileName))
    {
        DocsDb db(regNumber);
        QString msg;
        qint32 ver = 0;
        valid = db.db().isOpen() && db.checkVersion(ver, msg, true) == DocsDb::DATABASE_IS_VALID;
        if(valid)
        {
            QSqlQuery q(db.db());
            valid = q.exec("pragma integrity_check;");
            q.first();
        }
    }
    if(!valid)
    {
        QDir dir(cfg.dbDir());
        QFileInfo fi(fileName);
        QFileInfoList fil = dir.entryInfoList(QStringList() << fi.fileName() + "*", QDir::Files);

        for(const QFileInfo &f :fil)
        {
            QFile::remove(f.absoluteFilePath());
        }
    }
    lmWarning() << valid;
    if(!QFile::exists(fileName))
    {
        DocsDb dbc(regNumber);
        QSqlDatabase db = dbc.db();
        DocsDBDbInitializer dbi;
        bool res = dbi.createDatabase(&db, ":/docsdb/docsdb.sql");
        return res;
    }
    return true;
}

bool DocsDBDbInitializer::createDatabase(QSqlDatabase * const db, const QString &fileName)
{
    QString sql = getScript(fileName);
    if(sql.isEmpty() || !(db->isOpen() || db->open()))
    {
        QSqlError err = db->lastError();
        qCritical() << (err.isValid() ? err.text() : tr("Не задан скрипт для создания базы!"));
        return false;
    }
    db->exec("PRAGMA journal_mode = \"WAL\"");
//    lmCritical() << db->exec(QStringLiteral("").arg(CORE_CONFIG_DB_METADATA_VERSION)).lastError();
    return applyScript(db, sql);
}

bool DocsDBDbInitializer::updateDatabase(QSqlDatabase * const db, qint32 toMetaData)
{
    QString fileName = QString(":/docsdb/md_%1.sql").arg(toMetaData);
    QString sql = getScript(fileName);
    if(sql.isEmpty()) return true;
    if(!(db->isOpen() || db->open()))
    {
        QSqlError err = db->lastError();
        qCritical() << (err.isValid() ? err.text() : tr("Не задан скрипт для создания базы!"));
        return false;
    }
    return applyScript(db, sql);
}

QString DocsDBDbInitializer::getScript(const QString &fileName)
{
    QFile f(fileName);
    if(!f.exists() || !f.open(QIODevice::ReadOnly))
    {
        qCritical() << tr("Ошибка открытия файла %1 (%2)").arg(fileName).arg(f.errorString());
        return QString();
    }
    QTextStream ts(&f);
    ts.setCodec("UTF-8");
    QString res = ts.readAll();
    f.close();
    return res;
}

bool DocsDBDbInitializer::applyScript(QSqlDatabase * const db, const QString &sql)
{
    auto checkFk = [db](bool off) -> bool{
        QSqlQuery q(*db);
        bool res = q.exec(QStringLiteral("PRAGMA foreign_keys = %1").arg(off ? "off" : "on"));
        if(!res)
        {
            qCritical() << q.lastError() << logtab << q.lastQuery();
        }
        return res;
    };
    if(!checkFk(true))
    {
        return false;
    }
    bool res = db->transaction();
    QStringList script = sql.split(";;");
    QSqlQuery q(*db);

    for(QString s: script)
    {
        if(s.size() < 7) continue;
        res = res && q.exec(s);
        if(!res)
        {
            s.replace("\r\n","\n");
            qWarning() << loglist(s.split("\n"));
            qCritical() << q.lastError();
            break;
        }
    }
    res = res && db->commit();
    if(res)
    {
        return checkFk(false);
    }
    qCritical() << db->lastError();
    db->rollback();
    checkFk(false);
    return false;
}
