#include "frworkingthread.h"

#include "logmanip.h"
#include "abstractworkerthread.h"
#include "coreconfigdb.h"
#include "internalobjectsorage.h"
#include "androidapihelper.h"
#include "textprinterdocument.h"
#include "modeldatastorage.h"
#include "printdocumentformatter.h"
#include "fiscalizationcontroller.h"
#include "reportscontroller.h"
#include "docsdb.h"
#include "cashboxbuildconfig.h"

#include "hwmodeldatastorage.h"
#include "appdirectories.h"
#include "bbcodeparser.h"
#include "umkaauth.h"
#include "watcherdb.h"


#include <QTimer>
#include <QFile>
#include <QCoreApplication>
#include <QTextDecoder>
#include <QVariantList>

FrWorkingThread::FrWorkingThread()
    : AbstractWorker()
    , started_(false)
    , queue_(new ProcessingQueue())
    , loopTimer_(nullptr)
    , pdw_(nullptr)
    , printer_(nullptr)
    , fs_(nullptr)
    , receipts_(nullptr)
    , louncherTimer_(nullptr)
    , louncherCount_(0)


{
}

FrWorkingThread::~FrWorkingThread()
{
    delete queue_;
    queue_ = nullptr;
    doStop();
}

bool FrWorkingThread::doStart()
{
    started_ = false;
    doStop();
    lmWarning() << "START";


    fs_ = new FsWrapper(this);
    fs_->open();
    printer_ = new PrintersWrapper(this);
    pdw_ = new ProcessingDataWrapper(fs_, printer_, this);
    pdw_->watcher()->clean();
    connect(pdw_, &ProcessingDataWrapper::sendToOfd, this, &FrWorkingThread::sendToOfd);

    receipts_ = new ReceiptsController(fs_, pdw_, printer_, this);



    loopTimer_ =new QTimer(this);
    loopTimer_->setInterval(MAX_LOOP_INTERVAL);
    loopTimer_->setSingleShot(true);
    connect(loopTimer_, SIGNAL(timeout()), this, SLOT(execCmd()));
    loopTimer_->start();

    louncherCount_ = 0;
    louncherTimer_ = new QTimer(this);
    louncherTimer_->setInterval(10000);
    connect(louncherTimer_, &QTimer::timeout, this, &FrWorkingThread::sendSerialToLouncher);
    louncherTimer_->start();
    sendSerialToLouncher();

    return true;
}

bool FrWorkingThread::doStop()
{
    lmWarning() << "STOP";
    if(loopTimer_)
    {
        loopTimer_->blockSignals(true);
        loopTimer_->stop();
        delete loopTimer_;
        loopTimer_ = nullptr;
    }
    if(receipts_)
    {
        receipts_->deleteLater();
        receipts_ = nullptr;
    }
    if(printer_)
    {
        printer_->deleteLater();;
        printer_ = nullptr;
    }
    if(pdw_)
    {
        pdw_->deleteLater();
        pdw_ = nullptr;
    }
    if(louncherTimer_)
    {
        louncherTimer_->stop();
        delete louncherTimer_;
        louncherTimer_ = nullptr;
    }
    if(fs_)
    {
        fs_->close();
        fs_->deleteLater();
        fs_ = nullptr;
    }
    return true;
}

void FrWorkingThread::processTask(const CoreTransaction &task, CoreApiResult &result)
{
    lmWarning() << QString::number(static_cast<qint32>(task.operation()), 16);
    static const QSet<CoreApiConst::RemoteCommand> NEED_SEND_BROADCAST = {
        CoreApiConst::RemoteCommand::WRITE_SERIAL,
        CoreApiConst::RemoteCommand::FISCALIZATION,
        CoreApiConst::RemoteCommand::FS_CLEAN,
        CoreApiConst::RemoteCommand::OPEN_CYCLE,
        CoreApiConst::RemoteCommand::CLOSE_CYCLE,
        CoreApiConst::RemoteCommand::CALC_REPORT,
        CoreApiConst::RemoteCommand::FS_CLOSE,
        CoreApiConst::RemoteCommand::RECEIPT,
        CoreApiConst::RemoteCommand::CORRECTION,
        CoreApiConst::RemoteCommand::PRINT_NOT_PRINTED,
        CoreApiConst::RemoteCommand::FULL_CLEAR

    };
    if(task.operation() == CoreApiConst::RemoteCommand::FULL_CLEAR)
    {
        result = fullClear(task);
        return;
    }
    Cashier cashier;
    if(pdw_->watcher()->checkNotPrinted(task))
    {
        result = printNotPrinted(task);
    }
    if(pdw_->watcher()->find(task, result))
    {
        //NOTE: Специальный код ошибки для отметки, что была попытка дубля. пока уберу. буду возвращать ок а то у клиентов поломается
//        result.setCode(CoreApiConst::ErrorCode::ExternalIdAndTaskDublicated);

        return;
    }
    if(!checkAccess(task, result, cashier)) return;
    if(!checkModelData(task, result)) return;

    else switch (task.operation())
    {
    case CoreApiConst::RemoteCommand::GET_CASHIER_NAMES: result = getCashierNames();break;
    case CoreApiConst::RemoteCommand::LOGIN: result = getLogin(task);break;
    case CoreApiConst::RemoteCommand::CASHBOX_STATUS: result = getCashboxStatus(task);break;
    case CoreApiConst::RemoteCommand::WRITE_SERIAL: result = writeSerial(task);break;
    case CoreApiConst::RemoteCommand::FISCALIZATION: result = fiscalize(task);break;
    case CoreApiConst::RemoteCommand::PREPRINT_FISCALIZATION: result = fiscalizePreprint(task);break;
    case CoreApiConst::RemoteCommand::FS_CLEAN: result = fsClean(task);break;
    case CoreApiConst::RemoteCommand::FS_COUNTERS: result = fsCounters(task);break;
    case CoreApiConst::RemoteCommand::OPEN_CYCLE: result = receipts_->openCycle(task);break;
    case CoreApiConst::RemoteCommand::CALC_REPORT: result = receipts_->calcReport(task);break;
    case CoreApiConst::RemoteCommand::CLOSE_CYCLE: result = receipts_->closeCycle(task);break;
    case CoreApiConst::RemoteCommand::RECEIPT: result = receipts_->receipt(task, cashier);break;
    case CoreApiConst::RemoteCommand::CORRECTION: result = receipts_->correction(task, cashier);break;
    case CoreApiConst::RemoteCommand::CHECK_LABEL: result = receipts_->checkLabel(task);break;
    case CoreApiConst::RemoteCommand::CLEAN_LABELS: result = receipts_->cleanLabels(task);break;
    case CoreApiConst::RemoteCommand::OFD_2_FS_CMD: result = ofdFsCommand(task);break;
    case CoreApiConst::RemoteCommand::FIND_FS_DOC:
        result = ReportsController(fs_, pdw_, printer_, this).findFsDocument(task);
        break;
    case CoreApiConst::RemoteCommand::FIND_OFD_TICKET:
        result = ReportsController(fs_, pdw_, printer_, this).findOfdTicket(task);
        break;
    case CoreApiConst::RemoteCommand::GET_TLV_DOC:
        result = ReportsController(fs_, pdw_, printer_, this).tlvDocument(task);
        break;
    case CoreApiConst::RemoteCommand::GET_REG_ARCHIVE_DOC:
        result = ReportsController(fs_, pdw_, printer_, this).regArchiveDocument(task);
        break;
    case CoreApiConst::RemoteCommand::GET_REG_TLV_DOC:
        result = ReportsController(fs_, pdw_, printer_, this).regArchiveTlvDocument(task);
        break;
    case CoreApiConst::RemoteCommand::FS_CLOSE: result = closefs(task);break;
    case CoreApiConst::RemoteCommand::LOAD_OFD_SETTINGS: result = downloadOfdSettings(task);break;
    case CoreApiConst::RemoteCommand::SAVE_OFD_SETTINGS: result = saveOfdSettings(task);break;
    case CoreApiConst::RemoteCommand::GET_OFFLINE_NOTIFICATIONS: result = getOfflineNotifications(task);break;
    case CoreApiConst::RemoteCommand::DOWNLOAD_CASHIERS: result = downloadCashiers(task);break;
    case CoreApiConst::RemoteCommand::SAVE_CASHIER: result = saveCashier(task);break;
    case CoreApiConst::RemoteCommand::REMOVE_CASHIER: result = removeCashier(task);break;
    case CoreApiConst::RemoteCommand::SELF_TEST: result = selfTest(task);break;
    case CoreApiConst::RemoteCommand::PRINT_NOT_PRINTED: result = printNotPrinted(task);break;
    case CoreApiConst::RemoteCommand::GET_CASHBOX_MODELDATA: result = getCashboxModelData(task);break;
    case CoreApiConst::RemoteCommand::PRINT_TEXT: result = printText(task);break;
   case CoreApiConst::RemoteCommand::X_REPORT:
        result = ReportsController(fs_, pdw_, printer_, this).xreport(task, cashier);break;
   case CoreApiConst::RemoteCommand::FS_REPORT:
       result = ReportsController(fs_, pdw_, printer_, this).fsreport(task, cashier);break;
   case CoreApiConst::RemoteCommand::PRINT_DOC_COPY:
       result = ReportsController(fs_, pdw_, printer_, this).copyDocument(task);
       break;
    case CoreApiConst::RemoteCommand::LOAD_CORE_SETTINGS: result = loadConfig(task);break;
    case CoreApiConst::RemoteCommand::SAVE_CORE_SETTINGS: result = saveConfig(task);break;
   default: result = CoreApiResult(CoreApiConst::ErrorCode::UnsupportedApiCommand);break;
    }
    if(NEED_SEND_BROADCAST.contains(task.operation()) && pdw_) pdw_->sendBroadcastCashboxStatus();
    //    fs_->frcoreStatus();
}

bool FrWorkingThread::checkAccess(const CoreTransaction &task, CoreApiResult &result, Cashier &cashier)
{
    static const QSet<CoreApiConst::RemoteCommand> SKIP_ALL_CHECK = {
        CoreTransaction::Operation::GET_CASHIER_NAMES,
        CoreTransaction::Operation::CASHBOX_STATUS,
        CoreTransaction::Operation::LOGIN,
        CoreTransaction::Operation::OFD_2_FS_CMD,
        CoreTransaction::Operation::SELF_TEST,
        CoreTransaction::Operation::GET_CASHBOX_MODELDATA,
    };

    static const QSet<CoreApiConst::RemoteCommand> ANY_PERMISSION = {
        CoreTransaction::Operation::FS_COUNTERS,
        CoreTransaction::Operation::LOAD_OFD_SETTINGS,
        CoreTransaction::Operation::PRINT_NOT_PRINTED,
        CoreTransaction::Operation::PRINT_TEXT,
        CoreTransaction::Operation::LOAD_CORE_SETTINGS
    };

    static const QMap<CoreApiConst::RemoteCommand, Cashier::Permission> PERMISSIONS = {
        std::make_pair(CoreApiConst::RemoteCommand::WRITE_SERIAL, Cashier::Permission::WriteSerial),
        std::make_pair(CoreApiConst::RemoteCommand::PREPRINT_FISCALIZATION, Cashier::Permission::Fiscalization),
        std::make_pair(CoreApiConst::RemoteCommand::FISCALIZATION, Cashier::Permission::Fiscalization),
        std::make_pair(CoreApiConst::RemoteCommand::FS_CLEAN, Cashier::Permission::Fiscalization),
        std::make_pair(CoreApiConst::RemoteCommand::OPEN_CYCLE, Cashier::Permission::CycleOpen),
        std::make_pair(CoreApiConst::RemoteCommand::CALC_REPORT, Cashier::Permission::Reports),
        std::make_pair(CoreApiConst::RemoteCommand::CLOSE_CYCLE, Cashier::Permission::CycleClose),
        std::make_pair(CoreApiConst::RemoteCommand::FIND_FS_DOC, Cashier::Permission::Reports),
        std::make_pair(CoreApiConst::RemoteCommand::FIND_OFD_TICKET, Cashier::Permission::Reports),
        std::make_pair(CoreApiConst::RemoteCommand::GET_TLV_DOC, Cashier::Permission::Reports),
        std::make_pair(CoreApiConst::RemoteCommand::PRINT_DOC_COPY, Cashier::Permission::Reports),
        std::make_pair(CoreApiConst::RemoteCommand::GET_REG_ARCHIVE_DOC, Cashier::Permission::Reports),
        std::make_pair(CoreApiConst::RemoteCommand::GET_REG_TLV_DOC, Cashier::Permission::Reports),
        std::make_pair(CoreApiConst::RemoteCommand::FS_CLOSE, Cashier::Permission::Fiscalization),
        std::make_pair(CoreApiConst::RemoteCommand::SAVE_OFD_SETTINGS, Cashier::Permission::Settings),
        std::make_pair(CoreApiConst::RemoteCommand::GET_OFFLINE_NOTIFICATIONS,
        Cashier::Permission::OfflineNotifications),
        std::make_pair(CoreApiConst::RemoteCommand::DOWNLOAD_CASHIERS,
        Cashier::Permission::Settings),
        std::make_pair(CoreApiConst::RemoteCommand::SAVE_CASHIER,
        Cashier::Permission::Settings),
        std::make_pair(CoreApiConst::RemoteCommand::REMOVE_CASHIER,
        Cashier::Permission::Settings),
        std::make_pair(CoreApiConst::RemoteCommand::X_REPORT,
                       Cashier::Permission::Reports),
        std::make_pair(CoreApiConst::RemoteCommand::FS_REPORT,
                       Cashier::Permission::Reports),
        std::make_pair(CoreApiConst::RemoteCommand::SAVE_CORE_SETTINGS,
                       Cashier::Permission::Settings),
    };

    if(SKIP_ALL_CHECK.contains(task.operation())) return true;
    QString msg;
    CoreApiConst::ErrorCode err = CoreApiConst::ErrorCode::Ok;
    if(!task.isValid(err, &msg))
    {
        result.setCode(err).setDescr(msg);
        return false;
    }
    Cashier c;
    if(!getCashier(task.cashierLogin(), task.cashierPassword(), c) || c.password() != task.cashierPassword())
    {
        result.setCode(CoreApiConst::ErrorCode::AccessDenied)
                .setDescr(tr("Доступ запрещен. Неверный номер кассира или пароль"));
        return false;
    }
    cashier = c;

    if(ANY_PERMISSION.contains(task.operation())) return true;

    lmWarning() << QString::number(static_cast<qint32>(task.operation()), 16)
                << c.hasOneReceiptPermission() << c.hasOneCorrectionPermission()
                << QString::number(static_cast<quint32>(c.permissions()), 2);

    if(task.operation() == CoreTransaction::Operation::LOGIN) return true;

    if(task.operation() == CoreTransaction::Operation::RECEIPT )
    {
        if(c.hasOneReceiptPermission()) return true;
        result.setCode(CoreApiConst::ErrorCode::AccessDenied)
                .setDescr(tr("Доступ запрещен. Не достаточно прав"));
        return false;

    }
    if(task.operation() == CoreTransaction::Operation::CORRECTION)
    {
        if(c.hasOneCorrectionPermission()) return true;
        result.setCode(CoreApiConst::ErrorCode::AccessDenied)
                .setDescr(tr("Доступ запрещен. Не достаточно прав"));
        return false;

    }

    if(task.operation() == CoreTransaction::Operation::CHECK_LABEL ||
        task.operation() == CoreTransaction::Operation::CLEAN_LABELS)
    {
        if(c.hasOneCorrectionPermission() || c.hasOneReceiptPermission()) return true;
        result.setCode(CoreApiConst::ErrorCode::AccessDenied)
                .setDescr(tr("Доступ запрещен. Не достаточно прав"));
        return false;
    }

    if(!PERMISSIONS.contains(task.operation()) ||
            !c.hasPermission(PERMISSIONS[task.operation()]))
    {
        result.setCode(CoreApiConst::ErrorCode::AccessDenied)
                .setDescr(tr("Доступ запрещен. Не достаточно прав"));
        return false;
    }
    return true;
}

bool FrWorkingThread::getCashier(qint32 id, Cashier &cashier)
{
    return pdw_? pdw_->getCashier(id, cashier) : false;
}

bool FrWorkingThread::getCashier(const QString &login, const QString &password, Cashier &cashier)
{
    return pdw_? pdw_->getCashier(login, password, cashier) : false;
}

Cashier FrWorkingThread::getCashier(const QString &login, const QString &password)
{
    Cashier c;
    if(pdw_) pdw_->getCashier(login, password, c);
    return c;
}



void FrWorkingThread::execCmd()
{
    procesFrStart();

    CoreTransaction task;

    loopTimer_->stop();
    if(!queue_->dequeue(task))
    {
//        checkNotSendedZ();
//        if(fmDC_)
//        {
//            QElapsedTimer t;
//            t.start();
//            fmDC_->exec();
//            if(t.elapsed() > 1000) lmWarning() << t.elapsed();
//        }
        if(loopTimer_)
        {
            loopTimer_->setSingleShot(true);
            loopTimer_->start(MAX_LOOP_INTERVAL);
        }
    }
    else
    {
        lmWarning() << task.uid() << QString::number(static_cast<qint32>(task.operation()));
        QElapsedTimer testtime;
        testtime.start();
        CoreApiResult result;
        processTask(task, result);
        lmWarning() << "PROCESSED " << testtime.elapsed() << " / " << static_cast<qint32>(task.operation())
                    << "/"<< static_cast<qint32>(result.code());
        result.setUid(task.uid());
        queue_->storeResult(result);
        lmWarning() << "STORED " <<testtime.elapsed();

        //NOTE: пока перезапуск после очистки вставил сюда. а там будем посмотреть. выглядит коряво. needRestart?
        if(task.operation() == CoreTransaction::Operation::FULL_CLEAR && result.isOk())
        {
            doStop();
            doStart();
        }


        //TODO: Возможно обновить статус ФМ
        lmWarning() << "FS STATUS PROCESSED " <<testtime.elapsed();
        if(loopTimer_)
        {
            loopTimer_->setSingleShot(true);
            loopTimer_->start(MIN_LOOP_INTERVAL);
        }
    }
}

bool FrWorkingThread::procesFrStart()
{
    if(!started_)
    {
        started_ = true;

        FiscalizationController(fs_, pdw_, printer_, this).syncRegData();
        RegData rd = pdw_->getRegData();
        PrintDocumentFormatter f(rd);
        pdw_->getFsFullStatus(FsFullStatus::CLEAN_ALL);
        CashboxStatus st = pdw_->getCashboxStatus();
        TextPrinterDocument doc = f.welcome(st);
        lmWarning() << st.fs().isFiscalized() << rd.regNumber();
        if(st.fs().isFiscalized())
        {
            DocsDBDbInitializer::initializeDatabase(rd.regNumber());
        }

        pdw_->reloadOfd();
        printer_->printTextDocument(doc);


//        lmWarning() << ">>>>>>>>>>>>>>>>>>TEST>>>ARCHIVE";

//        for(int i = 1; i <= st.fs().status().lastFd(); ++i)
//        {
//            PFsArchiveReportBase d;
//            fs_->findDocument(i, d, QStringList());
//            if(d) lmWarning() << logvariant(d->toExternalMap());
//            else lmWarning() << i << " IS NULL";
//        }

//        lmWarning() << "<<<<<<<<<<<<<<<<<TEST<<<<ARCHIVE";

    }
    return true;
}


bool FrWorkingThread::checkModelData(const CoreTransaction &task, CoreApiResult &result)
{
    static const QSet<CoreApiConst::RemoteCommand> SKIP_ALL_CHECK = {
        CoreTransaction::Operation::GET_CASHIER_NAMES,
        CoreTransaction::Operation::LOGIN,
        CoreTransaction::Operation::CASHBOX_STATUS,
        CoreTransaction::Operation::FS_CLEAN,
        CoreTransaction::Operation::FS_COUNTERS,
        CoreTransaction::Operation::OFD_2_FS_CMD,
        CoreTransaction::Operation::LOAD_OFD_SETTINGS,
        CoreTransaction::Operation::SAVE_OFD_SETTINGS,
        CoreTransaction::Operation::DOWNLOAD_CASHIERS,
        CoreTransaction::Operation::SAVE_CASHIER,
        CoreTransaction::Operation::REMOVE_CASHIER,
        CoreTransaction::Operation::PRINT_NOT_PRINTED,
        CoreTransaction::Operation::GET_CASHBOX_MODELDATA,
        CoreTransaction::Operation::PRINT_TEXT,
        CoreTransaction::Operation::LOAD_CORE_SETTINGS,
        CoreTransaction::Operation::SAVE_CORE_SETTINGS
    };
    if(!pdw_ || !started_)
    {
        result = CoreApiResult(CoreApiConst::ErrorCode::ServiceDoesNotStarted, tr("Сервис еще не запущен"));
        return false;
    }

    if(SKIP_ALL_CHECK.contains(task.operation())) return true;

    if(AndroidApiHelper().cashbox105Installed())
    {
        result.setCode(CoreApiConst::ErrorCode::OldCashboxInstalled);
        return false;

    }

    if(task.operation() == CoreTransaction::Operation::WRITE_SERIAL)
    {
        AndroidApiHelper h;
        IOSG::cleanModelData();
        if(pdw_ && pdw_->getModelData().checkSign(h.getHardwareHash()))
        {
            result.setCode(CoreApiConst::ErrorCode::SerialExists)
                    .setDescr(tr("Касса уже имеет заводской номер: ") + IOSG::getModelData().serial());
            return false;
        }
        return true;
    }

    ModelData md = IOSG::getModelData();
    FsFullStatus fs = pdw_->getFsFullStatus();

//    lmWarning() << fs.isRelease() << fs.fsNumber() << fs.version() << logtab << logvariant(fs.toExternalMap());



    if(fs.isFiscalized() && !md.serial().isEmpty() &&
            !fs.cashboxSerial().isEmpty() && md.serial() != fs.cashboxSerial())
    {
        result = CoreApiResult(CoreApiConst::ErrorCode::AlienFs,
                               tr("Подключенный ФН зарегистрирован с ККТ %1. А ЗН этой ККТ - %2")
                               .arg(fs.cashboxSerial()).arg(md.serial()));
        return false;
    }
    //NOTE: тут выходит преждеврееный ответ на ошибку работы без фн? переделать или оставит?
    if(fs.supportedFfd() != fs::FFD::FFD1_2 && task.operation() != CoreTransaction::Operation::WRITE_SERIAL)
    {
        if(fs.error() == CoreApiConst::ErrorCode::Ok)
        {
            result = CoreApiResult(CoreApiConst::ErrorCode::UnsupportedFsFFDVersion,
                                   tr("Данная версия ККТ работает только с фн, поддерживающим ФФД 1.2"));
        }
        else
        {
            result = CoreApiResult(fs.error(), CoreApiConst::defaultErrorMsg(fs.error()));
        }
        return false;

    }
    if(DEBUG_BUILD_VERSION && fs.isRelease() && !fs.fsNumber().trimmed().isEmpty())
    {
        if(fs.error() == CoreApiConst::ErrorCode::Ok)
        {
            lmWarning() <<fs.fsNumber();
            result = CoreApiResult(CoreApiConst::ErrorCode::AlienFs,
                                   tr("Настоящая версия ККТ может работать только с МГМ. Подключен боевой ФН %1")
                                       .arg(fs.fsNumber()));
        }
        else
        {
            result = CoreApiResult(fs.error(), CoreApiConst::defaultErrorMsg(fs.error()));
        }
        return false;

    }

    if(fs.isFiscalized() && fs.currentFfd() != fs::FFD::FFD1_2 &&
            task.operation() != CoreTransaction::Operation::FISCALIZATION &&
            task.operation() != CoreTransaction::Operation::PREPRINT_FISCALIZATION &&
            task.operation() != CoreTransaction::Operation::FS_CLOSE &&
            task.operation() != CoreTransaction::Operation::CLOSE_CYCLE &&
            task.operation() != CoreTransaction::Operation::WRITE_SERIAL)
    {
        if(fs.error() == CoreApiConst::ErrorCode::Ok)
        {
            result = CoreApiResult(CoreApiConst::ErrorCode::UnsupportedFFDVersion,
                                   tr("Данная версия ККТ работает только ФФД 1.2. Необходимо выполнить перерегистрацию ККТ для поддержки ФФД 1.2"));
        }
        else
        {
            result = CoreApiResult(fs.error(), CoreApiConst::defaultErrorMsg(fs.error()));
        }
        return false;
    }

    if(task.operation() != CoreTransaction::Operation::WRITE_SERIAL)
    {
        AndroidApiHelper h;
        IOSG::cleanModelData();
        if(pdw_ && !pdw_->getModelData().checkSign(h.getHardwareHash()))
        {
            result.setCode(CoreApiConst::ErrorCode::NoCashboxSerial)
                    .setDescr(tr("Не задан ЗН ККТ"));
            return false;
        }
    }

    switch (task.operation()) {
    case CoreTransaction::Operation::WRITE_SERIAL:
    {
        AndroidApiHelper h;
        IOSG::cleanModelData();
        if(pdw_ && pdw_->getModelData().checkSign(h.getHardwareHash()))
        {
            result.setCode(CoreApiConst::ErrorCode::SerialExists)
                    .setDescr(tr("Касса уже имеет заводской номер: ") + IOSG::getModelData().serial());
            return false;
        }
        return true;
    }break;
    case CoreTransaction::Operation::FISCALIZATION:
        case CoreTransaction::Operation::PREPRINT_FISCALIZATION:
        case CoreTransaction::Operation::FS_CLOSE:
    {
        CashboxStatus cs = pdw_->getCashboxStatus();
        if(cs.fs().isFiscalized() && cs.fs().cycleIsOpen())
        {
            result.setCode(CoreApiConst::ErrorCode::CycleIsOpened)
                    .setDescr(tr("Перед перерегистрации ККТ необходимо закрыть смену"));
            return false;
        }
        return true;
    }break;
    case CoreTransaction::Operation::GET_OFFLINE_NOTIFICATIONS:
    {
        RegData rd = pdw_->getRegData();
        if(!rd.getOfflineMode())
        {
            result.setCode(CoreApiConst::ErrorCode::OfflineModeOnly)
                    .setDescr(tr("Команда выполняется только в автономном режиме"));
            return false;
        }
        return true;
    }break;
    case CoreTransaction::Operation::OPEN_CYCLE:
    case CoreTransaction::Operation::CALC_REPORT:
    case CoreTransaction::Operation::CLOSE_CYCLE:
    case CoreTransaction::Operation::RECEIPT:
    case CoreTransaction::Operation::CORRECTION:
    case CoreTransaction::Operation::CHECK_LABEL:
    case CoreTransaction::Operation::CLEAN_LABELS:
        return receipts_->checkState(task, result);
    case CoreTransaction::Operation::FIND_FS_DOC:
    case CoreTransaction::Operation::FIND_OFD_TICKET:
    case CoreTransaction::Operation::GET_TLV_DOC:
    case CoreTransaction::Operation::PRINT_DOC_COPY:
    case CoreTransaction::Operation::GET_REG_ARCHIVE_DOC:
    case CoreTransaction::Operation::GET_REG_TLV_DOC:
    case CoreTransaction::Operation::SELF_TEST:
    case CoreTransaction::Operation::X_REPORT:
    case CoreTransaction::Operation::FS_REPORT:
    {
        return true;
    }break;
    default:
        break;
    }
   //TODO: Проверка доступа по ЗН
    result.setCode(CoreApiConst::ErrorCode::UnsupportedApiCommand)
            .setDescr(tr("Обработчик команды не реализован"));
    return false;
}

void FrWorkingThread::sendSerialToLouncher()
{
    ModelData md = pdw_->getModelData();
    if(md.isValid())
    {
        if(AndroidApiHelper().sendAboutToLouncher(md.serial(), md.modelName(), md.model()) &&
            louncherTimer_)
        {
            if(louncherTimer_ && louncherTimer_->interval() == 10000 && louncherCount_ >=30)
            {
                louncherTimer_->stop();
                louncherTimer_->setInterval(60000);
                louncherCount_ = 0;
                louncherTimer_->start();
            }
            else if(louncherTimer_ && louncherTimer_->interval() == 60000 && louncherCount_ >= 100)
            {
                louncherTimer_->stop();
                louncherCount_ = 0;
            }
            louncherCount_ ++;
        }
    }
}

CoreApiResult FrWorkingThread::getCashierNames() const
{
    CoreConfigDB dbc(CoreConfigDB::CM_READ_ONLY, nullptr);
    QVariantList names;
    dbc.loadCashierNames(names);
    QVariantMap res;
    res.insert("cashiers", names);
    return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(), res);
}

CoreApiResult FrWorkingThread::getLogin(const CoreTransaction &task)
{
    Cashier cashier;
    lmWarning() << task.cashierLogin() << task.cashierPassword();
    if(getCashier(task.cashierLogin(), task.cashierPassword(), cashier))
    {
        return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(),
                             task.formatToExternal()?cashier.toExternalMap(false) : cashier.toMap());
    }
    return CoreApiResult(CoreApiConst::ErrorCode::AccessDenied, tr("Неверный пароль"));
}

CoreApiResult FrWorkingThread::getCashboxStatus(const CoreTransaction &task)
{
    Q_UNUSED(task)


    if(task.formatToExternal()) IOSG::cleanPrinterStatus();

    CashboxStatus status = pdw_->getCashboxStatus();

    return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(),
                         task.formatToExternal() ? status.toExternalMap() : status.toMap());
}

CoreApiResult FrWorkingThread::getCashboxModelData(const CoreTransaction &task)
{
    Q_UNUSED(task)

    ModelData md = pdw_->getModelData();

    return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(),
                         task.formatToExternal() ? md.toExternalMap() : md.toMap());
}

CoreApiResult FrWorkingThread::writeSerial(const CoreTransaction &task)
{
    ModelData md = pdw_->getModelData();
    QString serial = task.params()["serial"].toString().trimmed();
    QString sign = task.params()["sign"].toString().trimmed();
    bool ok = false;
    if(serial.length() != md.SERIAL_LENGTH || serial.toLongLong(&ok) <= 0 || !ok)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::SerialWriteError,
                             tr("Некорректный заводской номер"));
    }
    md.setSerial(serial);
    md.setSign(sign);
    if(!md.checkSign(AndroidApiHelper().getHardwareHash()))
    {
        return CoreApiResult(CoreApiConst::ErrorCode::SerialWriteError,
                             tr("Ошибка проверки подписи"));
    }
    ModelDataStorage mds;
    mds.setModelData(md);
    mds.setRegData(RegData());
    IOSG::cleanRegData();
    IOSG::cleanModelData();
    md = pdw_->getModelData();
    if(!md.checkSign(AndroidApiHelper().getHardwareHash()))
    {
        return CoreApiResult(CoreApiConst::ErrorCode::SerialWriteError,
                             tr("Ошибка записи заводского номера"));
    }
    if(md.isValid())
    {
        AndroidApiHelper().sendAboutToLouncher(md.serial(), md.modelName(), md.model());
    }

    FiscalizationController(fs_, pdw_, printer_, this).syncRegData();
    printer_->printTextDocument(PrintDocumentFormatter(pdw_->getRegData()).writeSerial(pdw_->getCashboxStatus()));
    return CoreApiResult(CoreApiConst::ErrorCode::Ok,
                         QString(), QVariantMap());
}

CoreApiResult FrWorkingThread::fiscalize(const CoreTransaction &task)
{

    return FiscalizationController(fs_, pdw_, printer_, this).fiscalize(task);
}

CoreApiResult FrWorkingThread::fiscalizePreprint(const CoreTransaction &task)
{
    return FiscalizationController(fs_, pdw_, printer_, this).preprint(task);
}

CoreApiResult FrWorkingThread::closefs(const CoreTransaction &task)
{
    return FiscalizationController(fs_, pdw_, printer_, this).closefs(task);
}

CoreApiResult FrWorkingThread::fsCounters(const CoreTransaction &task)
{
    FsTotalType type = static_cast<FsTotalType>(task.params()["type"].toInt());
    if(type != FsTotalType::CycleTotal) type = FsTotalType::FsTotalType;
    FsCounters counters;
    CoreApiConst::ErrorCode err = fs_->readFsCounters(type, counters);
    if(err != CoreApiConst::ErrorCode::Ok) return CoreApiResult(err);
    return CoreApiResult(err, QString(), task.formatToExternal() ? counters.toExternalMap() : counters.toMap());
}

CoreApiResult FrWorkingThread::fsClean(const CoreTransaction &task)
{
    Q_UNUSED(task)
    QString version;
    bool isRelease = false;
    CoreApiConst::ErrorCode err = fs_->getFsVersion(version, isRelease);
    lmWarning() << version << isRelease << static_cast<quint32>(err);
    if(err == CoreApiConst::ErrorCode::Ok)
    {
        if(isRelease) err= CoreApiConst::ErrorCode::UnknownFsError;
        else err  = fs_->cleanDebugFs();
        lmWarning() << version << isRelease << static_cast<quint32>(err);
    }
    IOSG::cleanFs();

//COMMENT ME
//    ModelDataStorage mds;
//    mds.cleanAll();
//    IOSG::cleanRegData();
//    IOSG::cleanModelData();
//----------
    pdw_->getCashboxStatus();
    return CoreApiResult(err);
}

CoreApiResult FrWorkingThread::downloadCashiers(const CoreTransaction &task)
{
    //NOTE: можно сделать через параметр выбор кассиров по реквизитам
    Q_UNUSED(task)
    CoreConfigDB dbc(CoreConfigDB::CM_READ_ONLY);
    QVariantList cashiers;
    if(!task.params().contains("cashierId"))
    {
        if(!dbc.loadCashiers(false, true, true, cashiers))
        {
            return CoreApiResult(CoreApiConst::ErrorCode::DatabaseLoadingError,
                                 tr("Ошибка загрузки данных"));
        }
    }
    else
    {
        if(!dbc.loadCashiers(task.params().value("cashierId").toInt(), false, true, true, cashiers))
        {
            return CoreApiResult(CoreApiConst::ErrorCode::DatabaseLoadingError,
                                 tr("Ошибка загрузки данных"));
        }

    }
    QVariantMap res;
    if(!task.formatToExternal())
    {
        res.insert("cashiers", cashiers);
    }
    else
    {
        QVariantList out;
        for(const QVariant &v : cashiers)
        {
            out << Cashier(v.toMap()).toExternalMap(true);
        }
        res.insert("cashiers", out);
    }
    return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(), res);

}

CoreApiResult FrWorkingThread::saveCashier(const CoreTransaction &task)
{
    CoreConfigDB dbc;
    Cashier c (task.params());
    IOSG::cleanActiveCashiers();
    QVariantMap out;
    if(!dbc.saveCashier(c, out) || out.isEmpty())
    {
        return CoreApiResult(CoreApiConst::ErrorCode::DatabaseSavingError,
                             tr("Ошибка сохранения данных кассира"));

    }
    if(task.formatToExternal())
    {
        out = Cashier(out).toExternalMap(true);
    }
    return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(), out);
}

CoreApiResult FrWorkingThread::removeCashier(const CoreTransaction &task)
{
    CoreConfigDB dbc;
    Cashier c (task.params());
    IOSG::cleanActiveCashiers();

    qint32 id =  task.params()["cashierId"].toInt();

    if(!dbc.removeCashier(id))
    {
        return CoreApiResult(CoreApiConst::ErrorCode::DatabaseSavingError,
                             tr("Ошибка удаления данных кассира"));

    }

    return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(), task.params());
}


CoreApiResult FrWorkingThread::downloadOfdSettings(const CoreTransaction &task)
{
    Q_UNUSED(task)
    OfdSettings ofd = pdw_->reloadOfd();
    return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(), ofd.toMap());

}

CoreApiResult FrWorkingThread::saveOfdSettings(const CoreTransaction &task)
{
    OfdSettings ofd (task.params());
    CoreConfigDB dbc;
    if(!dbc.saveCurrentOfd(ofd, true))
    {
        return CoreApiResult(CoreApiConst::ErrorCode::DatabaseSavingError, tr("Ошибка сохранения настроек ОФД"));
    }
    ofd = pdw_->reloadOfd();
    pdw_->getFsFullStatus(FsFullStatus::CLEAN_CYCLE);
    return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(), ofd.toMap());
}

CoreApiResult FrWorkingThread::ofdFsCommand(const CoreTransaction &task)
{
    FsFullStatus st = pdw_->getFsFullStatus();
//    lmWarning() << "OFD_CMD_START" << logtab << logvariant(task.toMap())
//                << logtab << logvariant(st.toMap());
    OfdFsCmd cmd(task.params());
    if(!cmd.isValid()) return CoreApiResult(CoreApiConst::ErrorCode::UnknownFsCmd);
    FsOfdResponse resp = fs_->ofdCmd(cmd);
    if(cmd.cmdIsTransport())
    {
        st = pdw_->getFsFullStatus(FsFullStatus::CLEAN_TRANSPORT);
    }
    if(cmd.cmdIsLabelTransport())
    {
        st = pdw_->getFsFullStatus(FsFullStatus::CLEAN_LABELS_TRANSPORT);
    }

//    lmWarning() << "OFD_CMD_STOP"<< logtab<< logvariant(st.toMap()) << logtab
//                << logvariant(task.toMap()) << logtab
//                << logvariant(resp.toMap());
    return resp.toResult();
}

CoreApiResult FrWorkingThread::getOfflineNotifications(const CoreTransaction &task)
{
    OfflineDSessionStatus status;
    CoreApiConst::ErrorCode err = fs_->d3(true, status);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmCritical() << CoreApiConst::defaultErrorMsg(err);
        return CoreApiResult(err);
    }
    lmWarning() << logvariant(status.toMap());
    if(status.notSendedCount() == 0u)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::NoDataRequested,
                             tr("В ФН нет уведомлений для выгрузки"));
    }
    QMap<quint32, quint16> crcMap;
    QString fn = APATH.offlineNotifications() + "/" + task.externalId();
    QFile f(fn);
    if(!f.open(QIODevice::WriteOnly))
    {
        return CoreApiResult(CoreApiConst::ErrorCode::OfflineFileCreationError,
                             tr("Ошибка создания файла: ") + f.errorString());
    }
    QByteArray header;
    QDataStream ds(&header, QIODevice::WriteOnly);
    ds.setByteOrder(QDataStream::LittleEndian);
    QString sbuf = "Отчет о реализации маркированного товара";
    QTextEncoder cp866(QTextCodec::codecForName("CP866"));
    QByteArray buf = cp866.fromUnicode(sbuf);
    if(buf.size() < 66) buf.append(QByteArray(66 - buf.size(), ' '));
    ds.writeRawData(buf.constData(), buf.size());
    {
        ModelData md = pdw_->getModelData();
        buf = cp866.fromUnicode(md.modelName() + ", версия " + md.version());
        if(buf.size() < 256) buf.append(QByteArray(256 - buf.size(), ' '));
        ds.writeRawData(buf.constData(), buf.size());
    }
    {
        RegData rd = pdw_->getRegData();
        buf = cp866.fromUnicode(rd.regNumber());
        if(buf.size() < 20) buf.append(QByteArray(20 - buf.size(), ' '));
        ds.writeRawData(buf.constData(), buf.size());
    }
    FsFullStatus fs = pdw_->getFsFullStatus();
    {
        buf = cp866.fromUnicode(fs.fsNumber().trimmed());
        if(buf.size() < 16) buf.append(QByteArray(16 - buf.size(), ' '));
        ds.writeRawData(buf.constData(), buf.size());
    }
    quint32 docsCount = status.sessionNotSendedCount();
    ds << static_cast<quint8>(4) << status.fistSessionNotSended() << static_cast<quint32>(0)
       << static_cast<quint32>(status.sessionNotSendedCount()) << static_cast<quint32>(0);

    f.write(header);
    Crc16CITT crc;
    Crc32 crc32;
    crc32.startCalculating();
    bool first = true;
    quint32 last = status.fistSessionNotSended();
    quint32 firstDoc = status.fistSessionNotSended();
    while(status.sessionNotSendedCount() > 0u)
    {
        QByteArray data;
        quint16 len = 0;
        quint32 num = 0;
        err = fs_->d4(!first, len, num);
        if(num > 0 ) last = num;

        if(err != CoreApiConst::ErrorCode::Ok)
        {
            if(err == CoreApiConst::ErrorCode::NoDataRequested) break;
            lmCritical() << CoreApiConst::defaultErrorMsg(err);
            f.close();
            QFile::remove(fn);
            return CoreApiResult(err);
        }
        for(quint16 offcet = 0; offcet < len; offcet += 1024)
        {
            quint16 sz = qMin(1024, len - offcet);
            QByteArray buf;
            quint16 invalidTag = 0;
            err = fs_->d5(offcet, sz, buf, invalidTag);
            if(err != CoreApiConst::ErrorCode::Ok)
            {
                QString msg = CoreApiConst::defaultErrorMsg(err);
                if(invalidTag > 0) msg += QString(" [ТЭГ: %1]").arg(invalidTag);
                lmCritical() << msg;
                return CoreApiResult(err);
            }            
            data.append(buf);
        }

        crcMap.insert(num, crc(data.mid(0, 2) + data.mid(4)));

        QByteArray hdata;
        QDataStream hdds(&hdata, QIODevice::WriteOnly);
        hdds.setByteOrder(QDataStream::LittleEndian);
        hdds << static_cast<quint16>(201) << static_cast<quint16>(data.size());
        data = hdata + data;

        f.write(data);
        crc32.addData(data);
        err = fs_->d3(false, status);
        first = false;
        lmWarning() << logvariant(status.toMap());
    }
    f.flush();
    header.clear();
    QDataStream sds(&header, QIODevice::WriteOnly);
    sds.setByteOrder(QDataStream::LittleEndian);
    sds << last << crc32.finishCalculating();
    f.seek(363);
    f.write(header.mid(0, 4));
    f.seek(371);
    f.write(header.mid(4, 4));
    f.close();
    lmWarning() << logbinary(header);
    lmWarning() << last;

    quint32 num = 0;
    quint16 notConfirmed = 0;
    err = fs_->d6_0(notConfirmed, num);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmCritical() << CoreApiConst::defaultErrorMsg(err);
        QFile::remove(fn);
        return CoreApiResult(err);
    }
    while(notConfirmed > 0)
    {
        err = fs_->d6_1(num, crcMap[num]);
        if(err != CoreApiConst::ErrorCode::Ok)
        {
            lmCritical() << CoreApiConst::defaultErrorMsg(err);
            if(err == CoreApiConst::ErrorCode::InvalidCmdParams) continue;
            QFile::remove(fn);
            return CoreApiResult(err);
        }
        err = fs_->d6_0(notConfirmed, num);
        if(err != CoreApiConst::ErrorCode::Ok)
        {
            lmCritical() << CoreApiConst::defaultErrorMsg(err);
            QFile::remove(fn);
            return CoreApiResult(err);
        }
    }


    QVariantMap res;
    res["fileName"] = fn;
//    QString baseName = QStringLiteral("Arc_%1_%2_%3.crpt").arg(fs.fsNumber()).arg(firstDoc).arg(last);
    QString baseName = QStringLiteral("%1_%2%3.fnm").arg(fs.fsNumber()).arg(firstDoc).arg(docsCount);
    res["baseName"] = baseName;
    return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(), res);
}

CoreApiResult FrWorkingThread::selfTest(const CoreTransaction &task)
{
    Q_UNUSED(task)
    CoreApiConst::ErrorCode err = printer_->test();
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        return CoreApiResult(err);
    }
    FsFullStatus fs = pdw_->getFsFullStatus(FsFullStatus::CLEAN_ALL);
    if(fs.error() != CoreApiConst::ErrorCode::Ok)
    {
        return CoreApiResult(fs.error());
    }

    PrintDocumentFormatter pdf(pdw_->getRegData());
    TextPrinterDocument doc = pdf.test(pdw_->getCashboxStatus());
    lmWarning() << doc.blocks().size();
    if(!printer_->printTextDocument(doc))
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(printer_->lastErrorCode());
        return CoreApiResult(printer_->lastErrorCode());
    }

    return CoreApiResult(CoreApiConst::ErrorCode::Ok);
}

CoreApiResult FrWorkingThread::printNotPrinted(const CoreTransaction &task)
{
    PrintDocWatcher *watcher = pdw_->watcher();
    lmWarning() << QString::number(static_cast<qint32>(task.operation())) << logvariant(task.toMap());
    if(task.operation() != CoreTransaction::Operation::PRINT_NOT_PRINTED)
    {
        CoreApiResult result;
        TextPrinterDocument printable;
        if(!watcher->find(task, result, printable))
        {
            return  result;
        }


        if((result.code() == CoreApiConst::ErrorCode::Ok) || printer_->printTextDocument(printable))
        {
            watcher->finalize(task.externalId(), CoreApiConst::ErrorCode::Ok);
        }
        result.setCode(printer_->lastErrorCode());
        return  result;
    }
    else
    {
        CoreApiResult result;
        TextPrinterDocument printable;
        QString externalId;
        if(!watcher->restoreLast(externalId, result, printable))
        {
            lmWarning() << externalId << CoreApiConst::defaultErrorMsg(result.code())
                        << static_cast<qint32>(result.code())
                        << logvariant(result.toMap());
            return  result;
        }
        lmWarning() << externalId << CoreApiConst::defaultErrorMsg(result.code())
                    << logvariant(result.toMap());
        // Не печатаем, если не было ошибок печати. в вотчере костыль при восстановлении
        if((result.code() == CoreApiConst::ErrorCode::Ok) || printer_->printTextDocument(printable))
        {
            watcher->finalize(externalId, CoreApiConst::ErrorCode::Ok);
        }
        result.setCode(printer_->lastErrorCode());
        return  result;
    }
}

CoreApiResult FrWorkingThread::printText(const CoreTransaction &task)
{
    if(task.params().contains("bbcode"))
    {
        lmWarning() << logvariant(task.params());
        BbCodeParser parser;
        parser.setImagesPath(APATH.imagesPath());
        parser.parseText(task.params().value("bbcode").toString());
        TextPrinterDocument doc = parser.document();
        lmWarning() << logvariant(doc.toMap());
        if(printer_->printTextDocument(doc)) return CoreApiResult(CoreApiConst::ErrorCode::Ok);
        return CoreApiResult(printer_->lastErrorCode());

    }
    return CoreApiResult(CoreApiConst::ErrorCode::InvalidParams);
}

CoreApiResult FrWorkingThread::fullClear(const CoreTransaction &task)
{
    CoreApiResult result;
    UmkaAuth *auth = new UmkaAuth(this);
    if(!auth->syncLogin(task.cashierLogin(), task.cashierPassword()))
    {
        result = CoreApiResult(CoreApiConst::ErrorCode::AccessDenied, auth->lastError());
    }
    else
    {
        umka365::SessionData sd = auth->lastData();
        if(sd.dealerId() != FULL_CLEAR_AUTH_DEALER_ID)
        {
            result = CoreApiResult(CoreApiConst::ErrorCode::AccessDenied, tr("Доступ разрешен только для пользователей кабинета разработчика"));
        }
        else
        {
            //FIXME: сюда надо впихнуть очистку нварм, если есть, и баз
            ModelDataStorage mds;
            mds.cleanAll();
            IOSG::cleanRegData();
            IOSG::cleanModelData();
            mds.cleanAll();
            WhatcherDb wdb(WhatcherDb::CM_READ_WRITE);
            wdb.resetDb();
            result = CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(), auth->lastData().toMap());
        }
    }
    auth->deleteLater();
    return result;
}

CoreApiResult FrWorkingThread::loadConfig(const CoreTransaction &task)
{
    return pdw_->coreConfig()->processTransaction(task);
}

CoreApiResult FrWorkingThread::saveConfig(const CoreTransaction &task)
{
    return pdw_->coreConfig()->processTransaction(task);
}





