#include "ofdworker.h"

#include "logmanip.h"
#include "internalobjectsorage.h"
#include "ofdmsgcontainer.h"

#include <QEventLoop>
#include <QLoggingCategory>

Q_DECLARE_LOGGING_CATEGORY(OFD)
Q_LOGGING_CATEGORY(OFD, "ofd")

OfdWorker::OfdWorker()
    : AbstractWorker()
    , queue_(new ProcessingQueue())
    , loopTimer_(nullptr)
    , socket_(nullptr)
{

}

OfdWorker::~OfdWorker()
{
    delete queue_;
}

void OfdWorker::send()
{
    lmCWarning(OFD) << "SEND";
    if(loopTimer_) loopTimer_->stop();
    loop();
}

bool OfdWorker::doStart()
{
    doStop();
    loopTimer_ = new QTimer(this);
    connect(loopTimer_, &QTimer::timeout, this, &OfdWorker::loop);
    loopTimer_->setInterval(LOOP_INTERVAL);
    loopTimer_->start();
    socket_ = new SyncSocketWrapper("ofd", this);
    return true;
}

bool OfdWorker::doStop()
{
    if(loopTimer_)
    {
        loopTimer_->stop();
        loopTimer_->deleteLater();
        loopTimer_ = nullptr;
    }
    if(socket_)
    {
        socket_->disconnectFromHost();
        socket_->deleteLater();
        socket_ = nullptr;
    }
    return true;
}

FsOfdResponse OfdWorker::exec(const OfdFsCmd &cmd)
{
    if(!cmd.isValid())
    {
        FsOfdResponse res(cmd);
        res.setError(CoreApiConst::ErrorCode::UnknownFsCmd);
        return res;
    }
    CoreTransaction transaction = cmd.transaction();
    QEventLoop loop;
    auto resultWaiter = [&loop] () ->void {
        QTimer::singleShot(5, &loop, SLOT(quit()));
    };
    bool savedToQueue =  queue_->enqueue(transaction, resultWaiter);
    lmCWarning(OFD) << savedToQueue;

    CoreApiResult res;
    if(savedToQueue)
    {
        loop.exec();
        res = queue_->takeResult(transaction.uid());
    }
    else
    {
        FsOfdResponse resp(cmd);
        resp.setError(CoreApiConst::ErrorCode::InvalidTransactionType);
        return resp;
    }
    FsOfdResponse resp(res.data());
    return resp;
}

bool OfdWorker::syncConnectedState(bool connected, const FsTransportState &st)
{
    if(connected || socket_)
    {
        lmCWarning(OFD) << connected << st.ofdConnected() << socket_;
    }
    if(!connected)
    {
        if(socket_)socket_->disconnectFromHost();
        if(st.ofdConnected())sendConnectionStatus(false);
        return true;
    }
    if(!socket_) return false;
    OfdSettings ofd = IOSG::getOfd();
    lmCWarning(OFD) << logvariant(ofd.toMap());
    if(!ofd.isValid()) return false;
    socket_->setExchangeTimeout(ofd.queryInterval());
    return (socket_->connectedToHost() || socket_->connectToHost(ofd.address(), ofd.port())) &&
            (st.ofdConnected() || sendConnectionStatus(true));
}

QByteArray OfdWorker::readMsg(QTcpSocket *socket) const
{
    if(!socket) return QByteArray();
    QByteArray res;
    qint32 len = 30;
    QElapsedTimer readTimer;
    readTimer.start();
    while(socket && res.size() < len && readTimer.elapsed() < 10000)
    {
        res.append(socket->read(len - res.size()));
        if(len == 30 && res.size() >= 30)
        {
            QDataStream ds(res.mid(24, 2));
            ds.setByteOrder(QDataStream::LittleEndian);
            quint16 buf = 0;
            ds >> buf;
            len += buf;
        }
    }
    lmCWarning(OFD) << len << logbinary(res);
    return res;
}

void OfdWorker::loop()
{
    if(loopTimer_) loopTimer_->stop();
    ReadFunc func = [this](QTcpSocket *socket) -> QByteArray {
        return readMsg(socket);
    };
    FsFullStatus st = IOSG::getFs();
//    lmCWarning(OFD) << st.hasTranspot() << st.transport().hasMessageForOfd() << st.transport().messagesCount();
    while (st.hasTranspot() && st.transport().hasMessageForOfd()/*st.transport().messagesCount() > 0*/)
    {
        QByteArray msg;
        if(!syncConnectedState(true, st.transport()) || !readOfdMessage(msg, st.fsNumber())) break;
        msg = socket_->sendMsg(msg, func);
        if(msg.isEmpty() || !writeOfdTicket(msg, st.fsNumber())) break;
        st = IOSG::getFs();
    }
    st = IOSG::getFs();
    syncConnectedState(false, st.transport());
//    lmCWarning(OFD) << "BROKE " << loopTimer_;
    if(loopTimer_)loopTimer_->start();
}

bool OfdWorker::sendConnectionStatus(bool connected)
{
    OfdFsCmd cmd(OfdFsCmd::Code::SendOfdConnectionState);
    cmd.setOfdConnected(connected);
    FsOfdResponse resp = exec(cmd);
    lmCWarning(OFD) << logvariant(resp.toMap());
    return resp.error() == CoreApiConst::ErrorCode::Ok;
}

bool OfdWorker::readOfdMessage(QByteArray &msg, const QString &fsNumber)
{
    OfdFsCmd cmd(OfdFsCmd::Code::ReadOfdMsg);
    FsOfdResponse resp = exec(cmd);
    lmCWarning(OFD) << logvariant(resp.toMap());
    if(resp.error() == CoreApiConst::ErrorCode::Ok)
    {
        OfdMsgContainer c(0x0016u, fsNumber.trimmed(), resp.ofdMessage());
        msg = c.serialize();
        lmCWarning(OFD) << fsNumber << logbinary(msg);
    }
    return !msg.isEmpty();
}

bool OfdWorker::cancelOfdMessage()
{
    OfdFsCmd cmd(OfdFsCmd::Code::CancelOfdMsgReading);
    FsOfdResponse resp = exec(cmd);
    lmCWarning(OFD) << logvariant(resp.toMap());
    return resp.error() == CoreApiConst::ErrorCode::Ok;
}

bool OfdWorker::writeOfdTicket(const QByteArray &msg, const QString &fsNumber)
{
    OfdMsgContainer c(fsNumber, msg);
    if(c.container().isEmpty()) return false;
    OfdFsCmd cmd(OfdFsCmd::Code::SendOfdReceipt);
    cmd.setOfdTicket(c.container());
    FsOfdResponse resp = exec(cmd);
    lmCWarning(OFD) << logvariant(resp.toMap()) << logtab << logbinary(c.container()) << logtab
                    << fsNumber << c.fsNumber();
    return resp.error() == CoreApiConst::ErrorCode::Ok;
}
