#include "oismworker.h"

#include "logmanip.h"
#include "internalobjectsorage.h"
#include "ofdmsgcontainer.h"

#include <QEventLoop>
#include <QLoggingCategory>

Q_DECLARE_LOGGING_CATEGORY(OISM)
Q_LOGGING_CATEGORY(OISM, "oism")

OismWorker::OismWorker()
    : AbstractWorker()
    , queue_(new ProcessingQueue())
    , loopTimer_(nullptr)
    , socket_(nullptr)
{

}

OismWorker::~OismWorker()
{
    delete queue_;
}

void OismWorker::send()
{
    lmCWarning(OISM) << "SEND";
    if(loopTimer_) loopTimer_->stop();
    loop();
}

bool OismWorker::doStart()
{
    doStop();
    loopTimer_ = new QTimer(this);
    connect(loopTimer_, &QTimer::timeout, this, &OismWorker::loop);
    loopTimer_->setInterval(LOOP_INTERVAL);
    loopTimer_->start();
    socket_ = new SyncSocketWrapper("oism", this);
    return true;
}

bool OismWorker::doStop()
{
    if(loopTimer_)
    {
        loopTimer_->stop();
        loopTimer_->deleteLater();
        loopTimer_ = nullptr;
    }
    if(socket_)
    {
        socket_->disconnectFromHost();
        socket_->deleteLater();
        socket_ = nullptr;
    }
    return true;
}

FsOfdResponse OismWorker::exec(const OfdFsCmd &cmd)
{
    if(!cmd.isValid())
    {
        FsOfdResponse res(cmd);
        res.setError(CoreApiConst::ErrorCode::UnknownFsCmd);
        return res;
    }
    CoreTransaction transaction = cmd.transaction();
    QEventLoop loop;
    auto resultWaiter = [&loop] () ->void {
        QTimer::singleShot(5, &loop, SLOT(quit()));
    };
    bool savedToQueue =  queue_->enqueue(transaction, resultWaiter);
    lmCWarning(OISM) << savedToQueue;

    CoreApiResult res;
    if(savedToQueue)
    {
        loop.exec();
        res = queue_->takeResult(transaction.uid());
    }
    else
    {
        FsOfdResponse resp(cmd);
        resp.setError(CoreApiConst::ErrorCode::InvalidTransactionType);
        return resp;
    }
    FsOfdResponse resp(res.data());
    return resp;
}

bool OismWorker::syncConnectedState(bool connectToHost)
{

    if(connectToHost || socket_)
    {
        lmCWarning(OISM) << connectToHost  << socket_;
    }
    if(!connectToHost)
    {
        if(socket_)socket_->disconnectFromHost();
        return true;
    }
    if(!socket_) return false;
    OfdSettings ofd = IOSG::getOfd();
    lmCWarning(OISM) << logvariant(ofd.toMap());
    if(!ofd.isValid()) return false;
    socket_->setExchangeTimeout(ofd.queryInterval());
    return (socket_->connectedToHost() || socket_->connectToHost(ofd.labelsAddress(), ofd.labelsPort()));
}

QByteArray OismWorker::readMsg(QTcpSocket *socket) const
{
    if(!socket) return QByteArray();
    QByteArray res;
    qint32 len = 30;
    QElapsedTimer readTimer;
    readTimer.start();
    while(socket && res.size() < len && readTimer.elapsed() < 10000)
    {
        res.append(socket->read(len - res.size()));
        if(len == 30 && res.size() >= 30)
        {
            QDataStream ds(res.mid(24, 2));
            ds.setByteOrder(QDataStream::LittleEndian);
            quint16 buf = 0;
            ds >> buf;
            len += buf;
        }
    }
    lmCWarning(OISM) << logbinary(res);
    return res;
}

void OismWorker::loop()
{
    if(loopTimer_) loopTimer_->stop();
    ReadFunc func = [this](QTcpSocket *socket) -> QByteArray {
        return readMsg(socket);
    };
    FsFullStatus st = IOSG::getFs();
//    lmCWarning(OISM) << st.hasLabelsTranspot() << st.labelsTransport().notificationCount();
    while (st.hasLabelsTranspot() && st.labelsTransport().notificationCount() > 0)
    {
        QByteArray msg;
        if(!syncConnectedState(true) || !readOismMessage(msg, st.fsNumber())) break;
        msg = socket_->sendMsg(msg, func);
        lmWarning() << logbinary(msg);
        if(msg.isEmpty() || !writeOismTicket(msg, st.fsNumber())) break;
        st = IOSG::getFs();
    }
    st = IOSG::getFs();
    syncConnectedState(false);
//    lmCWarning(OISM) << "BROKE " << loopTimer_;
    if(loopTimer_)loopTimer_->start();
}


bool OismWorker::readOismMessage(QByteArray &msg, const QString &fsNumber)
{
    OfdFsCmd cmd(OfdFsCmd::Code::ReadNotification);
    FsOfdResponse resp = exec(cmd);
    lmCWarning(OISM) << logvariant(resp.toMap());
    if(resp.error() == CoreApiConst::ErrorCode::Ok)
    {
        LabelNotifyContainer c(0x0016u, fsNumber.trimmed(), resp.ofdMessage());
        msg = c.serialize();
        lmCWarning(OISM) << fsNumber << logbinary(msg);
    }
    return !msg.isEmpty();
}

bool OismWorker::cancelOismMessage()
{
    OfdFsCmd cmd(OfdFsCmd::Code::CancelNofification);
    FsOfdResponse resp = exec(cmd);
    lmCWarning(OISM) << logvariant(resp.toMap());
    return resp.error() == CoreApiConst::ErrorCode::Ok;
}

bool OismWorker::writeOismTicket(const QByteArray &msg, const QString &fsNumber)
{
    LabelNotifyContainer c(fsNumber, msg);
    if(c.container().isEmpty()) return false;
    OfdFsCmd cmd(OfdFsCmd::Code::WriteNotificationTicket);
    cmd.setOfdTicket(c.container());
    FsOfdResponse resp = exec(cmd);
    lmCWarning(OISM) << logvariant(resp.toMap()) << logtab << logbinary(c.container()) << logtab
                    << fsNumber << c.fsNumber();
    return resp.error() == CoreApiConst::ErrorCode::Ok;
}
