#include "processingdatawrapper.h"

#include "modeldatastorage.h"
#include "internalobjectsorage.h"
#include "logmanip.h"
#include "androidapihelper.h"
#include "coreconfigdb.h"
#include "docsdb.h"

#include <QNetworkInterface>
#include <QtConcurrent>

ProcessingDataWrapper::ProcessingDataWrapper(FsWrapper *fs, PrintersWrapper *printer, QObject *parent)
    : QObject(parent)
    , fs_(fs)
    , printer_(printer)
    , watcher_(new PrintDocWatcher(this, fs))
    , coreConfig_(new CoreConfigStore(this))
{

}

ProcessingDataWrapper::~ProcessingDataWrapper()
{

}

ModelData ProcessingDataWrapper::getModelData() const
{
    ModelData res = IOSG::getModelData();
    if(!res.isValid())
    {
        ModelDataStorage md(nullptr);
        res = md.getModelData();
        if(res.isValid()) IOSG::setModelData(res);
    }
    return res;
}


DeviceStatus ProcessingDataWrapper::getDeviceStatus() const
{
    DeviceStatus info = IOSG::deviceStatus();
    if(!info.isValid())
    {
        AndroidApiHelper h;
        info.setDsn(h.getDsn());
        info.setHwHash(h.getHardwareHash());
        info.setVersionCode(h.versionCode());
        info.setVersionText(h.versionName());
        info.setAccumLevel(h.getAccumLevel());
        info.setIpAddress(getIpAddress());
        if(info.isValid())
        {
            IOSG::setDeviceStatus(info);
        }
    }
    if(!IOSG::printerStatus().has_value())
    {
        IOSG::setPrinterStatus(printer_->test());
    }
    info.setPrinterStatus(static_cast<qint32>(
        IOSG::printerStatus().value_or(CoreApiConst::ErrorCode::Ok)));
    return info;
}

FsFullStatus ProcessingDataWrapper::getFsFullStatus(FsFullStatus::CleanFlags f) const
{
    FsFullStatus status = IOSG::getFs();
//    lmWarning() << f << logvariant(status.toExternalMap());
    if(status.error() != CoreApiConst::ErrorCode::Ok) f = FsFullStatus::CLEAN_ALL;
    status.clean(f);
//    lmWarning() << f << logvariant(status.toExternalMap());
    RegData rd = readRegData();
    fs_->getFullFsStatus(status, rd.flags(), rd.extFlags().flags());
    IOSG::setFs(status);

//    lmWarning() << f << logvariant(IOSG::getFs().toExternalMap());
    return status;
}


ActiveCashiers ProcessingDataWrapper::activeCashiers() const
{
    if(!IOSG::hasActiveCashiers())
    {
        CoreConfigDB dbc(CoreConfigDB::CM_READ_ONLY, nullptr);
        QList<Cashier> cashiers;
        if(dbc.loadCashiers(true, true, true, cashiers))
        {
            IOSG::setActiveCashiers(cashiers);
        }
    }
    return IOSG::activeCashiers();
}

CashboxStatus ProcessingDataWrapper::getCashboxStatus() const
{
    CashboxStatus res;
    res.setDev(getDeviceStatus());
    res.setModelData(getModelData());
    res.setFs(getFsFullStatus());
    res.setRegData(getRegData());
    res.setHasNotPrinted(watcher()->hasNotPrinted());
    return res;
}

RegData ProcessingDataWrapper::getRegData(bool fromFs) const
{
    RegData res = IOSG::getRegData();
    if(!res.isValid())
    {
        ModelDataStorage md(nullptr);
        res = md.getRegData();
        if(res.isValid()) IOSG::setRegData(res);
    }
    FsFullStatus st = getFsFullStatus();
    if(fromFs && st.isFiscalized())
    {
        RegData rd;
        CoreApiConst::ErrorCode err = fs_->readRegData(st.lifeTime().regsCount(), rd);
        if(err == CoreApiConst::ErrorCode::Ok)
        {
            res = rd;
            ModelDataStorage md;
            md.setRegData(rd);
            if(res.isValid()) IOSG::setRegData(res);
        }
    }
    return res;
}

OfdSettings ProcessingDataWrapper::reloadOfd() const
{
    if(!getCashboxStatus().isFiscalized())
    {
        IOSG::cleanOfd();
        return OfdSettings();
    }
    CoreConfigDB dbc(CoreConfigDB::CM_READ_ONLY, nullptr);
    OfdSettings ofd;
    dbc.loadCurrentOfd(getRegData().ofdInn(), ofd);
    lmWarning() << logvariant(ofd.toMap());
    IOSG::setOfd(ofd);
    return ofd;
}

OfdSettings ProcessingDataWrapper::getOfd() const
{
    if(!getCashboxStatus().isFiscalized())
    {
        IOSG::cleanOfd();
        return OfdSettings();
    }
    if(!IOSG::getOfd().isValid())
    {
        CoreConfigDB dbc(CoreConfigDB::CM_READ_ONLY, nullptr);
        OfdSettings ofd;
        dbc.loadCurrentOfd(getRegData().ofdInn(), ofd);
        lmWarning() << logvariant(ofd.toMap());
        IOSG::setOfd(ofd);
    }
    return IOSG::getOfd();
}




bool ProcessingDataWrapper::getCashier(qint32 id, Cashier &cashier)
{
    cashier = Cashier();

    if(!IOSG::hasActiveCashiers())
    {
        CoreConfigDB dbc(CoreConfigDB::CM_READ_ONLY, this);
        QList<Cashier> cashiers;
        if(dbc.loadCashiers(true, true, true, cashiers))
        {
            IOSG::setActiveCashiers(cashiers);
        }
    }
    return IOSG::getCashier(id, cashier);
}

bool ProcessingDataWrapper::getCashier(const QString &login, const QString &password, Cashier &cashier)
{
    cashier = Cashier();
    if(!IOSG::hasActiveCashiers())
    {
        CoreConfigDB dbc(CoreConfigDB::CM_READ_ONLY, this);
        QList<Cashier> cashiers;
        if(dbc.loadCashiers(true, true, true, cashiers))
        {
            IOSG::setActiveCashiers(cashiers);
        }
    }
    return IOSG::getCashier(login, password, cashier);
}

void ProcessingDataWrapper::saveDocument(BaseDoc *doc, const CoreTransaction &trans)
{
    saveDocument(doc, trans.externalId(), trans.clientId());
}

void ProcessingDataWrapper::saveDocument(BaseDoc *doc, const QString &externalId, const QString &clientId)
{
    emit sendToOfd();
    //Пойдет для отправки на сервер
    doc->setExternalId(externalId);
    doc->setClientId(clientId);

    BaseDocDbSaver cont(doc);
    RegData rd = getRegData();
    //NOTE: как пойдет в поток? по идее должно скопировать
    auto save = [rd, cont]() -> void {
        DocsDb db(rd.regNumber());
        db.saveDocument(cont);
    };
    QtConcurrent::run(save);

}

PrintDocWatcher *ProcessingDataWrapper::watcher() const
{
    return watcher_;
}

void ProcessingDataWrapper::sendBroadcastCashboxStatus() const
{
    static const QString action = "ru.armax.cashbox012.status";
    CashboxStatus st = getCashboxStatus();
    AndroidApiHelper().sendBroadCast(action, "status", st.toExternalMap());
    AndroidApiHelper().sendBroadCast(action + "_internal", "status", st.toMap());
}

CoreConfigStore *ProcessingDataWrapper::coreConfig() const
{
    return coreConfig_;
}

QString ProcessingDataWrapper::getIpAddress() const
{
    QList<QNetworkInterface> interfaces = QNetworkInterface::allInterfaces();
    QStringList outRes;
    for(const QNetworkInterface &i :interfaces)
    {
        if(i.flags().testFlag(QNetworkInterface::IsLoopBack) ||
                !i.flags().testFlag(QNetworkInterface::IsUp)) continue;
        QList<QNetworkAddressEntry> entries = i.addressEntries();
        for(const QNetworkAddressEntry &ae: entries)
        {
            if(!ae.ip().isNull() && !ae.ip().isLoopback())
            {
                bool ok = false;
                ae.ip().toIPv4Address(&ok);
                if(ok && ae.ip().toString().contains(QRegExp("(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}"
                                                             "(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)")) &&
                        !ae.ip().toString().startsWith("169.254."))
                {
                    outRes << ae.ip().toString();
                }
            }
        }
    }
    return outRes.isEmpty() ? QString() : outRes.join(",");
}

RegData ProcessingDataWrapper::readRegData() const
{
    RegData res = IOSG::getRegData();
    if(!res.isValid())
    {
        ModelDataStorage md(nullptr);
        res = md.getRegData();
        if(res.isValid()) IOSG::setRegData(res);
    }
    return res;
}
