#include "receiptscontroller.h"

#include "printdocumentformatter.h"
#include "logmanip.h"
#include "internalobjectsorage.h"
#include "cycleopendoc.h"
#include "calcreportdoc.h"
#include "cycleclosedoc.h"
#include "fsmsg.h"
#include "receipt.h"
#include "receiptdoc.h"
#include "keysupdater.h"
#include "labelcheckdata.h"
#include "labelschecker.h"
#include "watcherator.h"
#include "watcherdb.h"

ReceiptsController::ReceiptsController(FsWrapper *fs,  ProcessingDataWrapper *dw,
                                       PrintersWrapper *p, QObject *parent)
    : BaseTaskProcessor(fs, dw, p, parent)
{

}

ReceiptsController::~ReceiptsController()
{

}

bool ReceiptsController::checkState(const CoreTransaction &task, CoreApiResult &result)
{
    if(!BaseTaskProcessor::checkState(task, result)) return false;

    CashboxStatus st = pdw_->getCashboxStatus();

    if(st.hasNotPrinted())
    {
        result = CoreApiResult(CoreApiConst::ErrorCode::HasNotPrinted,
                               tr("Последний документ не был напечатан.\n"
                                  "Напечатайте сначала последний ненапечатанный документ"));
        return false;

    }


    if(!st.isFiscalized())
    {
        result = CoreApiResult(CoreApiConst::ErrorCode::NotFiscalized);
        return false;
    }
    switch (task.operation()) {
    case CoreTransaction::Operation::OPEN_CYCLE: return canOpenCycle(result);
    case CoreTransaction::Operation::CLOSE_CYCLE: return canCloseCycle(result);
    case CoreTransaction::Operation::RECEIPT:
    case CoreTransaction::Operation::CORRECTION:return canMakeReceipt(result);
    case CoreTransaction::Operation::CHECK_LABEL: return canCheckLabel(result);
    case CoreTransaction::Operation::CLEAN_LABELS: return canCheckLabel(result);
    default:
        return true;
    }
    return true;
}

CoreApiResult ReceiptsController::openCycle(const CoreTransaction &task)
{
    CycleOpenIncoming incoming (task.params());

    lmWarning() << logvariant(task.params()) << logtab << logvariant(incoming.toMap());
    RegData rd = pdw_->getRegData();
    FsFullStatus fs = pdw_->getFsFullStatus();
    if(fs.error() != CoreApiConst::ErrorCode::Ok)
    {
        return CoreApiResult(fs.error());
    }
    Watcherator watcher(pdw_->watcher(), task, fs.status().lastFd() + 1);
    if(!watcher.init()) return  watcher.toErrorResult();
    QDateTime dt = QDateTime::currentDateTime();
    TextPrinterDocument keysDoc;
    updateKeys(keysDoc);
    if(!keysDoc.blocks().isEmpty())
    {
        printTextDoc(keysDoc);
    }
//    return CoreApiResult(CoreApiConst::ErrorCode::UnknownError, tr("ТЕСТ ОБНОВЛЕНИЯ КЛЮЧЕЙ"));

    CoreApiConst::ErrorCode err = fs_->openCycleStart(dt);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        return CoreApiResult(err);
    }
    err = fs_->sendDocData(incoming.tlv(rd));
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        return CoreApiResult(err);
    }
    quint16 cycle = 0;
    qint32 fd = 0;
    quint32 fiscCode = 0;
    if(!watcher.startFs())
    {
        fs_->cancelDoc();
        return  watcher.toErrorResult();
    }
    err = fs_->openCycleCommit(cycle, fd, fiscCode);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        std::optional<FsStatus> st;
        if(fs_->getFsStatus(st) == CoreApiConst::ErrorCode::Ok && st.has_value())
        {
            if(st.value().lastFd() < static_cast<quint32>(watcher.docNumber())) watcher.remove();
            //NOTE: возможно сюда сразу добавить восстановление и вычитать из фн как и другие документы
        }
        return CoreApiResult(err);
    }

    lmWarning() << cycle << fd << fiscCode;

    quint16 docType = 0;
    Tlv::Stlv list;
    err = fs_->readTlvDoc(fd, docType, list);
    CycleOpenDoc doc;
    doc.setDataFromTask(task);
    lmWarning() << cycle << fd << fiscCode << docType;

    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err);


        doc.setIncoming(incoming, true);
        doc.setDt(dt);
        doc.setCycle(cycle);
        doc.setFd(fd);
        doc.setFiscalCode(fiscCode);
        doc.setFsNumber(fs.fsNumber());
        doc.setRegData(rd);
    }
    else
    {

        doc.parseFromTlvList(list);
        doc.setFiscalCode(fiscCode);
        doc.setDt(dt);
        doc.setFd(fd);
        doc.setCycle(cycle);
        doc.setIncoming(incoming, true);
        doc.setRegData(rd, false);
    }
    PrintDocumentFormatter pf(rd, this);
    TextPrinterDocument pdoc = pf.openCycle(&doc);
    watcher.update(fd, task.formatToExternal() ? doc.toExternalMap() : doc.toMap(), pdoc);
    if(task.printResult())
    {
        err = printTextDoc(pdoc);
    }
    watcher.finalize(err);

    pdw_->saveDocument(&doc, task);
    QVariantMap resdata = task.formatToExternal() ? doc.toExternalMap() : doc.toMap();

    lmWarning() << logvariant(resdata);

    pdw_->getFsFullStatus(FsFullStatus::CLEAN_CYCLE);
    return CoreApiResult(err, QString(), resdata);
}

CoreApiResult ReceiptsController::closeCycle(const CoreTransaction &task)
{

    CycleCloseIncoming incoming (task.params());

    lmWarning() << logvariant(task.params()) << logtab << logvariant(incoming.toMap());
    RegData rd = pdw_->getRegData();
    FsFullStatus fs = pdw_->getFsFullStatus();
    Watcherator watcher(pdw_->watcher(), task, fs.status().lastFd() + 1);
    if(!watcher.init()) return  watcher.toErrorResult();

    QDateTime dt = QDateTime::currentDateTime();
    CoreApiConst::ErrorCode err = fs_->closeCycleStart(dt);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        return CoreApiResult(err);
    }
    //TODO: Возможно сделать получение откуда-то. когда будет работать обмен с честным знаком и если не возвращается из фн
    fdf::IncorrectLabelsFlags ilf;
    fdf::IncorrectNotificationsFlags inf;

    WhatcherDb wdbc(WhatcherDb::CM_READ_ONLY);
    wdbc.get2112And2113(fs.cycle().cycle(), ilf, inf);

    err = fs_->sendDocData(incoming.tlv(rd, ilf, inf));
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        return CoreApiResult(err);
    }
    quint16 cycle = 0;
    qint32 fd = 0;
    quint32 fiscCode = 0;
    fs::Warnings warnings;
    if(!watcher.startFs())
    {
        fs_->cancelDoc();
        return  watcher.toErrorResult();
    }
    err = fs_->closeCycleCommit(cycle, fd, fiscCode, warnings);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        std::optional<FsStatus> st;
        if(fs_->getFsStatus(st) == CoreApiConst::ErrorCode::Ok && st.has_value())
        {
            if(st.value().lastFd() < static_cast<quint32>(watcher.docNumber())) watcher.remove();
            //NOTE: возможно сюда сразу добавить восстановление и вычитать из фн как и другие документы
        }
        return CoreApiResult(err);
    }

    lmWarning() << cycle << fd << fiscCode;

    quint16 docType = 0;
    Tlv::Stlv list;
    err = fs_->readTlvDoc(fd, docType, list);
    CycleCloseDoc doc;
    doc.setDataFromTask(task);
    lmWarning() << cycle << fd << fiscCode << docType;

    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err);


        doc.setIncoming(incoming, true);
        doc.setDt(dt);
        doc.setCycle(cycle);
        doc.setFd(fd);
        doc.setFiscalCode(fiscCode);
        doc.setFsNumber(fs.fsNumber());
        doc.setRegData(rd);
        doc.setWarnings(warnings);
    }
    else
    {

        doc.parseFromTlvList(list);
        doc.setFiscalCode(fiscCode);
        doc.setDt(dt);
        doc.setFd(fd);
        doc.setCycle(cycle);
        doc.setIncoming(incoming, true);
        doc.setRegData(rd, false);
    }
    PrintDocumentFormatter pf(rd, this);
    TextPrinterDocument pdoc = pf.closeCycle(&doc, task.briefReceipt());
    watcher.update(fd, task.formatToExternal() ? doc.toExternalMap() : doc.toMap(), pdoc);
    if(task.printResult())
    {
        err = printTextDoc(pdoc);
    }
    watcher.finalize(err);

    pdw_->saveDocument(&doc, task);
    QVariantMap resdata = task.formatToExternal() ? doc.toExternalMap() : doc.toMap();

    lmWarning() << logvariant(resdata);

    pdw_->getFsFullStatus(FsFullStatus::CLEAN_CYCLE);
    return CoreApiResult(err, QString(), resdata);
}

CoreApiResult ReceiptsController::calcReport(const CoreTransaction &task)
{
    CalcReportIncoming incoming (task.params());

    lmWarning() << logvariant(task.params()) << logtab << logvariant(incoming.toMap());
    RegData rd = pdw_->getRegData();
    FsFullStatus fs = pdw_->getFsFullStatus();
    Watcherator watcher(pdw_->watcher(), task, fs.status().lastFd() + 1);
    if(!watcher.init()) return  watcher.toErrorResult();
    QDateTime dt = QDateTime::currentDateTime();
    CoreApiConst::ErrorCode err = fs_->calcReportStart(dt);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        return CoreApiResult(err);
    }
    err = fs_->sendDocData(incoming.tlv(rd));
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        return CoreApiResult(err);
    }
    qint32 fd = 0;
    quint32 fiscCode = 0, notSended = 0;
    QDate notSendedDt;
    if(!watcher.startFs())
    {
        fs_->cancelDoc();
        return  watcher.toErrorResult();
    }

    err = fs_->calcReportCommit(fd, fiscCode, notSended, notSendedDt);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        std::optional<FsStatus> st;
        if(fs_->getFsStatus(st) == CoreApiConst::ErrorCode::Ok && st.has_value())
        {
            if(st.value().lastFd() < static_cast<quint32>(watcher.docNumber())) watcher.remove();
            //NOTE: возможно сюда сразу добавить восстановление и вычитать из фн как и другие документы
        }
        return CoreApiResult(err);
    }


    lmWarning() << fd << fiscCode << notSended << notSendedDt.toString(Qt::ISODate);

    quint16 docType = 0;
    Tlv::Stlv list;
    err = fs_->readTlvDoc(fd, docType, list);
    CalcReportDoc doc;
    doc.setDataFromTask(task);
    lmWarning()  << fd << fiscCode << docType<< notSended << notSendedDt.toString(Qt::ISODate);

    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err);


        doc.setIncoming(incoming, true);
        doc.setDt(dt);
        doc.setCycle(fs.cycle().cycle());
        doc.setFd(fd);
        doc.setFiscalCode(fiscCode);
        doc.setFsNumber(fs.fsNumber());
        doc.setRegData(rd);
        doc.setNotSendedDocs(notSended);
        doc.setFirstNotSendedDocDt(notSendedDt);
    }
    else
    {

        doc.parseFromTlvList(list);
        doc.setFiscalCode(fiscCode);
        doc.setDt(dt);
        doc.setFd(fd);
        doc.setNotSendedDocs(notSended);
        doc.setFirstNotSendedDocDt(notSendedDt);
        doc.setIncoming(incoming, true);
        doc.setRegData(rd, false);
    }
    doc.setOfflineMode(rd.getOfflineMode());
    PrintDocumentFormatter pf(rd, this);
    TextPrinterDocument pdoc = pf.calcReport(&doc, task.briefReceipt());
    watcher.update(fd, task.formatToExternal() ? doc.toExternalMap() : doc.toMap(), pdoc);
    if(task.printResult())
    {
        err = printTextDoc(pdoc);
    }
    watcher.finalize(err);

    pdw_->saveDocument(&doc, task);
    QVariantMap resdata = task.formatToExternal() ? doc.toExternalMap() : doc.toMap();
    lmWarning() << logvariant(resdata);


    pdw_->getFsFullStatus(FsFullStatus::CleanFlag::Status | FsFullStatus::CleanFlag::Resource | FsFullStatus::CleanFlag::Transport);
    return CoreApiResult(err, QString(), resdata);
}

CoreApiResult ReceiptsController::receipt(const CoreTransaction &task, const Cashier &cashier)
{
    lmWarning() << logvariant(task.toMap());

    Receipt r(task.params());
    FixNumber maxAmount = pdw_->coreConfig()->receiptConfig().maxReceiptAmount();
    if(maxAmount.value() > 0 && r.amount().value() > maxAmount.value())
    {
        return CoreApiResult(CoreApiConst::ErrorCode::ReceiptAmountToLarge,
                             tr("Превышена максимальная сумма по чеку (разрешено %1)").arg(maxAmount.toString()));
    }

    if((r.isRefund() && !cashier.hasPermission(Cashier::Permission::ReturnReceipts)) ||
            (!r.isRefund() && !cashier.hasPermission(Cashier::Permission::Receipts)))
    {
        return CoreApiResult(CoreApiConst::ErrorCode::AccessDenied, tr("Доступ запрещен. Не достаточно прав"));
    }

    RegData rd = pdw_->getRegData();
    r.setCorrection(false);
    r.setFosa(rd.extFlags().getFosaCashbox());
    r.setInternetCashbox(rd.extFlags().getInternetOnly());
    r.setSupportLabledTrandes(rd.extFlags().getLabledProducts());
    r.setSupportExciseTrandes(rd.extFlags().getExcisableProducts());

    return makeReceipt(task, r, rd);

}

CoreApiResult ReceiptsController::correction(const CoreTransaction &task, const Cashier &cashier)
{
    lmWarning() << logvariant(task.toMap());

    Receipt r(task.params());

    if((r.isRefund() && !cashier.hasPermission(Cashier::Permission::ReturnCorrections)) ||
            (!r.isRefund() && !cashier.hasPermission(Cashier::Permission::Corrections)))
    {
        return CoreApiResult(CoreApiConst::ErrorCode::AccessDenied, tr("Доступ запрещен. Не достаточно прав"));
    }

    RegData rd = pdw_->getRegData();
    r.setCorrection(true);
    r.setFosa(rd.extFlags().getFosaCashbox());
    r.setInternetCashbox(rd.extFlags().getInternetOnly());
    r.setSupportLabledTrandes(rd.extFlags().getLabledProducts());
    r.setSupportExciseTrandes(rd.extFlags().getExcisableProducts());

    return makeReceipt(task, r, rd);

}

CoreApiResult ReceiptsController::checkLabel(const CoreTransaction &task)
{
    lmWarning() << logvariant(task.toMap());
    LabelCheckData lcd(task.params());
    CoreApiConst::ErrorCode err = CoreApiConst::ErrorCode::UnknownError;
    QString msg;
    lmWarning() << logvariant(lcd.toMap());
    if(!lcd.isValid(err, msg))
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err) << msg;
        return CoreApiResult(err, msg);
    }
    lmWarning() << "НАЧАЛО ПРОВЕРКИ МАРКИ";

    LabelsChecker c(fs_, pdw_, this);
    LabelCheckResult result;
    err = c.execute(lcd, result, msg);

    if(err != CoreApiConst::ErrorCode ::Ok) return CoreApiResult(err, msg);

    return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(),
                         task.formatToExternal() ? result.toExternalMap() : result.toMap());
}

CoreApiResult ReceiptsController::cleanLabels(const CoreTransaction &task)
{
    lmWarning() << logvariant(task.toMap());
    CoreApiConst::ErrorCode err = fs_->b3();
    return CoreApiResult(err);
}


bool ReceiptsController::canOpenCycle(CoreApiResult &result)
{
    CashboxStatus st = pdw_->getCashboxStatus();
    if(st.fs().cycleIsOpen())
    {
        result = CoreApiResult(CoreApiConst::ErrorCode::CycleIsOpened);
        return false;
    }

    return true;
}

bool ReceiptsController::canCloseCycle(CoreApiResult &result)
{
    CashboxStatus st = pdw_->getCashboxStatus();
    if(!st.fs().cycleIsOpen())
    {
        result = CoreApiResult(CoreApiConst::ErrorCode::CycleIsClosed);
        return false;
    }
    return true;
}

bool ReceiptsController::canMakeReceipt(CoreApiResult &result)
{
    CashboxStatus st = pdw_->getCashboxStatus();
    if(!st.fs().cycleIsOpen())
    {
        result = CoreApiResult(CoreApiConst::ErrorCode::CycleIsClosed);
        return false;
    }
    return true;
}

bool ReceiptsController::canCheckLabel(CoreApiResult &result)
{
    if(!canMakeReceipt(result)) return false;
    RegData rd = pdw_->getRegData();
    if(!rd.extFlags().getLabledProducts())
    {
        result = CoreApiResult(CoreApiConst::ErrorCode::LabelsNotSupported);
        return false;
    }
    return true;
}

CoreApiResult ReceiptsController::makeReceipt(const CoreTransaction &task, Receipt &receipt, const RegData &rd)
{
    Q_UNUSED(task)


    FsFullStatus fs = pdw_->getFsFullStatus();

    Watcherator watcher(pdw_->watcher(), task, fs.status().lastFd() + 1);
    if(!watcher.init()) return  watcher.toErrorResult();

    QDateTime dt = QDateTime::currentDateTime();

    CoreApiConst::ErrorCode err = CoreApiConst::ErrorCode::Ok;
    QString errMsg;
    if(!receipt.isValid(rd, err, errMsg))
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err) << errMsg;
        return CoreApiResult(err, errMsg);
    }

    err = receipt.isCorrection() ? fs_->correctionStart(dt) : fs_->receiptStart(dt);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        if(receipt.mayBeLabled()) fs_->b3();
        lmWarning() << CoreApiConst::defaultErrorMsg(err);
        return CoreApiResult(err);
    }
    if(receipt.containsLabels())
    {
        err = processLabledOperations(receipt, errMsg);
        if(err != CoreApiConst::ErrorCode::Ok)
        {
            fs_->cancelDoc();
            fs_->b3();
            lmWarning() << CoreApiConst::defaultErrorMsg(err) << errMsg;
            return CoreApiResult(err, errMsg);
        }
        ReceiptOperations ops = receipt.labledOperattions(false, true);
        Tlv::Stlv tops;
        std::for_each(ops.constBegin(), ops.constEnd(),
                      [&tops](const ReceiptOperation &o){tops << o.to1059();});
        if(!tops.isEmpty())
        {
            err = fs_->sendDocData(tops);
            if(err != CoreApiConst::ErrorCode::Ok)
            {
                fs_->cancelDoc();
                fs_->b3();
                lmWarning() << CoreApiConst::defaultErrorMsg(err);
                return CoreApiResult(err);
            }

        }
    }
    err = fs_->sendDocData(receipt.toTlv(rd));
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        if(receipt.mayBeLabled()) fs_->b3();
        lmWarning() << CoreApiConst::defaultErrorMsg(err);
        return CoreApiResult(err);
    }
    quint16 rec = 0;
    qint32 fd = 0;
    quint32 fiscalCode = 0;
    if(!watcher.startFs())
    {
        fs_->cancelDoc();
        if(receipt.mayBeLabled()) fs_->b3();
        return  watcher.toErrorResult();
    }
    err = fs_->receiptCommit(dt, receipt.paymentAttr(), receipt.amount(), rec, fd, fiscalCode);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        fs_->cancelDoc();
        if(receipt.mayBeLabled()) fs_->b3();
        lmWarning() << CoreApiConst::defaultErrorMsg(err);
        std::optional<FsStatus> st;
        if(fs_->getFsStatus(st) == CoreApiConst::ErrorCode::Ok && st.has_value())
        {
            if(st.value().lastFd() < static_cast<quint32>(watcher.docNumber())) watcher.remove();
            //NOTE: возможно сюда сразу добавить восстановление и вычитать из фн как и другие документы
        }
        return CoreApiResult(err);
    }

    BaseReceiptDoc *doc = BaseReceiptDoc::createReceiptDoc(receipt, this);    
    if(!doc)
    {
        lmCritical() << "UNKNOWN DOC ERROR";
        return CoreApiResult(CoreApiConst::ErrorCode::UnknownError);
    }
    doc->setDataFromTask(task);
    doc->setRegData(rd);
    doc->setDt(dt);
    doc->setFd(fd);
    doc->setCycle(fs.cycle().cycle());
    doc->setFiscalCode(fiscalCode);
    doc->setFsNumber(fs.fsNumber());
    doc->setReceiptNumber(rec);

    PrintDocumentFormatter pf(rd, this);
    TextPrinterDocument pdoc = pf.receipt(doc);
    //NOTE: может сохранять toMap, а потом переформатировать для наружи? Отложено до реализации апи
    watcher.update(fd, task.formatToExternal() ? doc->toExternalMap() : doc->toMap(), pdoc);
    if(task.printResult())
    {
        err = printTextDoc(pdoc);
    }
    watcher.finalize(err);

    pdw_->saveDocument(doc, task);
    QVariantMap resdata = task.formatToExternal() ? doc->toExternalMap() : doc->toMap();
    doc->deleteLater();
    lmWarning() << logvariant(resdata);
    pdw_->getFsFullStatus(FsFullStatus::CLEAN_CYCLE);
    return CoreApiResult(err, QString(), resdata);
}

CoreApiConst::ErrorCode ReceiptsController::processLabledOperations(Receipt &r, QString &msg)
{
    r.setT2107(255);
    ReceiptOperations ops = r.labledOperattions(true, false);
    if(ops.isEmpty()) return CoreApiConst::ErrorCode::Ok;
    quint8 t2107 = 0;
    CoreApiConst::ErrorCode err = CoreApiConst::ErrorCode::Ok;
    for(const ReceiptOperation &o: ops)
    {
        quint16 invalidTag = 0;
        if(!o.isValid(err, msg)) return err;
        err = fs_->b7_1(o.to2007(), invalidTag);
        if(err != CoreApiConst::ErrorCode::Ok)
        {
            msg = CoreApiConst::defaultErrorMsg(err);
            if(invalidTag > 0)
            {
                msg += QString(" [ТЭГ: %1]").arg(invalidTag);
            }
            lmCritical() << msg;
            return err;
        }
        err = fs_->b7_2(o.to1059(), invalidTag);
        if(err != CoreApiConst::ErrorCode::Ok)
        {
            msg = CoreApiConst::defaultErrorMsg(err);
            if(invalidTag > 0)
            {
                msg += QString(" [ТЭГ: %1]").arg(invalidTag);
            }
            lmCritical() << msg;
            return err;
        }
        if(o.checkLabelFlags() !=  fdf::CheckLabelFlags(0x0f) &&
                o.checkLabelFlags() != fdf::CheckLabelFlags(0x13))
        {
            t2107 = 1;
        }
    }
    quint16 invalidTag = 0;
    r.setT2107(t2107);
    err = fs_->b7_3(r.toTlvB7_3(pdw_->getRegData()), invalidTag);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        msg = CoreApiConst::defaultErrorMsg(err);
        if(invalidTag > 0)
        {
            msg += QString(" [ТЭГ: %1]").arg(invalidTag);
        }
        lmCritical() << msg;
        return err;
    }
    return err;

}

CoreApiConst::ErrorCode ReceiptsController::updateKeys(TextPrinterDocument &doc)
{
    KeysUpdater u(fs_, pdw_, this);
    DocumentBlock b;
    CoreApiConst::ErrorCode err = u.exec(b);
    lmWarning() << CoreApiConst::defaultErrorMsg(err) << loglist(doc.text());
    if(!b.text().isEmpty()) doc =TextPrinterDocument(DocumentBlocks() << b);
    return err;
}



