#include "reportscontroller.h"
#include "logmanip.h"
#include "printdocumentformatter.h"
#include "regdoc.h"
#include "cycleclosedoc.h"
#include "cycleopendoc.h"
#include "calcreportdoc.h"
#include "receiptdoc.h"
#include "xreport.h"
#include "docsdb.h"

ReportsController::ReportsController(FsWrapper *fs, ProcessingDataWrapper *dw, PrintersWrapper *p,
                                     QObject *parent)
    : BaseTaskProcessor(fs, dw, p, parent)
{

}

ReportsController::~ReportsController()
{

}

CoreApiResult ReportsController::findFsDocument(const CoreTransaction &task)
{
    CoreApiResult result;
    if(!checkState(task, result)) return result;
    qint32 fd = 0;
    if(task.params().contains("fd")) fd = task.params()["fd"].toInt();
    else if(task.params().contains("number")) fd = task.params()["number"].toInt();
    lmWarning() << fd << logvariant(task.toMap());
    if(fd == 0)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::InvalidFd, tr("Не указан номер документа"));
    }

    PFsArchiveReportBase doc;
    CoreApiConst::ErrorCode err = fs_->findDocument(fd, doc, createReportHeader());
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err);
        return CoreApiResult(err);
    }
    if(!doc)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::UnknownFsError);
    }
    if(doc->hasTicket())
    {
        PFsArchiveOfdTicket tdoc;
        err = fs_->getOfdTicketInfo(fd, tdoc);
        if(err == CoreApiConst::ErrorCode::Ok)doc->setTicket(tdoc->sign());
    }
    if(task.printResult())
    {
        err = printTextDoc(doc->printDocument());
    }
    return CoreApiResult(err, QString(), task.formatToExternal() ? doc->toExternalMap() : doc->toMap());
}

CoreApiResult ReportsController::findOfdTicket(const CoreTransaction &task)
{
    CoreApiResult result;
    if(!checkState(task, result)) return result;
    qint32 fd = 0;
    if(task.params().contains("fd")) fd = task.params()["fd"].toInt();
    else if(task.params().contains("number")) fd = task.params()["number"].toInt();
    lmWarning() << fd << logvariant(task.toMap());
    if(fd == 0)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::InvalidFd, tr("Не указан номер документа"));
    }

    PFsArchiveOfdTicket doc;
    CoreApiConst::ErrorCode err = fs_->getOfdTicketInfo(fd, doc);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err);
        return CoreApiResult(err);
    }
    if(!doc)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::UnknownFsError);
    }
    if(task.printResult())
    {
        err = printTextDoc(doc->printDocument(createReportHeader()));
    }
    return CoreApiResult(err, QString(), task.formatToExternal() ? doc->toExternalMap() : doc->toMap());
}

CoreApiResult ReportsController::tlvDocument(const CoreTransaction &task)
{
    CoreApiResult result;
    if(!checkState(task, result)) return result;
    qint32 fd = 0;
    if(task.params().contains("fd")) fd = task.params()["fd"].toInt();
    else if(task.params().contains("number")) fd = task.params()["number"].toInt();
    lmWarning() << fd << logvariant(task.toMap());
    if(fd == 0)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::InvalidFd, tr("Не указан номер документа"));
    }

    BaseDoc *doc = nullptr;
    quint16 docType = 0;
    Tlv::Stlv list;
    CoreApiConst::ErrorCode err = fs_->readTlvDoc(fd, docType, list);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err);
        return CoreApiResult(err);
    }
    lmWarning() << docType;
    switch (static_cast<fdf::DocType>(docType)) {
    case fdf::DocType::Registration: doc = new RegDoc(this);break;
    case fdf::DocType::ReRegistration: doc = new ReRegDoc(this);break;
    case fdf::DocType::CycleOpen: doc = new CycleOpenDoc(this);break;
    case fdf::DocType::CycleClose: doc = new CycleCloseDoc(this);break;
    case fdf::DocType::CalcReport: doc = new CalcReportDoc(this);break;
    case fdf::DocType::FsClose: doc = new FsCloseDoc(this);break;
    case fdf::DocType::Receipt: doc = new ReceiptDoc(this);break;
    case fdf::DocType::Fosa: doc = new FosaDoc(this);break;
    case fdf::DocType::Correction: doc = new CorrectionDoc(this);break;
    case fdf::DocType::FosaCorrection: doc = new CorrectionFosaDoc(this);break;
    default:
        break;
    }
    if(!doc)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::UnknownFsError);
    }
    doc->parseFromTlvList(list);
    doc->setIsCopy(true);
    if(doc->fsNumber().trimmed().isEmpty())doc->setFsNumber(pdw_->getFsFullStatus().fsNumber());

    BaseWorkDoc *wdoc = dynamic_cast<BaseWorkDoc *>(doc);

    //NOTE: Возможно нужно будет вытащить другие рег. данные. из предшествующего отчета.
    if(wdoc)
    {
        wdoc->setRegData(pdw_->getRegData());
    }

    if(task.printResult())
    {
        PrintDocumentFormatter pdf(pdw_->getRegData(), this);
        err = printTextDoc(pdf.document(doc, task.briefReceipt(), true));
    }
    CoreApiResult res =  CoreApiResult(err, QString(), task.formatToExternal() ? doc->toExternalMap() : doc->toMap());
    doc->deleteLater();
    return res;
}

CoreApiResult ReportsController::copyDocument(const CoreTransaction &task)
{
    CoreApiResult result;
    if(!checkState(task, result)) return result;
    qint32 fd = 0;
    QString externalId;
    if(task.params().contains("fd")) fd = task.params()["fd"].toInt();
    else if(task.params().contains("number")) fd = task.params()["number"].toInt();
    else if(task.params().contains("externalId")) externalId = task.params()["externalId"].toString();
    lmWarning() << fd << logvariant(task.toMap());
    if(fd == 0 && externalId.isEmpty())
    {
        return CoreApiResult(CoreApiConst::ErrorCode::InvalidFd, tr("Не указан номер документа"));
    }
    BaseDoc *doc = nullptr;
    DocsDb dbc(pdw_->getRegData().regNumber(), DocsDb::CM_READ_ONLY, nullptr);
    if(fd != 0)
    {
        doc = dbc.loadDocument(fd);
        if(!doc)
        {
            CoreTransaction t = task;
            t.setOperation(CoreTransaction::Operation::GET_TLV_DOC);
            return tlvDocument(task);
        }
    }
    else
    {
        doc = dbc.loadDocument(externalId);
    }
    if(!doc)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::NoDataRequested);
    }
    doc->setIsCopy(true);
    CoreApiConst::ErrorCode err = CoreApiConst::ErrorCode::Ok;
    if(task.printResult())
    {
        PrintDocumentFormatter pdf(pdw_->getRegData(), this);
        err = printTextDoc(pdf.document(doc, task.briefReceipt(), false));
    }
    CoreApiResult res =  CoreApiResult(err, QString(), task.formatToExternal() ? doc->toExternalMap() : doc->toMap());
    doc->deleteLater();
    return res;
}

CoreApiResult ReportsController::regArchiveDocument(const CoreTransaction &task)
{
    CoreApiResult result;
    if(!checkState(task, result)) return result;
    qint32 num = 0;
    if(task.params().contains("number")) num = task.params()["number"].toInt();
    lmWarning() << num << logvariant(task.toMap());

    PFsArchiveReportBase doc;
    CoreApiConst::ErrorCode err = fs_->findRegDoc(num, doc, createReportHeader());
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err);
        return CoreApiResult(err);
    }
    if(!doc)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::UnknownFsError);
    }
    if(task.printResult())
    {
        err = printTextDoc(doc->printDocument());
    }
    return CoreApiResult(err, QString(), task.formatToExternal() ? doc->toExternalMap() : doc->toMap());
}

CoreApiResult ReportsController::regArchiveTlvDocument(const CoreTransaction &task)
{
    CoreApiResult result;
    if(!checkState(task, result)) return result;
    qint32 num = 0;
    if(task.params().contains("number")) num = task.params()["number"].toInt();
    lmWarning() << num << logvariant(task.toMap());

    PFsArchiveReportBase doc;
    CoreApiConst::ErrorCode err = fs_->findRegDoc(num, doc, QStringList());
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err);
        return CoreApiResult(err);
    }
    if(!doc)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::UnknownFsError);
    }
    if(doc->dt().daysTo(QDateTime::currentDateTime()) < 30)
    {
        QVariantMap params;
        params["fd"] = doc->docNumber();
        CoreTransaction t = task;
        t.setOperation(CoreApiConst::RemoteCommand::GET_TLV_DOC);
        t.setParams(params);
        t.genUid();
        result = tlvDocument(t);
        if(result.isOk())
        {
            lmWarning() << "OK!";
            return result;
        }
    }
    BaseRegDoc *regdoc = nullptr;
    err = fs_->readRegDoc(num, regdoc);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err);
        return CoreApiResult(err);
    }
    if(!regdoc)
    {
        return CoreApiResult(CoreApiConst::ErrorCode::UnknownFsError);
    }
    if(regdoc->fsNumber().trimmed().isEmpty())regdoc->setFsNumber(pdw_->getFsFullStatus().fsNumber());
    QVariantMap data = task.formatToExternal() ? regdoc->toExternalMap() : regdoc->toMap();
    delete regdoc;
    return CoreApiResult(CoreApiConst::ErrorCode::Ok, QString(), data);
}

CoreApiResult ReportsController::xreport(const CoreTransaction &task, const Cashier &cashier)
{
    BaseWorkIncoming incoming(task.params());
    lmWarning() << logvariant(task.params()) << logtab << logvariant(incoming.toMap()) << logtab << logvariant(task.toMap());
    RegData rd = pdw_->getRegData();
    if(rd.address().isEmpty())rd.setAddress(incoming.address());
    if(rd.place().isEmpty()) rd.setPlace(incoming.place());
    FsFullStatus fs = pdw_->getFsFullStatus();
    ModelData md = pdw_->getModelData();

    FsCounters counters;
    CoreApiConst::ErrorCode err = fs_->readFsCounters(FsTotalType::CycleTotal, counters);
    if(err != CoreApiConst::ErrorCode::Ok) return CoreApiResult(err);
    if(counters.cycle() == 0)return CoreApiResult(CoreApiConst::ErrorCode::NoDataRequested);



    XReport rep(counters, nullptr);
    rep.setCashier(incoming.cashier().isEmpty() ? cashier.name() : incoming.cashier());
    rep.setCashier(incoming.cashierInn().isEmpty() ? cashier.inn() : incoming.cashier());
    rep.setClientId(task.clientId());
    rep.setExternalId(task.externalId());
    rep.setFooter(incoming.footer());
    rep.setHeader(incoming.header());
    rep.setModel(md);
    rep.setRegData(rd);
    rep.setStatus(fs);
    if(!incoming.cashier().isEmpty())
    {
        rep.setCashier(incoming.cashier());
        rep.setCashierInn(incoming.cashierInn());

    }
    else
    {
        rep.setCashier(cashier.name());
        rep.setCashierInn(cashier.inn());
    }
    rep.setAddress(incoming.address().isEmpty() ? rd.address() : incoming.address());
    rep.setPlace(incoming.place().isEmpty() ? rd.place() : incoming.place());

    PrintDocumentFormatter pf(rd, this);
    TextPrinterDocument prep = pf.xreport(&rep, task.briefReceipt());
    if(task.printResult())
    {
        err = printTextDoc(prep);
    }
    lmWarning() << task.formatToExternal();
    return CoreApiResult(err, QString(), task.formatToExternal() ? rep.toExternalMap() : rep.toMap());
}

CoreApiResult ReportsController::fsreport(const CoreTransaction &task, const Cashier &cashier)
{
    BaseWorkIncoming incoming(task.params());
    lmWarning() << logvariant(task.params()) << logtab << logvariant(incoming.toMap());
    RegData rd = pdw_->getRegData();
    if(rd.address().isEmpty())rd.setAddress(incoming.address());
    if(rd.place().isEmpty()) rd.setPlace(incoming.place());
    FsFullStatus fs = pdw_->getFsFullStatus();
    ModelData md = pdw_->getModelData();

    FsCounters counters;
    CoreApiConst::ErrorCode err = fs_->readFsCounters(FsTotalType::FsTotalType, counters);
    if(err != CoreApiConst::ErrorCode::Ok) return CoreApiResult(err);
    if(counters.cycle() == 0)return CoreApiResult(CoreApiConst::ErrorCode::NoDataRequested);



    XReport rep(counters, nullptr);
    rep.setCashier(incoming.cashier().isEmpty() ? cashier.name() : incoming.cashier());
    rep.setCashier(incoming.cashierInn().isEmpty() ? cashier.inn() : incoming.cashier());
    rep.setClientId(task.clientId());
    rep.setExternalId(task.externalId());
    rep.setFooter(incoming.footer());
    rep.setHeader(incoming.header());
    rep.setModel(md);
    rep.setRegData(rd);
    rep.setStatus(fs);
    if(!incoming.cashier().isEmpty())
    {
        rep.setCashier(incoming.cashier());
        rep.setCashierInn(incoming.cashierInn());
    }
    else
    {
        rep.setCashier(cashier.name());
        rep.setCashierInn(cashier.inn());
    }
    rep.setAddress(incoming.address().isEmpty() ? rd.address() : incoming.address());
    rep.setPlace(incoming.place().isEmpty() ? rd.place() : incoming.place());


    PrintDocumentFormatter pf(rd, this);
    TextPrinterDocument prep = pf.xreport(&rep, task.briefReceipt());
    if(task.printResult())
    {
        err = printTextDoc(prep);
    }

    return CoreApiResult(err, QString(), task.formatToExternal() ? rep.toExternalMap() : rep.toMap());
}

bool ReportsController::checkState(const CoreTransaction &task, CoreApiResult &result)
{
    if(!BaseTaskProcessor::checkState(task, result)) return false;
    FsFullStatus st = pdw_->getFsFullStatus();
    if(!st.isFiscalized())
    {
        result = CoreApiResult(CoreApiConst::ErrorCode::NotFiscalized);
        return false;
    }
    switch (task.operation()) {
    case CoreApiConst::RemoteCommand::FIND_FS_DOC:
    case CoreApiConst::RemoteCommand::FIND_OFD_TICKET:
    case CoreApiConst::RemoteCommand::GET_TLV_DOC    :
    {
        quint32 fd = 0;
        if(task.params().contains("fd")) fd = task.params()["fd"].toUInt();
        else if(task.params().contains("number")) fd = task.params()["number"].toUInt();
        lmWarning() << fd << logvariant(task.toMap());
        if(fd == 0)
        {
            result = CoreApiResult(CoreApiConst::ErrorCode::InvalidFd, tr("Не указан номер документа"));
            return false;
        }
        if(st.status().lastFd() < fd)
        {
            result = CoreApiResult(CoreApiConst::ErrorCode::InvalidFd, tr("В ФН нет такого документа"));
            return false;
        }
    }break;
    case CoreApiConst::RemoteCommand::GET_REG_ARCHIVE_DOC:
    case CoreApiConst::RemoteCommand::GET_REG_TLV_DOC:
    {
        qint32 num = 0;
        if(task.params().contains("number")) num = task.params()["number"].toInt();
        lmWarning() << num << logvariant(task.toMap());
        if(st.lifeTime().regsCount() < num)
        {
            result = CoreApiResult(CoreApiConst::ErrorCode::InvalidParameter, tr("В ФН не было столько перерегистраций"));
            return false;
        }
        return true;
    }break;
    default:
        break;
    }
    return true;
}

QStringList ReportsController::createReportHeader() const
{
    RegData reg = pdw_->getRegData();
    QStringList text;
    text << reg.ownerName() << reg.address();
    if(!reg.place().isEmpty()) text << tr("МЕСТО РАСЧЕТОВ %1").arg(reg.place());
    text << tr("ЗН ККТ\t%1").arg(reg.serial());
    text << tr("РНМ ККТ\t%2").arg(reg.regNumber());
    text << tr("ФН\t%3").arg(pdw_->getFsFullStatus().fsNumber());
    text << tr("ИНН\t%1").arg(reg.ownerInn());
    return text;
}
