#include "printdocwatcher.h"

#include "logmanip.h"
#include "processingdatawrapper.h"
#include "receipt.h"

const QSet<qint32> PrintDocWatcher::CHECKABLE_TRANSACTIONS = {
    static_cast<qint32>(CoreApiConst::RemoteCommand::OPEN_CYCLE),
    static_cast<qint32>(CoreApiConst::RemoteCommand::CLOSE_CYCLE),
    static_cast<qint32>(CoreApiConst::RemoteCommand::RECEIPT),
    static_cast<qint32>(CoreApiConst::RemoteCommand::CORRECTION),
};

PrintDocWatcher::~PrintDocWatcher()
{

}

bool PrintDocWatcher::init(const CoreTransaction &trans, qint32 docNumber)
{
    bool res = dbc_->initDocument(trans, docNumber);
    lastError_ = dbc_->lastError();
    lastErrorMsg_ = dbc_->lastErrorMsg();
    return res;
}

bool PrintDocWatcher::startFs(const QString &externalId)
{
    bool res =  dbc_->startFs(externalId);
    lastError_ = dbc_->lastError();
    lastErrorMsg_ = dbc_->lastErrorMsg();
    return res;

}

bool PrintDocWatcher::update(const QString &externalId, qint32 docNumber, const QVariantMap &doc,
                             const TextPrinterDocument &printable)
{
    bool res = dbc_->update(externalId, docNumber, doc, printable);
    lastError_ = dbc_->lastError();
    lastErrorMsg_ = dbc_->lastErrorMsg();
    return res;

}

bool PrintDocWatcher::finalize(const QString &externalId, CoreApiConst::ErrorCode result)
{
    bool res = dbc_->finalize(externalId, result);
    lastError_ = dbc_->lastError();
    lastErrorMsg_ = dbc_->lastErrorMsg();
    return res;

}

bool PrintDocWatcher::hasNotPrinted()
{
    bool res = dbc_->hasNotPrinted();
    lastError_ = dbc_->lastError();
    lastErrorMsg_ = dbc_->lastErrorMsg();
    return res;

}

bool PrintDocWatcher::checkNotPrinted(const CoreTransaction &trans)
{
    return  CHECKABLE_TRANSACTIONS.contains(static_cast<qint32>(trans.operation())) &&
            dbc_->checkNotPrinted(trans);
}

bool PrintDocWatcher::find(const CoreTransaction &trans, CoreApiResult &result, TextPrinterDocument &pdoc)
{
    int status = 0;
    if(!dbc_->find(trans, result, pdoc, status))
    {
        if(dbc_->lastError() != CoreApiConst::ErrorCode::Ok)
        {
            result = CoreApiResult(dbc_->lastError(), dbc_->lastErrorMsg());
        }
        else
        {
            result = CoreApiResult(CoreApiConst::ErrorCode::NoRequestedStoredData,
                                   tr("Нет запрашиваемых данных"));
        }
        return false;
    }
    if(status >= Status::BeforePrint) return true;
    if(!restore(trans, result, pdoc)) return false;


    return  true;
}

bool PrintDocWatcher::find(const CoreTransaction &trans, CoreApiResult &result)
{
    int status = 0;
    TextPrinterDocument pdoc;
    if(!dbc_->find(trans, result, pdoc, status))
    {
        if(dbc_->lastError() != CoreApiConst::ErrorCode::Ok)
        {
            result = CoreApiResult(dbc_->lastError(), dbc_->lastErrorMsg());
        }
        else
        {
            result = CoreApiResult(CoreApiConst::ErrorCode::NoRequestedStoredData,
                                   tr("Нет запрашиваемых данных"));
        }
        return false;
    }
    return  true;
}

bool PrintDocWatcher::remove(const QString &externalId)
{
    bool res = dbc_->remove(externalId);
    lastError_ = dbc_->lastError();
    lastErrorMsg_ = dbc_->lastErrorMsg();
    return res;
}

bool PrintDocWatcher::restoreLast(QString &externalId)
{
    QVariantMap doc;
    TextPrinterDocument pdoc;
    CoreApiResult result;
    return restoreLast(externalId, result, pdoc);
}

bool PrintDocWatcher::restoreLast(QString &externalId, CoreApiResult &result, TextPrinterDocument &pdoc)
{
    externalId.clear();
    if(!dbc_->getNotPrinterExternalId(externalId))
    {
        lastError_ = dbc_->lastError();
        lastErrorMsg_ = dbc_->lastErrorMsg();
        lmWarning() << externalId << lastErrorMsg_;
        return false;
    }
    return restore(externalId, result, pdoc);
}

bool PrintDocWatcher::restore(const CoreTransaction &trans, CoreApiResult &result, TextPrinterDocument &pdoc)
{
    return restore(trans.externalId(), result, pdoc);
}

void PrintDocWatcher::clean()
{
    if(dbc_)dbc_->clean();
}

CoreApiConst::ErrorCode PrintDocWatcher::lastError() const
{
    return lastError_;
}

const QString &PrintDocWatcher::lastErrorMsg() const
{
    return lastErrorMsg_;
}

PrintDocWatcher::PrintDocWatcher(ProcessingDataWrapper *parent, FsWrapper *fs)
    : QObject(parent)
    , fs_(fs)
    , dbc_(new WhatcherDb(WhatcherDb::CM_READ_WRITE, this))
    , lastError_(CoreApiConst::ErrorCode::Ok)
    , lastErrorMsg_()
    , formatResultToExternal_(false)
{

}

ProcessingDataWrapper *PrintDocWatcher::pdw()
{
    return dynamic_cast<ProcessingDataWrapper *>(parent());
}

bool PrintDocWatcher::restore(const QString &externalId, CoreApiResult &result, TextPrinterDocument &pdoc)
{
//TODO: Не забыть сделать pdw_->saveDoc
    //TODO: Не забыть вернуть ошибки для некоторых операций как CoreApiConst::errorcoce
    //TODO: Сделать восстановление при обрывве связи с ФН. тут сделана только печать
    int status = 0;
    lmWarning() << externalId;
    if(!dbc_->find(externalId, result, pdoc, status))
    {
        if(dbc_->lastError() != CoreApiConst::ErrorCode::Ok)
        {
            lastError_ = dbc_->lastError();
            lastErrorMsg_ = dbc_->lastErrorMsg();
        }
        else
        {
            lastError_ = CoreApiConst::ErrorCode::NoRequestedStoredData;
            lastErrorMsg_ = tr("Нет запрашиваемых данных");
        }
        lmWarning() << externalId << lastErrorMsg_;
        return false;
    }
    if(status >= BeforePrint)
    {
        return true;
    }
    lmWarning() << "GET INCOMING";
    WhatcherDb::Data data;
    if(!dbc_->getIncoming(externalId, data))
    {
        if(dbc_->lastError() != CoreApiConst::ErrorCode::Ok)
        {
            lastError_ = dbc_->lastError();
            lastErrorMsg_ = dbc_->lastErrorMsg();
        }
        else
        {
            lastError_ = CoreApiConst::ErrorCode::NoRequestedStoredData;
            lastErrorMsg_ = tr("Нет запрашиваемых данных");
        }
        return false;
    }
    lmWarning() << "FINISH RESTORE";
    return finishRestore(data, result, pdoc);
}

bool PrintDocWatcher::finishRestore(const WhatcherDb::Data &incoming, CoreApiResult &result,
                                    TextPrinterDocument &pdoc)
{
    PFsArchiveReportBase docSummary;
    CoreApiConst::ErrorCode err = fs_->findDocument(incoming.docNumber, docSummary, QStringList());
    if(err == CoreApiConst::ErrorCode::NoDataRequested)
    {
        dbc_->remove(incoming.externalId);
        lastError_ = CoreApiConst::ErrorCode::NoRequestedStoredData;
        lastErrorMsg_ = tr("Нет запрашиваемых данных");
        return false;
    }
    else if(err != CoreApiConst::ErrorCode::Ok)
    {
        lastError_ = err;
        return false;
    }
    else if(!needRestore(incoming, docSummary))
    {
        dbc_->remove(incoming.externalId);
        lastError_ = CoreApiConst::ErrorCode::NoRequestedStoredData;
        lastErrorMsg_ = tr("Нет запрашиваемых данных");
        return false;
    }
    if(incoming.operation == CoreTransaction::Operation::RECEIPT ||
        incoming.operation == CoreTransaction::Operation::CORRECTION)
    {
        return restoreReceipt(incoming, docSummary, result, pdoc);
    }
    BaseDoc *doc = nullptr;
    quint16 docType = 0;
    Tlv::Stlv list;
    err = fs_->readTlvDoc(incoming.docNumber, docType, list);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lmWarning() << CoreApiConst::defaultErrorMsg(err);
        lastError_ = err;
        return false;
    }
    lmWarning() << docType;
    switch (static_cast<fdf::DocType>(docType)) {
    case fdf::DocType::Registration: doc = new RegDoc(this);break;
    case fdf::DocType::ReRegistration: doc = new ReRegDoc(this);break;
    case fdf::DocType::CycleOpen: doc = new CycleOpenDoc(this);break;
    case fdf::DocType::CycleClose: doc = new CycleCloseDoc(this);break;
    case fdf::DocType::CalcReport: doc = new CalcReportDoc(this);break;
    case fdf::DocType::FsClose: doc = new FsCloseDoc(this);break;
    default:
        lastError_ = CoreApiConst::ErrorCode::NoRequestedStoredData;
        lastErrorMsg_ = tr("Нет запрашиваемых данных");
        return false;
        break;
    }
    if(!doc)
    {
        lastError_ = CoreApiConst::ErrorCode::UnknownFsError;
        return false;
    }
    doc->parseFromTlvList(list);
    doc->setIsCopy(false);
    BaseWorkIncoming i(incoming.incoming);
    doc->setFooter(i.footer());
    doc->setHeader(i.header());
    BaseWorkDoc *wdoc = dynamic_cast<BaseWorkDoc *>(doc);
    //NOTE: Возможно нужно будет вытащить другие рег. данные. из предшествующего отчета.
    if(wdoc)wdoc->setRegData(pdw()->getRegData());
    pdw()->saveDocument(doc, incoming.externalId, incoming.clientId);
    PrintDocumentFormatter pdf(pdw()->getRegData(), this);
    pdoc = pdf.document(doc, false, true);
    result =  CoreApiResult(CoreApiConst::ErrorCode::PrinterBaseError, QString(),
                           formatResultToExternal_ ? doc->toExternalMap() : doc->toMap());
    update(incoming.externalId, docSummary->docNumber(), result.data(), pdoc);
    doc->deleteLater();
    return true;
}

bool PrintDocWatcher::needRestore(const WhatcherDb::Data &incoming, const PFsArchiveReportBase &docSummary) const
{
    if(!docSummary || static_cast<quint32>(incoming.docNumber) != docSummary->docNumber()||
        qAbs(incoming.dt.secsTo(docSummary->dt())) > 90) return false;
    switch (incoming.operation) {
    case CoreTransaction::Operation::OPEN_CYCLE:
        return docSummary->type() == static_cast<quint8>(fdf::DocType::CycleOpen);
    case CoreTransaction::Operation::CLOSE_CYCLE:
        return docSummary->type() == static_cast<quint8>(fdf::DocType::CycleClose);
    case CoreTransaction::Operation::FISCALIZATION:
        return docSummary->type() == static_cast<quint8>(fdf::DocType::Registration) ||
            docSummary->type() == static_cast<quint8>(fdf::DocType::ReRegistration);
    case CoreTransaction::Operation::CALC_REPORT:
        return docSummary->type() == static_cast<quint8>(fdf::DocType::CalcReport);
    case CoreTransaction::Operation::FS_CLOSE:
        return docSummary->type() == static_cast<quint8>(fdf::DocType::FsClose);
    case CoreTransaction::Operation::RECEIPT:
        return (docSummary->type() == static_cast<quint8>(fdf::DocType::Receipt) ||
                (docSummary->type() == static_cast<quint8>(fdf::DocType::Fosa))) &&
               checkReceiptAmount(docSummary->amount(), incoming.incoming);
    case CoreTransaction::Operation::CORRECTION:
        return (docSummary->type() == static_cast<quint8>(fdf::DocType::Correction) ||
                (docSummary->type() == static_cast<quint8>(fdf::DocType::FosaCorrection))) &&
               checkReceiptAmount(docSummary->amount(), incoming.incoming);
    default: return false;
    }
    return false;
}

bool PrintDocWatcher::checkReceiptAmount(const FixNumber &amount, const QVariantMap &incoming) const
{
    Receipt r(incoming);
    return qAbs(amount.value() - r.amount().value()) < 2ll;
}

bool PrintDocWatcher::restoreReceipt(const WhatcherDb::Data &incoming, const PFsArchiveReportBase &docSummary,
                                     CoreApiResult &result, TextPrinterDocument &pdoc)
{
    if(!docSummary) return false;
    quint16 docType = 0;
    Tlv::Stlv list;
    CoreApiConst::ErrorCode err = fs_->readTlvDoc(incoming.docNumber, docType, list);
    if(err != CoreApiConst::ErrorCode::Ok)
    {
        lastError_ = err;
        return false;;
    }
    Receipt r(incoming.incoming);
    RegData rd = pdw()->getRegData();
    r.setCorrection(false);
    r.setFosa(rd.extFlags().getFosaCashbox());
    r.setInternetCashbox(rd.extFlags().getInternetOnly());
    r.setSupportLabledTrandes(rd.extFlags().getLabledProducts());
    r.setSupportExciseTrandes(rd.extFlags().getExcisableProducts());

    BaseReceiptDoc *doc = BaseReceiptDoc::createReceiptDoc(r, this);
    BaseReceiptDoc *tempdoc =  BaseReceiptDoc::createReceiptDoc(r, this);
    tempdoc->parseFromTlvList(list);
    if(!doc)
    {
        lmCritical() << "UNKNOWN DOC ERROR";
        lastError_ = CoreApiConst::ErrorCode::UnknownError;
        return false;
    }
    doc->setRegData(rd);
    doc->setDt(incoming.dt);
    doc->setFd(docSummary->docNumber());
    doc->setCycle(tempdoc->cycle());
    doc->setFiscalCode(docSummary->fiscalCode());
    doc->setFsNumber(tempdoc->fsNumber());
    doc->setReceiptNumber(tempdoc->receiptNumber());
    pdw()->saveDocument(doc, incoming.externalId, incoming.clientId);
    PrintDocumentFormatter pf(rd, this);
    pdoc = pf.receipt(doc);
    update(incoming.externalId, docSummary->docNumber(),
           formatResultToExternal_ ? doc->toExternalMap() : doc->toMap(), pdoc);
    result =  CoreApiResult(CoreApiConst::ErrorCode::PrinterBaseError, QString(),
                           formatResultToExternal_ ? doc->toExternalMap() : doc->toMap());
    doc->deleteLater();
    return true;
}

