#include "syncsocketwrapper.h"
#include "logmanip.h"

#include <QHostAddress>
#include<QHostInfo>
#include <QAuthenticator>

//Q_DECLARE_LOGGING_CATEGORY(log_)
//Q_LOGGING_CATEGORY(SOCKET, "sock")

SyncSocketWrapper::SyncSocketWrapper(const char *log, QObject *parent)
    : QObject(parent)
    , log_(log)
    , socket_(nullptr)
    , loop_(nullptr)
    , loopTimer_(nullptr)
    , timer_()
    , exchangeTimeout_(30000)
{

}

SyncSocketWrapper::~SyncSocketWrapper()
{
    if(socket_)
    {
        socket_->disconnectFromHost();
    }
    if(loopTimer_)
    {
        loopTimer_->stop();
    }
}

bool SyncSocketWrapper::connectedToHost() const
{
    return socket_ && socket_->state() == QTcpSocket::ConnectedState;
}

bool SyncSocketWrapper::connectToHost(const QString &host, quint16 port)
{
    QList<QHostAddress> addrss = QHostInfo::fromName(host).addresses();
    if(!addrss.isEmpty())
    {
        QHostAddress addr;
        addr = addrss.first();
        lmCWarning(log_) << host << port;
    }
    socket()->connectToHost(host, port);
    execLoop(exchangeTimeout_);
    return connectedToHost();
}

bool SyncSocketWrapper::disconnectFromHost()
{
    if(!socket_ || socket_->state() == QTcpSocket::UnconnectedState) return true;
    socket()->disconnectFromHost();
    lmCWarning(log_) <<socketState();
    if(socket_->state() == QTcpSocket::UnconnectedState) return true;
    execLoop(exchangeTimeout_);
    return !socket_ || socket_->state() == QTcpSocket::UnconnectedState;
}

QByteArray SyncSocketWrapper::sendMsg(const QByteArray &msg, ReadFunc reader)
{
    if(!connectedToHost()) return QByteArray();
    socket()->write(msg);
    socket()->flush();
    lmCWarning(log_) <<socketState();
    if(socket_ && socket_->bytesAvailable() <= 0)execLoop(exchangeTimeout_);
    lmCWarning(log_) <<socketState() << socket_  << (socket_ ? socket_->bytesAvailable() : 0);
    if(!socket_ || socket_->bytesAvailable() <= 0)
    {
        if(socket_)
        {
            lmCWarning(log_) <<socketState() << socket_->bytesAvailable();
        }
        return QByteArray();
    }
    return reader(socket_);
}

qint32 SyncSocketWrapper::exchangeTimeout() const
{
    return exchangeTimeout_;
}

void SyncSocketWrapper::setExchangeTimeout(qint32 newExchangeTimeout)
{
    exchangeTimeout_ = newExchangeTimeout;
    if(exchangeTimeout_ < 1000) exchangeTimeout_ = 10000;
}

QAbstractSocket::SocketState SyncSocketWrapper::socketState() const
{
    return socket_ ? socket_->state() : QTcpSocket::UnconnectedState;
}

QTcpSocket *SyncSocketWrapper::socket()
{
    if(!socket_)
    {
        socket_ = new QTcpSocket(this);
        connect(socket_, &QTcpSocket::connected, this, &SyncSocketWrapper::connected);
        connect(socket_, &QTcpSocket::disconnected, this, &SyncSocketWrapper::disconnected);
        connect(socket_, &QTcpSocket::errorOccurred, this, &SyncSocketWrapper::errorOccurred);
        connect(socket_, &QTcpSocket::hostFound, this, &SyncSocketWrapper::hostFound);
        connect(socket_, &QTcpSocket::proxyAuthenticationRequired, this,
                &SyncSocketWrapper::proxyAuthenticationRequired);
        connect(socket_, &QTcpSocket::stateChanged, this, &SyncSocketWrapper::stateChanged);
        connect(socket_, &QTcpSocket::readyRead, this, &SyncSocketWrapper::readyRead);
    }
    return socket_;
}

QEventLoop *SyncSocketWrapper::loop()
{
    if(!loop_) loop_ = new QEventLoop(this);
    return loop_;
}

QTimer *SyncSocketWrapper::timer()
{
    if(!loopTimer_)
    {
        loopTimer_ = new QTimer(this);
        connect(loopTimer_, &QTimer::timeout, this, &SyncSocketWrapper::timeout);
        loopTimer_->setSingleShot(false);
    }
    return loopTimer_;
}

void SyncSocketWrapper::execLoop(qint32 timeout)
{

    timer()->start(timeout);
    timer_.invalidate();
    timer_.start();
    lmCWarning(log_) << timeout << timer_.elapsed();
    loop()->exec();
    lmCWarning(log_) << timeout << timer_.elapsed();
    if(loopTimer_)
    {
        loopTimer_->stop();
        loopTimer_->deleteLater();
        loopTimer_ = nullptr;
    }
}

void SyncSocketWrapper::timeout()
{
    lmCWarning(log_) << "timeout " << timer_.elapsed();
    if(!socket_ || socket_->bytesAvailable() > 0 || timer_.elapsed() >= exchangeTimeout()) loop()->quit();
}

void SyncSocketWrapper::connected()
{
    lmCWarning(log_) << "connected" << timer_.elapsed();
}

void SyncSocketWrapper::disconnected()
{
    lmCWarning(log_) << "disconnected" << timer_.elapsed();
}

void SyncSocketWrapper::errorOccurred(QAbstractSocket::SocketError socketError)
{
    lmCWarning(log_) << "socket error" << socketError << timer_.elapsed() << logtab
                       << (socket_ ? socket_->errorString() : QString());
}

void SyncSocketWrapper::hostFound()
{
    lmCWarning(log_) << "hostFound" << timer_.elapsed() << logtab
                       << (socket_ ? socket_->peerAddress().toString() : QString());
}

void SyncSocketWrapper::proxyAuthenticationRequired(const QNetworkProxy &proxy, QAuthenticator *authenticator)
{
    Q_UNUSED(proxy)
    Q_UNUSED(authenticator)

    lmCCritical(log_) << "NEED_PROXY";
}

void SyncSocketWrapper::stateChanged(QAbstractSocket::SocketState socketState)
{
    lmCWarning(log_) << "socket state" << socketState << timer_.elapsed();
    if(socket_)lmCWarning(log_) << socket_->peerAddress() << socket_->peerPort();
    if(socketState == QTcpSocket::ConnectedState || socketState == QTcpSocket::UnconnectedState)
    {
        lmWarning() << "BEFORE LOOP QUIT";
        loop()->quit();
        lmWarning() << "AFTER LOOP QUIT";
    }
}

void SyncSocketWrapper::readyRead()
{
    lmCWarning(log_) << "socket ready read" << timer_.elapsed() << socket_->bytesAvailable();
    loop()->quit();
}
