#include "watcherdb.h"

#include "cashboxbuildconfig.h"
#include "appdirectories.h"
#include "dircreator.h"
#include "logmanip.h"
#include "fixnumber.h"
#include "printdocwatcher.h"
#include "formatutils.h"

#include <QThread>
#include <QSqlDatabase>
#include <QSqlQuery>
#include <QSqlError>
#include <QDebug>
#include <QFile>
#include <QDir>
#include <QJsonDocument>
#include <QJsonParseError>


WhatcherDb::ConnectionsCounter WhatcherDb::_rwopened;
WhatcherDb::ConnectionsCounter WhatcherDb::_roopened;

WhatcherDb::WhatcherDb(QObject *parent)
    : BaseSqliteDbConnector("WhatcherDb-", &_rwopened, parent)
    , lastError_(CoreApiConst::ErrorCode::Ok)
    , lastErrorMsg_()
{
    init();
}

WhatcherDb::WhatcherDb(ConnectionMode mode, QObject *parent)
    : BaseSqliteDbConnector("WhatcherDb-", mode,
                            mode == CM_READ_ONLY ? &_roopened : &_rwopened,
                            parent)
    , lastError_(CoreApiConst::ErrorCode::Ok)
    , lastErrorMsg_()
{
    init();
}

WhatcherDb::~WhatcherDb()
{
    release();
}

bool WhatcherDb::initDocument(const CoreTransaction &trans, qint32 docNumber)
{
    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    q.prepare("select count(id) from docs where externalId = :externalId");
    q.bindValue(":externalId", trans.externalId());
    if(!q.exec() || !q.first())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    if(q.value(0).toUInt() > 0)
    {
        lastError_ = CoreApiConst::ErrorCode::ExternalIdDublicated;
        lastErrorMsg_ = tr("Документ с таким externalId уже существует");
        db_.rollback();
        return false;
    }
    q.prepare("insert into docs (externalId, transactionType, clientId, docNumber, incoming, dt, status)"
              "values (:externalId, :transactionType, :clientId, :docNumber, :incoming, :dt, :status)");
    q.bindValue(":externalId", trans.externalId());
    q.bindValue(":transactionType", static_cast<qint32>(trans.operation()));
    q.bindValue(":clientId", trans.clientId());
    q.bindValue(":docNumber", docNumber);
    QJsonDocument jdoc = QJsonDocument::fromVariant(trans.params());
    q.bindValue(":incoming", QString::fromUtf8(jdoc.toJson(QJsonDocument::Compact)));
    q.bindValue(":dt", QDateTime::currentDateTime().toString(Qt::ISODate));
    q.bindValue(":status", PrintDocWatcher::Status::Initial);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        lastErrorMsg_ = tr("Ошибка кэширования входных данных документа");
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        lmCritical() << db_.lastError();
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    return  true;
}

bool WhatcherDb::startFs(const QString &externalId)
{
    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    q.prepare("update docs set status = :status where externalId = :externalId");
    q.bindValue(":status", PrintDocWatcher::Status::BeforeFsFinalize);
    q.bindValue(":externalId", externalId);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        lastErrorMsg_ = tr("Ошибка кэширования входных данных документа");
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        lmCritical() << db_.lastError();
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    return  true;
}

bool WhatcherDb::update(const QString &externalId, qint32 docNumber, const QVariantMap &doc,
                        const TextPrinterDocument &printable)
{
    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    q.prepare("update docs set status = :status, docNumber = :docNumber, outgoing = :outgoing, "
              "printable = :printable where externalId = :externalId");
    q.bindValue(":status", PrintDocWatcher::Status::BeforePrint);
    q.bindValue(":externalId", externalId);
    q.bindValue(":docNumber", docNumber);
    q.bindValue(":outgoing", QString::fromUtf8(QJsonDocument::fromVariant(doc).toJson(QJsonDocument::Compact)));
    q.bindValue(":printable", QString::fromUtf8(QJsonDocument::fromVariant(printable.toMap())
                                                .toJson(QJsonDocument::Compact)));
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        lastErrorMsg_ = tr("Ошибка кэширования входных данных документа");
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        lmCritical() << db_.lastError();
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    return  true;
}

bool WhatcherDb::finalize(const QString &externalId, CoreApiConst::ErrorCode result)
{
    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    if(result == CoreApiConst::ErrorCode::Ok)
    {
        q.prepare("update docs set status = :status, printable = :printable, result =:result "
                  "where externalId = :externalId");
        q.bindValue(":status", PrintDocWatcher::Status::FinalSuccess);
        q.bindValue(":printable", QVariant());
    }
    else
    {
        q.prepare("update docs set status = :status, result =:result where externalId = :externalId");
        q.bindValue(":status", PrintDocWatcher::Status::Final);
    }
    q.bindValue(":result", static_cast<qint32>(result));
    q.bindValue(":externalId", externalId);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        lastErrorMsg_ = tr("Ошибка кэширования входных данных документа");
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        lmCritical() << db_.lastError();
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    return  true;
}

bool WhatcherDb::hasNotPrinted()
{
    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    q.prepare("select count(id) from docs where status < :status");
    q.bindValue(":status", PrintDocWatcher::Status::FinalSuccess);
    if(!q.exec() || !q.first())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    bool res = q.value(0).toUInt() > 0;
    db_.commit() || db_.rollback();
    return res;
}

bool WhatcherDb::checkNotPrinted(const CoreTransaction &trans)
{
    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    q.prepare("select count(id) from docs where externalId = :externalId and transactionType = :transactionType");
    q.bindValue(":externalId", trans.externalId());
    q.bindValue(":transactionType", static_cast<qint32>(trans.operation()));

    if(!q.exec() || !q.first())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    bool res = q.value(0).toUInt() > 0;
    db_.commit() || db_.rollback();
    return res;
}

bool WhatcherDb::find(const CoreTransaction &trans, CoreApiResult &result,
                      TextPrinterDocument &printable, int &status)
{
    result = CoreApiResult();
    printable = TextPrinterDocument();

    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    q.prepare("select docNumber, incoming, status, result, outgoing, result, printable from docs "
              "where externalId = :externalId and transactionType = :transactionType");
    q.bindValue(":externalId", trans.externalId());
    q.bindValue(":transactionType", static_cast<qint32>(trans.operation()));
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    if(!q.first())
    {
        lastError_ = CoreApiConst::ErrorCode::NoRequestedStoredData;
        db_.rollback();
        return false;
    }
    status = q.value("status").toInt();
    result = CoreApiResult(static_cast<CoreApiConst::ErrorCode>(q.value("result").toUInt()),
                           QString(),
                           QJsonDocument::fromJson(q.value("outgoing")
                                                   .toString().toUtf8()).toVariant().toMap());
    printable.setMap(QJsonDocument::fromJson(q.value("printable")
                                             .toString().toUtf8()).toVariant().toMap());

    db_.commit() || db_.rollback();
    return true;
}

bool WhatcherDb::find(const QString &externalId, CoreApiResult &result, TextPrinterDocument &printable, int &status)
{
    result = CoreApiResult();
    printable = TextPrinterDocument();

    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    q.prepare("select docNumber, incoming, status, result, outgoing, result, printable from docs "
              "where externalId = :externalId");
    q.bindValue(":externalId", externalId);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    if(!q.first())
    {
        lastError_ = CoreApiConst::ErrorCode::NoRequestedStoredData;
        db_.rollback();
        return false;
    }
    status = q.value("status").toInt();
    result = CoreApiResult(static_cast<CoreApiConst::ErrorCode>(q.value("result").toUInt()),
                           QString(),
                           QJsonDocument::fromJson(q.value("outgoing")
                                                   .toString().toUtf8()).toVariant().toMap());
    printable.setMap(QJsonDocument::fromJson(q.value("printable")
                                             .toString().toUtf8()).toVariant().toMap());

    db_.commit() || db_.rollback();
    return true;
}

bool WhatcherDb::remove(const QString &externalId)
{
    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    q.prepare("delete from docs where externalId = :externalId");
    q.bindValue(":externalId", externalId);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        lmCritical() << db_.lastError();
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    return  true;
}

bool WhatcherDb::clean()
{
    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    if(!q.exec("select max (id) from docs") || !q.first())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    qint64 id = q.value(0).toLongLong() - 1000;
    QDateTime dt = QDateTime::currentDateTime().addDays(-30);
    q.prepare("delete from docs where (id < :id or dt < :dt) and status in (0, 4)");
    q.bindValue(":id", id);
    q.bindValue(":dt", dt.toString(Qt::ISODate));


    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        lastErrorMsg_ = tr("Ошибка кэширования входных данных документа");
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        lmCritical() << db_.lastError();
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    return  true;
}

bool WhatcherDb::getIncoming(const QString &externalId, Data &data)
{
    data = Data();
    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    q.prepare("select id, transactionType, clientId, docNumber, incoming, dt, status from docs "
              "where externalId = :externalId");
    q.bindValue(":externalId", externalId);
    if(!q.exec())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    if(!q.first())
    {
        lastError_ = CoreApiConst::ErrorCode::NoRequestedStoredData;
        db_.rollback();
        return false;
    }
    data.id = q.value("id").toLongLong();
    data.operation = static_cast<CoreTransaction::Operation>(q.value("transactionType").toUInt());
    data.clientId = q.value("clientId").toString();
    data.externalId = externalId;
    data.docNumber = q.value("docNumber").toInt();
    data.incoming = QJsonDocument::fromJson(q.value("incoming").toString().toUtf8()).toVariant().toMap();
    data.dt = STR2DT_(q.value("dt").toString());
    data.status = q.value("status").toInt();
    db_.commit() || db_.rollback();
    return true;
}

bool WhatcherDb::getNotPrinterExternalId(QString &externalId)
{
    cleanErrors();
    if(!db_.transaction())
    {
        lmCritical() << tr("TRANSACTION ERROR");
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        return  false;
    }
    QSqlQuery q(db_);
    if(!q.exec("select max (id) from docs") || !q.first())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    qint64 id = q.value(0).toLongLong();
    q.prepare("select externalId from docs where status < :status and id = :id");
    q.bindValue(":status", PrintDocWatcher::Status::FinalSuccess);
    q.bindValue(":id", id);
    if(!q.exec() || !q.first())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    externalId = q.value("externalId").toString().trimmed();
    db_.commit() || db_.rollback();
    return true;
}

bool WhatcherDb::update2112(fdf::IncorrectLabelsFlag flag, qint32 cycle)
{
    if(!db_.transaction())
    {
        return  false;
    }
    fdf::IncorrectLabelsFlags flags;
    QSqlQuery q(db_);

    q.prepare("select flag2112 from labelsFlags where id = :cycle");
    q.bindValue(":cycle", cycle);
    if(!q.exec() )
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    if(q.first())
    {
        flags = fdf::IncorrectLabelsFlags(q.value("flag2112").toUInt());
        flags.setFlag(flag, true);
        q.prepare("update labelsFlags set flag2112 = :flag2112 where id = :cycle");
    }
    else
    {
        q.prepare("insert into labelsFlags (id, flag2112, flag2113) values(:cycle, :flag2112, 0)");
    }
    q.bindValue(":flag2112" , static_cast<qint32>(flags));
    q.bindValue(":cycle", cycle);
    if(!q.exec() )
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        lmCritical() << db_.lastError();
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    return  true;
}

bool WhatcherDb::update2113(fdf::IncorrectNotificationsFlag flag, qint32 cycle)
{
    if(!db_.transaction())
    {
        return  false;
    }
    fdf::IncorrectNotificationsFlags flags;
    QSqlQuery q(db_);

    q.prepare("select flag2113 from labelsFlags where id = :cycle");
    q.bindValue(":cycle", cycle);
    if(!q.exec() )
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    if(q.first())
    {
        flags = fdf::IncorrectNotificationsFlags(q.value("flag2113").toUInt());
        flags.setFlag(flag, true);
        q.prepare("update labelsFlags set flag2113 = :flag2113 where id = :cycle");
    }
    else
    {
        q.prepare("insert into labelsFlags (id, flag2112, flag2113) values(:cycle, 0, :flag2113)");
    }
    q.bindValue(":flag2113" , static_cast<qint32>(flags));
    q.bindValue(":cycle", cycle);
    if(!q.exec() )
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        lmCritical() << db_.lastError();
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    return  true;
}

bool WhatcherDb::get2112And2113(qint32 cycle, fdf::IncorrectLabelsFlags &f2112,
                                fdf::IncorrectNotificationsFlags &f2113)
{
    if(!db_.transaction())
    {
        return  false;
    }
    QSqlQuery q(db_);

    q.prepare("select flag2112, flag2113 from labelsFlags where id = :cycle");
    q.bindValue(":cycle", cycle);
    if(!q.exec() || !q.first())
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }

    f2112 = fdf::IncorrectLabelsFlags(q.value("flag2112").toUInt());
    f2113 = fdf::IncorrectNotificationsFlags(q.value("flag2113").toUInt());

    db_.commit() || db_.rollback();

    return  true;
}


bool WhatcherDb::resetDb()
{
    QStringList tables  = db_.tables();

    for(int i = tables.size() - 1; i >= 0; --i)
    {
        if(tables[i].toLower() == "metadata")
        {
            tables.removeAt(i);
            break;
        }
    }

    if(!db_.transaction())
    {
        lmWarning() << "TRANSACTION ERROR " << db_.lastError();
        return false;
    }
    bool res =  resetDbTables(tables);


    if(res && !db_.commit())
    {
        lmCritical() << db_.lastError();
        db_.rollback();
        return  false;
    }
    return  true;
}

bool WhatcherDb::updateKeysStatus(const QString &fsNumber)
{
    if(!db_.transaction())
    {
        return  false;
    }
    QSqlQuery q(db_);

    q.prepare("update metadata set fsNumber = :fs, keysLastUpdatesDt = :dt");
    q.bindValue(":dt", QDate::currentDate().toString(Qt::ISODate));
    q.bindValue(":fs", fsNumber.trimmed());
    if(!q.exec() )
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    if(!db_.commit())
    {
        lmCritical() << db_.lastError();
        lastError_ = CoreApiConst::ErrorCode::InternalStorageError;
        db_.rollback();
        return false;
    }
    return  true;
}

bool WhatcherDb::getKeysStatus(const QString &fsNumber, fs::KeysStatus &ks)
{
    if(!db_.transaction())
    {
        return  false;
    }
    QSqlQuery q(db_);
    q.prepare("select keysLastUpdatesDt from metadata where fsNumber = :fs");
    q.bindValue(":fs", fsNumber.trimmed());
    if(!q.exec() )
    {
        lmCritical() << q.lastError() << logtab << q.executedQuery() << logtab << logvariant(q.boundValues());
        db_.rollback();
        return false;
    }
    if(!q.first())
    {
        ks = fs::KeysStatus::Keys60;
    }
    else
    {
        QDate dt = QDate::fromString(q.value("keysLastUpdatesDt").toString(), Qt::ISODate);
        if(!dt.isValid() || dt.daysTo(QDate::currentDate()) > 60) ks = fs::KeysStatus::Keys60;
        else if(dt.daysTo(QDate::currentDate()) > 15)ks = fs::KeysStatus::Keys15;
        else ks = fs::KeysStatus::KeysUpdated;
    }
    db_.commit() || db_.rollback();
    return true;
}

CoreApiConst::ErrorCode WhatcherDb::lastError() const
{
    return lastError_;
}

const QString &WhatcherDb::lastErrorMsg() const
{
    return lastErrorMsg_;
}



QString WhatcherDb::databaseFileName() const
{
    return APATH.watcherDir() +"/w_4.db";
}

QString WhatcherDb::connectionName() const
{
    return objectName();
}

bool WhatcherDb::getMetaData(qint32 &metadataVersion, BaseSqliteDbConnector::EMetadataState &state)
{
    Q_UNUSED(state)
    metadataVersion = 0;
    if(!db_.isOpen()) return false;
    QSqlQuery q(db_);
    if(!q.exec("select metadataVersion from metadata") )
    {
        qCritical() << q.lastError();
        return false;
    }
    if(!q.first())
    {
        qCritical() << tr("Касса не зарегистрирована");
        return false;
    }
    metadataVersion =  q.value(0).toInt();
    return true;
}

qint32 WhatcherDb::actualMetadataVersion() const
{
    return DB_METADATA_VERSION;
}

bool WhatcherDb::updateDatabase(qint32 from)
{
    WhatcherDbDbInitializer dbi;
    bool res = true;
    for(int i = from + 1; i <= DB_METADATA_VERSION;++i)
    {
        res = res && dbi.updateDatabase(&db_, i);
    }
    return res;
}

void WhatcherDb::cleanErrors()
{
    lastErrorMsg_.clear();
    lastError_ = CoreApiConst::ErrorCode::Ok;
}


//-----------------------------------------------------------------------------------------

WhatcherDbDbInitializer::WhatcherDbDbInitializer(QObject *parent)
    : QObject(parent)
{
    QString thrId = QString::number(reinterpret_cast<qint64>(QThread::currentThreadId()), 16);
    setObjectName(QString("WhatcherDbDbInitializer-")+ thrId);
}

WhatcherDbDbInitializer::~WhatcherDbDbInitializer()
{

}

bool WhatcherDbDbInitializer::initializeDatabase()
{
    AppDirectories cfg;
    QString fileName = cfg.watcherDir() +"/w_4.db";
    lmWarning() << fileName;
    bool valid = false;
    if(QFile::exists(fileName))
    {
        WhatcherDb db;
        QString msg;
        qint32 ver = 0;
        valid = db.db().isOpen() && db.checkVersion(ver, msg, true) == WhatcherDb::DATABASE_IS_VALID;
        if(valid)
        {
            QSqlQuery q(db.db());
            valid = q.exec("pragma integrity_check;");
            q.first();
        }
    }
    if(!valid)
    {
        QDir dir(cfg.watcherDir());
        QFileInfo fi(fileName);
        QFileInfoList fil = dir.entryInfoList(QStringList() << fi.fileName() + "*", QDir::Files);

        for(const QFileInfo &fi :fil)
        {
            QFile::remove(fi.absoluteFilePath());
        }
    }
    lmWarning() << valid;
    if(!QFile::exists(fileName))
    {
        WhatcherDb dbc;
        QSqlDatabase db = dbc.db();
        WhatcherDbDbInitializer dbi;
        bool res = dbi.createDatabase(&db, ":/watcherdb/watcherdb.sql");
        return res;
    }
    return true;
}

bool WhatcherDbDbInitializer::createDatabase(QSqlDatabase * const db, const QString &fileName)
{
    QString sql = getScript(fileName);
    if(sql.isEmpty() || !(db->isOpen() || db->open()))
    {
        QSqlError err = db->lastError();
        qCritical() << (err.isValid() ? err.text() : tr("Не задан скрипт для создания базы!"));
        return false;
    }
    db->exec("PRAGMA journal_mode = \"WAL\"");
//    lmCritical() << db->exec(QStringLiteral("").arg(CORE_CONFIG_DB_METADATA_VERSION)).lastError();
    return applyScript(db, sql);
}

bool WhatcherDbDbInitializer::updateDatabase(QSqlDatabase * const db, qint32 toMetaData)
{
    QString fileName = QString(":/watcherdb/md_%1.sql").arg(toMetaData);
    QString sql = getScript(fileName);
    if(sql.isEmpty()) return true;
    if(!(db->isOpen() || db->open()))
    {
        QSqlError err = db->lastError();
        qCritical() << (err.isValid() ? err.text() : tr("Не задан скрипт для создания базы!"));
        return false;
    }
    return applyScript(db, sql);
}

QString WhatcherDbDbInitializer::getScript(const QString &fileName)
{
    QFile f(fileName);
    if(!f.exists() || !f.open(QIODevice::ReadOnly))
    {
        qCritical() << tr("Ошибка открытия файла %1 (%2)").arg(fileName).arg(f.errorString());
        return QString();
    }
    QTextStream ts(&f);
    ts.setCodec("UTF-8");
    QString res = ts.readAll();
    f.close();
    return res;
}

bool WhatcherDbDbInitializer::applyScript(QSqlDatabase * const db, const QString &sql)
{
    auto checkFk = [db](bool off) -> bool{
        QSqlQuery q(*db);
        bool res = q.exec(QStringLiteral("PRAGMA foreign_keys = %1").arg(off ? "off" : "on"));
        if(!res)
        {
            qCritical() << q.lastError() << logtab << q.lastQuery();
        }
        return res;
    };
    if(!checkFk(true))
    {
        return false;
    }
    bool res = db->transaction();
    QStringList script = sql.split(";;");
    QSqlQuery q(*db);

    for(QString s: script)
    {
        if(s.size() < 7) continue;
        res = res && q.exec(s);
        if(!res)
        {
            s.replace("\r\n","\n");
            qWarning() << loglist(s.split("\n"));
            qCritical() << q.lastError();
            break;
        }
    }
    res = res && db->commit();
    if(res)
    {
        return checkFk(false);
    }
    qCritical() << db->lastError();
    db->rollback();
    checkFk(false);
    return false;
}
