#include "printerswrapper.h"
#include "cashboxbuildconfig.h"
#include "logmanip.h"
#include "appdirectories.h"
#include "internalobjectsorage.h"

#include <QtAndroidExtras/QAndroidJniObject>
#include <QtAndroidExtras/QtAndroid>
#include <QAndroidJniEnvironment>
#include <QAndroidIntent>
#include <QCryptographicHash>

#include <QPainter>
#include <QBuffer>
#include <QBitmap>

PrintersWrapper::PrintersWrapper(QObject *parent)
    : QObject(parent)
    , obj_(new QAndroidJniObject(
          "ru/armax/cashbox/hw/PrinterWrapper",
          "(Landroid/content/Context;)V",
          QtAndroid::androidContext().object()))
{

}

PrintersWrapper::~PrintersWrapper()
{

}

bool PrintersWrapper::printLine(qint32 format, const QString &line)
{
//    lmWarning() << line;
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    QAndroidJniObject text = QAndroidJniObject::fromString(line);
    jboolean res = obj_->callMethod<jboolean>("printLine", "(ILjava/lang/String;)Z",
                                              format,  text.object<jstring>());
    IOSG::setPrinterStatus(lastErrorCode());
    return res;
}

bool PrintersWrapper::printBarCode(qint32 format, qint32 type, const QString &line)
{
    lmWarning()<< format << type << line;
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    QAndroidJniObject text = QAndroidJniObject::fromString(line);
    jboolean res = obj_->callMethod<jboolean>("printBarCode", "(Ljava/lang/String;II)Z",
                                              text.object<jstring>(), type, format);
    IOSG::setPrinterStatus(lastErrorCode());
    return res;
}

bool PrintersWrapper::printQrCode(qint32 format, const QString &line)
{
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    QAndroidJniObject text = QAndroidJniObject::fromString(line);
    jboolean res = obj_->callMethod<jboolean>("printQrCode", "(Ljava/lang/String;I)Z",
                                              text.object<jstring>(), format);
    IOSG::setPrinterStatus(lastErrorCode());
    return res;
}

bool PrintersWrapper::printBitMap(const QByteArray &data, qint32 format)
{
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    auto qByteArray2jbyteArray =[] (QByteArray buf) -> jbyteArray{
        QAndroidJniEnvironment env;
        jbyteArray array = env->NewByteArray(buf.length());
        env->SetByteArrayRegion (array, 0, buf.length(), reinterpret_cast<jbyte*>(buf.data()));
        return array;
    };
    jbyteArray jarray = qByteArray2jbyteArray(data);

    QAndroidJniObject param = QAndroidJniObject::fromLocalRef(jarray);

    jboolean res = obj_->callMethod<jboolean>("printBitMap", "([BI)Z",
                                              param.object<jbyteArray>(), format);
    IOSG::setPrinterStatus(lastErrorCode());
    return res;
}

qint32 PrintersWrapper::paperWidth()
{
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    jint res = obj_->callMethod<jint>("paperWidth");
    return res;
}

qint32 PrintersWrapper::imgBlockSize()
{
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    jint res = obj_->callMethod<jint>("imgBlockSize");
    return res;
}

bool PrintersWrapper::feed()
{
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    jboolean res = obj_->callMethod<jboolean>("feed");
    return res;
}

bool PrintersWrapper::printTextDocument(const TextPrinterDocument &doc)
{
    return startDocument() && printBlocks(doc.blocks()) && feed() && finishDocument();
}

CoreApiConst::ErrorCode PrintersWrapper::test()
{
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    jint res = obj_->callMethod<jint>("getStatus", "()I");
    IOSG::setPrinterStatus(lastErrorCode());
    lmWarning() << res;
    switch (res) {
    case 0: return CoreApiConst::ErrorCode::Ok;
    case 1: return CoreApiConst::ErrorCode::PrinterBusy;
    case 2: return CoreApiConst::ErrorCode::PrinterPaperLack;
    case 3: return CoreApiConst::ErrorCode::PrintreTooHot;
    case 4: return CoreApiConst::ErrorCode::PrinterPrintFail;
    case 5: return CoreApiConst::ErrorCode::PrinterFault;
    case 6: return CoreApiConst::ErrorCode::PrinterNoFontLib;
    case 7: return CoreApiConst::ErrorCode::PrinterOutOfMemory;
    default: return  CoreApiConst::ErrorCode::PrinterOtherError;
    }
}

bool PrintersWrapper::printBlock(const DocumentBlock &block)
{

    switch (block.type()) {
    case DocumentBlock::Text    : return printTextBlock(block);
    case DocumentBlock::Image   :
    {
        if(block.img().fileName().isEmpty())
        {
            return printFormattedImage(block.img(), 32, block.format());
        }
        else if(!QFile::exists(block.img().fileName())) return true;
        else
        {
            QImage img;
            if(!img.load(block.img().fileName())) return true;
            const qint32 ppwidth = paperWidth();
            qint32 width = qMin(img.width(), ppwidth);
            if(width < img.width()) img = img.scaledToWidth(width);
            return printQImage(img.convertToFormat(QImage::Format_Mono), block.format());
        }
    }break;
    case DocumentBlock::BarCode :
    {
        lmWarning() << logbinary(block.barcode()) << logtab << loglist(block.text());
        return printBarCode(
            block.format().format(),
            block.barCodeType(), QString::fromUtf8(block.barcode()));
    }
    case DocumentBlock::QrCode  : return printQrCode(
                    block.format().format(), QString::fromUtf8(block.qrCode()));
    }
}

bool PrintersWrapper::printBlocks(const DocumentBlocks &blocks)
{
    return std::all_of(blocks.constBegin(), blocks.constEnd(),
                       [this](const DocumentBlock &b) {
                           bool res =  printBlock(b);
                           lmWarning() << res << b.type() << logtab << loglist(b.text());
                           return res;
                       });
}

bool PrintersWrapper::hasPaper()
{
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    jboolean res = obj_->callMethod<jboolean>("ifHavePaper");
    return res;
}

/*
        PrinterBaseError                        = 0x000003E8,
        PrinterPrintFail                        = 0x000003E9,
        PrinterAddPrnStrFail                    = 0x000003EA,
        PrinterAddImgFail                       = 0x000003EB,
        PrinterBusy                             = 0x000003EC,
        PrinterPaperLack                        = 0x000003ED,
        PrinterWrongPackage                     = 0x000003EE,
        PrinterFault                            = 0x000003EF,
        PrintreTooHot                           = 0x000003F0,
        PrinterUnFinished                       = 0x000003F1,
        PrinterNoFontLib                        = 0x000003F2,
        PrinterOutOfMemory                      = 0x000003F3,
        PrinterOtherError                       = 0x000003CF,*/

CoreApiConst::ErrorCode PrintersWrapper::lastErrorCode()
{
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    jint res = obj_->callMethod<jint>("lastErrorCode");
    lmWarning() << res;
    if(res == 0) return CoreApiConst::ErrorCode::Ok;
    qint32 code = res;
    lmWarning() << res << code;
    switch (code) {
    case 0: return CoreApiConst::ErrorCode::Ok;
    case 1: return CoreApiConst::ErrorCode::PrinterBusy;
    case 2: return CoreApiConst::ErrorCode::PrinterPaperLack;
    case 3: return CoreApiConst::ErrorCode::PrintreTooHot;
    case 4: return CoreApiConst::ErrorCode::PrinterPrintFail;
    case 5: return CoreApiConst::ErrorCode::PrinterFault;
    case 6: return CoreApiConst::ErrorCode::PrinterNoFontLib;
    case 7: return CoreApiConst::ErrorCode::PrinterOutOfMemory;
    default: return  CoreApiConst::ErrorCode::PrinterOtherError;
    }
    return CoreApiConst::ErrorCode::Ok;
}

bool PrintersWrapper::printTextBlock(const DocumentBlock &block)
{
    qint32 spl = block.format().symbolsPerLine();
    QStringList text;
    std::function<void(QString)> processLine;
    processLine = [spl, &text, &processLine](QString line) -> void {
        if(line.isEmpty()) text << line;
        else if(line.contains("\n"))
        {
            QStringList sl = line.split("\n");
            for(const QString &l :sl) processLine(l);
        }
        else if(line.contains("\t"))
        {
            if(line.size() > spl)
            {
                QStringList sl = line.split("\t");
                if(sl.size() != 2 || sl.last().size() >= spl) processLine(line.replace("\t", " "));
                else
                {
                    if(sl.first().size() <= spl)
                    {
                        text << sl.first();
                        processLine("\t" + sl.last());
                    }
                    else
                    {
                        processLine(sl.first());
                        QString s = text.last();
                        if(s.size() + 1 +sl.last().size() > spl)processLine(line.replace("\t", " "));
                        else
                        {
                            s += "\t" + sl.last();
                            text.pop_back();
                            processLine(s);
                        }
                    }
                }
            }
            else
            {
                QStringList sl = line.split("\t");
                qint32 spcnt = (spl - line.size() + (sl.size() - 1)) /(sl.size() - 1);
                if(sl.size() == 2)
                {
                    text << (sl.first() + QString(spcnt, QLatin1Char(' '))+ sl.last()) ;
                }
                else
                {
                    line = sl.first();
                    for(int i = 1; i < sl.size(); ++i)
                    {
                        if(i == sl.size() - 1)
                        {
                            line += QString(spl - line.size() - sl[i].size(), QLatin1Char(' ')) + sl[i];
                        }
                        else
                        {
                            line += QString(spcnt, QLatin1Char(' ')) + sl[i];
                        }
                    }
                    text << line;
                }
            }
        }
        else if(line.size() <= spl) text << line;
        else if(!line.contains(" "))
        {
            text << line.mid(0, spl);
            processLine(line.mid(spl));
        }
        else
        {
            QStringList sl = line.split(" ");
            line.clear();
            for(int i = 0; i < sl.size(); ++i)
            {
                if(line.isEmpty())
                {
                    if(sl[i].size() > spl) processLine(sl[i]);
                    else if(sl[i].size() == spl) text << sl[i];
                    else line = sl[i];
                }
                else
                {
                    int sz = line.size() + sl[i].size() + 1;
                    if( sz < spl) line += " " + sl[i];
                    else if(sz == spl)
                    {
                        text << (line + " " + sl[i]);
                        line.clear();
                    }
                    else
                    {
                        text << line;
                        line.clear();
                        if(sl[i].size() > spl) processLine(sl[i]);
                        else if(sl[i].size() == spl) text << sl[i];
                        else line = sl[i];
                    }
                }
            }
            if(!line.isEmpty()) text << line;
//            lmWarning() << line << logtab << loglist(text);
        }
    };
    text.clear();
    for(const QString &s: block.text())
    {
//        lmWarning() << s;
        processLine(s);
//        lmWarning() << loglist(text);
    }
//    lmWarning() << block.format().format()<< loglist(text);
    return printLine(block.format().format(), text.join("\n"));
}

bool PrintersWrapper::printFormattedImage(const MonochromeImage &img, qint32 rawword, const LineFormat &f)
{
    const qint32 ppwidth = paperWidth();
    qint32 width = qMin(img.width(), ppwidth);

//	QBitmap bitmap(width, img.height());
    QImage qrImg(img.width(), img.height(), QImage::Format_RGB32);

    for(int i = 0; i < img.width(); ++i)
    {
        for(int j = 0; j < img.height(); ++j)
        {
            if(img.pixel(i, j, rawword))
            {
                qrImg.setPixelColor(i, j, QColor(Qt::black));
            }
            else
            {
                qrImg.setPixelColor(i, j, QColor(Qt::white));
            }
        }
    }
    if(qrImg.width() > width)
    {
        qrImg = qrImg.scaledToWidth(width, Qt::SmoothTransformation);
    }

    QBitmap * bufImage = new QBitmap(ppwidth, img.height() + ((img.height() % 8) ? 8 - (img.height() % 8)  : 0));
    bufImage->fill(QColor(Qt::white));
    QPainter painter;
    painter.begin(bufImage);
    QRect source(0, 0, width, img.height());

    qint32 newWidth = ppwidth;
    newWidth = qMin(width, newWidth);
    qint32 printableWidth = ppwidth;


    QRect dest((printableWidth - newWidth) / 2, 0, newWidth, img.height());
    painter.drawImage(dest, qrImg.scaledToWidth(newWidth), source);
    painter.end();

    bool res = printQImage(bufImage->toImage().convertToFormat(QImage::Format_Mono), f);
    delete bufImage;

    return res;
}

bool PrintersWrapper::printQImage(const QImage &img, const LineFormat &f)
{
    const qint32 bs = imgBlockSize();
    QList <QImage> imageCuts;

    for(int i = 0; i < img.height(); i += bs)
    {

        QImage block = img.copy(0, i, img.width(), qMin(bs, img.height() - i));
        if(img.sizeInBytes() <= 0) break;
        imageCuts.append(block);
    }


    for (const QImage &img: imageCuts)
    {
        QBuffer buffer;
        img.save(&buffer, "BMP", -1);
        if(!printBitMap(buffer.data(), f.format()))
        {
            return false;
        }
    }
    return true;
}

bool PrintersWrapper::startDocument()
{
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    jboolean res = obj_->callMethod<jboolean>("startDocument");
    IOSG::setPrinterStatus(lastErrorCode());
    return res;
}

bool PrintersWrapper::finishDocument()
{
    QAndroidJniExceptionCleaner cleaner(QAndroidJniExceptionCleaner::OutputMode::Silent);
    Q_UNUSED(cleaner)
    jboolean res = obj_->callMethod<jboolean>("finishDocument");
    IOSG::setPrinterStatus(lastErrorCode());
    return res;
}
