#include "approothandler.h"
#include "androidapihelper.h"
#include "cashboxroot.h"
#include "regdata.h"
#include "appdirectories.h"
#include "logmanip.h"
#include "cashboxbuildconfig.h"

#include <QCoreApplication>
#include <QFile>
#include <QCryptographicHash>
#include <QStandardPaths>
#include <QFutureWatcher>
#include <QtConcurrent>
#include <QEventLoop>

AppRootHandler::AppRootHandler(QObject *parent)
    : BaseCoreHandler(parent)
    , android_(new AndroidApiHelper(this))
    , status_(new CashboxStatusWrapper(this))
    , firmwareTestText_()
{
    connect(status_, &CashboxStatusWrapper::statusChanged, this, &AppRootHandler::onStatusChanged);
    connect(appRoot, &CashboxRoot::popPage, this, &AppRootHandler::popPage);
}

AppRootHandler::~AppRootHandler()
{

}

QString AppRootHandler::getDsn() const
{
    return android_ ? android_->getDsn() : QString();
}

bool AppRootHandler::loginned() const
{
    return session_ && session_->cashier().isValid();
    //    return Umka365Session::currentSession().isLoggined();
}

QString AppRootHandler::homePage() const
{
    if(!status_->hasSerial()) return "WriteSerialPage.qml";
    if(status_->status().needFiscalization()) return "FiscalizationPage.qml";
    if(status_->fiscalized() && status_->cycleIsOpen())
    {
        if(status_->cycleIsOver() && session_->cashier().hasPermission(Cashier::Permission::CycleClose))
            return "CycleClosePage.qml";
        if(session_->cashier().hasPermission(Cashier::Permission::Receipts))
            return "SaleView.qml";
        if(session_->cashier().hasPermission(Cashier::Permission::ReturnReceipts))
            return "RefundView.qml";
    }
    return "HomeForm.qml";
}

QString AppRootHandler::appHeader() const
{
    if(!status_) return QString();
    ModelData md = status_->status().modelData();
    RegData rd = status_->status().regData();
    QString res = md.modelName();
    if(md.isValid()) res += " " + md.serial();
//    if(rd.isValid()) res+= "\n" + rd.orgName();
    return res;
}

QString AppRootHandler::testText() const
{
    QStringList res;
    res << "<html><head>"
        << "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">"
        << "</head><body>" ;
    CashboxStatus st = status_->status();
    res.append(st.modelData().toHtml().split("\n"));
    res << QStringLiteral("<h3 align=\"center\">ТЕСТИРОВАНИЕ ПО</h3>");
    if(st.isFiscalized()) res.append(st.regData().toHtml().split("\n"));
    res.append(st.fs().toHtml());

    res << QStringLiteral("<div>IP: %1</div>").arg(st.dev().ipAddress());
    QList<QNetworkInterface> interfaces = QNetworkInterface::allInterfaces();
    QString inet = "Нет связи с Интернет";
    for(const QNetworkInterface &i :interfaces)
    {
        if(i.flags().testFlag(QNetworkInterface::IsLoopBack) ||
                !i.flags().testFlag(QNetworkInterface::IsUp)) continue;
        QString ifaceTypeName;
        switch (i.type()) {
        case QNetworkInterface::Virtual     : ifaceTypeName = "виртуальный"; break;
        case QNetworkInterface::Ethernet    : ifaceTypeName = "ethernet"; break;
        case QNetworkInterface::Slip        : ifaceTypeName = "SLIP"; break;
        case QNetworkInterface::CanBus      : ifaceTypeName = "CanBas"; break;
        case QNetworkInterface::Ppp         : ifaceTypeName = "PPP"; break;
        case QNetworkInterface::Fddi        : ifaceTypeName = "FDDI"; break;
        case QNetworkInterface::Wifi        : ifaceTypeName = "WIFI"; break;
        case QNetworkInterface::Phonet      : ifaceTypeName = "Модем"; break;
        case QNetworkInterface::Ieee802154  : ifaceTypeName = "WPAN"; break;
        case QNetworkInterface::SixLoWPAN   : ifaceTypeName = "WPAN"; break;
        case QNetworkInterface::Ieee80216   : ifaceTypeName = "WiMAX"; break;
        case QNetworkInterface::Ieee1394    : ifaceTypeName = "FireWire"; break;
        default:
            ifaceTypeName = ifaceTypeName = AndroidApiHelper().currentNetworkInterfaceName();
            lmWarning() << i.type() << i.flags() << i.name() << i.hardwareAddress() << i.humanReadableName();
            break;
        }
        if(!ifaceTypeName.isEmpty()) ifaceTypeName = " (" + ifaceTypeName + ")";
        inet = QStringLiteral("Интернет:\t%1%2").arg(i.humanReadableName())
                .arg(ifaceTypeName);
    }
    res << QStringLiteral("<div align=\"center\">%1</div>").arg(inet);
    QStorageInfo si(APATH.workDir());
    res <<  QStringLiteral("<div align=\"center\"> Доступно %1/%2 байт<div>").arg(si.bytesFree()).arg(si.bytesTotal());
    res << QStringLiteral("<h3 align=\"center\">УСПЕШНО!</h3>");
    res	<< "</body></html>";
//    lmWarning() << loglist(res);
    return res.join("\n");
}

void AppRootHandler::reloadStatus()
{
    emit showBusy(true);
    std::function<QPair<CoreTransaction, CoreApiResult>()> task =
            [this]() -> QPair<CoreTransaction, CoreApiResult> {
        CoreApiResult res = api_->loadCashboxStatus();
        CoreTransaction trans;
        trans.setOperation(CoreTransaction::Operation::CASHBOX_STATUS);
        return qMakePair(trans, res);
    };
    execTask(task);
}

void AppRootHandler::selfTest()
{
    emit showBusy(true);
    std::function<QPair<CoreTransaction, CoreApiResult>()> task =
            [this]() -> QPair<CoreTransaction, CoreApiResult> {
        CoreTransaction transaction;
        transaction.setCashierLogin(session_->cashier().login());
        transaction.setCashierPassword(session_->cashier().password());
        transaction.setOperation(CoreTransaction::Operation::SELF_TEST);
        transaction.genUid();
        CoreApiResult res = api_->exec(transaction);
        lmWarning() << logvariant(res.toMap());
        api_->loadCashboxStatus();
        return qMakePair(transaction, res);
    };
    execTask(task);
}

void AppRootHandler::restart()
{
    qApp->exit(CashboxRoot::RESTART_CODE);
}

void AppRootHandler::testFirmware(const QString &fn)
{
    firmwareTestText_.clear();
    if(fn.isEmpty()) emit error("Эталонная прошивка не выбрана!");
    else
    {
        emit showBusy(true);
        QString err;
        QEventLoop loop;

        auto exec = [&err, this, fn] () -> bool {
            QString path = fn.trimmed();
            if(path.startsWith("file://")) path = path.remove("file://");
            else if(!QFile::exists(path))
            {
                path = SDCARD_PATH + path;
            }
            AndroidApiHelper ap(nullptr);
            QString etVer = ap.fileVersionName(path).trimmed();
            QString etPkg = ap.filePackageName(path).trimmed();
            qint64 etVerCode = ap.fileVersionCode(path);

            QStringList ettext;
            QStringList curtext;
            ettext << "Эталонная прошивка:"
                << QStringLiteral("Пакет: %1").arg(etPkg)
                << QStringLiteral("Версия: %1").arg(etVer);
            curtext << "Текущая прошивка:"
                << QStringLiteral("Пакет: %1").arg(ap.packageName())
                << QStringLiteral("Версия: %1").arg(ap.versionName());

            auto createText = [this](const QStringList &s1, const QStringList &s2, const QString &text) {
                QStringList sl = s1 + s2;
                sl << text;
                firmwareTestText_ = sl.join("\n");
            };


            lmWarning() << ap.packageName() << ap.versionCode() << ap.versionName()  << ap.packagePath()<< logtab
                        << ap.filePackageName(path) << ap.fileVersionCode(path) << ap.fileVersionName(path);
            if(etVer.isEmpty() || etPkg.isEmpty() || etVerCode <= 0)
            {
                err = "Ошибка обработки эталонной прошивки. Возможно это не APK";
                createText(ettext, curtext, err);
                return false;
            }
            else if(etPkg != ap.packageName())
            {
                err = QString("Ошибка обработки эталонной прошивки: \nпредоставлен APK файл другого приложения.\n"
                            "(Должен быть %1, а предоставлен %2)").arg(ap.packageName()).arg(etPkg);
                createText(ettext, curtext, err);
                return false;
            }
            else if(etVerCode != ap.versionCode())
            {
                err = "Ошибка обработки эталонной прошивки: \nВерсия текущей прошивки отличается от версии эталонной прошивки\n";
                createText(ettext, curtext, err);
                return false;
            }
            else
            {
                QFile etF(path);
                QByteArray h1, h2;
                if(!etF.open(QIODevice::ReadOnly))
                {
                    err = "Ошибка чтения файла эталонной прошивки";
                    return false;
                }
                else
                {
                    h1 = QCryptographicHash::hash(etF.readAll(), QCryptographicHash::Md5);
                    etF.close();
                    QFile f(ap.packagePath().trimmed());
                    if(!f.open(QIODevice::ReadOnly))
                    {
                        err = "Ошибка чтения файла прошивки";
                        createText(ettext, curtext, err);
                        return false;
                    }
                    h2 = QCryptographicHash::hash(f.readAll(), QCryptographicHash::Md5);
                    f.close();
                    lmWarning() << h1.toHex() << logtab << h2.toHex();
                    ettext << QStringLiteral("MD5: %1").arg(QString::fromLatin1(h1.toHex()));
                    curtext << QStringLiteral("MD5: %1").arg(QString::fromLatin1(h2.toHex()));
                    if(h1 == h2)
                    {
                        createText(ettext, curtext, "Текушая прошивка идентична эталонной");
                        return true;
                    }
                    else
                    {
                        err = "Хеш суммы текущей прошивки и эталонной не совпадают";
                        createText(ettext, curtext, err);
                        return false;
                    }
                }

            }
        };

        QFutureWatcher<bool> *watcher = new QFutureWatcher<bool>();
        connect(watcher, &QFutureWatcher<bool>::finished, &loop, &QEventLoop::quit);
        QFuture<bool> res = QtConcurrent::run(exec);
        watcher->setFuture(res);
        loop.exec();
        res = watcher->future ();
        watcher->deleteLater();
        if(res.result()) emit successTest();
        else emit error(err);
    }
}

void AppRootHandler::updateFirmware(const QString &fn)
{
    firmwareTestText_.clear();
    emit showBusy(true);
    if(fn.isEmpty()) emit error("Новая прошивка не выбрана!");
    else
    {
        QString path = fn.trimmed();
        if(path.startsWith("file://")) path = path.remove("file://");
        else if(!QFile::exists(path))
        {
            path = SDCARD_PATH + path;
        }
        AndroidApiHelper ap(nullptr);
        QString etVer = ap.fileVersionName(path).trimmed();
        QString etPkg = ap.filePackageName(path).trimmed();
        qint64 etVerCode = ap.fileVersionCode(path);

        QStringList ettext;
        QStringList curtext;
        ettext << "Новая прошивка:"
            << QStringLiteral("Пакет: %1").arg(etPkg)
            << QStringLiteral("Версия: %1").arg(etVer);
        curtext << "Текущая прошивка:"
            << QStringLiteral("Пакет: %1").arg(ap.packageName())
            << QStringLiteral("Версия: %1").arg(ap.versionName());

        auto createText = [this](const QStringList &s1, const QStringList &s2, const QString &text) {
            QStringList sl = s1 + s2;
            sl << text;
            firmwareTestText_ = sl.join("\n");
        };


        lmWarning() << ap.packageName() << ap.versionCode() << ap.versionName()  << ap.packagePath()<< logtab
                    << ap.filePackageName(path) << ap.fileVersionCode(path) << ap.fileVersionName(path);
        if(etVer.isEmpty() || etPkg.isEmpty() || etVerCode <= 0)
        {
            createText(ettext, curtext, "Ошибка обработки новой прошивки. Возможно это не APK");
            emit error("Ошибка обработки новой прошивки. Возможно это не APK");
        }
        else if(etPkg != ap.packageName())
        {
            createText(ettext, curtext, QString("Ошибка обработки новой прошивки: \nпредоставлен APK файл другого приложения.\n"
                        "(Должен быть %1, а предоставлен %2)").arg(ap.packageName()).arg(etPkg));
            emit error(QString("Ошибка обработки новой прошивки: \nпредоставлен APK файл другого приложения.\n"
                        "(Должен быть %1, а предоставлен %2)").arg(ap.packageName()).arg(etPkg));
        }
        else if(etVerCode <= ap.versionCode())
        {
            createText(ettext, curtext,
                       "Ошибка обработки новой прошивки: \nВерсия текущей прошивки больше версии прошивки\n");
            emit error("Ошибка обработки новой прошивки: \nВерсия текущей прошивки больше версии прошивки\n");
        }
        else
        {
            ap.installPackage(path);
        }
        emit successTest();
    }
}

void AppRootHandler::firmwareToSdcard()
{
    emit showBusy(true);

    QString err;
    QEventLoop loop;

    auto exec = [&err] () -> bool {
        err.clear();
        AndroidApiHelper ap(nullptr);
        QStringList pathes = ap.sdcardStorage().split(":::");
        QString path = SDCARD_PATH;

        for(const QString &s: pathes)
        {
            if(s.startsWith(SDCARD_PATH))
            {
                lmWarning() << s;
                path = s + "/";
                break;
            }
        }

        path = path + ap.packageName() + QDateTime::currentDateTime().toString("_ddMMyyhhmmss") + ".apk";
        lmWarning() << ap.sdcardStorage() << logtab  <<  path << logtab <<loglist(pathes);
        QFile f(ap.packagePath().trimmed());
        if(!f.copy(path))
        {
            err =  tr("Ошибка копирования файла:") + f.errorString();
            return false;
        }
        err = "Файл прошивки скопирован на внешнее ЗУ";
        return true;
    };

    QFutureWatcher<bool> *watcher = new QFutureWatcher<bool>();
    connect(watcher, &QFutureWatcher<bool>::finished, &loop, &QEventLoop::quit);
    QFuture<bool> res = QtConcurrent::run(exec);
    watcher->setFuture(res);
    loop.exec();
    watcher->deleteLater();
    emit error(err);
}

void AppRootHandler::printNotPrinted()
{
    emit showBusy(true);
    std::function<QPair<CoreTransaction, CoreApiResult>()> task =
            [this]() -> QPair<CoreTransaction, CoreApiResult> {
        CoreTransaction trans;
        trans.setCashierLogin(session_->cashier().login());
        trans.setCashierPassword(session_->cashier().password());
        trans.setOperation(CoreTransaction::Operation::PRINT_NOT_PRINTED);
        trans.genUid();
        CoreApiResult res = api_->exec(trans);
        QPair<CoreTransaction, CoreApiResult>resp = qMakePair(trans, res);
        api_->loadCashboxStatus(false);
        return resp;
    };
    execTask(task);
}

void AppRootHandler::fsClean()
{
    emit showBusy(true);
    std::function<QPair<CoreTransaction, CoreApiResult>()> task =
            [this]() -> QPair<CoreTransaction, CoreApiResult> {
        CoreTransaction trans;
        trans.setCashierLogin(session_->cashier().login());
        trans.setCashierPassword(session_->cashier().password());
        trans.setOperation(CoreTransaction::Operation::FS_CLEAN);
        trans.genUid();
        CoreApiResult res = api_->exec(trans);
        QPair<CoreTransaction, CoreApiResult>resp = qMakePair(trans, res);
        api_->loadCashboxStatus(false);
        return resp;
    };
    execTask(task);
}

void AppRootHandler::totalClean(const QString &login, const QString &password)
{
    emit showBusy(true);
    std::function<QPair<CoreTransaction, CoreApiResult>()> task =
        [this, login, password]() -> QPair<CoreTransaction, CoreApiResult> {
        CoreTransaction trans;
        trans.setCashierLogin(login);
        trans.setCashierPassword(password);
        trans.setOperation(CoreTransaction::Operation::FULL_CLEAR);
        trans.genUid();
        CoreApiResult res = api_->exec(trans);
        QPair<CoreTransaction, CoreApiResult>resp = qMakePair(trans, res);
        api_->loadCashboxStatus(false);
        return resp;
    };
    execTask(task);
}

QString AppRootHandler::firmwareTestText() const
{
    return firmwareTestText_;
}

QString AppRootHandler::sdcardPath() const
{
    QString res = "file://" +SDCARD_PATH;
    while (res.endsWith("/")) {
        res = res.mid(0, res.size() - 1);
    }
    return res;
}

void AppRootHandler::processResult(const CoreTransaction &source, const CoreApiResult &res)
{
    Q_UNUSED(source)
    Q_UNUSED(res)
    emit showBusy(false);
    if(source.operation() == CoreTransaction::Operation::SELF_TEST)
    {
        if(!res.isOk()) emit error("Ошибка тестирования: " + res.descr());
        else emit successTest();
    }
    if(source.operation() == CoreTransaction::Operation::PRINT_NOT_PRINTED)
    {
        lmWarning() << res.isOk() << res.descr();
        if(!res.isOk())
        {
            QAndroidJniObject param = QAndroidJniObject::fromString(res.descr());
            QtAndroid::androidActivity().callMethod<void>("showToast", ("(Ljava/lang/String;Z)V"),
                                                                param.object<jstring>(), true);
        }
        else emit successNotPrinted();
    }
    if(source.operation() == CoreTransaction::Operation::FS_CLEAN)
    {
        lmWarning() << res.isOk() << res.descr();
        if(res.isOk()) emit successClean();
    }
    if(source.operation() == CoreTransaction::Operation::FULL_CLEAR)
    {
        lmWarning() << res.isOk() << res.descr();
        if(res.isOk()) emit successTotalClean();
        else emit error(res.descr());
    }
}

void AppRootHandler::onStatusChanged()
{
    emit appHeaderChanged(appHeader());
}
