#include "receipthandler.h"
#include "cashboxbuildconfig.h"
#include "logmanip.h"

ReceiptHandler::ReceiptHandler(QObject *parent)
    : BaseCoreHandler(parent)
    , uiRuntime_(new RuntimeUiSettings(this))
    , skipPrint_(false)
{

}

ReceiptHandler::~ReceiptHandler()
{

}

bool ReceiptHandler::skipPrint() const
{
    return skipPrint_;
}

void ReceiptHandler::setSkipPrint(bool newSkipPrint)
{
    if(newSkipPrint != skipPrint_)
    {
        skipPrint_ = newSkipPrint;
        emit skipPrintChanged();
    }
}

void ReceiptHandler::receiptFast(int type, const QVariantMap &data, double cash, double card)
{
    receipt(type, data, cash, card, 0.0, 0.0, 0.0);
}

void ReceiptHandler::correctionFast(int type, const QVariantMap &data, double cash, double card)
{
    correction(type, data, cash, card, 0.0, 0.0, 0.0);
}

void ReceiptHandler::receipt(int type, const QVariantMap &data, double cash, double card, double prepay, double postpay, double barter)
{
    Receipt r(data);
    uiRuntime_->setLastTaxSystem(r.tax());
    r.setPaymentAttr(static_cast<fdf::PaymentType>(type));
    if(!qFuzzyIsNull(cash))r.setCash(FixNumber(DEFAULT_AMOUNT_MULT, cash));
    if(!qFuzzyIsNull(card))r.setCard(FixNumber(DEFAULT_AMOUNT_MULT, card));
    if(!qFuzzyIsNull(prepay))r.setPrepay(FixNumber(DEFAULT_AMOUNT_MULT, prepay));
    if(!qFuzzyIsNull(postpay))r.setPostpay(FixNumber(DEFAULT_AMOUNT_MULT, postpay));
    if(!qFuzzyIsNull(barter))r.setBarter(FixNumber(DEFAULT_AMOUNT_MULT, barter));

    lmWarning() << r.amount().toString() << logtab
                << cash << logtab
                   << card << logtab
                      << prepay << logtab
                         << postpay << logtab
                << barter  << logtab
                << skipPrint_;
    emit showBusy(true);
    CoreApiConst::ErrorCode err = CoreApiConst::ErrorCode::Ok;
    QString msg;
    if(!r.isValid(CashboxStatusWrapper().status().regData(), err, msg))
    {
        lmWarning() << r.cashier() << logvariant(data);
        emit showBusy(false);
        emit error(msg);
        return;
    }
    std::function<QPair<CoreTransaction, CoreApiResult>()> task =
            [this, r]() -> QPair<CoreTransaction, CoreApiResult> {
        CoreTransaction transaction;
        transaction.setCashierLogin(session_->cashier().login());
        transaction.setCashierPassword(session_->cashier().password());
        transaction.setOperation(CoreTransaction::Operation::RECEIPT);
        transaction.setParams(r.toMap());
        transaction.setPrintResult(!skipPrint_);
        transaction.genUid();
        lmWarning() << logvariant(transaction.toMap());
        CoreApiResult res = api_->exec(transaction);
        lmWarning() << logvariant(res.toMap());
        api_->loadCashboxStatus();
        return qMakePair(transaction, res);
    };
    execTask(task);
}

void ReceiptHandler::correction(int type, const QVariantMap &data, double cash, double card, double prepay, double postpay, double barter)
{
    Receipt r(data);
    uiRuntime_->setLastTaxSystem(r.tax());
    r.setPaymentAttr(static_cast<fdf::PaymentType>(type));
    if(!qFuzzyIsNull(cash))r.setCash(FixNumber(DEFAULT_AMOUNT_MULT, cash));
    if(!qFuzzyIsNull(card))r.setCard(FixNumber(DEFAULT_AMOUNT_MULT, card));
    if(!qFuzzyIsNull(prepay))r.setPrepay(FixNumber(DEFAULT_AMOUNT_MULT, prepay));
    if(!qFuzzyIsNull(postpay))r.setPostpay(FixNumber(DEFAULT_AMOUNT_MULT, postpay));
    if(!qFuzzyIsNull(barter))r.setBarter(FixNumber(DEFAULT_AMOUNT_MULT, barter));
    lmWarning() << r.amount().toString() << logtab
                << cash << logtab
                   << card << logtab
                      << prepay << logtab
                         << postpay << logtab
                            << barter;
    emit showBusy(true);
    CoreApiConst::ErrorCode err = CoreApiConst::ErrorCode::Ok;
    QString msg;
    if(!r.isValid(CashboxStatusWrapper().status().regData(), err, msg))
    {
        emit showBusy(false);
        emit error(msg);
        return;
    }
    std::function<QPair<CoreTransaction, CoreApiResult>()> task =
            [this, r]() -> QPair<CoreTransaction, CoreApiResult> {
        CoreTransaction transaction;
        transaction.setCashierLogin(session_->cashier().login());
        transaction.setCashierPassword(session_->cashier().password());
        transaction.setOperation(CoreTransaction::Operation::CORRECTION);
        transaction.setParams(r.toMap());
        transaction.setPrintResult(!skipPrint_);
        transaction.genUid();
        lmWarning() << logvariant(transaction.toMap());
        CoreApiResult res = api_->exec(transaction);
        lmWarning() << logvariant(res.toMap());
        api_->loadCashboxStatus();
        return qMakePair(transaction, res);
    };
    execTask(task);
}

void ReceiptHandler::processResult(const CoreTransaction &source, const CoreApiResult &res)
{
    lmWarning() << res.isOk();
    if(!res.isOk())emit error(res.descr());
    emit showBusy(false);
    switch (source.operation()) {
    case CoreApiConst::RemoteCommand::RECEIPT:
    case CoreApiConst::RemoteCommand::CORRECTION:
    {
        if(res.isOk())
        {
            emit success();
            QAndroidJniObject param = QAndroidJniObject::fromString(tr("Успешно!"));
            QtAndroid::androidActivity().callMethod<void>("showToast", ("(Ljava/lang/String;Z)V"),
                                                          param.object<jstring>(), true);
        }
    }break;
    default:
        break;
    }
}
