#include "cashiersmodel.h"

#include "logmanip.h"


CashiersModel::CashiersModel(QObject *parent)
    : BaseCatalogModel(parent)
    , cashiers_()
{

}

CashiersModel::~CashiersModel()
{

}

QHash<int, QByteArray> CashiersModel::roleNames() const
{
    return {
        { RoleId, "cid" },
        { RoleServerId, "cserverid" },
        { RoleLogin, "clogin" },
        { RoleName, "cname" },
        { RolePassword, "cpassword" },
        { RoleInn, "cinn" },
        { RolePermissions, "cpermissions" }
    };
}

QVariant CashiersModel::data(const QModelIndex &idx, int role) const
{
    if(!idx.isValid() || idx.row() >= cashiers_.size())return QVariant();
    Cashier c = cashiers_[idx.row()];
    switch (role) {
    case RoleId     : return c.id();
    case RoleServerId     : return c.serverId();
    case RoleLogin       : return c.login();
    case RoleName       : return c.name();
    case RolePassword   : return c.password();
    case RoleInn   : return c.inn();
    case RolePermissions: return static_cast<qint32>(c.permissions());
    }
    return QVariant();
}

int CashiersModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return cashiers_.size();
}

bool CashiersModel::canRemoveCashier(int idx)
{
    if(idx < 0 || idx >= cashiers_.size()) return false;
    Cashier c = cashiers_[idx];
    return c.serverId() <= 0 && !c.isSuperUser();
}

void CashiersModel::removeCashier(int index)
{
    if(index < 0 || index >= rowCount()) return;
    Cashier c = cashiers_[index];
    emit showBusy(true);

    std::function<QPair<CoreTransaction, CoreApiResult>()> task =
            [this, c]() -> QPair<CoreTransaction, CoreApiResult> {
        CoreTransaction transaction;
        transaction.setCashierLogin(session_->cashier().login());
        transaction.setCashierPassword(session_->cashier().password());
        transaction.setOperation(CoreTransaction::Operation::REMOVE_CASHIER);
        QVariantMap params;
        params["cashierId"] = c.id();
        transaction.setParams(params);
        transaction.genUid();
        lmWarning() << logvariant(transaction.toMap());
        CoreApiResult res = api_->exec(transaction);
        lmWarning() << logvariant(res.toMap());
        return qMakePair(transaction, res);
    };
    execTask(task);
}

void CashiersModel::doReload()
{
    emit showBusy(true);
    std::function<QPair<CoreTransaction, CoreApiResult>()> task =
            [this]() -> QPair<CoreTransaction, CoreApiResult> {
        CoreTransaction transaction;
        transaction.setCashierLogin(session_->cashier().login());
        transaction.setCashierPassword(session_->cashier().password());
        transaction.setOperation(CoreTransaction::Operation::DOWNLOAD_CASHIERS);
        transaction.genUid();
        lmWarning() << logvariant(transaction.toMap());
        CoreApiResult res = api_->exec(transaction);
        lmWarning() << logvariant(res.toMap());
        return qMakePair(transaction, res);
    };
    execTask(task);
}

void CashiersModel::processResult(const CoreTransaction &source, const CoreApiResult &res)
{
    lmWarning() << res.isOk();
    switch (source.operation()) {
    case CoreApiConst::RemoteCommand::DOWNLOAD_CASHIERS:
    {
        if(!res.isOk())emit error(res.descr());
        else
        {
            beginResetModel();
            ActiveCashiersList cashiers = Cashier::parseCashiers(res.data());
            cashiers_.clear();
            for(int i = 0; i < cashiers.size(); ++i)
            {
                if(cashiers[i].isActive())
                {
                    cashiers_ << cashiers[i];
                    cashiers.removeAt(i--);
                }
            }
            std::sort(cashiers_.begin(), cashiers_.end(),
                      [](const Cashier &a, const Cashier &b) -> bool {
                return a.id() < b.id();
            });
            std::sort(cashiers.begin(), cashiers.end(),
                      [](const Cashier &a, const Cashier &b) -> bool {
                return a.id() < b.id();
            });
            cashiers_.append(cashiers);
            endResetModel();
        }
    }break;
    case CoreApiConst::RemoteCommand::SAVE_CASHIER:
    {
        if(!res.isOk())emit itemSavingError(res.descr());
        else
        {
            beginResetModel();
            ActiveCashiersList cashiers = cashiers_;
            Cashier cashier(res.data());
            cashiers_.clear();
            bool changed = false;
            for(int i = 0; i < cashiers.size(); ++i)
            {
                if(cashiers[i].id() == cashier.id())
                {
                    cashiers[i] = cashier;
                    changed = true;
                }
                if(cashiers[i].isActive())
                {
                    cashiers_ << cashiers[i];
                    cashiers.removeAt(i--);
                }
            }
            if(!changed)
            {
                if(cashier.isActive()) cashiers_ << cashier;
                else cashiers << cashier;
            }
            std::sort(cashiers_.begin(), cashiers_.end(),
                      [](const Cashier &a, const Cashier &b) -> bool {
                return a.id() < b.id();
            });
            std::sort(cashiers.begin(), cashiers.end(),
                      [](const Cashier &a, const Cashier &b) -> bool {
                return a.id() < b.id();
            });
            cashiers_.append(cashiers);
            endResetModel();
            emit itemSaved();
        }
    }break;
    case CoreApiConst::RemoteCommand::REMOVE_CASHIER:
    {
        if(!res.isOk())emit itemSavingError(res.descr());
        else
        {
            beginResetModel();
            qint32 id = res.data()["cashierId"].toInt();
            for(int i = 0; i < cashiers_.size(); ++i)
            {
                if(cashiers_[i].id() == id)
                {
                    cashiers_.removeAt(i);
                    break;
                }
            }
            endResetModel();
        }
    }break;
    default:
        break;
    }
    emit showBusy(false);
}

QVariantMap CashiersModel::prepareItem(int index) const
{
    if(index < 0 || index >= rowCount()) return QVariantMap();
    return cashiers_[index].toMap();
}

