#include "industrypropsmodel.h"

#include "logmanip.h"

IndustryPropsModel::IndustryPropsModel(QObject *parent)
    : QAbstractListModel(parent)
    , props_()
{

}

IndustryPropsModel::~IndustryPropsModel()
{

}

QHash<int, QByteArray> IndustryPropsModel::roleNames() const
{
    return {
        {RoleFedId     , "fedId"},
        {RoleDocDate   , "docDate"},
        {RoleDocNumber , "docNumber"},
        {RoleValue     , "propValue"},
    };
}

QVariant IndustryPropsModel::data(const QModelIndex &idx, int role) const
{
    if(idx.row() >=0 && idx.row() < props_.count())
    {
        const IndustryProperty &p = props_[idx.row()];
        switch (role)
        {
        case RoleFedId    : return p.fedId();
        case RoleDocDate  : return p.docDt().toString("dd.MM.yyyy");
        case RoleDocNumber: return p.docNumber();
        case RoleValue    : return p.value();
        }
    }
    return QVariant();
}

int IndustryPropsModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return props_.count();
}

QVariantList IndustryPropsModel::props() const
{
    QVariantList res;
    std::for_each(props_.constBegin(), props_.constEnd(),
                  [&res](const IndustryProperty &p){res << p.toMap();});
    return res;
}

int IndustryPropsModel::count() const
{
    return props_.count();
}

QVariantMap IndustryPropsModel::item(int idx)
{
    if(idx >= 0 && idx < props_.size())
    {
        return props_[idx].toMap();
    }
    return QVariantMap();
}

void IndustryPropsModel::save()
{
    if(props_.isEmpty())
    {
        emit saveData(QVariantList());
        emit finish();
        return;
    }
    CoreApiConst::ErrorCode err = CoreApiConst::ErrorCode::Ok;
    QString msg;
    if(!std::all_of(props_.constBegin(), props_.constEnd(),
                    [&err, &msg](const IndustryProperty &p) ->bool {
        return p.isValid(false, err, msg);
    }))
    {
        emit error(msg);
    }
    else
    {
        emit saveData(props());
        emit finish();
    }
}

void IndustryPropsModel::clean()
{
    beginResetModel();
    props_.clear();
    endResetModel();
}

void IndustryPropsModel::addItem()
{
    beginInsertRows(QModelIndex(), props_.size(), props_.size());
    props_ << IndustryProperty();
    endInsertRows();
}

void IndustryPropsModel::setItem(int idx, const QVariantMap &data)
{
    lmWarning() << idx << logvariant(data);
    if(idx >= 0 && idx < props_.size())
    {
        beginResetModel();
        props_[idx] = IndustryProperty(data);
        endResetModel();
    }
    else
    {
        beginInsertRows(QModelIndex(), props_.size(), props_.size());
        props_.push_back(IndustryProperty(data));
        endInsertRows();
    }
}

void IndustryPropsModel::removeItem(int idx)
{
    if(idx >= 0 && idx < props_.size())
    {
        beginRemoveRows(QModelIndex(), idx, idx);
        props_.removeAt(idx);
    }
}

void IndustryPropsModel::setProps(const QVariantList &data)
{
    beginResetModel();
    props_.clear();
    std::for_each(data.constBegin(), data.constEnd(),
                  [this](const QVariant &v){props_ << IndustryProperty(v.toMap());});
    endResetModel();
}

//--------------------------------------------------------------------------------------------------
IndustryPropertyWrapper::IndustryPropertyWrapper(QObject *parent)
    : QObject(parent)
    , prop_()
{

}

IndustryPropertyWrapper::~IndustryPropertyWrapper()
{

}

bool IndustryPropertyWrapper::isValid() const
{
    return  prop_.isValid();
}

QVariantMap IndustryPropertyWrapper::prop() const
{
    return prop_.toMap();
}

void IndustryPropertyWrapper::setProp(const QVariantMap &newProp)
{
    IndustryProperty p(newProp);
    if(p != prop_)
    {
        prop_ = p;
        emit fedIdChanged();
        emit docDtChanged();
        emit docNumberChanged();
        emit valueChanged();
        emit propChanged();
        emit isValidChanged();
    }
}

const QString &IndustryPropertyWrapper::fedId() const
{
    return prop_.fedId();
}

void IndustryPropertyWrapper::setFedId(const QString &newFedId)
{
    if (prop_.fedId() == newFedId)
        return;
    prop_.setFedId(newFedId);
    emit fedIdChanged();
    emit propChanged();
    emit isValidChanged();

}

QString IndustryPropertyWrapper::docDt() const
{
    return prop_.docDt().toString("dd.MM.yyyyy");
}

void IndustryPropertyWrapper::setDocDt(const QString &newDocDt)
{
    QDate d =QDate::fromString(newDocDt, "dd.MM.yyyy");
    if (prop_.docDt() == d)
        return;
    prop_.setDocDt(d);
    emit docDtChanged();
    emit propChanged();
    emit isValidChanged();

}

const QString &IndustryPropertyWrapper::docNumber() const
{
    return prop_.docNumber();
}

void IndustryPropertyWrapper::setDocNumber(const QString &newDocNumber)
{
    if (prop_.docNumber() == newDocNumber)
        return;
    prop_.setDocNumber(newDocNumber);
    emit docNumberChanged();
    emit propChanged();
    emit isValidChanged();

}

const QString &IndustryPropertyWrapper::value() const
{
    return prop_.value();
}

void IndustryPropertyWrapper::setValue(const QString &newValue)
{
    if (prop_.value() == newValue)
        return;
    prop_.setValue(newValue);
    emit valueChanged();
    emit propChanged();
    emit isValidChanged();
}
