#include "receiptmodel.h"
#include "logmanip.h"
#include "cashboxbuildconfig.h"
#include "runtimeuisettings.h"

ReceiptModel::ReceiptModel(QObject *parent)
    : QAbstractListModel(parent)
    , status_(new CashboxStatusWrapper(this))
    , session_(new SessionWrapper(this))
{
    calcModels();
    receipt_.setCashier(session_->cashierName());
    receipt_.setCashierInn(session_->cashier().inn());

}

ReceiptModel::~ReceiptModel()
{

}

QHash<int, QByteArray> ReceiptModel::roleNames() const
{
    return {
        {RoleIcon, "ricon"},
        {RoleName, "rname"},
        {RolePrice, "rprice"},
        {RoleQuantity, "rquantity"},
        {RoleType, "rtype"}
    };
}

QVariant ReceiptModel::data(const QModelIndex &idx, int role) const
{
    if(!idx.isValid() || idx.row() >= receipt_.count() ) return QVariant();
    PSimpleReceiptOperation op = receipt_[idx.row()];
    switch (role) {
    case RoleIcon: return "qrc:/icons/default_operation_64.png";
    case RoleName: return op->name();
    case RolePrice: return op->price().toString();
    case RoleQuantity: return op->quantity().toString();
    case RoleType: return fdf::opTypeCaption(op->type());
    default:
        break;
    }

    return QVariant();
}

int ReceiptModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return receipt_.count();
}

ReceiptModel::ReceiptType ReceiptModel::receiptType() const
{
    return static_cast<ReceiptType>(receipt_.paymentAttr());
}

void ReceiptModel::setReceiptType(ReceiptType type)
{
    fdf::PaymentType tp = static_cast<fdf::PaymentType>(type);
    if(tp != receipt_.paymentAttr())
    {
        receipt_.setPaymentAttr(tp);
        ReceiptOptions options = receipt_.options();
        RegData rd = status_->status().regData();
        options.setFlag(ReceiptOption::IsFosa, rd.extFlags().getFosaCashbox());
        options.setFlag(ReceiptOption::InternetCashbox, rd.extFlags().getInternetOnly());
        options.setFlag(ReceiptOption::SupportLabledTrandes, rd.extFlags().getLabledProducts());
        options.setFlag(ReceiptOption::SupportExciseTrades, rd.extFlags().getExcisableProducts());
        receipt_.setOptions(options);

        lmWarning() << static_cast<qint32> (options )<< static_cast<qint32>(receipt_.options());

        emit receiptTypeChanged();
        emit receiptNameChanged();
        emit supportLabelsChanged();
    }
}

bool ReceiptModel::isCorrection() const
{
    return receipt_.isCorrection();
}

void ReceiptModel::setCorrection(bool set)
{
    if(set != receipt_.isCorrection())
    {
        receipt_.setCorrection(set);
        ReceiptOptions options = receipt_.options();
        RegData rd = status_->status().regData();
        options.setFlag(ReceiptOption::IsFosa, rd.extFlags().getFosaCashbox());
        options.setFlag(ReceiptOption::InternetCashbox, rd.extFlags().getInternetOnly());
        options.setFlag(ReceiptOption::SupportExciseTrades, rd.extFlags().getLabledProducts());
        options.setFlag(ReceiptOption::SupportExciseTrades, rd.extFlags().getExcisableProducts());
        receipt_.setOptions(options);
        lmWarning() << static_cast<qint32> (options) << static_cast<qint32>(receipt_.options());
        emit correctionChanged();
        emit receiptNameChanged();
        emit supportLabelsChanged();
    }
}

bool ReceiptModel::supportLabels() const
{
    RegData rd = status_->status().regData();

    lmWarning() << rd.extFlags().getLabledProducts()
                << rd.extFlags().getFosaCashbox()
                << receipt_.mayBeLabled();

    return rd.extFlags().getLabledProducts() &&
            !rd.extFlags().getFosaCashbox() &&
            receipt_.mayBeLabled();

}

bool ReceiptModel::containsLabels() const
{
    return supportLabels() && receipt_.containsLabels();
}

QString ReceiptModel::totalAmount() const
{
    return receipt_.amount().toString();
}

QString ReceiptModel::receiptName() const
{
    switch (receipt_.paymentAttr()) {
    case fdf::PaymentType::Debit:
            return isCorrection() ? tr("КОРР. ПРИХ."): tr("ПРИХОД");
        break;
    case fdf::PaymentType::DebitReturn:
            return isCorrection() ? tr("КОРР. ВОЗВР. ПРИХ."): tr("ВОЗВРАТ ПРИХОДА");
        break;
    case fdf::PaymentType::Credit:
            return isCorrection() ? tr("КОРР. РАСХ."): tr("РАСХОД");
        break;
    case fdf::PaymentType::CreditReturn:
            return isCorrection() ? tr("КОРР. ВОЗВР. РАСХ. "): tr("ВОЗВРАТ РАСХОДА");
        break;
    default: return QString();
    }
}
bool ReceiptModel::canChangeOwnerName() const
{
    return false;//status_->status().regData().ownerName().isEmpty();
}

bool ReceiptModel::canChangeAddress() const
{
    return true;//status_->status().regData().address().isEmpty();
}

bool ReceiptModel::canChangePlace() const
{
    return true;// status_->status().regData().place().isEmpty();
}

bool ReceiptModel::canChangeBoxId() const
{
    const RegData &rd = status_->status().regData();
    return rd.getAutoMode() && rd.boxId().isEmpty();
}


QVariantMap ReceiptModel::operationData(int index) const
{
    if(index < 0 || index >= receipt_.count()) return QVariantMap();
    return receipt_[index]->toMap();
}

bool ReceiptModel::operationIsLabled(int index) const
{
    return index >= 0 && index <= receipt_.count() &&
            receipt_[index].dynamicCast<ReceiptOperation>();
}

double ReceiptModel::receiptAmount() const
{
    return receipt_.amount().toDouble();
}

double ReceiptModel::cash() const
{
    return receipt_.cash().value_or(FixNumber(DEFAULT_AMOUNT_MULT)).toDouble();
}

double ReceiptModel::card() const
{
    return receipt_.card().value_or(FixNumber(DEFAULT_AMOUNT_MULT)).toDouble();
}

double ReceiptModel::prepayment() const
{
    return receipt_.prepay().value_or(FixNumber(DEFAULT_AMOUNT_MULT)).toDouble();
}

double ReceiptModel::postpayment() const
{
    return receipt_.postpay().value_or(FixNumber(DEFAULT_AMOUNT_MULT)).toDouble();
}

double ReceiptModel::barter() const
{
    return receipt_.barter().value_or(FixNumber(DEFAULT_AMOUNT_MULT)).toDouble();
}

QVariantMap ReceiptModel::receiptData() const
{
    lmWarning() << receipt_.mayBeLabled() << cashier() << receipt_.cashier() << logvariant(receipt_.toMap());
    Receipt r = receipt_;
    if(!fdf::taxIsValid(r.tax()) && taxes_.size() == 1)
    {
        r.setTax(static_cast<fdf::Tax>(taxes_[0].toMap()["code"].toInt()));
    }
    return r.toMap();
}

QVariantList ReceiptModel::taxSystemModel() const
{
    return taxes_;
}

int ReceiptModel::taxIndex() const
{

    qint32 t = static_cast<qint32>(receipt_.tax());
    lmWarning() << t;
    if(!fdf::taxIsValid(receipt_.tax()))
    {
        lmWarning() << t;
        t = static_cast<qint32>(status_->status().regData().taxes());
        lmWarning() << t;
        if(!fdf::taxIsValid(static_cast<fdf::Tax>(t)))
        {
            RuntimeUiSettings rus;
            t = static_cast<qint32>(rus.lastTaxSystem());
        }
    }
    if(!fdf::taxIsValid(static_cast<fdf::Tax>(t)))   return -1;
    for(int i = 0; i < taxes_.size(); ++i)
    {
        if(taxes_[i].toMap()["code"].toInt() == t) return i;
    }
    return -1;
}

void ReceiptModel::setTaxIndex(int idx)
{
    fdf::Tax t = receipt_.tax();
    if(idx >= 0 && idx <= taxes_.size())
    {
        t = static_cast<fdf::Tax>(taxes_[idx].toMap()["code"].toInt());
    }
    if(t != receipt_.tax())
    {
        receipt_.setTax(t);
        emit taxIndexChanged();
    }
}

QString ReceiptModel::cashier() const
{
    return !receipt_.cashier().isEmpty() ?
                receipt_.cashier() :
                session_->cashier().name();

}

void ReceiptModel::setCashier(const QString &c)
{
    if(receipt_.cashier() != c)
    {
        receipt_.setCashier(c);
        emit cashierChanged();
    }
}

QString ReceiptModel::cashierInn() const
{
    return !receipt_.cashierInn().isEmpty() ?
                receipt_.cashierInn() :
                session_->cashier().inn();
}

void ReceiptModel::setCashierInn(const QString &c)
{
    if(receipt_.cashierInn() != c)
    {
        receipt_.setCashierInn(c);
        emit cashierInnChanged();
    }

}
QString ReceiptModel::ownerName() const
{
    return !receipt_.ownerName().isEmpty() ?
                receipt_.ownerName() :
                status_->status().regData().ownerName();

}

void ReceiptModel::setOwnerName(const QString &newUserName)
{
    if(canChangeOwnerName() && receipt_.ownerName() != newUserName)
    {
        receipt_.setOwnerName(newUserName);
        emit ownerNameChanged();
    }
}

QString ReceiptModel::address() const
{
    return !receipt_.address().isEmpty() ?
                receipt_.address() :
                status_->status().regData().address();
}

void ReceiptModel::setAddress(const QString &newAddress)
{
    if(canChangeAddress() && receipt_.address() != newAddress)
    {
        receipt_.setAddress(newAddress);
        emit addressChanged();
    }
}

QString ReceiptModel::place() const
{
    return !receipt_.place().isEmpty() ?
                receipt_.place() :
                status_->status().regData().place();
}

void ReceiptModel::setPlace(const QString &newPlace)
{
    if(canChangePlace() && receipt_.place() != newPlace)
    {
        receipt_.setPlace(newPlace);
        emit placeChanged();
    }
}

QString ReceiptModel::buyerPhone() const
{
    return receipt_.buyerPhone();
}

void ReceiptModel::setBuyerPhone(const QString &phone)
{
    if(receipt_.buyerPhone() != phone)
    {
        receipt_.setBuyerPhone(phone);
        emit buyerPhoneChanged();
    }
}

QString ReceiptModel::boxId() const
{
    return receipt_.boxId().isEmpty() ?
                status_->status().regData().boxId() :
                receipt_.boxId();
}

void ReceiptModel::setBoxId(const QString &id)
{
    if(canChangeBoxId() && receipt_.boxId() != id)
    {
        receipt_.setBoxId(id);
        emit boxIdChanged();
    }
}

QVariantList ReceiptModel::industryProps() const
{
    QVariantList res;
    QList<IndustryProperty> props = receipt_.industryProperties();
    std::for_each(props.constBegin(), props.constEnd(),
                  [&res](const IndustryProperty &p){res << p.toMap();});
    return  res;
}

void ReceiptModel::addOperation(int index, const QVariantMap &map)
{

    if(index < 0 || index > receipt_.count())
    {
        beginInsertRows(QModelIndex(), receipt_.count(), receipt_.count());
        receipt_.addOperation(SimpleReceiptOperation(map));
        endInsertRows();
    }
    else
    {
        beginResetModel();
        receipt_.setOperation(SimpleReceiptOperation(map), index);
        endResetModel();
    }
    emit totalAmountChanged();

}

void ReceiptModel::addLabledOperation(int index, const QVariantMap &map)
{
    if(index < 0 || index > receipt_.count())
    {
        beginInsertRows(QModelIndex(), receipt_.count(), receipt_.count());
        receipt_.addOperation(ReceiptOperation(map));
        endInsertRows();
    }
    else
    {
        beginResetModel();
        receipt_.setOperation(ReceiptOperation(map), index);
        endResetModel();
    }
    emit totalAmountChanged();
    emit containsLabelsChanged();
}

void ReceiptModel::removeOperation(int index)
{
    if(index >= 0 && index < receipt_.count())
    {
        beginRemoveRows(QModelIndex(), index, index);
        receipt_.removeOpearation(index);
        receipt_.setCashier(session_->cashierName());
        receipt_.setCashierInn(session_->cashier().inn());
        endRemoveRows();
        emit totalAmountChanged();
        emit containsLabelsChanged();
    }
}

void ReceiptModel::clean()
{
    lmWarning() << "CLEAN";
    ReceiptOptions options = receipt_.options();
    beginResetModel();
    fdf::PaymentType pt = receipt_.paymentAttr();
    bool corr = isCorrection();
    receipt_.clean();
    endResetModel();
    RegData rd = status_->status().regData();
    options.setFlag(ReceiptOption::IsFosa, rd.extFlags().getFosaCashbox());
    options.setFlag(ReceiptOption::InternetCashbox, rd.extFlags().getInternetOnly());
    options.setFlag(ReceiptOption::SupportExciseTrades, rd.extFlags().getLabledProducts());
    options.setFlag(ReceiptOption::SupportExciseTrades, rd.extFlags().getExcisableProducts());
    receipt_.setOptions(options);
    receipt_.setCorrection(corr);
    receipt_.setPaymentAttr(pt);
    receipt_.setCashier(session_->cashier().name());
    receipt_.setCashierInn(session_->cashier().inn());
    RuntimeUiSettings rus;
    receipt_.setTax(rus.lastTaxSystem());

    emit receiptTypeChanged();
    emit correctionChanged();
    emit supportLabelsChanged();
    emit containsLabelsChanged();
    emit totalAmountChanged();
    emit receiptNameChanged();
    emit canChangeOwnerNameChanged();
    emit canChangeAddressChanged();
    emit canChangePlaceChanged();
    emit canChangeBoxIdChanged();
    emit taxSystemModelChanged();
    emit taxIndexChanged();
    emit cashierChanged();
    emit cashierInnChanged();
    emit ownerNameChanged();
    emit addressChanged();
    emit placeChanged();
    emit buyerPhoneChanged();
    emit boxIdChanged();
    emit industryPropsChanged();
    emit additionalParamChanged();
    emit operationCodeChanged();
    emit operationDtChanged();
    emit operationParamDataChanged();
    emit additionalUserParamNameChanged();
    emit additionalUserParamValChanged();
    emit buyerNameChanged();
    emit buyerInnChanged();
    emit buyerBirthDayChanged();
    emit buyerCitizenshipChanged();
    emit buyerDocCodeChanged();
    emit buyerDocCodeIndexChanged();
    emit buyerDocDataChanged();
    emit buyerAddressChanged();
    emit correctionDateChanged();
    emit correctionDocNumberChanged();
}

void ReceiptModel::calcModels()
{
    fdf::Taxes taxes = status_->status().regData().taxes();
    if(!static_cast<qint32>(taxes)) taxes = fdf::Taxes(0x7F);
    taxes_.clear();
    auto setTax = [this](fdf::Tax t) {
            taxes_ << QVariantMap {
            {"code", static_cast<qint32>(t)},
            {"name", fdf::taxToString(t)}
        };
    };

    if(taxes.testFlag(fdf::Tax::Basic)) setTax(fdf::Tax::Basic);
    if(taxes.testFlag(fdf::Tax::SimpleProfit)) setTax(fdf::Tax::SimpleProfit);
    if(taxes.testFlag(fdf::Tax::SimpleProfitExpense)) setTax(fdf::Tax::SimpleProfitExpense);
//    if(taxes.testFlag(fdf::Tax::SingleOnImputedProfit)) setTax(fdf::Tax::SingleOnImputedProfit);
    if(taxes.testFlag(fdf::Tax::SingleAgriculturalTax)) setTax(fdf::Tax::SingleAgriculturalTax);
    if(taxes.testFlag(fdf::Tax::PatentSystemOfTaxation)) setTax(fdf::Tax::PatentSystemOfTaxation);
    emit taxSystemModelChanged();
}

void ReceiptModel::setIndustryProps(const QVariantList &props)
{
    if(props.isEmpty()) receipt_.setIndustryProperties(QList<IndustryProperty>());
    else
    {
        QList<IndustryProperty> res;
        std::for_each(props.constBegin(), props.constEnd(),
                      [&res] (const QVariant &v){res << IndustryProperty(v.toMap());});
        receipt_.setIndustryProperties(res);
    }
    emit industryPropsChanged();
}

void ReceiptModel::cleanOperationalParam()
{
    receipt_.setOperationalParam(std::optional<ReceiptOperationalParam>());
}

void ReceiptModel::cleanUserParam()
{
    receipt_.setUserParam(std::optional<ReceiptAdditionalUserParam>());
}

void ReceiptModel::cleanBuyerInfo()
{
    receipt_.setBuyerInfo(std::optional<BuyerInfo>());
}
void ReceiptModel::cleanCorrectionReason ()
{
    receipt_.setCorrectionReason(std::optional<CorrectionReason>());
}

void ReceiptModel::addQuantity(qint32 index)
{
    beginResetModel();
    receipt_.addQuantity(index);
    endResetModel();
    emit totalAmountChanged();
}

void ReceiptModel::subQuantity(qint32 index)
{
    beginResetModel();
    receipt_.subQuantity(index);
    endResetModel();
    emit totalAmountChanged();

}
QString ReceiptModel::correctionDocNumber() const
{
    return receipt_.correctionReason().value_or(CorrectionReason()).docNumber();
}

void ReceiptModel::setCorrectionDocNumber(const QString &newCorrectionDocNumber)
{
    if (receipt_.correctionReason().has_value() &&
            receipt_.correctionReason().value().docNumber() == newCorrectionDocNumber)
        return;
    CorrectionReason r = receipt_.correctionReason().value_or(CorrectionReason());
    r.setDocNumber(newCorrectionDocNumber);
    receipt_.setCorrectionReason( r);
    emit correctionDocNumberChanged();
}

bool ReceiptModel::correctionIsIndependent() const
{
    return receipt_.correctionReason().value_or(CorrectionReason()).isIndependent();
}

void ReceiptModel::setCorrectionIsIndependent(bool independent)
{
    if (receipt_.correctionReason().has_value() &&
        receipt_.correctionReason().value().isIndependent() == independent)
        return;
    CorrectionReason r = receipt_.correctionReason().value_or(CorrectionReason());
    r.setIsIndependent(independent);
    receipt_.setCorrectionReason( r);
    emit correctionIsIndependentChanged();
}

QString ReceiptModel::correctionDate() const
{
    return receipt_.correctionReason().value_or(CorrectionReason()).date().toString("dd.MM.yyyy");
}

void ReceiptModel::setCorrectionDate(const QString &newCorrectionDate)
{
    QDate d = QDate::fromString(newCorrectionDate, "dd.MM.yyyy");
    if (receipt_.correctionReason().has_value() &&
            receipt_.correctionReason().value().date() == d)
        return;
    CorrectionReason r = receipt_.correctionReason().value_or(CorrectionReason());
    r.setDate(d);
    receipt_.setCorrectionReason(r);
    emit correctionDateChanged();
}

QString ReceiptModel::buyerAddress() const
{
    return receipt_.buyerInfo().value_or(BuyerInfo()).address();
}

void ReceiptModel::setBuyerAddress(const QString &newBuyerAddress)
{
    if (receipt_.buyerInfo().has_value() &&
            receipt_.buyerInfo().value().address() == newBuyerAddress)
        return;
    BuyerInfo bi = receipt_.buyerInfo().value_or(BuyerInfo());
    bi.setAddress(newBuyerAddress);
    receipt_.setBuyerInfo(bi);
    emit buyerAddressChanged();
}

QString ReceiptModel::buyerDocData() const
{
    return receipt_.buyerInfo().value_or(BuyerInfo()).docData();
}

void ReceiptModel::setBuyerDocData(const QString &newBuyerDocData)
{
    if (receipt_.buyerInfo().has_value() &&
            receipt_.buyerInfo().value().docData() == newBuyerDocData)
        return;
    BuyerInfo bi = receipt_.buyerInfo().value_or(BuyerInfo());
    bi.setDocData(newBuyerDocData);
    receipt_.setBuyerInfo(bi);
    emit buyerDocDataChanged();
}

QString ReceiptModel::buyerDocCode() const
{
    return receipt_.buyerInfo().value_or(BuyerInfo()).docCode();
}

void ReceiptModel::setBuyerDocCode(const QString &newBuyerDocCode)
{
    if (receipt_.buyerInfo().has_value() &&
            receipt_.buyerInfo().value().docCode() == newBuyerDocCode)
        return;
    BuyerInfo bi = receipt_.buyerInfo().value_or(BuyerInfo());
    bi.setDocCode(newBuyerDocCode);
    receipt_.setBuyerInfo(bi);
    emit buyerDocCodeChanged();
}

int ReceiptModel::buyerDocCodeIndex() const
{
    static const QHash<QString, int> VALID = {
        {"21", 1}, {"22", 2}, {"26", 3}, {"27", 4},
        {"28", 5}, {"31", 6}, {"32", 7}, {"33", 8},
        {"34", 9}, {"35", 10}, {"36", 11}, {"37", 12},
        {"38", 13}, {"40", 14}
    };
    QString code = buyerDocCode().trimmed();
    if(code.size() != 2) return 0;
    return  VALID.contains(code) ? VALID[code] : 0;
}

QString ReceiptModel::buyerCitizenship() const
{
    return receipt_.buyerInfo().value_or(BuyerInfo()).citizenship();
}

void ReceiptModel::setBuyerCitizenship(const QString &newBuyerCitizenship)
{
    if (receipt_.buyerInfo().has_value() &&
            receipt_.buyerInfo().value().citizenship() == newBuyerCitizenship)
        return;
    BuyerInfo bi = receipt_.buyerInfo().value_or(BuyerInfo());
    bi.setCitizenship(newBuyerCitizenship);
    receipt_.setBuyerInfo(bi);
    emit buyerCitizenshipChanged();
}

QString ReceiptModel::buyerBirthDay() const
{
    return receipt_.buyerInfo().value_or(BuyerInfo()).birthDay().toString("dd.MM.yyyy");
}

void ReceiptModel::setBuyerBirthDay(const QString &newBuyerBirthDay)
{
    QDate d = QDate::fromString(newBuyerBirthDay, "dd.MM.yyyy");
    if (receipt_.buyerInfo().has_value() &&
            receipt_.buyerInfo().value().birthDay() == d)
        return;
    BuyerInfo bi = receipt_.buyerInfo().value_or(BuyerInfo());
    bi.setBirthDay(d);
    receipt_.setBuyerInfo(bi);
    emit buyerBirthDayChanged();
}

QString ReceiptModel::buyerInn() const
{
    return receipt_.buyerInfo().value_or(BuyerInfo()).address();
}

void ReceiptModel::setBuyerInn(const QString &newBuyerInn)
{
    if (receipt_.buyerInfo().has_value() &&
            receipt_.buyerInfo().value().inn() == newBuyerInn)
        return;
    BuyerInfo bi = receipt_.buyerInfo().value_or(BuyerInfo());
    bi.setInn(newBuyerInn);
    receipt_.setBuyerInfo(bi);
    emit buyerInnChanged();
}

QString ReceiptModel::buyerName() const
{
    return receipt_.buyerInfo().value_or(BuyerInfo()).address();
}

void ReceiptModel::setBuyerName(const QString &newBuyerName)
{
    if (receipt_.buyerInfo().has_value() &&
            receipt_.buyerInfo().value().name() == newBuyerName)
        return;
    BuyerInfo bi = receipt_.buyerInfo().value_or(BuyerInfo());
    bi.setName(newBuyerName);
    receipt_.setBuyerInfo(bi);
    emit buyerNameChanged();
}

QString ReceiptModel::additionalUserParamVal() const
{
    return receipt_.userParam().value_or(ReceiptAdditionalUserParam()).value();
}

void ReceiptModel::setAdditionalUserParamVal(const QString &newAdditionalUserParamVal)
{
    if (receipt_.userParam().has_value() &&
            receipt_.userParam().value().value() == newAdditionalUserParamVal)
        return;

    ReceiptAdditionalUserParam p = receipt_.userParam().value_or(ReceiptAdditionalUserParam());
    p.setValue(newAdditionalUserParamVal);
    receipt_.setUserParam(p);
    emit additionalUserParamValChanged();
}

QString ReceiptModel::additionalUserParamName() const
{
    return receipt_.userParam().value_or(ReceiptAdditionalUserParam()).name();
}

void ReceiptModel::setAdditionalUserParamName(const QString &newAdditionalUserParamName)
{
    if (receipt_.userParam().has_value() &&
            receipt_.userParam().value().name() == newAdditionalUserParamName)
        return;

    ReceiptAdditionalUserParam p = receipt_.userParam().value_or(ReceiptAdditionalUserParam());
    p.setName(newAdditionalUserParamName);
    receipt_.setUserParam(p);
    emit additionalUserParamNameChanged();
}

QString ReceiptModel::operationParamData() const
{
    return receipt_.operationalParam().value_or(ReceiptOperationalParam()).data();
}

void ReceiptModel::setOperationParamData(const QString &newOperationParamData)
{
    if (receipt_.operationalParam().has_value() &&
            receipt_.operationalParam().value().data() == newOperationParamData)
        return;
    ReceiptOperationalParam p = receipt_.operationalParam().value_or(ReceiptOperationalParam());
    p.setData(newOperationParamData);
    receipt_.setOperationalParam(p);
    emit operationParamDataChanged();
}

QString ReceiptModel::operationDt() const
{
    return receipt_.operationalParam().value_or(ReceiptOperationalParam())
            .dt().toString("dd.MM.yyyy hh:mm");
}

void ReceiptModel::setOperationDt(const QString &newOperationDt)
{
    QDateTime dt = QDateTime::fromString(newOperationDt, "dd.MM.yyyy hh:mm");
    if (receipt_.operationalParam().has_value() &&
            receipt_.operationalParam().value().dt() == dt)
        return;
    ReceiptOperationalParam p = receipt_.operationalParam().value_or(ReceiptOperationalParam());
    p.setDt(dt);

    receipt_.setOperationalParam(p);
    emit operationDtChanged();
}

qint32 ReceiptModel::operationCode() const
{
    return receipt_.operationalParam().value_or(ReceiptOperationalParam()).code();
}

void ReceiptModel::setOperationCode(qint32 newOperationCode)
{
    quint8 c = static_cast<quint8>(newOperationCode);
    if (receipt_.operationalParam().has_value() &&
            receipt_.operationalParam().value().code() == c)
        return;
    ReceiptOperationalParam p = receipt_.operationalParam().value_or(ReceiptOperationalParam());
    p.setCode(c);

    receipt_.setOperationalParam(p);
    emit operationCodeChanged();
}

const QString &ReceiptModel::additionalParam() const
{
    return receipt_.additionalParam();
}

void ReceiptModel::setAdditionalParam(const QString &newAdditionalParam)
{
    if (receipt_.additionalParam() == newAdditionalParam)
        return;
    receipt_.setAdditionalParam(newAdditionalParam);
    emit additionalParamChanged();
}

