#include "cashboxstatuswrapper.h"

#include "logmanip.h"
#include <QJsonDocument>

#include <QtAndroidExtras/QAndroidJniObject>
#include <QtAndroidExtras/QtAndroid>
#include <QAndroidIntent>
#include <QCryptographicHash>


QPointer<CashboxStatusWrapper> CashboxStatusWrapper::_globalStatus = nullptr;
QMutex CashboxStatusWrapper::_statusMutex;
CashboxStatus CashboxStatusWrapper::_status;


void updateCashboxStatus(const QVariantMap &data)
{
    CashboxStatus st;
    st.parseMap(data);
    CashboxStatusWrapper::setGlobalStatus(st);
}

extern "C" {
void Java_ru_armax_cashbox_BroadcastProcessApi_processStatusIntent(JNIEnv *env, jobject thiz, jobject intent)
{
    Q_UNUSED(env)
    Q_UNUSED(thiz)
    if(intent)
    {
        QAndroidJniObject obj(intent);
        if(obj.isValid())
        {
            QAndroidIntent i (obj);
            QVariantMap data = i.extraVariant("status").toMap();
            lmWarning() << logvariant(data);
            updateCashboxStatus(data);
        }
    }
}
}

CashboxStatusWrapper::CashboxStatusWrapper(QObject *parent)
    : QObject(parent)
{
    _statusMutex.lock();
    if(!_globalStatus)
    {
        _globalStatus = new CashboxStatusWrapper(false);
    }
    connect(_globalStatus, &CashboxStatusWrapper::statusInfoChanged,
            this, &CashboxStatusWrapper::statusInfoChanged);
    connect(_globalStatus, &CashboxStatusWrapper::statusChanged,
            this, &CashboxStatusWrapper::statusChanged);
    connect(_globalStatus, &CashboxStatusWrapper::hasSerialChanged,
            this, &CashboxStatusWrapper::hasSerialChanged);

    connect(_globalStatus, &CashboxStatusWrapper::fiscalizedChanged,
            this, &CashboxStatusWrapper::fiscalizedChanged);
    connect(_globalStatus, &CashboxStatusWrapper::cycleIsOpenChanged,
            this, &CashboxStatusWrapper::cycleIsOpenChanged);
    connect(_globalStatus, &CashboxStatusWrapper::cycleIsOverChanged,
            this, &CashboxStatusWrapper::cycleIsOverChanged);

    connect(_globalStatus, &CashboxStatusWrapper::archiveIsClosedChanged,
            this, &CashboxStatusWrapper::archiveIsClosedChanged);
    connect(_globalStatus, &CashboxStatusWrapper::serialChanged,
            this, &CashboxStatusWrapper::serialChanged);
    connect(_globalStatus, &CashboxStatusWrapper::orgNameChanged,
            this, &CashboxStatusWrapper::orgNameChanged);

    connect(_globalStatus, &CashboxStatusWrapper::orgAddressChanged,
            this, &CashboxStatusWrapper::orgAddressChanged);
    connect(_globalStatus, &CashboxStatusWrapper::placeChanged,
            this, &CashboxStatusWrapper::placeChanged);
    connect(_globalStatus, &CashboxStatusWrapper::ipAddressChanged,
            this, &CashboxStatusWrapper::ipAddressChanged);


    connect(_globalStatus, &CashboxStatusWrapper::cycleOpenedHoursChanged,
            this, &CashboxStatusWrapper::cycleOpenedHoursChanged);
    connect(_globalStatus, &CashboxStatusWrapper::cycleNumberChanged,
            this, &CashboxStatusWrapper::cycleNumberChanged);
    connect(_globalStatus, &CashboxStatusWrapper::isDebugChanged,
            this, &CashboxStatusWrapper::isDebugChanged);

    connect(_globalStatus, &CashboxStatusWrapper::isDebugChanged,
            this, &CashboxStatusWrapper::isDebugChanged);
    connect(_globalStatus, &CashboxStatusWrapper::inAutoModeChanged,
            this, &CashboxStatusWrapper::inAutoModeChanged);
    connect(_globalStatus, &CashboxStatusWrapper::inOfflineModeChanged,
            this, &CashboxStatusWrapper::inOfflineModeChanged);
    connect(_globalStatus, &CashboxStatusWrapper::lastDocNumberChanged,
            this, &CashboxStatusWrapper::lastDocNumberChanged);
    connect(_globalStatus, &CashboxStatusWrapper::hasNotPrintedChanged,
            this, &CashboxStatusWrapper::hasNotPrintedChanged);

    _statusMutex.unlock();
}

CashboxStatusWrapper::~CashboxStatusWrapper()
{

}

CashboxStatus CashboxStatusWrapper::status() const
{
    _statusMutex.lock();
    CashboxStatus  st = _status;
    _statusMutex.unlock();;
    return st;
}


QString CashboxStatusWrapper::statusInfo() const
{
    CashboxStatus st = status();
//    QJsonDocument jdoc = QJsonDocument::fromVariant(st.fiscalModuleInfo().toMap());
    return st.toHtml();//QString::fromUtf8(jdoc.toJson());
}

bool CashboxStatusWrapper::hasSerial() const
{
    CashboxStatus st = status();
    return !st.modelData().serial().trimmed().isEmpty() && st.modelData().isValid();
}

bool CashboxStatusWrapper::fiscalized() const
{
    CashboxStatus st = status();
    return st.isFiscalized();
}

bool CashboxStatusWrapper::cycleIsOpen() const
{
    CashboxStatus st = status();
    return st.cycleIsOpen();
}

bool CashboxStatusWrapper::cycleIsOver() const
{
    CashboxStatus st = status();


    return st.cycleIsOver();
}

bool CashboxStatusWrapper::archiveIsClosed() const
{
    CashboxStatus st = status();
    return st.fs().archiveIsClosed();
}

bool CashboxStatusWrapper::isDebug() const
{
    CashboxStatus st = status();
    return !st.fs().isRelease();
}

bool CashboxStatusWrapper::inAutoMode() const
{
    CashboxStatus st = status();
    return st.isFiscalized() && st.regData().getAutoMode();
}

bool CashboxStatusWrapper::inOfflineMode() const
{
    CashboxStatus st = status();
    return st.isFiscalized() && st.regData().getOfflineMode();
}

bool CashboxStatusWrapper::hasNotPrinted() const
{
    CashboxStatus st = status();
    return st.hasNotPrinted();
}

QString CashboxStatusWrapper::dsn() const
{
    CashboxStatus st = status();
    return st.dev().dsn().trimmed();
}

QString CashboxStatusWrapper::serial() const
{
    CashboxStatus st = status();
    return st.modelData().serial();
}

QString CashboxStatusWrapper::orgName() const
{
    CashboxStatus st = status();
    return st.regData().ownerName();
}

QString CashboxStatusWrapper::orgAddress() const
{
    CashboxStatus st = status();
    return st.regData().address();
}

QString CashboxStatusWrapper::place() const
{
    CashboxStatus st = status();
    return st.regData().place();
}

QString CashboxStatusWrapper::cycleOpenedHours() const
{
    CashboxStatus st = status();
    if(!st.cycleIsOpen()) return QString();
    qint32 res = st.fs().cycleOpenDt().secsTo(QDateTime::currentDateTime()) / 60;
    if(res < 60) return tr("%1 м").arg(res);
    res = (res % 60 < 30) ? res / 60 : res / 60 + 1;
    return tr("%1 ч").arg(res);
}

QString CashboxStatusWrapper::cycleNumber() const
{
    CashboxStatus st = status();
    if(!st.cycleIsOpen()) return QString();
    return QString::number(st.fs().cycle().cycle());
//    lmWarning() << QStringLiteral("%1").arg(st.fiscalCycleInfo().number(), 3, 10, QLatin1Char('0'));
    //    return QStringLiteral("%1").arg(st.fiscalCycleInfo().number(), 3, 10, QLatin1Char('0'));
}

qint32 CashboxStatusWrapper::lastDocNumber() const
{
    CashboxStatus st = status();
    return st.fs().status().lastFd();
}

qint32 CashboxStatusWrapper::regsCount() const
{
    CashboxStatus st = status();
    return st.fs().lifeTime().regsCount();
}

QString CashboxStatusWrapper::ipAddress() const
{
    CashboxStatus st = status();
    return st.dev().ipAddresses().join(" ");
}


CashboxStatusWrapper::CashboxStatusWrapper(bool destroyStatic)
    : QObject(nullptr)
{
    if(destroyStatic && _globalStatus) delete _globalStatus;
    _globalStatus = this;
}

void CashboxStatusWrapper::setGlobalStatus(const CashboxStatus &status)
{
    _statusMutex.lock();
    if(!_globalStatus)
    {
        _globalStatus = new CashboxStatusWrapper(false);
    }
    CashboxStatus oldStatus = _status;
    _status = status;
    _statusMutex.unlock();
    //TODO: Хорошо бы проверять изменения полей. может интерфейс будет поменьше дергать
    emit _globalStatus->statusInfoChanged();
    emit _globalStatus->statusChanged();
    emit _globalStatus->hasSerialChanged();
    emit _globalStatus->fiscalizedChanged();
    emit _globalStatus->isDebugChanged();
    emit _globalStatus->inAutoModeChanged();
    emit _globalStatus->inOfflineModeChanged();
    emit _globalStatus->cycleIsOpenChanged();
    emit _globalStatus->cycleIsOverChanged();
    emit _globalStatus->archiveIsClosedChanged();
    emit _globalStatus->serialChanged();
    emit _globalStatus->cycleOpenedHoursChanged();
    emit _globalStatus->cycleNumberChanged();
    emit _globalStatus->hasNotPrintedChanged();
    emit _globalStatus->placeChanged();
    emit _globalStatus->ipAddressChanged();
    emit _globalStatus->orgNameChanged();
    emit _globalStatus->orgAddressChanged();
    if(oldStatus.modelData() != status.modelData())
    {
        emit _globalStatus->hasSerialChanged();
    }
}

