#include "receiptoperationwrapper.h"
#include "cashboxbuildconfig.h"
#include "logmanip.h"
#include "productcodecreator.h"
#include "labelcheckdata.h"

ReceiptOperationWrapper::ReceiptOperationWrapper(QObject *parent)
    : BaseCoreHandler(parent)
    , index_(-1)
    , opTypeModel_()
    , itemUnitModel_()
    , operation_()
    , status_(new CashboxStatusWrapper(this))
    , isSale_(true)
    , promisStatus_(Piece)
    , label_()

{
    createOpTypeModel();
    createItemUnitModel();
}

ReceiptOperationWrapper::~ReceiptOperationWrapper()
{

}

qint32 ReceiptOperationWrapper::index() const
{
    return index_;
}

void ReceiptOperationWrapper::setIndex(qint32 newIndex)
{
    index_ = newIndex;
}

QVariantMap ReceiptOperationWrapper::data() const
{
    return operation_.toMap();
}

void ReceiptOperationWrapper::setData(const QVariantMap &map)
{
    operation_.parseMap(map);
    lmWarning() << opType() << opTypeIndex() << logvariant(map);
    if(operation_.labelCheckResult().has_value())
    {
        label_ = operation_.labelCheckResult().value().labelCode();
        switch (operation_.labelCheckResult().value().actualStatus())
        {
        case fdf::LabelProductStatus::Piece         : promisStatus_ = Piece;break;
        case fdf::LabelProductStatus::Measured      : promisStatus_ = Measured;break;
        case fdf::LabelProductStatus::PieceRefund   : promisStatus_ = Piece;break;
        case fdf::LabelProductStatus::MeasuredRefund: promisStatus_ = Measured;break;
        case fdf::LabelProductStatus::NotChanged    : promisStatus_ = OtherPromisStatus;break;
        default: break;
        }
    }
    emit promisStatusChanged();
    emit indexChanged();
    emit dataChanged();
    emit opTypeModelChanged();
    emit vatModelChanged();
    emit payTypeModelChanged();
    emit opTypeChanged();
    emit opTypeIndexChanged();
    emit nameChanged();
    emit vatRateChanged();
    emit payTypeChanged();
    emit priceChanged();
    emit priceForUnitChanged();
    emit quantityChanged();
    emit industryPropsChanged();
    emit additionalParamChanged();
    emit exciseTaxChanged();
    emit countryCodeChanged();
    emit declarationNumberChanged();
    emit isExciseChanged();
    emit isSaleChanged();
    emit itemUnitModelChanged();
    emit itemUnitChanged();
    emit hasFractialChanged();
    emit nominatorChanged();
    emit denominatorChanged();
    emit labelChanged();

}

bool ReceiptOperationWrapper::fractionQuantity() const
{
    return false;
}

QVariantList ReceiptOperationWrapper::opTypeModel() const
{
    return opTypeModel_;
}

QVariantList ReceiptOperationWrapper::vatModel() const
{
    QVariantList res;
    for(int i = Vat20; i <= VatNone; ++i)
    {
        res << QVariantMap {
        {"name", fdf::vatLabel(static_cast<fdf::VatRate>(i))},
        {"code", i}
        };
    }
    return res;

}

QVariantList ReceiptOperationWrapper::payTypeModel() const
{
    QVariantList res;
    for(int i = FullPrepayment; i <= CrediPayment; ++i)
    {
        res << QVariantMap {
        {"name", fdf::itemPaymentTypeToString(static_cast<fdf::ItemPaymentType>(i))},
        {"code", i}
        };
    }
    return res;
}

QVariantMap ReceiptOperationWrapper::agentData() const
{
    if(!fdf::itemPayAgentFlagIsValid(operation_.agentFlag())) return QVariantMap();
    QVariantMap res = {
        {"agentFlag", static_cast<qint32>(operation_.agentFlag())},
        {"providerInn", operation_.providerInn().trimmed()},
        {"providerData", operation_.providerData().value_or(ProviderData()).toMap()}
    };
    if(operation_.agentFlag() > fdf::ItemPayAgentFlag::PaySubAgent) return res;
    res["agentData"] = operation_.agentData().value_or(AgentData()).toMap();
    return res;
}

int ReceiptOperationWrapper::opTypeIndex() const
{
    if(!fdf::opTypeIsValid(operation_.type())) return -1;
    for(int i = 0; i < opTypeModel_.size(); ++i)
    {
        if(opTypeModel_[i].toMap()["code"].toInt() == opType()) return i;
    }
    return -1;
}

void ReceiptOperationWrapper::setOpTypeIndex(int idx)
{
    if(idx >= 0 && idx < opTypeModel_.size())
        setOpType(static_cast<ROperationType>(opTypeModel_[idx].toMap()["code"].toInt()));
}

ReceiptOperationWrapper::ROperationType ReceiptOperationWrapper::opType() const
{
    return static_cast<ROperationType>(operation_.type());
}

void ReceiptOperationWrapper::setOpType(ROperationType tp)
{
    fdf::ReceiptOperationType t = static_cast<fdf::ReceiptOperationType>(tp);
    if(operation_.type() != t)
    {
        operation_.setType(t);
        emit opTypeChanged();
        emit opTypeIndexChanged();
        emit isExciseChanged();
    }
}

bool ReceiptOperationWrapper::isExcise() const
{
    return opType() == ExcisableLabledProduct || opType() == ExcisableLabledWitoutCodeProduct;
}

QString ReceiptOperationWrapper::name() const
{
    return operation_.name();
}

void ReceiptOperationWrapper::setName(const QString &n)
{
    if(operation_.name() != n)
    {
        operation_.setName(n);
        emit nameChanged();
    }
}

ReceiptOperationWrapper::VatRate ReceiptOperationWrapper::vatRate() const
{
    return static_cast<VatRate>(operation_.vatRate());
}

void ReceiptOperationWrapper::setVatRate(VatRate rate)
{
    fdf::VatRate r = static_cast<fdf::VatRate>(rate);
    if(operation_.vatRate() != r)
    {
        operation_.setVatRate(r);
        emit vatRateChanged();
    }
}

ReceiptOperationWrapper::ItemPaymentType ReceiptOperationWrapper::payType() const
{
    return static_cast<ItemPaymentType>(operation_.paymentType());
}

void ReceiptOperationWrapper::setPayType(ItemPaymentType tp)
{
    fdf::ItemPaymentType t = static_cast<fdf::ItemPaymentType>(tp);
    if(operation_.paymentType() != t)
    {
        operation_.setPaymentType(t);
        emit payTypeChanged();
    }
}

QString ReceiptOperationWrapper::price() const
{
    return operation_.price().toString();
}

void ReceiptOperationWrapper::setPrice(const QString &price)
{
    FixNumber p = FixNumber(DEFAULT_AMOUNT_MULT, price);
    if(operation_.price() == p && !operation_.priceForUnit().has_value())
        return;
    operation_.setPrice(p);
    operation_.setPriceForUnit(std::optional<FixNumber>());
    emit priceChanged();
    emit priceForUnitChanged();
}

QString ReceiptOperationWrapper::priceForUnit() const
{
    return operation_.priceForUnit().has_value() ? operation_.priceForUnit().value().toString() : QString();
}

void ReceiptOperationWrapper::setPriceForUnit(const QString &price)
{
    FixNumber p = FixNumber(DEFAULT_AMOUNT_MULT, price);
    if(operation_.priceForUnit().has_value() && operation_.priceForUnit().value() == p)
        return;
    operation_.setPrice(std::optional<FixNumber>());
    operation_.setPriceForUnit(p);
    emit priceChanged();
    emit priceForUnitChanged();
}

QString ReceiptOperationWrapper::quantity() const
{
    if(fractionQuantity())
    {
        QuantityFraction f = operation_.fraction().value_or(QuantityFraction());
        if(!f.isValid())
        {
            f.setDenominator(1);
            f.setNominator(1);
        }
        return QStringLiteral("%1/%2").arg(f.nominator()).arg(f.denominator());
    }
    else
    {
        return operation_.quantity().value() > 0ll ? operation_.quantity().toShortString() : QString();
    }
}

void ReceiptOperationWrapper::setQuantity(const QString &q)
{
    if(fractionQuantity())
    {
        operation_.setQuantity(FixNumber(DEFAULT_QUANTITY_MULT,
                                         static_cast<qint64>(DEFAULT_QUANTITY_MULT)));
    }
    else
    {
        FixNumber qu (DEFAULT_QUANTITY_MULT, q);
        if(operation_.quantity() != qu)
        {
            operation_.setQuantity(qu);
            emit quantityChanged();
        }
    }
}

QVariantList ReceiptOperationWrapper::industryProps() const
{
    QVariantList res;
    QList<IndustryProperty> props = operation_.industryProperties();
    std::for_each(props.constBegin(), props.constEnd(),
                  [&res](const IndustryProperty &p){res << p.toMap();});
    return  res;
}

void ReceiptOperationWrapper::parseBarcode(const QString &code)
{
    lmWarning() << "TEST_LABELS_BEGIN" << code;
    operation_.setCode(std::optional<ProductCode>());
    setLabel(code.trimmed());

//    setLabel("010123456789012321M,7aL0JDGbJCWa91808B92CuE2b4wBhPv9XeoBQDEux9wOKeNR4vf4I+q/QbhqzhRGyYQymkkpgtAZUtPHlfp0THGVN6i+D8ZxZQcbTnvEMg==");
//    setLabel("010123456789012321M,7aL0JDGbJCWa\u001d91808B\u001d92CuE2b4wBhPv9XeoBQDEux9wOKeNR4vf4I+q/QbhqzhRGyYQymkkpgtAZUtPHlfp0THGVN6i+D8ZxZQcbTnvEMg==");

    operation_.setLabelCheckResult(std::optional<LabelCheckResult>());

}

void ReceiptOperationWrapper::save()
{
    CoreApiConst::ErrorCode err;
    QString msg;

    if(label_.tradeMark().isEmpty())
    {
        emit error(tr("Осканируйте или введите марку"));
        return;
    }
    if(label_.isEncodedTradeMark() && !operation_.isLabled()) emit needCheckLabel();
    else if(!operation_.isValid(err, msg))
    {
        emit error(msg);
    }
    else
    {
        lmWarning() << index_ << logvariant(operation_.toMap());

        emit saveData(index_, operation_.toMap());
        emit finish();
    }
}

void ReceiptOperationWrapper::setAgentData(const QVariantMap &map)
{
    operation_.setAgentFlag(fdf::ItemPayAgentFlag::NotAgent);
    operation_.setProviderInn(QString());
    operation_.setProviderData(std::optional<ProviderData>());
    operation_.setAgentData(std::optional<AgentData>());
    if(map.isEmpty()) return;
    fdf::ItemPayAgentFlag f = static_cast<fdf::ItemPayAgentFlag>(map["agentFlag"].toInt());
    if(!fdf::itemPayAgentFlagIsValid(f)) return;
    operation_.setAgentFlag(f);
    operation_.setProviderInn(map["providerInn"].toString());
    operation_.setProviderData(ProviderData(map["providerData"].toMap()));
    if(f > fdf::ItemPayAgentFlag::PaySubAgent) return;
    operation_.setAgentData(AgentData(map["agentData"].toMap()));
}

void ReceiptOperationWrapper::setIndustryProps(const QVariantList &props)
{
    if(props.isEmpty()) operation_.setIndustryProperties(QList<IndustryProperty>());
    else
    {
        QList<IndustryProperty> res;
        std::for_each(props.constBegin(), props.constEnd(),
                      [&res] (const QVariant &v){res << IndustryProperty(v.toMap());});
        operation_.setIndustryProperties(res);
    }
    emit industryPropsChanged();
}

void ReceiptOperationWrapper::checkLabel(int flags)
{
    //TODO: Предварительная проверка

//    ProductCodeCreator pcc;
//    LabelCode lc = pcc.create(label_);
//    lmWarning() << "TEST_LABELS_END" << logtab
//                << lc.tradeMark() << static_cast<qint32>(lc.type()) << logtab
//                << lc.productCode().data() << lc.t91() << lc.t92() << logtab
//                << lc.tailCrc() << QString::number(lc.tailCrc(), 16)
//                << logbinary(lc.t92().toLatin1()) << logtab
//                << logbinary(QByteArray::fromBase64(lc.t92().toLatin1()))  << logtab
//                << lc.t93() << logbinary(logbinary(QByteArray::fromBase64(lc.t93().toLatin1()))) << logtab
//                << static_cast<qint32>(lc.lct2100());
    if(label_.isEncodedTradeMark())
    {
        LabelCheckData lcd;
        lcd.setLabelCode(label_);
        lcd.setCheckFlags(LabelCheckData::LabelCheckFlags(flags));
        lmWarning() << flags << lcd.labelCode().tradeMark() << logtab << label_.tradeMark();
        fdf::LabelProductStatus st = fdf::LabelProductStatus::Invalid;

        switch (promisStatus()) {
        case Piece               : st = isSale_ ?
                        fdf::LabelProductStatus::Piece :
                        fdf::LabelProductStatus::PieceRefund;
            lcd.setUnit(operation_.unit());
            lcd.setQuantity(operation_.quantity());
            lcd.setFraction(operation_.fraction());
//            if(operation_.unit() == fdf::ItemUnit::Ut)
//            {
//            }
            break;
        case Measured            : st = isSale_ ?
                        fdf::LabelProductStatus::Measured :
                        fdf::LabelProductStatus::MeasuredRefund;
            lcd.setUnit(operation_.unit());
            lcd.setQuantity(operation_.quantity());
//            if(operation_.unit() == fdf::ItemUnit::Ut)
//            {
//            }
            lcd.setFraction(operation_.fraction());
            break;
        case OtherPromisStatus   : st =  fdf::LabelProductStatus::NotChanged;break;
        }

        lcd.setExcpectedStatus(st);

        lmWarning() << promisStatus_ << isSale_ << logtab << static_cast<qint32>(lcd.excpectedStatus());


        CoreApiConst::ErrorCode err;
        QString msg;
        if(!lcd.isValid(err, msg)) emit error(msg);
        else
        {
            emit showBusy(true);
            std::function<QPair<CoreTransaction, CoreApiResult>()> task =
                    [lcd, this]() -> QPair<CoreTransaction, CoreApiResult> {
                CoreTransaction transaction;
                transaction.setCashierLogin(session_->cashier().login());
                transaction.setCashierPassword(session_->cashier().password());
                transaction.setOperation(CoreTransaction::Operation::CHECK_LABEL);
                transaction.setParams(lcd.toMap());
                transaction.genUid();
                lmWarning() << logvariant(transaction.toMap());
                CoreApiResult res = api_->exec(transaction);
                return qMakePair(transaction, res);
            };
            execTask(task);
        }
    }
}

bool ReceiptOperationWrapper::canSkipB5() const
{
    RegData rd = status_->status().regData();
    return rd.extFlags().getAutoCashboxForLabled() ||
           rd.extFlags().getRestaurant() ||
           rd.extFlags().getWholesale();
}


QString ReceiptOperationWrapper::label() const
{
    return label_.tradeMark();
}

void ReceiptOperationWrapper::setLabel(QString newLabel)
{
    if(newLabel.endsWith("\r\n"))newLabel = newLabel.mid(0, newLabel.size() - 2);
    if(newLabel.endsWith("\n"))newLabel = newLabel.mid(0, newLabel.size() - 1);
    if(newLabel.startsWith("\u001d"))newLabel = newLabel.mid(1);

    lmWarning() << newLabel << logtab << ProductCodeCreator::checkAndPrepareGs1(newLabel);
    newLabel =  ProductCodeCreator::checkAndPrepareGs1(newLabel);
    newLabel = ProductCodeCreator::checkAndPrepareKis(newLabel);
    lmWarning() << newLabel;


    if (label() == newLabel)
        return;


    ProductCodeCreator pcc;
    label_ = pcc.create(newLabel);
    lmWarning() << "TEST_LABELS_END" << logtab
                << label_.tradeMark() << static_cast<qint32>(label_.type()) << logtab
                << label_.productCode().data() << label_.t91() << label_.t92() << logtab
                << label_.tailCrc() << QString::number(label_.tailCrc(), 16)
                << logbinary(label_.t92().toLatin1()) << logtab
                << logbinary(QByteArray::fromBase64(label_.t92().toLatin1()))  << logtab
                << label_.t93() << logbinary(logbinary(QByteArray::fromBase64(label_.t93().toLatin1()))) << logtab
                << static_cast<qint32>(label_.lct2100());
    operation_.setControlCode(label_.calc2115());
    if(!label_.isEncodedTradeMark()) operation_.setCode(label_.productCode());
    emit labelChanged();
}

QString ReceiptOperationWrapper::labelText() const
{
    return label_.typeName() + "\n" + label_.productCode().data();
}

qint32 ReceiptOperationWrapper::denominator() const
{
    return operation_.fraction().value_or(QuantityFraction()).denominator();
}

void ReceiptOperationWrapper::setDenominator(qint32 newDenominator)
{
    if (operation_.fraction().has_value() && operation_.fraction().value().denominator() == newDenominator)
        return;
    if(operation_.fraction().has_value())
    {
        QuantityFraction f = operation_.fraction().value();
        f.setDenominator(newDenominator);
        operation_.setFraction(f);
    }
    emit denominatorChanged();
}

qint32 ReceiptOperationWrapper::nominator() const
{
    return operation_.fraction().value_or(QuantityFraction()).nominator();
}

void ReceiptOperationWrapper::setNominator(qint32 newNominator)
{
    if (operation_.fraction().has_value() && operation_.fraction().value().nominator() == newNominator)
        return;
    if(operation_.fraction().has_value())
    {
        QuantityFraction f = operation_.fraction().value();
        f.setNominator(newNominator);
        operation_.setFraction(f);
    }

    emit nominatorChanged();
}

bool ReceiptOperationWrapper::hasFractial() const
{
    return/* promisStatus_ == Measured &&*/ itemUnit() == 0 && operation_.fraction().has_value();
}

void ReceiptOperationWrapper::setHasFractial(bool newHasFractial)
{
    bool oldVal = hasFractial();
    if(!newHasFractial || /*promisStatus_ != Measured ||*/ itemUnit() != 0 )
    {
        operation_.setFraction(std::optional<QuantityFraction>());
    }
    else
    {
        if(!operation_.fraction().has_value())
        {
            QuantityFraction f;
            f.setDenominator(1);
            f.setNominator(1);
            operation_.setFraction(f);
        }
    }
    if (oldVal != newHasFractial) emit hasFractialChanged();
}

const QVariantList &ReceiptOperationWrapper::itemUnitModel() const
{
    return itemUnitModel_;
}

int ReceiptOperationWrapper::itemUnit() const
{
    if(!operation_.unit().has_value() || !fdf::unitIsValid(operation_.unit().value())) return -1;
    int u = static_cast<int>(operation_.unit().value());
    for(int i = 0; i < itemUnitModel_.size(); ++i)
    {
        if(itemUnitModel_[i].toMap()["code"].toInt() == u) return i;
    }
    return -1;
}

void ReceiptOperationWrapper::setItemUnit(int idx)
{
    if(idx >= 0 && idx < itemUnitModel_.count())
    {
        fdf::ItemUnit u = static_cast<fdf::ItemUnit>(itemUnitModel_[idx].toMap()["code"].toInt());
        if(operation_.unit() != u)
        {
            operation_.setUnit(u);
            emit itemUnitChanged();
            emit hasFractialChanged();
        }
    }
    else
    {
        if(operation_.unit().has_value())
        {
            operation_.setUnit(std::optional<fdf::ItemUnit>());
            emit itemUnitChanged();
            emit hasFractialChanged();
        }
    }
}

int ReceiptOperationWrapper::promisStatus() const
{
    return promisStatus_;
}

void ReceiptOperationWrapper::setPromisStatus(int newPromisStatus)
{
    if (promisStatus_ == newPromisStatus)
        return;
    promisStatus_ = newPromisStatus;
    emit promisStatusChanged();
    emit hasFractialChanged();

}

bool ReceiptOperationWrapper::isSale() const
{
    return isSale_;
}

void ReceiptOperationWrapper::setIsSale(bool newIsSale)
{
    if (isSale_ == newIsSale)
        return;
    isSale_ = newIsSale;
    emit isSaleChanged();
}

QString ReceiptOperationWrapper::declarationNumber() const
{
    return operation_.declarationNumber();
}

void ReceiptOperationWrapper::setDeclarationNumber(const QString &newDeclarationNumber)
{
    if (operation_.declarationNumber() == newDeclarationNumber)
        return;
    operation_.setDeclarationNumber(newDeclarationNumber);
    emit declarationNumberChanged();
}

QString ReceiptOperationWrapper::countryCode() const
{
    return operation_.countryCode();
}

void ReceiptOperationWrapper::setCountryCode(const QString &newCountryCode)
{
    if (operation_.countryCode() == newCountryCode)
        return;
    operation_.setCountryCode(newCountryCode);
    emit countryCodeChanged();
}

QString ReceiptOperationWrapper::exciseTax() const
{
    return operation_.exciseTax().has_value() ? operation_.exciseTax().value().toString() : QString();
}

void ReceiptOperationWrapper::setExciseTax(const QString &newExciseTax)
{
    if(newExciseTax.isEmpty())
    {
        if(operation_.exciseTax().has_value())
        {
            operation_.setExciseTax(std::optional<FixNumber>());
            emit exciseTaxChanged();
        }
    }
    else
    {
        FixNumber v(DEFAULT_AMOUNT_MULT, newExciseTax);
        if (operation_.exciseTax().has_value() && operation_.exciseTax().value() == v)
            return;
        operation_.setExciseTax(v);
        emit exciseTaxChanged();
    }
}

QString ReceiptOperationWrapper::additionalParam() const
{
    return operation_.additionalParam();
}

void ReceiptOperationWrapper::setAdditionalParam(const QString &newAdditionalParam)
{
    if (operation_.additionalParam() == newAdditionalParam)
        return;
    operation_.setAdditionalParam(newAdditionalParam);
    emit additionalParamChanged();
}

void ReceiptOperationWrapper::createOpTypeModel()
{
    opTypeModel_.clear();
    RegistrationFlags1290 flags = status_->status().regData().extFlags();
    if(flags.getFosaCashbox() /*|| flags.getServicesOnly() - возникли замечания у атласа. и рапкат сказал убрать*/
        || !flags.getLabledProducts())  return;

    if(flags.getExcisableProducts())
    {
        opTypeModel_ << QVariantMap {
            {"code", ExcisableLabledWitoutCodeProduct},
            {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::ExcisableLabledWitoutCodeProduct)}
        };
        opTypeModel_ << QVariantMap {
            {"code", ExcisableLabledProduct},
            {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::ExcisableLabledProduct)}
        };
    }
    opTypeModel_ << QVariantMap {
        {"code", LabledWitoutCodeProduct},
        {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::LabledWitoutCodeProduct)}
    };
    opTypeModel_ << QVariantMap {
        {"code", LabledProduct},
        {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::LabledProduct)}
    };
    emit opTypeModelChanged();
}

void ReceiptOperationWrapper::createItemUnitModel()
{
    static const QList<fdf::ItemUnit> SUPPORTED = {
        fdf::ItemUnit::Ut     ,fdf::ItemUnit::g      ,fdf::ItemUnit::Kg     ,
        fdf::ItemUnit::t      ,fdf::ItemUnit::sm     ,fdf::ItemUnit::dm     ,
        fdf::ItemUnit::m      ,fdf::ItemUnit::sm2    ,fdf::ItemUnit::dm2    ,
        fdf::ItemUnit::m2     ,fdf::ItemUnit::ml     ,fdf::ItemUnit::l      ,
        fdf::ItemUnit::m3     ,fdf::ItemUnit::KWth   ,fdf::ItemUnit::GCal   ,
        fdf::ItemUnit::day    ,fdf::ItemUnit::hour   ,fdf::ItemUnit::minute ,
        fdf::ItemUnit::second ,fdf::ItemUnit::Kb     ,fdf::ItemUnit::Mb     ,
        fdf::ItemUnit::Gb     ,fdf::ItemUnit::Tb     ,fdf::ItemUnit::Another,
    };
    itemUnitModel_.clear();
    for(fdf::ItemUnit i : SUPPORTED)
    {
        QVariantMap m = {
            {"code", static_cast<qint32>(i)},
            {"name", fdf::unitText(i)}
        };
        itemUnitModel_ << m;
    }
    emit itemUnitModelChanged();
}

void ReceiptOperationWrapper::processResult(const CoreTransaction &source, const CoreApiResult &res)
{
    lmWarning() << res.isOk();
    if(!res.isOk())emit error(res.descr());
    emit showBusy(false);
    switch (source.operation()) {
    case CoreApiConst::RemoteCommand::CHECK_LABEL:
    {
        if(res.isOk())
        {
            LabelCheckResult r(res.data());
            operation_.setLabelCheckResult(r);
            save();
        }
    }break;
    default:
        break;
    }
}
