#include "simplereceiptoperationwrapper.h"
#include "cashboxbuildconfig.h"
#include "logmanip.h"
#include "labelcode.h"
#include "productcodecreator.h"

SimpleReceiptOperationWrapper::SimpleReceiptOperationWrapper(QObject *parent)
    : QObject(parent)
    , index_(-1)
    , opTypeModel_()
    , operation_()
    , status_(new CashboxStatusWrapper(this))
    , label_()
{
    lmWarning() << itemUnit();
    createOpTypeModel();
    lmWarning() << itemUnit();
    createItemUnitModel();
    lmWarning() << itemUnit();
}

SimpleReceiptOperationWrapper::~SimpleReceiptOperationWrapper()
{

}

qint32 SimpleReceiptOperationWrapper::index() const
{
    return index_;
}

void SimpleReceiptOperationWrapper::setIndex(qint32 newIndex)
{
    index_ = newIndex;
}

QVariantMap SimpleReceiptOperationWrapper::data() const
{
    return operation_.toMap();
}

void SimpleReceiptOperationWrapper::setData(const QVariantMap &map)
{
    lmWarning() << opType() << opTypeIndex() << logvariant(map);
    operation_.parseMap(map);
    emit opTypeChanged();
    emit opTypeIndexChanged();
    emit nameChanged();
    emit vatRateChanged();
    emit payTypeChanged();
    emit priceChanged();
    emit quantityChanged();
    emit industryPropsChanged();
}

QVariantList SimpleReceiptOperationWrapper::opTypeModel() const
{
    return opTypeModel_;
}

QVariantList SimpleReceiptOperationWrapper::vatModel() const
{
    QVariantList res;
    for(int i = Vat20; i <= VatNone; ++i)
    {
        res << QVariantMap {
        {"name", fdf::vatLabel(static_cast<fdf::VatRate>(i))},
        {"code", i}
        };
    }
    return res;
}

QVariantList SimpleReceiptOperationWrapper::payTypeModel() const
{
    QVariantList res;
    for(int i = FullPrepayment; i <= CrediPayment; ++i)
    {
        res << QVariantMap {
        {"name", fdf::itemPaymentTypeToString(static_cast<fdf::ItemPaymentType>(i))},
        {"code", i}
        };
    }
    return res;
}
QVariantList SimpleReceiptOperationWrapper::itemUnitModel() const
{
    return itemUnitModel_;
}

QVariantMap SimpleReceiptOperationWrapper::agentData() const
{
    if(!fdf::itemPayAgentFlagIsValid(operation_.agentFlag())) return QVariantMap();
    QVariantMap res = {
        {"agentFlag", static_cast<qint32>(operation_.agentFlag())},
        {"providerInn", operation_.providerInn().trimmed()},
        {"providerData", operation_.providerData().value_or(ProviderData()).toMap()}
    };
    if(operation_.agentFlag() > fdf::ItemPayAgentFlag::PaySubAgent) return res;
    res["agentData"] = operation_.agentData().value_or(AgentData()).toMap();
    return res;
}

int SimpleReceiptOperationWrapper::opTypeIndex() const
{
    if(!fdf::opTypeIsValid(operation_.type())) return -1;
    for(int i = 0; i < opTypeModel_.size(); ++i)
    {
        if(opTypeModel_[i].toMap()["code"].toInt() == opType()) return i;
    }
    return -1;
}

void SimpleReceiptOperationWrapper::setOpTypeIndex(int idx)
{
    if(idx >= 0 && idx < opTypeModel_.size())
        setOpType(static_cast<ROperationType>(opTypeModel_[idx].toMap()["code"].toInt()));
}

int SimpleReceiptOperationWrapper::itemUnit() const
{
    if(!operation_.unit().has_value() || !fdf::unitIsValid(operation_.unit().value())) return 0;
    int u = static_cast<int>(operation_.unit().value());
    for(int i = 0; i < itemUnitModel_.size(); ++i)
    {
        if(itemUnitModel_[i].toMap()["code"].toInt() == u)
        {
            return i;
        }
    }
    return 0;
}

void SimpleReceiptOperationWrapper::setItemUnit(int idx)
{
    if(idx > 0 && idx < itemUnitModel_.count())
    {
        fdf::ItemUnit u = static_cast<fdf::ItemUnit>(itemUnitModel_[idx].toMap()["code"].toInt());
        if(operation_.unit() != u)
        {
            operation_.setUnit(u);
            emit itemUnitChanged();
        }
    }
    else
    {
        if(operation_.unit().has_value())
        {
            operation_.setUnit(std::optional<fdf::ItemUnit>());
            emit itemUnitChanged();
        }
    }
}

SimpleReceiptOperationWrapper::ROperationType SimpleReceiptOperationWrapper::opType() const
{
    return static_cast<ROperationType>(operation_.type());
}

void SimpleReceiptOperationWrapper::setOpType(ROperationType tp)
{
    fdf::ReceiptOperationType t = static_cast<fdf::ReceiptOperationType>(tp);
    if(operation_.type() != t)
    {
        operation_.setType(t);
        emit opTypeChanged();
        emit opTypeIndexChanged();
    }
}

QString SimpleReceiptOperationWrapper::name() const
{
    return operation_.name();
}

void SimpleReceiptOperationWrapper::setName(const QString &n)
{
    if(operation_.name() != n)
    {
        operation_.setName(n);
        emit nameChanged();
    }
}

SimpleReceiptOperationWrapper::VatRate SimpleReceiptOperationWrapper::vatRate() const
{
    return static_cast<VatRate>(operation_.vatRate());
}

void SimpleReceiptOperationWrapper::setVatRate(VatRate rate)
{
    fdf::VatRate r = static_cast<fdf::VatRate>(rate);
    if(operation_.vatRate() != r)
    {
        operation_.setVatRate(r);
        emit vatRateChanged();
    }
}

SimpleReceiptOperationWrapper::ItemPaymentType SimpleReceiptOperationWrapper::payType() const
{
    return static_cast<ItemPaymentType>(operation_.paymentType());
}

void SimpleReceiptOperationWrapper::setPayType(ItemPaymentType tp)
{
    fdf::ItemPaymentType t = static_cast<fdf::ItemPaymentType>(tp);
    if(operation_.paymentType() != t)
    {
        operation_.setPaymentType(t);
        emit payTypeChanged();
    }
}

QString SimpleReceiptOperationWrapper::price() const
{
    return operation_.price().value() <=0 ? QString() : operation_.price().toString();
}

void SimpleReceiptOperationWrapper::setPrice(const QString &price)
{
    FixNumber p = FixNumber(DEFAULT_AMOUNT_MULT, price);
    if(operation_.price() != p)
    {
        operation_.setPrice(p);
        emit priceChanged();
    }
}

QString SimpleReceiptOperationWrapper::quantity() const
{
    return operation_.quantity().value() > 0ll ? operation_.quantity().toShortString() : QString();
}

void SimpleReceiptOperationWrapper::setQuantity(const QString &q)
{
    FixNumber qu (DEFAULT_QUANTITY_MULT, q);
    if(operation_.quantity() != qu)
    {
        operation_.setQuantity(qu);
        emit quantityChanged();
    }
}

QString SimpleReceiptOperationWrapper::additionalParam() const
{
    return operation_.additionalParam();
}

void SimpleReceiptOperationWrapper::setAdditionalParam(const QString &newAdditionalParam)
{
    if (operation_.additionalParam() == newAdditionalParam)
        return;
    operation_.setAdditionalParam(newAdditionalParam);
    emit additionalParamChanged();
}

QString SimpleReceiptOperationWrapper::exciseTax() const
{
    return operation_.exciseTax().has_value() ? operation_.exciseTax().value().toString() : QString();
}

void SimpleReceiptOperationWrapper::setExciseTax(const QString &newExciseTax)
{
    if(newExciseTax.isEmpty())
    {
        if(operation_.exciseTax().has_value())
        {
            operation_.setExciseTax(std::optional<FixNumber>());
            emit exciseTaxChanged();
        }
    }
    else
    {
        FixNumber v(DEFAULT_AMOUNT_MULT, newExciseTax);
        if (operation_.exciseTax().has_value() && operation_.exciseTax().value() == v)
            return;
        operation_.setExciseTax(v);
        emit exciseTaxChanged();
    }
}

QString SimpleReceiptOperationWrapper::countryCode() const
{
    return operation_.countryCode();
}

void SimpleReceiptOperationWrapper::setCountryCode(const QString &newCountryCode)
{
    if (operation_.countryCode() == newCountryCode)
        return;
    operation_.setCountryCode(newCountryCode);
    emit countryCodeChanged();
}

QString SimpleReceiptOperationWrapper::declarationNumber() const
{
    return operation_.declarationNumber();
}

void SimpleReceiptOperationWrapper::setDeclarationNumber(const QString &newDeclarationNumber)
{
    if (operation_.declarationNumber() == newDeclarationNumber)
        return;
    operation_.setDeclarationNumber(newDeclarationNumber);
    emit declarationNumberChanged();
}

QVariantList SimpleReceiptOperationWrapper::industryProps() const
{
    QVariantList res;
    QList<IndustryProperty> props = operation_.industryProperties();
    std::for_each(props.constBegin(), props.constEnd(),
                  [&res](const IndustryProperty &p){res << p.toMap();});
    return  res;
}

QString SimpleReceiptOperationWrapper::labelText() const
{
    return label_.typeName() + "\n" + label_.productCode().data();
}

void SimpleReceiptOperationWrapper::parseBarcode(const QString &code)
{
    lmWarning() << code;
    QString newLabel = code;
    if(newLabel.endsWith("\r\n"))newLabel = newLabel.mid(0, newLabel.size() - 2);
    if(newLabel.endsWith("\n"))newLabel = newLabel.mid(0, newLabel.size() - 1);
    if(newLabel.startsWith("\u001d"))newLabel = newLabel.mid(1);

    lmWarning() << newLabel << logtab << ProductCodeCreator::checkAndPrepareGs1(newLabel);
    newLabel =  ProductCodeCreator::checkAndPrepareGs1(newLabel);
    newLabel = ProductCodeCreator::checkAndPrepareKis(newLabel);
    lmWarning() << newLabel;



    label_ = ProductCodeCreator().create(newLabel);
    if(label_.isGs())
    {
        label_ = LabelCode();
        emit error(tr("Отсканирован код маркировки. Воспользуйтесь интерфейсом для продажи маркированных товаров!"));
    }
    else
    {
        operation_.setCode(label_.productCode());
    }
    emit barcodeChanged();
}

void SimpleReceiptOperationWrapper::save()
{
    CoreApiConst::ErrorCode err;
    QString msg;
    if(!operation_.isValid(err, msg))
    {
        emit error(msg);
    }
    else
    {
        lmWarning() << index_ << logvariant(operation_.toMap());
        emit saveData(index_, operation_.toMap());
        emit finish();
    }
}

void SimpleReceiptOperationWrapper::setAgentData(const QVariantMap &map)
{
    operation_.setAgentFlag(fdf::ItemPayAgentFlag::NotAgent);
    operation_.setProviderInn(QString());
    operation_.setProviderData(std::optional<ProviderData>());
    operation_.setAgentData(std::optional<AgentData>());
    if(map.isEmpty()) return;
    fdf::ItemPayAgentFlag f = static_cast<fdf::ItemPayAgentFlag>(map["agentFlag"].toInt());
    if(!fdf::itemPayAgentFlagIsValid(f)) return;
    operation_.setAgentFlag(f);
    operation_.setProviderInn(map["providerInn"].toString());
    operation_.setProviderData(ProviderData(map["providerData"].toMap()));
    if(f > fdf::ItemPayAgentFlag::PaySubAgent) return;
    operation_.setAgentData(AgentData(map["agentData"].toMap()));
}

void SimpleReceiptOperationWrapper::setIndustryProps(const QVariantList &props)
{
    if(props.isEmpty()) operation_.setIndustryProperties(QList<IndustryProperty>());
    else
    {
        QList<IndustryProperty> res;
        std::for_each(props.constBegin(), props.constEnd(),
                      [&res] (const QVariant &v){res << IndustryProperty(v.toMap());});
        operation_.setIndustryProperties(res);
    }
    emit industryPropsChanged();
}

void SimpleReceiptOperationWrapper::createOpTypeModel()
{
    opTypeModel_.clear();
    static const std::array<int, 19> TYPES = {
        RiaRights, PrepaymentOp, AgentCommission, Fine, Another,
        PropertyRight, NonReleaseIncome, AnotherPayments, TradeFee, SpaFee,
        Deposit, Cconsumption, PensionContribution, PensionContribution2, MedicalContribution,
        MedicalContribution2, SocialContribution, CasinoPayment, PaymentOfMoney
    };
    QVariantList res;

    RegistrationFlags1290 flags = status_->status().regData().extFlags();
    if(!flags.getFosaCashbox() /*&& !flags.getServicesOnly() - возникли замечания у атласа. и рапкат тоже сказал убрать*/)
    {
        res << QVariantMap {
            {"code", Product},
            {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::Product)}
        };
        if(flags.getExcisableProducts())
        {
            res << QVariantMap {
                {"code", ExcisableProduct},
                {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::ExcisableProduct)}
            };
        }
    }
    res << QVariantMap {
        {"code", Job},
        {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::Job)}
    };
    res << QVariantMap {
        {"code", Service},
        {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::Service)}
    };
    if(flags.getGamesCashbox())
    {
        res << QVariantMap {
            {"code", GameRate},
            {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::GameRate)}
        };
        res << QVariantMap {
            {"code", GamePrize},
            {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::GamePrize)}
        };
    }
    if(flags.getLotteryCashbox())
    {
        res << QVariantMap {
            {"code", Lottery},
            {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::Lottery)}
        };
        res << QVariantMap {
            {"code", LotteryPrize},
            {"name", fdf::opTypeCaption(fdf::ReceiptOperationType::LotteryPrize)}
        };
    }

    for(int i : TYPES)
    {
        res << QVariantMap {
            {"code", i},
            {"name", fdf::opTypeCaption(static_cast<fdf::ReceiptOperationType>(i))}
        };
    }
    opTypeModel_ = res;
    emit opTypeModelChanged();
}

void SimpleReceiptOperationWrapper::createItemUnitModel()
{
    static const QList<fdf::ItemUnit> SUPPORTED = {
        fdf::ItemUnit::Ut     ,fdf::ItemUnit::g      ,fdf::ItemUnit::Kg     ,
        fdf::ItemUnit::t      ,fdf::ItemUnit::sm     ,fdf::ItemUnit::dm     ,
        fdf::ItemUnit::m      ,fdf::ItemUnit::sm2    ,fdf::ItemUnit::dm2    ,
        fdf::ItemUnit::m2     ,fdf::ItemUnit::ml     ,fdf::ItemUnit::l      ,
        fdf::ItemUnit::m3     ,fdf::ItemUnit::KWth   ,fdf::ItemUnit::GCal   ,
        fdf::ItemUnit::day    ,fdf::ItemUnit::hour   ,fdf::ItemUnit::minute ,
        fdf::ItemUnit::second ,fdf::ItemUnit::Kb     ,fdf::ItemUnit::Mb     ,
        fdf::ItemUnit::Gb     ,fdf::ItemUnit::Tb     ,fdf::ItemUnit::Another,
    };
    itemUnitModel_.clear();
    for(fdf::ItemUnit i : SUPPORTED)
    {
        QVariantMap m = {
            {"code", static_cast<qint32>(i)},
            {"name", fdf::unitText(i)}
        };
        itemUnitModel_ << m;
    }
    itemUnitModel_.push_front(QVariantMap{
                                  {"code", static_cast<qint32>(-1)},
                                  {"name", QStringLiteral("Не указывать")}
                              });
    emit itemUnitModelChanged();
}
